// Copyright (c) 2022 The Foundry Visionmongers Ltd. All Rights Reserved.

#ifndef USG_PRIVATE_USDATTRIMPL_H
#define USG_PRIVATE_USDATTRIMPL_H

#include "usg/base/Value.h"
#include "usg/geom/Attribute.h"

#include "pxr/pxr.h"
#include "pxr/usd/sdf/layer.h"


namespace usg
{

//--------------------------------------------------------------------


/*! Attribute type info.
    Stores key intrisic properties as local vars for fast referencing
    and mapping to the corresponding Usd types.
*/
struct ValueTypeInfo
{
  // Intrinsic type info:
  PXR_NS::TfToken          name;                     //!< ValueType name  (may be different than the Usd Sdf typename!)
  Value::Type              usg_value_type;           //!< ValueType enum index
  Value::Type              usg_base_value_type;      //!< Base ValueType enum index  (Float3 instead of Color3f)
  Value::DataType          usg_base_data_type;       //!< Base ValueType element data type (float, int, etc)
  PXR_NS::SdfValueTypeName sdf_value_typename;       //!< Usd Sdf corresponding typename
  PXR_NS::TfType           tf_data_type;             //!< Usd Tf corresponding data type
  int                      bytes_per_data_element;   //!< Byte count of an individual data element (ie float=4, double=8)
  int                      num_data_elements;        //!< How many data elements in value type (ie Vec3=3, Mat4=16)

  // Derived flags for fast testing:
  bool                  is_array;     //!< Array variant of another type
  bool                  is_int;       //!<
  bool                  is_float;     //!<
  bool                  is_double;    //!<
  bool                  is_string;    //!< Arbitrary string with no implied meaning
  bool                  is_asset;     //!< Asset path that can be resolved vs. a raw string


  bool   isDefault() const { return (usg_value_type == Value::DefaultType); }
  bool   isValid() const { return (usg_value_type != Value::InvalidType); }
  size_t bytesPerValue() const { return (bytes_per_data_element*num_data_elements); }


  ValueTypeInfo() {}
  ValueTypeInfo(const char*                     _name,
                Value::Type                     _usg_value_type,
                const PXR_NS::SdfValueTypeName& _sdf_value_typename,
                Value::DataType                 _usg_base_data_type,
                int                             _num_data_elements) :
    name(_name),
    usg_value_type(_usg_value_type),
    usg_base_value_type(Value::InvalidType), // << updated in BuildTypeMaps ctor
    usg_base_data_type(_usg_base_data_type),
    sdf_value_typename(_sdf_value_typename),
    num_data_elements(_num_data_elements),
    //
    is_array(false),
    is_int(false),
    is_float(false),
    is_double(false),
    is_string(false),
    is_asset(false)
  {
    // Only bother deriving info if this is a concrete type:
    if (num_data_elements > 0) {

      switch (usg_base_data_type) {
      //-----------------------------------------------------------------------------
      case Value::STRING_DATA:
        bytes_per_data_element = sizeof(std::string);
        is_string = true;
        usg_base_value_type = Value::String;
        break;

      case Value::TOKEN_DATA:
        bytes_per_data_element = sizeof(Token);
        is_string = true;
        usg_base_value_type = Value::Token;
        break;

      case Value::ASSETPATH_DATA:
        bytes_per_data_element = sizeof(AssetPath);
        is_asset = true;
        usg_base_value_type = Value::AssetPath;
        break;

      //-----------------------------------------------------------------------------
      case Value::INT8_DATA:
        bytes_per_data_element = 1;
        is_int = true;
        usg_base_value_type = Value::UChar;
        break;

      case Value::INT16_DATA:
        bytes_per_data_element = sizeof(int16_t);
        is_int = true;
        //usg_base_value_type = Value::UShort;
        break;

      case Value::INT32_DATA:
        bytes_per_data_element = sizeof(int32_t);
        is_int = true;
        if      (_num_data_elements == 2) usg_base_value_type = Value::Int2;
        else if (_num_data_elements == 3) usg_base_value_type = Value::Int3;
        else if (_num_data_elements == 4) usg_base_value_type = Value::Int4;
        else usg_base_value_type = Value::Int;
        break;

      case Value::INT64_DATA:
        bytes_per_data_element = sizeof(int64_t);
        is_int = true;
        usg_base_value_type = Value::Int64;
        break;

      //-----------------------------------------------------------------------------
      case Value::HALF_DATA:
        bytes_per_data_element = sizeof(fdk::half);
        is_float = true;
        if      (_num_data_elements == 2) usg_base_value_type = Value::Half2;
        else if (_num_data_elements == 3) usg_base_value_type = Value::Half3;
        else if (_num_data_elements == 4)
        {
          if (_usg_value_type == Value::Quath || _usg_value_type == Value::QuathArray) usg_base_value_type = Value::Quath;
          else usg_base_value_type = Value::Half4;
        }
        else usg_base_value_type = Value::Half;
        break;

      case Value::FLOAT_DATA:
        bytes_per_data_element = sizeof(float);
        is_float = true;
        if      (_num_data_elements == 2) usg_base_value_type = Value::Float2;
        else if (_num_data_elements == 3) usg_base_value_type = Value::Float3;
        else if (_num_data_elements == 4)
        {
          if (_usg_value_type == Value::Quatf || _usg_value_type == Value::QuatfArray) usg_base_value_type = Value::Quatf;
          else usg_base_value_type = Value::Float4;
        }
        else usg_base_value_type = Value::Float;
        break;

      case Value::DOUBLE_DATA:
        bytes_per_data_element = sizeof(double);
        is_double = true;
        if      (_num_data_elements == 2) usg_base_value_type = Value::Double2;
        else if (_num_data_elements == 3) usg_base_value_type = Value::Double3;
        else if (_num_data_elements == 4)
        {
          if      (_usg_value_type == Value::Matrix2d || _usg_value_type == Value::Matrix2dArray) usg_base_value_type = Value::Matrix2d;
          else if (_usg_value_type == Value::Quatd    || _usg_value_type == Value::QuatdArray)    usg_base_value_type = Value::Quatd;
          else usg_base_value_type = Value::Double4;
        }
        else if (_num_data_elements == 9) usg_base_value_type = Value::Matrix3d;
        else if (_num_data_elements ==16) usg_base_value_type = Value::Matrix4d;
        else usg_base_value_type = Value::Double;
        break;

      //-----------------------------------------------------------------------------
      default:
        break; // should throw error
      }

      // Get some info from Usd:
      if (isValid()) {
        is_array     = sdf_value_typename.IsArray();
        tf_data_type = sdf_value_typename.GetType();
      }

    }

  }

};

  //--------------------------------------------------------------------------------------------------


  /*----------------------------------*/
  /*        Static operations         */
  /*----------------------------------*/


inline
std::ostream& operator << (std::ostream& o, const ValueTypeInfo& info)
{
  if (info.isValid()) {
    o << info.name << "(";
    if (info.is_int)
      o << "int";
    else if (info.is_float)
      o << "float";
    else if (info.is_double)
      o << "double";
    else if (info.is_string)
      o << "string";
    if (info.is_array)
      o << "[]";
    o << " bytesPerElem=" << info.bytes_per_data_element;
    o << ", nElems=" << info.num_data_elements;
    //o << ", TfType=" << info.tf_data_type;

    o << ")";
  }
  else {
    o << "<invalid>";
  }

  return o;
}



} // namespace usg

#endif
