// Copyright (c) 2022 The Foundry Visionmongers Ltd. All Rights Reserved.

#ifndef USG_PRIVATE_USDHDIMPL_H
#define USG_PRIVATE_USDHDIMPL_H

#include "usg/private/UsdImpl.h"

#include "pxr/imaging/hd/renderBuffer.h"
#include "pxr/imaging/hd/renderDelegate.h"
#include "pxr/imaging/hd/rendererPlugin.h"
#include "pxr/imaging/hd/rendererPluginRegistry.h"
#include "pxr/imaging/hd/engine.h"

#include "pxr/imaging/hgiInterop/hgiInterop.h"

#include "pxr/imaging/hdx/taskController.h"

#include "pxr/usdImaging/usdImaging/version.h"
#if USD_IMAGING_API_VERSION >= 13
# include "pxr/imaging/hgi/hgi.h"
# include "pxr/imaging/hgi/tokens.h"
#endif

#include "pxr/usdImaging/usdImaging/delegate.h"

namespace usg
{


//--------------------------------------------------------------------------------------------------


class RenderOutputLayerHandle
{
public:
  PXR_NS::HdRenderBuffer* buffer;

  // Derived from HdFormat of HdRenderBuffer:
  int32_t  nchans;      //!< Number of channels packed (interleaved) in buffer
  int32_t  bpchan;      //!< Buffer data type is 1,2, or 4 bytes per chan
  bool     is_signed;   //!< Buffer data type is signed (not normalized)
  bool     is_float;    //!< Buffer data type is floating-point


public:
  RenderOutputLayerHandle() : nchans(0), bpchan(0), is_signed(false), is_float(false) {}

  RenderOutputLayerHandle(const RenderOutputLayerHandle& b) :
    buffer(b.buffer),
    nchans(b.nchans),
    bpchan(b.bpchan),
    is_signed(b.is_signed),
    is_float(b.is_float)
  {
    //
  }

  RenderOutputLayerHandle(PXR_NS::HdRenderBuffer* p) :
    buffer(p)
  {
    getDataTypeInfo();
  }
  RenderOutputLayerHandle(const PXR_NS::HdRenderBuffer* p) :
    buffer(const_cast<PXR_NS::HdRenderBuffer*>(p))
  {
    getDataTypeInfo();
  }

  void getDataTypeInfo()
  {
    nchans = bpchan = 0;
    is_signed = is_float = false;
    if (!buffer)
      return;

    switch (buffer->GetFormat())
    {
      default: break;
      // UNorm8 - a 1-byte value representing a float between 0 and 1.
      // float value = (unorm / 255.0f);
      case PXR_NS::HdFormatUNorm8:     nchans = 1; bpchan = 1; is_signed = false; is_float = false; break;
      case PXR_NS::HdFormatUNorm8Vec2: nchans = 2; bpchan = 1; is_signed = false; is_float = false; break;
      case PXR_NS::HdFormatUNorm8Vec3: nchans = 3; bpchan = 1; is_signed = false; is_float = false; break;
      case PXR_NS::HdFormatUNorm8Vec4: nchans = 4; bpchan = 1; is_signed = false; is_float = false; break;
      // SNorm8 - a 1-byte value representing a float between -1 and 1.
      // float value = max(snorm / 127.0f, -1.0f);
      case PXR_NS::HdFormatSNorm8:     nchans = 1; bpchan = 1; is_signed = true; is_float = false; break;
      case PXR_NS::HdFormatSNorm8Vec2: nchans = 2; bpchan = 1; is_signed = true; is_float = false; break;
      case PXR_NS::HdFormatSNorm8Vec3: nchans = 3; bpchan = 1; is_signed = true; is_float = false; break;
      case PXR_NS::HdFormatSNorm8Vec4: nchans = 4; bpchan = 1; is_signed = true; is_float = false; break;
      // Float16 - a 2-byte IEEE half-precision float.
      case PXR_NS::HdFormatFloat16:     nchans = 1; bpchan = 2; is_signed = true; is_float = true; break;
      case PXR_NS::HdFormatFloat16Vec2: nchans = 2; bpchan = 2; is_signed = true; is_float = true; break;
      case PXR_NS::HdFormatFloat16Vec3: nchans = 3; bpchan = 2; is_signed = true; is_float = true; break;
      case PXR_NS::HdFormatFloat16Vec4: nchans = 4; bpchan = 2; is_signed = true; is_float = true; break;
      // Float32 - a 4-byte IEEE float.
      case PXR_NS::HdFormatFloat32:     nchans = 1; bpchan = 4; is_signed = true; is_float = true; break;
      case PXR_NS::HdFormatFloat32Vec2: nchans = 2; bpchan = 4; is_signed = true; is_float = true; break;
      case PXR_NS::HdFormatFloat32Vec3: nchans = 3; bpchan = 4; is_signed = true; is_float = true; break;
      case PXR_NS::HdFormatFloat32Vec4: nchans = 4; bpchan = 4; is_signed = true; is_float = true; break;
      // UInt16 - a 2-byte unsigned integer
      case PXR_NS::HdFormatUInt16:     nchans = 1; bpchan = 2; is_signed = false; is_float = false; break;
      case PXR_NS::HdFormatUInt16Vec2: nchans = 2; bpchan = 2; is_signed = false; is_float = false; break;
      case PXR_NS::HdFormatUInt16Vec3: nchans = 3; bpchan = 2; is_signed = false; is_float = false; break;
      case PXR_NS::HdFormatUInt16Vec4: nchans = 4; bpchan = 2; is_signed = false; is_float = false; break;
      // Int32 - a 4-byte signed integer
      case PXR_NS::HdFormatInt32:     nchans = 1; bpchan = 4; is_signed = true; is_float = false; break;
      case PXR_NS::HdFormatInt32Vec2: nchans = 2; bpchan = 4; is_signed = true; is_float = false; break;
      case PXR_NS::HdFormatInt32Vec3: nchans = 3; bpchan = 4; is_signed = true; is_float = false; break;
      case PXR_NS::HdFormatInt32Vec4: nchans = 4; bpchan = 4; is_signed = true; is_float = false; break;
      // Depth-stencil format
      case PXR_NS::HdFormatFloat32UInt8: nchans = 2; bpchan = 5; is_signed = true; is_float = true; break;
    }
  }

};


//--------------------------------------------------------------------------------------------------


class UsgSceneDelegate : public PXR_NS::UsdImagingDelegate
{
  class Impl;
  public:
    UsgSceneDelegate(PXR_NS::HdRenderIndex* renderIndex, const PXR_NS::SdfPath& delegateId);
    ~UsgSceneDelegate();

    void                    populate(StageRef stage);
    PXR_NS::SdfPath         GetMaterialId(const PXR_NS::SdfPath &rprim) override;
    PXR_NS::VtValue         GetMaterialResource(const PXR_NS::SdfPath &materialId) override;
    PXR_NS::HdReprSelector  GetReprSelector(PXR_NS::SdfPath const &id) override;
    PXR_NS::VtValue         Get(PXR_NS::SdfPath const& id, PXR_NS::TfToken const& paramName) override;
    PXR_NS::VtValue         GetLightParamValue(PXR_NS::SdfPath const& id, PXR_NS::TfToken const& paramName) override;

    void            updateSelection(const PathSet& selected, const PathSet& deselected);
    const PathSet&  materialsUsingTexture(const std::string& texture);
    void            clearMaterialUsingTexture(const std::string& texture);

    Impl* impl() { return _pImpl.get(); };
  private:
    StageRef _stage;
    PathSet _selected;
    PathSet _deselected;
    std::unordered_map<std::string, PathSet> _texturesUsed;
    std::unique_ptr<Impl> _pImpl;
};
using UsgSceneDelegatePtr = std::unique_ptr<UsgSceneDelegate>;


class RenderInterfaceHandle
{
public:
  PXR_NS::UsdStageRefPtr            stage;
  UsgSceneDelegatePtr               sceneDelegate;       //!< The scene delegate feeding the render delegate
  //
  PXR_NS::HdRenderIndex*            renderIndex;         //!<
  PXR_NS::HdRenderDelegate*         renderDelegate;      //!<
  PXR_NS::HdRendererPlugin*         rendererPlugin;      //!<
  PXR_NS::HdxTaskController*        taskController;      //!<
  //
  PXR_NS::HdEngine*                 engine;              //!<
  //
  PXR_NS::TfRefPtr<PXR_NS::GlfSimpleLightingContext> lightingContext;
  PXR_NS::HdxSelectionTrackerSharedPtr              selTracker;
  //
  bool                           initialized;         //!< Should be set true only if all the allocations are successful


  RenderInterfaceHandle() :
    sceneDelegate(nullptr),
    //
    renderIndex(nullptr),
    renderDelegate(nullptr),
    rendererPlugin(nullptr),
    taskController(nullptr),
    engine(nullptr),
    //
    initialized(false)
  {
    //
  }

  ~RenderInterfaceHandle()
  {
    destroyHydra();
  }


  /*!
  */
  bool initHydra(const PXR_NS::TfToken& pluginId)
  {
    destroyHydra();

    // Look for our renderer plugin
    auto& pluginRegistry = PXR_NS::HdRendererPluginRegistry::GetInstance();
    if (!pluginRegistry.IsRegisteredPlugin(pluginId)) {
      return false;
    }

    rendererPlugin = pluginRegistry.GetRendererPlugin(pluginId);
    if (rendererPlugin == nullptr) {
      return false;
    }

    if (!rendererPlugin->IsSupported()) {
      pluginRegistry.ReleasePlugin(rendererPlugin);
      return false;
    }

    // Got the delegate plugin, initialize a new instance:
    renderDelegate = rendererPlugin->CreateRenderDelegate();

#if USD_IMAGING_API_VERSION >= 13
# if USD_IMAGING_API_VERSION >= 14
    static std::unique_ptr<PXR_NS::Hgi> hgi(PXR_NS::Hgi::CreatePlatformDefaultHgi());
# else
    static std::unique_ptr<PXR_NS::Hgi> hgi(PXR_NS::Hgi::GetPlatformDefaultHgi());
# endif
    static PXR_NS::HdDriver driver{ PXR_NS::HgiTokens->renderDriver, PXR_NS::VtValue(hgi.get()) };
    renderIndex = PXR_NS::HdRenderIndex::New(renderDelegate, {&driver});
#else
    renderIndex = PXR_NS::HdRenderIndex::New(renderDelegate);
#endif

    static PXR_NS::SdfPath taskControllerId("/HdNuke_TaskController");
    taskController = new PXR_NS::HdxTaskController(renderIndex, taskControllerId);

    // We need to create a new HdEngine to clear the TaskContextData from the old TaskController:
    if (engine != nullptr) {
      delete engine;
    }
    engine = new PXR_NS::HdEngine();

    initialized = true;

    return true;
  }


  /*!
  */
  void destroyHydra()
  {
    // These must be destroyed before the renderIndex
    sceneDelegate.reset();

    if (engine != nullptr) {
      delete engine;
      engine = nullptr;
    }

    if (taskController != nullptr) {
      delete taskController;
      taskController = nullptr;
    }

    if (renderIndex != nullptr) {
      delete renderIndex;
      renderIndex = nullptr;
    }

    if (rendererPlugin != nullptr) {
      if (renderDelegate != nullptr) {
        rendererPlugin->DeleteRenderDelegate(renderDelegate);
      }

      PXR_NS::HdRendererPluginRegistry::GetInstance().ReleasePlugin(rendererPlugin);

      renderDelegate = nullptr;
      rendererPlugin = nullptr;
    }

    initialized = false;
  }

};


//--------------------------------------------------------------------------------------------------


} // namespace usg

#endif

