// Copyright (c) 2022 The Foundry Visionmongers Ltd. All Rights Reserved.

#include "usg/base/ArgSet.h"
#include "usg/base/ArrayTypes.h"

namespace usg {

extern std::string empty_string;


//-------------------------------------------------------------------------


void
ArgSet::print(std::ostream& o) const
{
  if (_args.size() == 0) {
    return;
  }
  TokenValueMap::const_iterator it = _args.begin();
  o << "'" << it->first.asString() << "'=[" << it->second << "]";
  for (++it; it != _args.end(); ++it) {
    o << ", '" << it->first.asString() << "'=[" << it->second << "]";
  }
}

//-------------------------------------------------------------------------

namespace {
  template<class T>
  Value::Type typeFor();

  template<> Value::Type typeFor<std::string>() { return Value::Type::String; }
  template<> Value::Type typeFor<int32_t>() { return Value::Type::Int; }
  template<> Value::Type typeFor<uint64_t>() { return Value::Type::UInt64; }
  template<> Value::Type typeFor<double>() { return Value::Type::Double; }
  template<> Value::Type typeFor<bool>() { return Value::Type::Bool; }
  template<> Value::Type typeFor<fdk::Hash>() { return Value::Type::UInt64; }
  template<> Value::Type typeFor<fdk::Vec2d>() { return Value::Type::Float2; }
  template<> Value::Type typeFor<fdk::Vec3d>() { return Value::Type::Float3; }
  template<> Value::Type typeFor<fdk::Vec4d>() { return Value::Type::Float4; }
  template<> Value::Type typeFor<fdk::Mat4d>() { return Value::Type::Matrix4d; }
  template<> Value::Type typeFor<fdk::Box2d>() { return Value::Type::Float2Array; }
  template<> Value::Type typeFor<fdk::Box3d>() { return Value::Type::Float3Array; }

  template<class T>
  const T getFromTokenValueMap(const TokenValueMap& map, const Token& key, const T& dflt_val)
  {
    const TokenValueMap::const_iterator it = map.find(key);
    if (it != map.end() && it->second.getType() == typeFor<T>()) {
      return it->second.get<T>();
    }
    return dflt_val;
  }

  template<>
  const fdk::Box2d getFromTokenValueMap<fdk::Box2d>(const TokenValueMap& map, const Token& key, const fdk::Box2d& dflt_val)
  {
    const TokenValueMap::const_iterator it = map.find(key);
    if (it != map.end() && it->second.getType() == typeFor<fdk::Box2d>()) {
      Vec2dArray values;
      it->second.get<Vec2dArray>(values, {dflt_val.min, dflt_val.max});
      return fdk::Box2d(values[0], values[1]);
    }
    return dflt_val;
  }

  template<>
  const fdk::Box3d getFromTokenValueMap<fdk::Box3d>(const TokenValueMap& map, const Token& key, const fdk::Box3d& dflt_val)
  {
    const TokenValueMap::const_iterator it = map.find(key);
    if (it != map.end() && it->second.getType() == typeFor<fdk::Box3d>()) {
      Vec3dArray values;
      it->second.get<Vec3dArray>(values, {dflt_val.min, dflt_val.max});
      return fdk::Box3d(values[0], values[1]);
    }
    return dflt_val;
  }

  template<class T>
  void setToTokenValueMap(TokenValueMap& map, const Token& key, const T& value)
  {
    Value v(value);
    auto iter = map.emplace(key, v);
    if (!iter.second) {
      iter.first->second = v;
    }
  }

  template<>
  void setToTokenValueMap<fdk::Box2d>(TokenValueMap& map, const Token& key, const fdk::Box2d& value)
  {
    Value v(Vec2dArray{value.min, value.max});
    auto iter = map.emplace(key, v);
    if (!iter.second) {
      iter.first->second = v;
    }
  }

  template<>
  void setToTokenValueMap<fdk::Box3d>(TokenValueMap& map, const Token& key, const fdk::Box3d& value)
  {
    Value v(Vec3dArray{value.min, value.max});
    auto iter = map.emplace(key, v);
    if (!iter.second) {
      iter.first->second = v;
    }
  }
}

template<>
std::string ArgSet::get<std::string>(const Token& key, const std::string& dflt_val) const
{
  return getFromTokenValueMap<std::string>(_args, key, dflt_val);
}

template<>
int32_t ArgSet::get<int32_t>(const Token& key, const int32_t& dflt_val) const
{
  return getFromTokenValueMap<int32_t>(_args, key, dflt_val);
}

template<>
double ArgSet::get<double>(const Token& key, const double& dflt_val) const
{
  return getFromTokenValueMap<double>(_args, key, dflt_val);
}

template<>
bool ArgSet::get<bool>(const Token& key, const bool& dflt_val) const
{
  return getFromTokenValueMap<bool>(_args, key, dflt_val);
}

template<>
fdk::Hash ArgSet::get<fdk::Hash>(const Token& key, const fdk::Hash& dflt_val) const
{
  return fdk::Hash(getFromTokenValueMap<uint64_t>(_args, key, dflt_val.value()));
}

template<>
fdk::Vec2d ArgSet::get<fdk::Vec2d>(const Token& key, const fdk::Vec2d& dflt_val) const
{
  return getFromTokenValueMap<fdk::Vec2d>(_args, key, dflt_val);
}

template<>
fdk::Vec3d ArgSet::get<fdk::Vec3d>(const Token& key, const fdk::Vec3d& dflt_val) const
{
  return getFromTokenValueMap<fdk::Vec3d>(_args, key, dflt_val);
}

template<>
fdk::Vec4d ArgSet::get<fdk::Vec4d>(const Token& key, const fdk::Vec4d& dflt_val) const
{
  return getFromTokenValueMap<fdk::Vec4d>(_args, key, dflt_val);
}

template<>
fdk::Mat4d ArgSet::get<fdk::Mat4d>(const Token& key, const fdk::Mat4d& dflt_val) const
{
  return getFromTokenValueMap<fdk::Mat4d>(_args, key, dflt_val);
}

template<>
fdk::Box2d ArgSet::get<fdk::Box2d>(const Token& key, const fdk::Box2d& dflt_val) const
{
  return getFromTokenValueMap<fdk::Box2d>(_args, key, dflt_val);
}

template<>
fdk::Box3d ArgSet::get<fdk::Box3d>(const Token& key, const fdk::Box3d& dflt_val) const
{
  return getFromTokenValueMap<fdk::Box3d>(_args, key, dflt_val);
}

std::string
ArgSet::getString(const Token& key, const std::string& dflt_val) const
{
  return getFromTokenValueMap<std::string>(_args, key, dflt_val);
}

int32_t
ArgSet::getInt(const Token& key, int32_t dflt_val) const
{
  return getFromTokenValueMap<int32_t>(_args, key, dflt_val);
}

double
ArgSet::getDouble(const Token& key, double dflt_val) const
{
  return getFromTokenValueMap<double>(_args, key, dflt_val);
}

bool
ArgSet::getBool(const Token& key, bool dflt_val) const
{
  return getFromTokenValueMap<bool>(_args, key, dflt_val);
}

fdk::Hash
ArgSet::getHash(const Token& key, uint64_t dflt_val) const
{
  return fdk::Hash(getFromTokenValueMap<uint64_t>(_args, key, dflt_val));
}


//-------------------------------------------------------------------------


fdk::Vec2d
ArgSet::getVec2d(const Token& key, const fdk::Vec2d& dflt_val) const
{
  return getFromTokenValueMap<fdk::Vec2d>(_args, key, dflt_val);
}

fdk::Vec3d
ArgSet::getVec3d(const Token& key, const fdk::Vec3d& dflt_val) const
{
  return getFromTokenValueMap<fdk::Vec3d>(_args, key, dflt_val);
}

fdk::Vec4d
ArgSet::getVec4d(const Token& key, const fdk::Vec4d& dflt_val) const
{
  return getFromTokenValueMap<fdk::Vec4d>(_args, key, dflt_val);
}

fdk::Mat4d
ArgSet::getMat4d(const Token& key, const fdk::Mat4d& dflt_val) const
{
  return getFromTokenValueMap<fdk::Mat4d>(_args, key, dflt_val);
}

fdk::Box2d
ArgSet::getBox2d(const Token& key, const fdk::Box2d& dflt_val) const
{
  return getFromTokenValueMap<fdk::Box2d>(_args, key, dflt_val);
}

fdk::Box3d
ArgSet::getBox3d(const Token& key, const fdk::Box3d& dflt_val) const
{
  return getFromTokenValueMap<fdk::Box3d>(_args, key, dflt_val);
}

//-------------------------------------------------------------------------


void
ArgSet::appendString(const Token& key, char separator, const std::string& value)
{
  TokenValueMap::iterator it = _args.find(key);
  if (it == _args.end())
  {
    _args.insert(std::make_pair(key, value));
  }
  else if (it->second.getType() == Value::Type::String)
  {
    const auto oldStr = it->second.get<std::string>();
    it->second = Value(oldStr + separator + value);
  }
}

template<>
void ArgSet::set<std::string>(const Token& key, const std::string& value)
{
  setToTokenValueMap<std::string>(_args, key, value);
}

template<>
void ArgSet::set<int32_t>(const Token& key, const int32_t& value)
{
  setToTokenValueMap<int32_t>(_args, key, value);
}

template<>
void ArgSet::set<double>(const Token& key, const double& value)
{
  setToTokenValueMap<double>(_args, key, value);
}

template<>
void ArgSet::set<bool>(const Token& key, const bool& value)
{
  setToTokenValueMap<bool>(_args, key, value);
}

template<>
void ArgSet::set<fdk::Hash>(const Token& key, const fdk::Hash& value)
{
  setToTokenValueMap(_args, key, value.value());
}

template<>
void ArgSet::set<fdk::Vec2d>(const Token& key, const fdk::Vec2d& value)
{
  setToTokenValueMap<fdk::Vec2d>(_args, key, value);
}

template<>
void ArgSet::set<fdk::Vec3d>(const Token& key, const fdk::Vec3d& value)
{
  setToTokenValueMap<fdk::Vec3d>(_args, key, value);
}

template<>
void ArgSet::set<fdk::Vec4d>(const Token& key, const fdk::Vec4d& value)
{
  setToTokenValueMap<fdk::Vec4d>(_args, key, value);
}

template<>
void ArgSet::set<fdk::Mat4d>(const Token& key, const fdk::Mat4d& value)
{
  setToTokenValueMap<fdk::Mat4d>(_args, key, value);
}

template<>
void ArgSet::set<fdk::Box2d>(const Token& key, const fdk::Box2d& value)
{
  setToTokenValueMap<fdk::Box2d>(_args, key, value);
}

template<>
void ArgSet::set<fdk::Box3d>(const Token& key, const fdk::Box3d& value)
{
  setToTokenValueMap<fdk::Box3d>(_args, key, value);
}

#define DBL_STR_SIZE 64

void
ArgSet::setInt(const Token& key, int32_t value)
{
  setToTokenValueMap(_args, key, value);
}

void
ArgSet::setDouble(const Token& key, double value)
{
  setToTokenValueMap(_args, key, value);
}

void
ArgSet::setBool(const Token& key, bool value)
{
  setToTokenValueMap(_args, key, value);
}

void
ArgSet::setHash(const Token& key, const fdk::Hash& value)
{
  setToTokenValueMap(_args, key, value.value());
}

//
void
ArgSet::setVec2d(const Token& key, const fdk::Vec2d& v)
{
  setToTokenValueMap(_args, key, v);
}

void
ArgSet::setVec3d(const Token& key, const fdk::Vec3d& v)
{
  setToTokenValueMap(_args, key, v);
}

void
ArgSet::setVec4d(const Token& key, const fdk::Vec4d& v)
{
  setToTokenValueMap(_args, key, v);
}

void
ArgSet::setMat4d(const Token& key, const fdk::Mat4d& m)
{
  setToTokenValueMap(_args, key, m);
}

void
ArgSet::setBox2d(const Token& key, const fdk::Box2d& b)
{
  setToTokenValueMap(_args, key, b);
}

void
ArgSet::setBox3d(const Token& key, const fdk::Box3d& b)
{
  setToTokenValueMap(_args, key, b);
}

} // namespace usg
