// Copyright (c) 2022 The Foundry Visionmongers Ltd. All Rights Reserved.

#include "usg/base/Value.h"
#include "usg/base/Token.h"
#include "usg/base/ArrayTypes.h"

#include "usg/geom/Path.h"
#include "usg/geom/AssetPath.h"

#include "usg/private/UsdImpl.h"
#include "usg/private/UsdAttrImpl.h"

#include "fdk/math/Box2.h"
#include "fdk/math/Box3.h"
#include "fdk/math/Quat.h"

#include <pxr/base/gf/matrix4f.h>

namespace usg
{

//-----------------------------------------------------------------------

int usdAPIMajorVersion() { return (int)PXR_MAJOR_VERSION; }
int usdAPIMinorVersion() { return (int)PXR_MINOR_VERSION; }
int usdAPIPathVersion()  { return (int)PXR_PATCH_VERSION; }

int usdAPIVersion()      { return (int)PXR_VERSION; }

//-----------------------------------------------------------------------

std::string      empty_string;
fdk::StringList  empty_string_array;
//
usg::Token       empty_token("");
TokenArray       empty_token_array;
//
usg::Path        empty_geompath; // do not set to '', it will cause a Tf error
PathArray        empty_geompath_array;
//
usg::AssetPath   empty_assetpath("");
AssetPathArray   empty_assetpath_array;
//
bool             empty_bool(false);
BoolArray        empty_bool_array;
//
fdk::half        empty_half(0);
HalfArray        empty_half_array;
float            empty_float(0);
FloatArray       empty_float_array;
double           empty_double(0);
DoubleArray      empty_double_array;
//
char             empty_char(0);
CharArray        empty_char_array;
unsigned char    empty_uchar(0);
UCharArray       empty_uchar_array;
int32_t          empty_int32(0);
Int32Array       empty_int32_array;
uint32_t         empty_uint32(0);
Uint32Array      empty_uint32_array;
int64_t          empty_int64(0);
Int64Array       empty_int64_array;
uint64_t         empty_uint64(0);
Uint64Array      empty_uint64_array;
//
fdk::Vec2h       empty_vec2h(0,0);
Vec2hArray       empty_vec2h_array;
fdk::Vec2f       empty_vec2f(0.0f, 0.0f);
Vec2fArray       empty_vec2f_array;
fdk::Vec2d       empty_vec2d(0.0, 0.0);
Vec2dArray       empty_vec2d_array;
fdk::Vec2i       empty_vec2i(0,0);
Vec2iArray       empty_vec2i_array;
//
fdk::Vec3h       empty_vec3h(0,0,0);
Vec3hArray       empty_vec3h_array;
fdk::Vec3f       empty_vec3f(0.0f, 0.0f, 0.0f);
Vec3fArray       empty_vec3f_array;
fdk::Vec3d       empty_vec3d(0.0, 0.0, 0.0);
Vec3dArray       empty_vec3d_array;
fdk::Vec3i       empty_vec3i(0,0,0);
Vec3iArray       empty_vec3i_array;
//
fdk::Vec4h       empty_vec4h(0,0,0,0);
Vec4hArray       empty_vec4h_array;
fdk::Vec4f       empty_vec4f(0.0f, 0.0f, 0.0f, 0.0f);
Vec4fArray       empty_vec4f_array;
fdk::Vec4d       empty_vec4d(0.0, 0.0, 0.0);
Vec4dArray       empty_vec4d_array;
fdk::Vec4i       empty_vec4i(0,0,0,0);
Vec4iArray       empty_vec4i_array;
//
fdk::Mat4d       empty_mat4d(1.0);
Mat4dArray       empty_mat4d_array;
//
fdk::Quath       empty_quath;
QuathArray       empty_quath_array;
fdk::Quatf       empty_quatf;
QuatfArray       empty_quatf_array;
fdk::Quatd       empty_quatd;
QuatdArray       empty_quatd_array;
//
fdk::Box2h       empty_box2h(0,0);
Box2hArray       empty_box2h_array;
fdk::Box2f       empty_box2f(0.0f, 0.0f);
Box2fArray       empty_box2f_array;
fdk::Box2d       empty_box2d(0.0, 0.0);
Box2dArray       empty_box2d_array;
fdk::Box2i       empty_box2i(0,0);
Box2iArray       empty_box2i_array;
//
fdk::Box3h       empty_box3h(0,0,0);
Box3hArray       empty_box3h_array;
fdk::Box3f       empty_box3f(0.0f, 0.0f, 0.0f);
Box3fArray       empty_box3f_array;
fdk::Box3d       empty_box3d(0.0, 0.0, 0.0);
Box3dArray       empty_box3d_array;
fdk::Box3i       empty_box3i(0,0,0);
Box3iArray       empty_box3i_array;


//-----------------------------------------------------------------------
//-----------------------------------------------------------------------

/*! Mappings from Value::Types to Sdf types, and vice-versa.

    TODO: The number of entries here is usually the same as the Value::Type enums though they do
    not need to be in the same order. This could be made more flexible to allow dynamic adding of
    types at runtime.

    Note - we could copy the ValueType names directly from SdfValueTypeNames, but hardcoding the
    strings this way allows them to be altered (for whatever reason) while still maintaining their
    mapping to a specific SdfValueType. ie instead of Usd 'asset' we could decide to use 'path' as
    the type name instead.

    We could also extract the number of elements and element type from Usd, but it's easy enough
    to provide here explicitly.
*/
static std::vector<ValueTypeInfo> typeinfo_list
{
  //--------------------------------------------------------------------------------------------------
  //             ValueType name  ValueType index         SdfValueTypeName                         ElemDataType        nElems
  { ValueTypeInfo("default"     , Value::DefaultType     , PXR_NS::SdfValueTypeName()                , Value::STRING_DATA , 0) },
  //--------------------------------------------------------------------------------------------
  { ValueTypeInfo("string"      , Value::String          , PXR_NS::SdfValueTypeNames->String         , Value::STRING_DATA , 1) },
  { ValueTypeInfo("string[]"    , Value::StringArray     , PXR_NS::SdfValueTypeNames->StringArray    , Value::STRING_DATA , 1) },
  { ValueTypeInfo("token"       , Value::Token           , PXR_NS::SdfValueTypeNames->Token          , Value::TOKEN_DATA  , 1) },
  { ValueTypeInfo("token[]"     , Value::TokenArray      , PXR_NS::SdfValueTypeNames->TokenArray     , Value::TOKEN_DATA  , 1) },
  { ValueTypeInfo("asset"       , Value::AssetPath       , PXR_NS::SdfValueTypeNames->Asset          , Value::ASSETPATH_DATA , 1) },
  { ValueTypeInfo("asset[]"     , Value::AssetPathArray  , PXR_NS::SdfValueTypeNames->AssetArray     , Value::ASSETPATH_DATA , 1) },
  //--------------------------------------------------------------------------------------------
  { ValueTypeInfo("bool"        , Value::Bool            , PXR_NS::SdfValueTypeNames->Bool           , Value::INT8_DATA   , 1) },
  { ValueTypeInfo("bool[]"      , Value::BoolArray       , PXR_NS::SdfValueTypeNames->BoolArray      , Value::INT8_DATA   , 1) },
  { ValueTypeInfo("int"         , Value::Int             , PXR_NS::SdfValueTypeNames->Int            , Value::INT32_DATA  , 1) },
  { ValueTypeInfo("int[]"       , Value::IntArray        , PXR_NS::SdfValueTypeNames->IntArray       , Value::INT32_DATA  , 1) },
  { ValueTypeInfo("int64"       , Value::Int64           , PXR_NS::SdfValueTypeNames->Int64          , Value::INT64_DATA  , 1) },
  { ValueTypeInfo("int64[]"     , Value::Int64Array      , PXR_NS::SdfValueTypeNames->Int64Array     , Value::INT64_DATA  , 1) },
  { ValueTypeInfo("uchar"       , Value::UChar           , PXR_NS::SdfValueTypeNames->UChar          , Value::INT8_DATA   , 1) },
  { ValueTypeInfo("uchar[]"     , Value::UCharArray      , PXR_NS::SdfValueTypeNames->UCharArray     , Value::INT8_DATA   , 1) },
  { ValueTypeInfo("uint"        , Value::UInt            , PXR_NS::SdfValueTypeNames->UInt           , Value::INT32_DATA  , 1) },
  { ValueTypeInfo("uint[]"      , Value::UIntArray       , PXR_NS::SdfValueTypeNames->UIntArray      , Value::INT32_DATA  , 1) },
  { ValueTypeInfo("uint64"      , Value::UInt64          , PXR_NS::SdfValueTypeNames->UInt64         , Value::INT64_DATA  , 1) },
  { ValueTypeInfo("uint64[]"    , Value::UInt64Array     , PXR_NS::SdfValueTypeNames->UInt64Array    , Value::INT64_DATA  , 1) },
  { ValueTypeInfo("half"        , Value::Half            , PXR_NS::SdfValueTypeNames->Half           , Value::HALF_DATA   , 1) },
  { ValueTypeInfo("half[]"      , Value::HalfArray       , PXR_NS::SdfValueTypeNames->HalfArray      , Value::HALF_DATA   , 1) },
  { ValueTypeInfo("float"       , Value::Float           , PXR_NS::SdfValueTypeNames->Float          , Value::FLOAT_DATA  , 1) },
  { ValueTypeInfo("float[]"     , Value::FloatArray      , PXR_NS::SdfValueTypeNames->FloatArray     , Value::FLOAT_DATA  , 1) },
  { ValueTypeInfo("double"      , Value::Double          , PXR_NS::SdfValueTypeNames->Double         , Value::DOUBLE_DATA , 1) },
  { ValueTypeInfo("double[]"    , Value::DoubleArray     , PXR_NS::SdfValueTypeNames->DoubleArray    , Value::DOUBLE_DATA , 1) },
  { ValueTypeInfo("timecode"    , Value::TimeCode        , PXR_NS::SdfValueTypeNames->TimeCode       , Value::DOUBLE_DATA , 1) },
  { ValueTypeInfo("timecode[]"  , Value::TimeCodeArray   , PXR_NS::SdfValueTypeNames->TimeCodeArray  , Value::DOUBLE_DATA , 1) },
  //--------------------------------------------------------------------------------------------
  { ValueTypeInfo("int2"        , Value::Int2            , PXR_NS::SdfValueTypeNames->Int2           , Value::INT32_DATA  , 2) },
  { ValueTypeInfo("int2[]"      , Value::Int2Array       , PXR_NS::SdfValueTypeNames->Int2Array      , Value::INT32_DATA  , 2) },
  { ValueTypeInfo("half2"       , Value::Half2           , PXR_NS::SdfValueTypeNames->Half2          , Value::HALF_DATA   , 2) },
  { ValueTypeInfo("half2[]"     , Value::Half2Array      , PXR_NS::SdfValueTypeNames->Half2Array     , Value::HALF_DATA   , 2) },
  { ValueTypeInfo("float2"      , Value::Float2          , PXR_NS::SdfValueTypeNames->Float2         , Value::FLOAT_DATA  , 2) },
  { ValueTypeInfo("float2[]"    , Value::Float2Array     , PXR_NS::SdfValueTypeNames->Float2Array    , Value::FLOAT_DATA  , 2) },
  { ValueTypeInfo("double2"     , Value::Double2         , PXR_NS::SdfValueTypeNames->Double2        , Value::DOUBLE_DATA , 2) },
  { ValueTypeInfo("double2[]"   , Value::Double2Array    , PXR_NS::SdfValueTypeNames->Double2Array   , Value::DOUBLE_DATA , 2) },
  //--------------------------------------------------------------------------------------------
  { ValueTypeInfo("int3"        , Value::Int3            , PXR_NS::SdfValueTypeNames->Int3           , Value::INT32_DATA  , 3) },
  { ValueTypeInfo("int3[]"      , Value::Int3Array       , PXR_NS::SdfValueTypeNames->Int3Array      , Value::INT32_DATA  , 3) },
  { ValueTypeInfo("half3"       , Value::Half3           , PXR_NS::SdfValueTypeNames->Half3          , Value::HALF_DATA   , 3) },
  { ValueTypeInfo("half3[]"     , Value::Half3Array      , PXR_NS::SdfValueTypeNames->Half3Array     , Value::HALF_DATA   , 3) },
  { ValueTypeInfo("float3"      , Value::Float3          , PXR_NS::SdfValueTypeNames->Float3         , Value::FLOAT_DATA  , 3) },
  { ValueTypeInfo("float3[]"    , Value::Float3Array     , PXR_NS::SdfValueTypeNames->Float3Array    , Value::FLOAT_DATA  , 3) },
  { ValueTypeInfo("double3"     , Value::Double3         , PXR_NS::SdfValueTypeNames->Double3        , Value::DOUBLE_DATA , 3) },
  { ValueTypeInfo("double3[]"   , Value::Double3Array    , PXR_NS::SdfValueTypeNames->Double3Array   , Value::DOUBLE_DATA , 3) },
  //--------------------------------------------------------------------------------------------
  { ValueTypeInfo("int4"        , Value::Int4            , PXR_NS::SdfValueTypeNames->Int4           , Value::INT32_DATA  , 4) },
  { ValueTypeInfo("int4[]"      , Value::Int4Array       , PXR_NS::SdfValueTypeNames->Int4Array      , Value::INT32_DATA  , 4) },
  { ValueTypeInfo("half4"       , Value::Half4           , PXR_NS::SdfValueTypeNames->Half4          , Value::HALF_DATA   , 4) },
  { ValueTypeInfo("half4[]"     , Value::Half4Array      , PXR_NS::SdfValueTypeNames->Half4Array     , Value::HALF_DATA   , 4) },
  { ValueTypeInfo("float4"      , Value::Float4          , PXR_NS::SdfValueTypeNames->Float4         , Value::FLOAT_DATA  , 4) },
  { ValueTypeInfo("float4[]"    , Value::Float4Array     , PXR_NS::SdfValueTypeNames->Float4Array    , Value::FLOAT_DATA  , 4) },
  { ValueTypeInfo("double4"     , Value::Double4         , PXR_NS::SdfValueTypeNames->Double4        , Value::DOUBLE_DATA , 4) },
  { ValueTypeInfo("double4[]"   , Value::Double4Array    , PXR_NS::SdfValueTypeNames->Double4Array   , Value::DOUBLE_DATA , 4) },
  //--------------------------------------------------------------------------------------------
  { ValueTypeInfo("matrix2d"    , Value::Matrix2d        , PXR_NS::SdfValueTypeNames->Matrix2d       , Value::DOUBLE_DATA , 4) },
  { ValueTypeInfo("matrix2d[]"  , Value::Matrix2dArray   , PXR_NS::SdfValueTypeNames->Matrix2dArray  , Value::DOUBLE_DATA , 4) },
  { ValueTypeInfo("matrix3d"    , Value::Matrix3d        , PXR_NS::SdfValueTypeNames->Matrix3d       , Value::DOUBLE_DATA , 9) },
  { ValueTypeInfo("matrix3d[]"  , Value::Matrix3dArray   , PXR_NS::SdfValueTypeNames->Matrix3dArray  , Value::DOUBLE_DATA , 9) },
  { ValueTypeInfo("matrix4d"    , Value::Matrix4d        , PXR_NS::SdfValueTypeNames->Matrix4d       , Value::DOUBLE_DATA ,16) },
  { ValueTypeInfo("matrix4d[]"  , Value::Matrix4dArray   , PXR_NS::SdfValueTypeNames->Matrix4dArray  , Value::DOUBLE_DATA ,16) },
  //--------------------------------------------------------------------------------------------
  // Aliases / interpretations:
  //--------------------------------------------------------------------------------------------
  { ValueTypeInfo("channel"     , Value::Channel         , PXR_NS::SdfValueTypeNames->Int            , Value::INT32_DATA  , 1) },
  { ValueTypeInfo("channel[]"   , Value::ChannelArray    , PXR_NS::SdfValueTypeNames->IntArray       , Value::INT32_DATA  , 1) },
  //--------------------------------------------------------------------------------------------
  { ValueTypeInfo("color"       , Value::Color           , PXR_NS::SdfValueTypeNames->Color3f        , Value::FLOAT_DATA  , 3) },
  { ValueTypeInfo("color[]"     , Value::ColorArray      , PXR_NS::SdfValueTypeNames->Color3fArray   , Value::FLOAT_DATA  , 3) },
  { ValueTypeInfo("color3h"     , Value::Color3h         , PXR_NS::SdfValueTypeNames->Color3h        , Value::HALF_DATA   , 3) },
  { ValueTypeInfo("color3h[]"   , Value::Color3hArray    , PXR_NS::SdfValueTypeNames->Color3hArray   , Value::HALF_DATA   , 3) },
  { ValueTypeInfo("color3f"     , Value::Color3f         , PXR_NS::SdfValueTypeNames->Color3f        , Value::FLOAT_DATA  , 3) },
  { ValueTypeInfo("color3f[]"   , Value::Color3fArray    , PXR_NS::SdfValueTypeNames->Color3fArray   , Value::FLOAT_DATA  , 3) },
  { ValueTypeInfo("color3d"     , Value::Color3d         , PXR_NS::SdfValueTypeNames->Color3d        , Value::DOUBLE_DATA , 3) },
  { ValueTypeInfo("color3d[]"   , Value::Color3dArray    , PXR_NS::SdfValueTypeNames->Color3dArray   , Value::DOUBLE_DATA , 3) },
  //--------------------------------------------------------------------------------------------
  { ValueTypeInfo("normal"      , Value::Normal          , PXR_NS::SdfValueTypeNames->Normal3f       , Value::FLOAT_DATA  , 3) },
  { ValueTypeInfo("normal[]"    , Value::NormalArray     , PXR_NS::SdfValueTypeNames->Normal3fArray  , Value::FLOAT_DATA  , 3) },
  { ValueTypeInfo("normal3h"    , Value::Normal3h        , PXR_NS::SdfValueTypeNames->Normal3h       , Value::HALF_DATA   , 3) },
  { ValueTypeInfo("normal3h[]"  , Value::Normal3hArray   , PXR_NS::SdfValueTypeNames->Normal3hArray  , Value::HALF_DATA   , 3) },
  { ValueTypeInfo("normal3f"    , Value::Normal3f        , PXR_NS::SdfValueTypeNames->Normal3f       , Value::FLOAT_DATA  , 3) },
  { ValueTypeInfo("normal3f[]"  , Value::Normal3fArray   , PXR_NS::SdfValueTypeNames->Normal3fArray  , Value::FLOAT_DATA  , 3) },
  { ValueTypeInfo("normal3d"    , Value::Normal3d        , PXR_NS::SdfValueTypeNames->Normal3d       , Value::DOUBLE_DATA , 3) },
  { ValueTypeInfo("normal3d[]"  , Value::Normal3dArray   , PXR_NS::SdfValueTypeNames->Normal3dArray  , Value::DOUBLE_DATA , 3) },
  //--------------------------------------------------------------------------------------------
  { ValueTypeInfo("point"       , Value::Point           , PXR_NS::SdfValueTypeNames->Point3f        , Value::FLOAT_DATA  , 3) },
  { ValueTypeInfo("point[]"     , Value::PointArray      , PXR_NS::SdfValueTypeNames->Point3fArray   , Value::FLOAT_DATA  , 3) },
  { ValueTypeInfo("point3h"     , Value::Point3h         , PXR_NS::SdfValueTypeNames->Point3h        , Value::HALF_DATA   , 3) },
  { ValueTypeInfo("point3h[]"   , Value::Point3hArray    , PXR_NS::SdfValueTypeNames->Point3hArray   , Value::HALF_DATA   , 3) },
  { ValueTypeInfo("point3f"     , Value::Point3f         , PXR_NS::SdfValueTypeNames->Point3f        , Value::FLOAT_DATA  , 3) },
  { ValueTypeInfo("point3f[]"   , Value::Point3fArray    , PXR_NS::SdfValueTypeNames->Point3fArray   , Value::FLOAT_DATA  , 3) },
  { ValueTypeInfo("point3d"     , Value::Point3d         , PXR_NS::SdfValueTypeNames->Point3d        , Value::DOUBLE_DATA , 3) },
  { ValueTypeInfo("point3d[]"   , Value::Point3dArray    , PXR_NS::SdfValueTypeNames->Point3dArray   , Value::DOUBLE_DATA , 3) },
  //--------------------------------------------------------------------------------------------
  { ValueTypeInfo("vector"      , Value::Vector          , PXR_NS::SdfValueTypeNames->Vector3f       , Value::FLOAT_DATA  , 3) },
  { ValueTypeInfo("vector[]"    , Value::VectorArray     , PXR_NS::SdfValueTypeNames->Vector3fArray  , Value::FLOAT_DATA  , 3) },
  { ValueTypeInfo("vector3h"    , Value::Vector3h        , PXR_NS::SdfValueTypeNames->Vector3h       , Value::HALF_DATA   , 3) },
  { ValueTypeInfo("vector3h[]"  , Value::Vector3hArray   , PXR_NS::SdfValueTypeNames->Vector3hArray  , Value::HALF_DATA   , 3) },
  { ValueTypeInfo("vector3f"    , Value::Vector3f        , PXR_NS::SdfValueTypeNames->Vector3f       , Value::FLOAT_DATA  , 3) },
  { ValueTypeInfo("vector3f[]"  , Value::Vector3fArray   , PXR_NS::SdfValueTypeNames->Vector3fArray  , Value::FLOAT_DATA  , 3) },
  { ValueTypeInfo("vector3d"    , Value::Vector3d        , PXR_NS::SdfValueTypeNames->Vector3d       , Value::DOUBLE_DATA , 3) },
  { ValueTypeInfo("vector3d[]"  , Value::Vector3dArray   , PXR_NS::SdfValueTypeNames->Vector3dArray  , Value::DOUBLE_DATA , 3) },
  //--------------------------------------------------------------------------------------------
  { ValueTypeInfo("color4h"     , Value::Color4h         , PXR_NS::SdfValueTypeNames->Color4h        , Value::HALF_DATA   , 4) },
  { ValueTypeInfo("color4h[]"   , Value::Color4hArray    , PXR_NS::SdfValueTypeNames->Color4hArray   , Value::HALF_DATA   , 4) },
  { ValueTypeInfo("color4f"     , Value::Color4f         , PXR_NS::SdfValueTypeNames->Color4f        , Value::FLOAT_DATA  , 4) },
  { ValueTypeInfo("color4f[]"   , Value::Color4fArray    , PXR_NS::SdfValueTypeNames->Color4fArray   , Value::FLOAT_DATA  , 4) },
  { ValueTypeInfo("color4d"     , Value::Color4d         , PXR_NS::SdfValueTypeNames->Color4d        , Value::DOUBLE_DATA , 4) },
  { ValueTypeInfo("color4d[]"   , Value::Color4dArray    , PXR_NS::SdfValueTypeNames->Color4dArray   , Value::DOUBLE_DATA , 4) },
  //--------------------------------------------------------------------------------------------
  { ValueTypeInfo("texCoord2h"  , Value::TexCoord2h      , PXR_NS::SdfValueTypeNames->TexCoord2h     , Value::HALF_DATA   , 2) },
  { ValueTypeInfo("texCoord2h[]", Value::TexCoord2hArray , PXR_NS::SdfValueTypeNames->TexCoord2hArray, Value::HALF_DATA   , 2) },
  { ValueTypeInfo("texCoord2f"  , Value::TexCoord2f      , PXR_NS::SdfValueTypeNames->TexCoord2f     , Value::FLOAT_DATA  , 2) },
  { ValueTypeInfo("texCoord2f[]", Value::TexCoord2fArray , PXR_NS::SdfValueTypeNames->TexCoord2fArray, Value::FLOAT_DATA  , 2) },
  { ValueTypeInfo("texCoord2d"  , Value::TexCoord2d      , PXR_NS::SdfValueTypeNames->TexCoord2d     , Value::DOUBLE_DATA , 2) },
  { ValueTypeInfo("texCoord2d[]", Value::TexCoord2dArray , PXR_NS::SdfValueTypeNames->TexCoord2dArray, Value::DOUBLE_DATA , 2) },
  { ValueTypeInfo("texCoord3h"  , Value::TexCoord3h      , PXR_NS::SdfValueTypeNames->TexCoord3h     , Value::HALF_DATA   , 3) },
  { ValueTypeInfo("texCoord3h[]", Value::TexCoord3hArray , PXR_NS::SdfValueTypeNames->TexCoord3hArray, Value::HALF_DATA   , 3) },
  { ValueTypeInfo("texCoord3f"  , Value::TexCoord3f      , PXR_NS::SdfValueTypeNames->TexCoord3f     , Value::FLOAT_DATA  , 3) },
  { ValueTypeInfo("texCoord3f[]", Value::TexCoord3fArray , PXR_NS::SdfValueTypeNames->TexCoord3fArray, Value::FLOAT_DATA  , 3) },
  { ValueTypeInfo("texCoord3d"  , Value::TexCoord3d      , PXR_NS::SdfValueTypeNames->TexCoord3d     , Value::DOUBLE_DATA , 3) },
  { ValueTypeInfo("texCoord3d[]", Value::TexCoord3dArray , PXR_NS::SdfValueTypeNames->TexCoord3dArray, Value::DOUBLE_DATA , 3) },
  //--------------------------------------------------------------------------------------------
  { ValueTypeInfo("matrix"      , Value::Matrix          , PXR_NS::SdfValueTypeNames->Matrix4d       , Value::DOUBLE_DATA ,16) },
  { ValueTypeInfo("matrix[]"    , Value::MatrixArray     , PXR_NS::SdfValueTypeNames->Matrix4dArray  , Value::DOUBLE_DATA ,16) },
  { ValueTypeInfo("frame4d"     , Value::Frame4d         , PXR_NS::SdfValueTypeNames->Frame4d        , Value::DOUBLE_DATA ,16) },
  { ValueTypeInfo("frame4d[]"   , Value::Frame4dArray    , PXR_NS::SdfValueTypeNames->Frame4dArray   , Value::DOUBLE_DATA ,16) },
  //--------------------------------------------------------------------------------------------
  { ValueTypeInfo("quath"       , Value::Quath           , PXR_NS::SdfValueTypeNames->Quath          , Value::HALF_DATA   , 4) },
  { ValueTypeInfo("quath[]"     , Value::QuathArray      , PXR_NS::SdfValueTypeNames->QuathArray     , Value::HALF_DATA   , 4) },
  { ValueTypeInfo("quatf"       , Value::Quatf           , PXR_NS::SdfValueTypeNames->Quatf          , Value::FLOAT_DATA  , 4) },
  { ValueTypeInfo("quatf[]"     , Value::QuatfArray      , PXR_NS::SdfValueTypeNames->QuatfArray     , Value::FLOAT_DATA  , 4) },
  { ValueTypeInfo("quatd"       , Value::Quatd           , PXR_NS::SdfValueTypeNames->Quatd          , Value::DOUBLE_DATA , 4) },
  { ValueTypeInfo("quatd[]"     , Value::QuatdArray      , PXR_NS::SdfValueTypeNames->QuatdArray     , Value::DOUBLE_DATA , 4) },
  //--------------------------------------------------------------------------------------------
  //--------------------------------------------------------------------------------------------
  // Additional abstract type aliases used in SdrShaderProperties and UsdShade:
//===========================================
// TODO: these need to be better defined!
  { ValueTypeInfo("struct"      , Value::Struct          , PXR_NS::SdfValueTypeNames->Token          , Value::TOKEN_DATA  , 1) },
  { ValueTypeInfo("struct[]"    , Value::StructArray     , PXR_NS::SdfValueTypeNames->TokenArray     , Value::TOKEN_DATA  , 1) },
  { ValueTypeInfo("terminal"    , Value::Terminal        , PXR_NS::SdfValueTypeNames->Token          , Value::TOKEN_DATA  , 1) },
  { ValueTypeInfo("terminal[]"  , Value::TerminalArray   , PXR_NS::SdfValueTypeNames->TokenArray     , Value::TOKEN_DATA  , 1) },
  { ValueTypeInfo("vstruct"     , Value::Vstruct         , PXR_NS::SdfValueTypeNames->Token          , Value::TOKEN_DATA  , 1) },
  { ValueTypeInfo("vstruct[]"   , Value::VstructArray    , PXR_NS::SdfValueTypeNames->TokenArray     , Value::TOKEN_DATA  , 1) },
//===========================================
  //--------------------------------------------------------------------------------------------
  //--------------------------------------------------------------------------------------------
  { ValueTypeInfo("invalid"     , Value::InvalidType     , PXR_NS::SdfValueTypeName()                , Value::STRING_DATA , 0) },
  { ValueTypeInfo("unknown"     , Value::InvalidType     , PXR_NS::SdfValueTypeName()                , Value::STRING_DATA , 0) },
  //--------------------------------------------------------------------------------------------
  //--------------------------------------------------------------------------------------------
};


/*! Build static maps of Attribute types to Usd equivalents and vice-versa.

    Provides fast map lookups rather than a series if/then comparisons.

    For usd-to-attr mapping we use the hash value from SdfValueTypeName::GetHash() or
    TfType::GetTypeid().hash_code() rather than the string name for increased lookup speed.
*/
using StringToValueTypeMap = std::unordered_map<std::string, Value::Type>;
using HashToValueTypeMap   = std::unordered_map<size_t,      Value::Type>;


//using TfTypeToValueTypeMap = std::map<PXR_NS::TfType, Value::Type>; // unordered fails due to lack of type hashable support
// TODO: provide a method to hash TfType?


static std::vector<int>     type_to_typeinfo_list;        // Value::Type enum to typeinfo_list entry index
static StringToValueTypeMap typename_to_usgtype_map;      // Reverse lookup from ValueTypeInfo::name

static HashToValueTypeMap   usd_sdftype_to_usgtype_map;
static HashToValueTypeMap   usd_tftype_to_usgtype_map;


/*! TODO: this could become a singleton so that the maps are created at runtime
    rather than lib load.
*/
struct BuildTypeMaps
{
  BuildTypeMaps()
  {
    type_to_typeinfo_list.resize(Value::NumValueTypes+1, -1);

    for (size_t i=0; i < typeinfo_list.size(); ++i) {
      const ValueTypeInfo& info = typeinfo_list[i];

      // There should be one ValueTypeInfo entry per Value::Type enum, so if there's
      // any extra or they're out of range they'll be ignored:
      if (info.usg_value_type < 0 ||
          info.usg_value_type >= type_to_typeinfo_list.size() ||
          type_to_typeinfo_list[info.usg_value_type] != -1) {
        continue;
      }

      type_to_typeinfo_list[info.usg_value_type] = i;

      typename_to_usgtype_map[info.name.GetString()] = info.usg_value_type;

      // SdfValueTypeName to attr type - pick first 'base' mapping and skip later type aliases with alternate
      // usage interpretations.
      // Use SdfValueTypeName's hash value rather than string name for lookup speed:
      {
        const HashToValueTypeMap::const_iterator it = usd_sdftype_to_usgtype_map.find(info.sdf_value_typename.GetHash());

        if (it == usd_sdftype_to_usgtype_map.end()) {
          usd_sdftype_to_usgtype_map[info.sdf_value_typename.GetHash()] = info.usg_value_type;
        }
      }

      // TfType to attr type - pick first 'base' mapping and skip later type aliases with alternate
      // usage interpretations:
      {
        const HashToValueTypeMap::const_iterator it = usd_tftype_to_usgtype_map.find(info.tf_data_type.GetTypeid().hash_code());

        // TODO: this is bit of a hack - skip later Sdf matches as those are types with interpretations
        // overlaying the data type, like 'Normal3', 'Color3' or 'Point3'.
        // Maybe we should have a more explicit TfType->Value::Type mapping matrix, but atm I'm
        // not sure how to get the TfType list since it's a runtime thing...
        if (it == usd_tftype_to_usgtype_map.end()) {
          usd_tftype_to_usgtype_map[info.tf_data_type.GetTypeid().hash_code()] = info.usg_value_type;
        }
      }

    }

#if 0
for (size_t i=0; i < typeinfo_list.size(); ++i) {
  ValueTypeInfo& info = typeinfo_list[i];
  std::cout << " usg='" << info.name << "', usg_base_type='" << Value::getTypeName(info.usg_base_value_type) << "'";
  std::cout << ", sdf='" << info.sdf_value_typename << "'";
  std::cout << ", sdfcpp='" << info.sdf_value_typename.GetCPPTypeName() << "'";
  std::cout << ", sdfvaluetype=0x" << std::hex << info.sdf_value_typename.GetType().GetTypeid().hash_code() << std::dec;
  std::cout << ", tftype=0x" << std::hex << info.tf_data_type.GetTypeid().hash_code() << std::dec;
  std::cout << " " << info << std::endl;
}

std::cout << " usg::Value for String='" << Value::getTypeName(getUsgValueType(PXR_NS::SdfValueTypeNames->String)) << "'" << std::endl;
std::cout << " usg::Value for TfToken='" << Value::getTypeName(getUsgValueType(PXR_NS::SdfValueTypeNames->Token)) << "'" << std::endl;
std::cout << " usg::Value for SdfAssetPath='" << Value::getTypeName(getUsgValueType(PXR_NS::SdfValueTypeNames->Asset)) << "'" << std::endl;
std::cout << " usg::Value for Int='" << Value::getTypeName(getUsgValueType(PXR_NS::SdfValueTypeNames->Int)) << "'" << std::endl;
#endif

  }
} type_maps_builder;


//-----------------------------------------------------------------------


inline bool isValidDataType(Value::Type type) { return !(type < Value::String || type >= Value::InvalidType); }


const usg::ValueTypeInfo&
getValueTypeInfo(Value::Type type)
{
  return (isValidDataType(type)) ?
            typeinfo_list[type_to_typeinfo_list[type]] :
            typeinfo_list[Value::InvalidType];
}

Value::Type      getBaseType(Value::Type type)     { return usg::getValueTypeInfo(type).usg_base_value_type; }
Value::DataType  getBaseDataType(Value::Type type) { return usg::getValueTypeInfo(type).usg_base_data_type; }

/*! Get the corresponding Value::Type for a SdfValueTypeName.

    Sdf types are different than TfTypes which are storage types.
*/
Value::Type
getUsgValueType(const PXR_NS::SdfValueTypeName& sdf_typename)
{
  const HashToValueTypeMap::const_iterator it = usd_sdftype_to_usgtype_map.find(sdf_typename.GetHash());
  assert(it != usd_sdftype_to_usgtype_map.end()); // shouldn't happen... If it does then we're missing some types!

  //std::cout << "              getValueType() SdfValueTypeName='" << sdf_typename << "'";
  //std::cout << ", valuetype='" << Value::getTypeName(it->second) << "'" << std::endl;

  return (it != usd_sdftype_to_usgtype_map.end()) ? it->second : Value::InvalidType;
}


/*! Get the corresponding Value::Type for a TfType.
*/
Value::Type
getUsgValueType(const PXR_NS::TfType& tftype)
{
  const HashToValueTypeMap::const_iterator it = usd_tftype_to_usgtype_map.find(tftype.GetTypeid().hash_code());

  //std::cout << "    getValueType() TfType=0x" << std::hex << tftype.GetTypeid().hash_code() << std::dec;
  //std::cout << ", name='" << tftype.GetTypeName() << "'";
  //std::cout << ", valuetype='" << Value::getTypeName((it != usd_tftype_to_usgtype_map.end()) ? it->second : Value::InvalidType) << "'" << std::endl;

  return (it != usd_tftype_to_usgtype_map.end()) ? it->second : Value::InvalidType;
}


/*! Utility to get Value::Type name from enum, as a TfToken.
*/
const PXR_NS::TfToken&
getUsgTypeNameAsToken(Value::Type type)
{
  return (isValidDataType(type)) ?
            usg::getValueTypeInfo(type).name :
            usg::getValueTypeInfo(Value::InvalidType).name;
}


/*! Get the corresponding SdfValueTypeName for a Value::Type.
*/
const PXR_NS::SdfValueTypeName&
getSdfValueTypeName(Value::Type type)
{
  return (isValidDataType(type)) ?
            usg::getValueTypeInfo(type).sdf_value_typename :
            usg::getValueTypeInfo(Value::InvalidType).sdf_value_typename;
}


//-----------------------------------------------------------------------
//-----------------------------------------------------------------------


/*! Default ctor is invalid.
*/
Value::Value() :
  _pImpl()
{
  //
}


Value::Value(const Value& b) :
  _pImpl((b._pImpl.get() != nullptr) ? new Value::Impl(*b._pImpl.get()) : nullptr)
{
  //
}

Value& Value::operator = (const Value& b)
{
  if (_pImpl.get() != b._pImpl.get()) {
    _pImpl.reset(new Value::Impl(*b._pImpl.get()));
  }
  return *this;
}


Value::Value(const Value::Impl& b) :
  _pImpl(new Value::Impl(b))
{
  //
}
Value& Value::operator = (const Value::Impl& b)
{
  if (_pImpl.get() != &b) {
    _pImpl.reset(new Value::Impl(b));
  }
  return *this;
}


bool
Value::setFromImpl(Value::Impl&& b)
{
  _pImpl.reset(new Value::Impl(std::move(b)));
  return _pImpl->isValid();
}


Value::~Value()
{
  //
}

bool               Value::isValid()      const { return (_pImpl && _pImpl->isValid()); }
bool               Value::isArray()      const { return (_pImpl && _pImpl->vtvalue.IsArrayValued()); }
size_t             Value::getArraySize() const { return (_pImpl) ? _pImpl->vtvalue.GetArraySize() : 0; }
size_t             Value::getHash()      const { return (_pImpl) ? _pImpl->vtvalue.GetHash() : 0; }
Value::Type        Value::getType()      const { return (_pImpl) ? _pImpl->valuetype : Value::InvalidType; }
const std::string& Value::getTypeName()  const { return Value::getTypeName((_pImpl) ? _pImpl->valuetype : Value::InvalidType); }

//std::type_info Value::GetTypeid()        const;
//std::type_info Value::GetElementTypeid() const;

Value::Handle* Value::getVtValue(int usd_version)
{
  return (_pImpl && usdVersionMatchesUsg(usd_version)) ?
            reinterpret_cast<Value::Handle*>(&_pImpl->vtvalue) :
            nullptr;
}

const Value::Handle* Value::getVtValue(int usd_version) const
{
  return (_pImpl && usdVersionMatchesUsg(usd_version)) ?
            reinterpret_cast<const Value::Handle*>(&_pImpl->vtvalue) :
            nullptr;
}

std::ostream& operator << (std::ostream& o, const Value& val)
{
  if (val._pImpl) {
#if 1
    o << val._pImpl->vtvalue;
#else
    const ValueTypeInfo& info = usg::getValueTypeInfo(val._pImpl->valuetype);
    o << info;
    if (val.isArray()) {
      o << " size=" << val.getArraySize();
    }
#endif
  }
  else {
    o << "<invalid>";
  }
  return o;
}

bool operator == (const usg::Value& lhs, const usg::Value& rhs)
{
  if (lhs._pImpl && rhs._pImpl)
  {
    return lhs._pImpl->vtvalue == rhs._pImpl->vtvalue;
  }
  else
  {
    return lhs._pImpl == rhs._pImpl;
  }
}



//--------------------------------------------

/*static*/ Value::Type      Value::getBaseType(Value::Type type)     { return usg::getBaseType(type); }
/*static*/ Value::DataType  Value::getBaseDataType(Value::Type type) { return usg::getBaseDataType(type); }
/*static*/ size_t           Value::getBytesPerDataElement(Type type) { return usg::getValueTypeInfo(type).bytes_per_data_element; }
/*static*/ size_t           Value::getNumDataElements(Type type)     { return usg::getValueTypeInfo(type).num_data_elements; }

/*static*/ const std::string&
Value::getTypeName(Type type)
{
  return (isValidDataType(type)) ? usg::getValueTypeInfo(type).name.GetString() : usg::getValueTypeInfo(Value::InvalidType).name.GetString();
}

/*! Utility to get ValueType enum from string type name.

    Type name may be arbitrary and not guaranteed to match the pre-existing types. In that
    case additional logic may be applied to map them to declared types.
*/
/*static*/ Value::Type
Value::getTypeFromName(const std::string& type_name)
{
  const StringToValueTypeMap::const_iterator it = typename_to_usgtype_map.find(type_name);
  if (it != typename_to_usgtype_map.end()) {
    return it->second; // match!
  }


  //-----------------------------------------------------------------------------------------
  // TODO: If there's no exact match then we may want consider adding add'l comparison logic
  // here similar to what's in SdrShaderProperty::GetTypeAsSdfType().
  //
  // I suppose there should be a registration mechanism so that shader schema plugins can
  // declare new type names like Usd supports. I think that may be easy to do since it's
  // just adding to the 'typeinfo_list' vector and adding new map entries. 'typeinfo_list' would
  // need to become a singleton var in BuildTypeMaps.

  //if (type_name == "myTerminal")
  //  return Value::Terminal

  //-----------------------------------------------------------------------------------------


  return Value::InvalidType; // no match
}

//--------------------------------------------

// ***************************************************************************************
//
// TODO: we're only supporting a subset of possible types enumerated in Value::Type!!!
//   So anytime a type that's not explicitly handled here is passed to a ctor causes
//   a linker error.
//
//   How to make this cleaner...?
//
// Casting a VtValue to a specific Sdf type: PXR_NS::VtValue::Cast<PXR_NS::VtVec2fArray>(vtvalue)
//
// ***************************************************************************************


Value::Value(const char* data) : _pImpl(new Value::Impl(PXR_NS::VtValue(std::string(data)))) {}

template <> USG_API Value::Value(const std::string   & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(data))) {}
template <> USG_API Value::Value(const usg::Token    & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(toTfToken(data)))) {}
template <> USG_API Value::Value(const usg::Path     & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(toSdfPath(data)))) {}
template <> USG_API Value::Value(const usg::AssetPath& data) : _pImpl(new Value::Impl(PXR_NS::VtValue(toSdfAssetPath(data)))) {}
//
template <> USG_API Value::Value(const bool       & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(data))) {}
template <> USG_API Value::Value(const char       & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(data))) {}
template <> USG_API Value::Value(const unsigned char& data) : _pImpl(new Value::Impl(PXR_NS::VtValue(data))) {}
template <> USG_API Value::Value(const fdk::half  & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(reinterpret_cast<const PXR_NS::GfHalf&>(data)))) {}
template <> USG_API Value::Value(const float      & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(data))) {}
template <> USG_API Value::Value(const double     & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(data))) {}
template <> USG_API Value::Value(const int32_t    & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(data))) {}
template <> USG_API Value::Value(const int64_t    & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(data))) {}
template <> USG_API Value::Value(const uint32_t   & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(data))) {}
template <> USG_API Value::Value(const uint64_t   & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(data))) {}
//
template <> USG_API Value::Value(const fdk::Vec2h & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(reinterpret_cast<const PXR_NS::GfVec2h&>(data)))) {}
template <> USG_API Value::Value(const fdk::Vec2f & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(reinterpret_cast<const PXR_NS::GfVec2f&>(data)))) {}
template <> USG_API Value::Value(const fdk::Vec2d & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(reinterpret_cast<const PXR_NS::GfVec2d&>(data)))) {}
template <> USG_API Value::Value(const fdk::Vec2i & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(reinterpret_cast<const PXR_NS::GfVec2i&>(data)))) {}
//
template <> USG_API Value::Value(const fdk::Vec3h & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(reinterpret_cast<const PXR_NS::GfVec3h&>(data)))) {}
template <> USG_API Value::Value(const fdk::Vec3f & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(reinterpret_cast<const PXR_NS::GfVec3f&>(data)))) {}
template <> USG_API Value::Value(const fdk::Vec3d & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(reinterpret_cast<const PXR_NS::GfVec3d&>(data)))) {}
template <> USG_API Value::Value(const fdk::Vec3i & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(reinterpret_cast<const PXR_NS::GfVec3i&>(data)))) {}
//
template <> USG_API Value::Value(const fdk::Vec4h & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(reinterpret_cast<const PXR_NS::GfVec4h&>(data)))) {}
template <> USG_API Value::Value(const fdk::Vec4f & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(reinterpret_cast<const PXR_NS::GfVec4f&>(data)))) {}
template <> USG_API Value::Value(const fdk::Vec4d & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(reinterpret_cast<const PXR_NS::GfVec4d&>(data)))) {}
template <> USG_API Value::Value(const fdk::Vec4i & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(reinterpret_cast<const PXR_NS::GfVec4i&>(data)))) {}
//
template <> USG_API Value::Value(const fdk::Quath & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(reinterpret_cast<const PXR_NS::GfQuath&>(data)))) {}
template <> USG_API Value::Value(const fdk::Quatf & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(reinterpret_cast<const PXR_NS::GfQuatf&>(data)))) {}
template <> USG_API Value::Value(const fdk::Quatd & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(reinterpret_cast<const PXR_NS::GfQuatd&>(data)))) {}
//
#if 0
// Disabling Mat4f version to avoid creating an invalid Value::Type. TODO: make this throw a static_assert()...?
template <> USG_API Value::Value(const fdk::Mat4f & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(reinterpret_cast<const PXR_NS::GfMatrix4d&>(data)))) {}
#endif
template <> USG_API Value::Value(const fdk::Mat4d & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(reinterpret_cast<const PXR_NS::GfMatrix4d&>(data)))) {}
//
template <> USG_API Value::Value(const fdk::Box3f & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(PXR_NS::VtVec3fArray { reinterpret_cast<const PXR_NS::GfVec3f&>(data.min), reinterpret_cast<const PXR_NS::GfVec3f&>(data.max) }))) {}
template <> USG_API Value::Value(const fdk::Box3d & data) : _pImpl(new Value::Impl(PXR_NS::VtValue(PXR_NS::VtVec3dArray { reinterpret_cast<const PXR_NS::GfVec3d&>(data.min), reinterpret_cast<const PXR_NS::GfVec3d&>(data.max) }))) {}
//-------------------------------------------

// Copy a usg::Array to a VtArray, wrapped in a VtValue.
template <typename T, typename usdT>
inline PXR_NS::VtValue  setVtArrayValue(const Array<T>& usg_array)
{
  const Array<usdT>& src_list = reinterpret_cast<const Array<usdT>&>(usg_array);
  return PXR_NS::VtValue(PXR_NS::VtArray<usdT>(src_list.data(), src_list.data() + src_list.size()));
}


template <> USG_API Value::Value(const usg::StringArray   & array) : _pImpl(new Value::Impl(setVtArrayValue<std::string   , std::string         >(array))) {}
template <> USG_API Value::Value(const usg::TokenArray    & array) : _pImpl(new Value::Impl(setVtArrayValue<usg::Token    , PXR_NS::TfToken     >(array))) {}
template <> USG_API Value::Value(const usg::PathArray     & array) : _pImpl(new Value::Impl(setVtArrayValue<usg::Path     , PXR_NS::SdfPath     >(array))) {}
template <> USG_API Value::Value(const usg::AssetPathArray& array) : _pImpl(new Value::Impl(setVtArrayValue<usg::AssetPath, PXR_NS::SdfAssetPath>(array))) {}
//
//template <> USG_API Value::Value(const usg::BoolArray     & array) : _pImpl(new Value::Impl(setVtArrayValue<bool       , bool          >(array))) {}
template <> USG_API Value::Value(const usg::CharArray     & array) : _pImpl(new Value::Impl(setVtArrayValue<char       , char          >(array))) {}
template <> USG_API Value::Value(const usg::UCharArray    & array) : _pImpl(new Value::Impl(setVtArrayValue<unsigned char, unsigned char>(array))) {}
template <> USG_API Value::Value(const usg::HalfArray     & array) : _pImpl(new Value::Impl(setVtArrayValue<fdk::half  , PXR_NS::GfHalf>(array))) {}
template <> USG_API Value::Value(const usg::FloatArray    & array) : _pImpl(new Value::Impl(setVtArrayValue<float      , float         >(array))) {}
template <> USG_API Value::Value(const usg::DoubleArray   & array) : _pImpl(new Value::Impl(setVtArrayValue<double     , double        >(array))) {}
template <> USG_API Value::Value(const usg::Int32Array    & array) : _pImpl(new Value::Impl(setVtArrayValue<int32_t    , int32_t       >(array))) {}
template <> USG_API Value::Value(const usg::Int64Array    & array) : _pImpl(new Value::Impl(setVtArrayValue<int64_t    , int64_t       >(array))) {}
template <> USG_API Value::Value(const usg::Uint32Array   & array) : _pImpl(new Value::Impl(setVtArrayValue<uint32_t   , uint32_t      >(array))) {}
template <> USG_API Value::Value(const usg::Uint64Array   & array) : _pImpl(new Value::Impl(setVtArrayValue<uint64_t   , uint64_t      >(array))) {}
//
template <> USG_API Value::Value(const usg::Vec2hArray    & array) : _pImpl(new Value::Impl(setVtArrayValue<fdk::Vec2h , PXR_NS::GfVec2h>(array))) {}
template <> USG_API Value::Value(const usg::Vec2fArray    & array) : _pImpl(new Value::Impl(setVtArrayValue<fdk::Vec2f , PXR_NS::GfVec2f>(array))) {}
template <> USG_API Value::Value(const usg::Vec2dArray    & array) : _pImpl(new Value::Impl(setVtArrayValue<fdk::Vec2d , PXR_NS::GfVec2d>(array))) {}
template <> USG_API Value::Value(const usg::Vec2iArray    & array) : _pImpl(new Value::Impl(setVtArrayValue<fdk::Vec2i , PXR_NS::GfVec2i>(array))) {}
//
template <> USG_API Value::Value(const usg::Vec3hArray    & array) : _pImpl(new Value::Impl(setVtArrayValue<fdk::Vec3h , PXR_NS::GfVec3h>(array))) {}
template <> USG_API Value::Value(const usg::Vec3fArray    & array) : _pImpl(new Value::Impl(setVtArrayValue<fdk::Vec3f , PXR_NS::GfVec3f>(array))) {}
template <> USG_API Value::Value(const usg::Vec3dArray    & array) : _pImpl(new Value::Impl(setVtArrayValue<fdk::Vec3d , PXR_NS::GfVec3d>(array))) {}
template <> USG_API Value::Value(const usg::Vec3iArray    & array) : _pImpl(new Value::Impl(setVtArrayValue<fdk::Vec3i , PXR_NS::GfVec3i>(array))) {}
//
template <> USG_API Value::Value(const usg::Vec4hArray    & array) : _pImpl(new Value::Impl(setVtArrayValue<fdk::Vec4h , PXR_NS::GfVec4h>(array))) {}
template <> USG_API Value::Value(const usg::Vec4fArray    & array) : _pImpl(new Value::Impl(setVtArrayValue<fdk::Vec4f , PXR_NS::GfVec4f>(array))) {}
template <> USG_API Value::Value(const usg::Vec4dArray    & array) : _pImpl(new Value::Impl(setVtArrayValue<fdk::Vec4d , PXR_NS::GfVec4d>(array))) {}
template <> USG_API Value::Value(const usg::Vec4iArray    & array) : _pImpl(new Value::Impl(setVtArrayValue<fdk::Vec4i , PXR_NS::GfVec4i>(array))) {}
//
template <> USG_API Value::Value(const usg::QuathArray    & array) : _pImpl(new Value::Impl(setVtArrayValue<fdk::Quath , PXR_NS::GfQuath>(array))) {}
template <> USG_API Value::Value(const usg::QuatfArray    & array) : _pImpl(new Value::Impl(setVtArrayValue<fdk::Quatf , PXR_NS::GfQuatf>(array))) {}
template <> USG_API Value::Value(const usg::QuatdArray    & array) : _pImpl(new Value::Impl(setVtArrayValue<fdk::Quatd , PXR_NS::GfQuatd>(array))) {}
//
template <> USG_API Value::Value(const usg::Mat4dArray    & array) : _pImpl(new Value::Impl(setVtArrayValue<fdk::Mat4d , PXR_NS::GfMatrix4d>(array))) {}

//-------------------------------------------

// Copy a std::vector array to a VtArray, wrapped in a VtValue.
template <typename T, typename usdT>
inline PXR_NS::VtValue  copyVecToVtArray(const std::vector<T>& vector)
{
  const std::vector<usdT>& src_list = reinterpret_cast<const std::vector<usdT>&>(vector);
  return PXR_NS::VtValue(PXR_NS::VtArray<usdT>(src_list.data(), src_list.data() + src_list.size()));
}

//inline PXR_NS::VtValue  setVtArrayValue(const std::vector<usg::Token&>& usg_array)
//{
//  const std::vector<usdT>& src_list = reinterpret_cast<const std::vector<usdT>&>(usg_array);
//  return PXR_NS::VtValue(PXR_NS::VtArray<PXR_NS::TfToken>(src_list.data(), src_list.data() + src_list.size()));
//}




template <> USG_API Value::Value(const std::vector<std::string   >& vec) : _pImpl(new Value::Impl(copyVecToVtArray<std::string   , std::string         >(vec))) {}
template <> USG_API Value::Value(const std::vector<usg::Token    >& vec) : _pImpl(new Value::Impl(copyVecToVtArray<usg::Token    , PXR_NS::TfToken     >(vec))) {}
template <> USG_API Value::Value(const std::vector<usg::Path     >& vec) : _pImpl(new Value::Impl(copyVecToVtArray<usg::Path     , PXR_NS::SdfPath     >(vec))) {}
template <> USG_API Value::Value(const std::vector<usg::AssetPath>& vec) : _pImpl(new Value::Impl(copyVecToVtArray<usg::AssetPath, PXR_NS::SdfAssetPath>(vec))) {}
//
//template <> USG_API Value::Value(const std::vector<bool     >& vec) : _pImpl(new Value::Impl(copyVecToVtArray<bool      , bool          >(vec))) {}
template <> USG_API Value::Value(const std::vector<char     >& vec) : _pImpl(new Value::Impl(copyVecToVtArray<char      , char          >(vec))) {}
template <> USG_API Value::Value(const std::vector<unsigned char>& vec) : _pImpl(new Value::Impl(copyVecToVtArray<unsigned char, unsigned char>(vec))) {}
template <> USG_API Value::Value(const std::vector<fdk::half>& vec) : _pImpl(new Value::Impl(copyVecToVtArray<fdk::half , PXR_NS::GfHalf>(vec))) {}
template <> USG_API Value::Value(const std::vector<float    >& vec) : _pImpl(new Value::Impl(copyVecToVtArray<float     , float         >(vec))) {}
template <> USG_API Value::Value(const std::vector<double   >& vec) : _pImpl(new Value::Impl(copyVecToVtArray<double    , double        >(vec))) {}
template <> USG_API Value::Value(const std::vector<int32_t  >& vec) : _pImpl(new Value::Impl(copyVecToVtArray<int32_t   , int32_t       >(vec))) {}
template <> USG_API Value::Value(const std::vector<int64_t  >& vec) : _pImpl(new Value::Impl(copyVecToVtArray<int64_t   , int64_t       >(vec))) {}
template <> USG_API Value::Value(const std::vector<uint32_t >& vec) : _pImpl(new Value::Impl(copyVecToVtArray<uint32_t  , uint32_t      >(vec))) {}
template <> USG_API Value::Value(const std::vector<uint64_t >& vec) : _pImpl(new Value::Impl(copyVecToVtArray<uint64_t  , uint64_t      >(vec))) {}
//
template <> USG_API Value::Value(const std::vector<fdk::Vec2h>& vec) : _pImpl(new Value::Impl(copyVecToVtArray<fdk::Vec2h, PXR_NS::GfVec2h>(vec))) {}
template <> USG_API Value::Value(const std::vector<fdk::Vec2f>& vec) : _pImpl(new Value::Impl(copyVecToVtArray<fdk::Vec2f, PXR_NS::GfVec2f>(vec))) {}
template <> USG_API Value::Value(const std::vector<fdk::Vec2d>& vec) : _pImpl(new Value::Impl(copyVecToVtArray<fdk::Vec2d, PXR_NS::GfVec2d>(vec))) {}
template <> USG_API Value::Value(const std::vector<fdk::Vec2i>& vec) : _pImpl(new Value::Impl(copyVecToVtArray<fdk::Vec2i, PXR_NS::GfVec2i>(vec))) {}
//
template <> USG_API Value::Value(const std::vector<fdk::Vec3h>& vec) : _pImpl(new Value::Impl(copyVecToVtArray<fdk::Vec3h, PXR_NS::GfVec3h>(vec))) {}
template <> USG_API Value::Value(const std::vector<fdk::Vec3f>& vec) : _pImpl(new Value::Impl(copyVecToVtArray<fdk::Vec3f, PXR_NS::GfVec3f>(vec))) {}
template <> USG_API Value::Value(const std::vector<fdk::Vec3d>& vec) : _pImpl(new Value::Impl(copyVecToVtArray<fdk::Vec3d, PXR_NS::GfVec3d>(vec))) {}
template <> USG_API Value::Value(const std::vector<fdk::Vec3i>& vec) : _pImpl(new Value::Impl(copyVecToVtArray<fdk::Vec3i, PXR_NS::GfVec3i>(vec))) {}
//
template <> USG_API Value::Value(const std::vector<fdk::Vec4h>& vec) : _pImpl(new Value::Impl(copyVecToVtArray<fdk::Vec4h, PXR_NS::GfVec4h>(vec))) {}
template <> USG_API Value::Value(const std::vector<fdk::Vec4f>& vec) : _pImpl(new Value::Impl(copyVecToVtArray<fdk::Vec4f, PXR_NS::GfVec4f>(vec))) {}
template <> USG_API Value::Value(const std::vector<fdk::Vec4d>& vec) : _pImpl(new Value::Impl(copyVecToVtArray<fdk::Vec4d, PXR_NS::GfVec4d>(vec))) {}
template <> USG_API Value::Value(const std::vector<fdk::Vec4i>& vec) : _pImpl(new Value::Impl(copyVecToVtArray<fdk::Vec4i, PXR_NS::GfVec4i>(vec))) {}
//
template <> USG_API Value::Value(const std::vector<fdk::Quath>& vec) : _pImpl(new Value::Impl(copyVecToVtArray<fdk::Quath, PXR_NS::GfQuath>(vec))) {}
template <> USG_API Value::Value(const std::vector<fdk::Quatf>& vec) : _pImpl(new Value::Impl(copyVecToVtArray<fdk::Quatf, PXR_NS::GfQuatf>(vec))) {}
template <> USG_API Value::Value(const std::vector<fdk::Quatd>& vec) : _pImpl(new Value::Impl(copyVecToVtArray<fdk::Quatd, PXR_NS::GfQuatd>(vec))) {}
//
template <> USG_API Value::Value(const std::vector<fdk::Mat4d>& vec) : _pImpl(new Value::Impl(copyVecToVtArray<fdk::Mat4d, PXR_NS::GfMatrix4d>(vec))) {}

//----------------------------------------------------------------


// Specialized versions handling types which can't return const references:
usg::Token      Value::getToken()     const { return (_pImpl) ? fromTfToken(_pImpl->vtvalue.UncheckedGet<PXR_NS::TfToken>()) : empty_token; }
usg::Path       Value::getPath()      const { return (_pImpl) ? fromSdfPath(_pImpl->vtvalue.UncheckedGet<PXR_NS::SdfPath>()) : empty_geompath; }
usg::AssetPath  Value::getAssetPath() const { return (_pImpl) ? fromSdfAssetPath(_pImpl->vtvalue.UncheckedGet<PXR_NS::SdfAssetPath>()) : empty_assetpath; }


template <> USG_API const std::string& Value::get() const { return (_pImpl) ? _pImpl->vtvalue.UncheckedGet<PXR_NS::TfToken>().GetString() : empty_string; }
//
template <> USG_API const bool       & Value::get() const { return (_pImpl) ? _pImpl->vtvalue.UncheckedGet<bool    >() : empty_bool;   }
template <> USG_API const char       & Value::get() const { return (_pImpl) ? _pImpl->vtvalue.UncheckedGet<char    >() : empty_char;   }
template <> USG_API const unsigned char & Value::get() const { return (_pImpl) ? _pImpl->vtvalue.UncheckedGet<unsigned char>() : empty_uchar;   }
template <> USG_API const fdk::half  & Value::get() const { return (_pImpl) ? reinterpret_cast<const fdk::half&>(_pImpl->vtvalue.UncheckedGet<PXR_NS::GfHalf>()) : empty_half; }
template <> USG_API const float      & Value::get() const { return (_pImpl) ? _pImpl->vtvalue.UncheckedGet<float   >() : empty_float;  }
template <> USG_API const double     & Value::get() const { return (_pImpl) ? _pImpl->vtvalue.UncheckedGet<double  >() : empty_double; }
template <> USG_API const int32_t    & Value::get() const { return (_pImpl) ? _pImpl->vtvalue.UncheckedGet<int32_t >() : empty_int32;  }
template <> USG_API const int64_t    & Value::get() const { return (_pImpl) ? _pImpl->vtvalue.UncheckedGet<int64_t >() : empty_int64;  }
template <> USG_API const uint32_t   & Value::get() const { return (_pImpl) ? _pImpl->vtvalue.UncheckedGet<uint32_t>() : empty_uint32; }
template <> USG_API const uint64_t   & Value::get() const { return (_pImpl) ? _pImpl->vtvalue.UncheckedGet<uint64_t>() : empty_uint64; }
//
template <> USG_API const fdk::Vec2h & Value::get() const { return (_pImpl) ? reinterpret_cast<const fdk::Vec2h&>(_pImpl->vtvalue.UncheckedGet<PXR_NS::GfVec2h>()) : empty_vec2h; }
template <> USG_API const fdk::Vec2f & Value::get() const { return (_pImpl) ? reinterpret_cast<const fdk::Vec2f&>(_pImpl->vtvalue.UncheckedGet<PXR_NS::GfVec2f>()) : empty_vec2f; }
template <> USG_API const fdk::Vec2d & Value::get() const { return (_pImpl) ? reinterpret_cast<const fdk::Vec2d&>(_pImpl->vtvalue.UncheckedGet<PXR_NS::GfVec2d>()) : empty_vec2d; }
template <> USG_API const fdk::Vec2i & Value::get() const { return (_pImpl) ? reinterpret_cast<const fdk::Vec2i&>(_pImpl->vtvalue.UncheckedGet<PXR_NS::GfVec2i>()) : empty_vec2i; }
//
template <> USG_API const fdk::Vec3h & Value::get() const { return (_pImpl) ? reinterpret_cast<const fdk::Vec3h&>(_pImpl->vtvalue.UncheckedGet<PXR_NS::GfVec3h>()) : empty_vec3h; }
template <> USG_API const fdk::Vec3f & Value::get() const { return (_pImpl) ? reinterpret_cast<const fdk::Vec3f&>(_pImpl->vtvalue.UncheckedGet<PXR_NS::GfVec3f>()) : empty_vec3f; }
template <> USG_API const fdk::Vec3d & Value::get() const { return (_pImpl) ? reinterpret_cast<const fdk::Vec3d&>(_pImpl->vtvalue.UncheckedGet<PXR_NS::GfVec3d>()) : empty_vec3d; }
template <> USG_API const fdk::Vec3i & Value::get() const { return (_pImpl) ? reinterpret_cast<const fdk::Vec3i&>(_pImpl->vtvalue.UncheckedGet<PXR_NS::GfVec3i>()) : empty_vec3i; }
//
template <> USG_API const fdk::Vec4h & Value::get() const { return (_pImpl) ? reinterpret_cast<const fdk::Vec4h&>(_pImpl->vtvalue.UncheckedGet<PXR_NS::GfVec4h>()) : empty_vec4h; }
template <> USG_API const fdk::Vec4f & Value::get() const { return (_pImpl) ? reinterpret_cast<const fdk::Vec4f&>(_pImpl->vtvalue.UncheckedGet<PXR_NS::GfVec4f>()) : empty_vec4f; }
template <> USG_API const fdk::Vec4d & Value::get() const { return (_pImpl) ? reinterpret_cast<const fdk::Vec4d&>(_pImpl->vtvalue.UncheckedGet<PXR_NS::GfVec4d>()) : empty_vec4d; }
template <> USG_API const fdk::Vec4i & Value::get() const { return (_pImpl) ? reinterpret_cast<const fdk::Vec4i&>(_pImpl->vtvalue.UncheckedGet<PXR_NS::GfVec4i>()) : empty_vec4i; }
//
template <> USG_API const fdk::Quath & Value::get() const { return (_pImpl) ? reinterpret_cast<const fdk::Quath&>(_pImpl->vtvalue.UncheckedGet<PXR_NS::GfQuath>()) : empty_quath; }
template <> USG_API const fdk::Quatf & Value::get() const { return (_pImpl) ? reinterpret_cast<const fdk::Quatf&>(_pImpl->vtvalue.UncheckedGet<PXR_NS::GfQuatf>()) : empty_quatf; }
template <> USG_API const fdk::Quatd & Value::get() const { return (_pImpl) ? reinterpret_cast<const fdk::Quatd&>(_pImpl->vtvalue.UncheckedGet<PXR_NS::GfQuatd>()) : empty_quatd; }
//
#if 0
// Disabling Mat4f version to avoid creating an invalid Value::Type. TODO: make this throw a static_assert()...?
template <> USG_API const fdk::Mat4f & Value::get() const { return (_pImpl) ? reinterpret_cast<const fdk::Mat4f&>(_pImpl->vtvalue.UncheckedGet<PXR_NS::GfMatrix4f>()) : empty_mat4f; }
#endif
template <> USG_API const fdk::Mat4d & Value::get() const { return (_pImpl) ? reinterpret_cast<const fdk::Mat4d&>(_pImpl->vtvalue.UncheckedGet<PXR_NS::GfMatrix4d>()) : empty_mat4d; }
//
template <> USG_API const fdk::Box3f & Value::get() const {
  if (_pImpl) {
    // Extents in USD are implemented as a 2-element array of float3
    const auto array = _pImpl->vtvalue.UncheckedGet<PXR_NS::VtVec3fArray>();
    if (array.size() == 2) {
      return *reinterpret_cast<const fdk::Box3f*>(array.data());
    }
  }
  return empty_box3f;
}
template <> USG_API const fdk::Box3d & Value::get() const {
  if (_pImpl) {
    // Extents in USD are implemented as a 2-element array of double3
    const auto array = _pImpl->vtvalue.UncheckedGet<PXR_NS::VtVec3dArray>();
    if (array.size() == 2) {
      return *reinterpret_cast<const fdk::Box3d*>(array.data());
    }
  }
  return empty_box3d;
}
//-------------------------------------------
// Note - USD does not have a specific 'GfStringArray' typedef, so we use NdrStringVec:
// TODO: Check that we don't need UncheckedGet<VtArray> translation!!!!
template <> USG_API const fdk::StringList& Value::get() const { return (_pImpl) ? _pImpl->vtvalue.UncheckedGet<PXR_NS::NdrStringVec>() : empty_string_array; }
//
//template <> USG_API const usg::BoolArray  & Value::get() const { return (_pImpl) ? _pImpl->vtvalue.UncheckedGet<usg::BoolArray>() : empty_bool_array;   }
template <> USG_API const usg::CharArray  & Value::get() const { return (_pImpl) ? _pImpl->vtvalue.UncheckedGet<usg::CharArray>() : empty_char_array;   }
template <> USG_API const usg::UCharArray  & Value::get() const { return (_pImpl) ? _pImpl->vtvalue.UncheckedGet<usg::UCharArray>() : empty_uchar_array;   }
template <> USG_API const usg::HalfArray  & Value::get() const { return (_pImpl) ? reinterpret_cast<const usg::HalfArray &>(_pImpl->vtvalue.UncheckedGet<std::vector<PXR_NS::GfHalf>>()) : empty_half_array;  }
template <> USG_API const usg::FloatArray & Value::get() const { return (_pImpl) ? _pImpl->vtvalue.UncheckedGet<usg::FloatArray>() : empty_float_array;  }
template <> USG_API const usg::DoubleArray& Value::get() const { return (_pImpl) ? _pImpl->vtvalue.UncheckedGet<usg::DoubleArray>() : empty_double_array; }
template <> USG_API const usg::Int32Array & Value::get() const { return (_pImpl) ? _pImpl->vtvalue.UncheckedGet<usg::Int32Array>() : empty_int32_array;  }
//template <> USG_API const usg::Int64Array & Value::get() const { return (_pImpl) ? _pImpl->vtvalue.UncheckedGet<usg::Int64Array>() : empty_int64_array;  }
template <> USG_API const usg::Uint32Array& Value::get() const { return (_pImpl) ? _pImpl->vtvalue.UncheckedGet<usg::Uint32Array>() : empty_uint32_array; }
template <> USG_API const usg::Uint64Array& Value::get() const { return (_pImpl) ? _pImpl->vtvalue.UncheckedGet<usg::Uint64Array>() : empty_uint64_array; }
//
template <> USG_API const usg::Vec2hArray& Value::get() const { return (_pImpl) ? reinterpret_cast<const usg::Vec2hArray &>(_pImpl->vtvalue.UncheckedGet<std::vector<PXR_NS::GfVec2h>>()) : empty_vec2h_array; }
template <> USG_API const usg::Vec2fArray& Value::get() const { return (_pImpl) ? reinterpret_cast<const usg::Vec2fArray &>(_pImpl->vtvalue.UncheckedGet<std::vector<PXR_NS::GfVec2f>>()) : empty_vec2f_array; }
template <> USG_API const usg::Vec2dArray& Value::get() const { return (_pImpl) ? reinterpret_cast<const usg::Vec2dArray &>(_pImpl->vtvalue.UncheckedGet<std::vector<PXR_NS::GfVec2d>>()) : empty_vec2d_array; }
template <> USG_API const usg::Vec2iArray& Value::get() const { return (_pImpl) ? reinterpret_cast<const usg::Vec2iArray &>(_pImpl->vtvalue.UncheckedGet<std::vector<PXR_NS::GfVec2i>>()) : empty_vec2i_array; }
//
template <> USG_API const usg::Vec3hArray& Value::get() const { return (_pImpl) ? reinterpret_cast<const usg::Vec3hArray &>(_pImpl->vtvalue.UncheckedGet<std::vector<PXR_NS::GfVec3h>>()) : empty_vec3h_array; }
template <> USG_API const usg::Vec3fArray& Value::get() const { return (_pImpl) ? reinterpret_cast<const usg::Vec3fArray &>(_pImpl->vtvalue.UncheckedGet<std::vector<PXR_NS::GfVec3f>>()) : empty_vec3f_array; }
template <> USG_API const usg::Vec3dArray& Value::get() const { return (_pImpl) ? reinterpret_cast<const usg::Vec3dArray &>(_pImpl->vtvalue.UncheckedGet<std::vector<PXR_NS::GfVec3d>>()) : empty_vec3d_array; }
template <> USG_API const usg::Vec3iArray& Value::get() const { return (_pImpl) ? reinterpret_cast<const usg::Vec3iArray &>(_pImpl->vtvalue.UncheckedGet<std::vector<PXR_NS::GfVec3i>>()) : empty_vec3i_array; }
//
template <> USG_API const usg::Vec4hArray& Value::get() const { return (_pImpl) ? reinterpret_cast<const usg::Vec4hArray &>(_pImpl->vtvalue.UncheckedGet<std::vector<PXR_NS::GfVec4h>>()) : empty_vec4h_array; }
template <> USG_API const usg::Vec4fArray& Value::get() const { return (_pImpl) ? reinterpret_cast<const usg::Vec4fArray &>(_pImpl->vtvalue.UncheckedGet<std::vector<PXR_NS::GfVec4f>>()) : empty_vec4f_array; }
template <> USG_API const usg::Vec4dArray& Value::get() const { return (_pImpl) ? reinterpret_cast<const usg::Vec4dArray &>(_pImpl->vtvalue.UncheckedGet<std::vector<PXR_NS::GfVec4d>>()) : empty_vec4d_array; }
template <> USG_API const usg::Vec4iArray& Value::get() const { return (_pImpl) ? reinterpret_cast<const usg::Vec4iArray &>(_pImpl->vtvalue.UncheckedGet<std::vector<PXR_NS::GfVec4i>>()) : empty_vec4i_array; }
//
template <> USG_API const usg::QuathArray& Value::get() const { return (_pImpl) ? reinterpret_cast<const usg::QuathArray &>(_pImpl->vtvalue.UncheckedGet<std::vector<PXR_NS::GfQuath>>()) : empty_quath_array; }
template <> USG_API const usg::QuatfArray& Value::get() const { return (_pImpl) ? reinterpret_cast<const usg::QuatfArray &>(_pImpl->vtvalue.UncheckedGet<std::vector<PXR_NS::GfQuatf>>()) : empty_quatf_array; }
template <> USG_API const usg::QuatdArray& Value::get() const { return (_pImpl) ? reinterpret_cast<const usg::QuatdArray &>(_pImpl->vtvalue.UncheckedGet<std::vector<PXR_NS::GfQuatd>>()) : empty_quatd_array; }
//
template <> USG_API const usg::Mat4dArray& Value::get() const { return (_pImpl) ? reinterpret_cast<const usg::Mat4dArray &>(_pImpl->vtvalue.UncheckedGet<std::vector<PXR_NS::GfMatrix4d>>()) : empty_mat4d_array; }


//----------------------------------------------------------------


/*static*/ Value
Value::getDefaultValueForType(Value::Type type)
{
  static std::map<Value::Type, Value> type_default_values_map
  {
    //  type             default value
    //---------------------------------------------------------------------------------------
    { DefaultType , Value()                 },
    //---------------------------------------------------------------------------------------
    { String      , Value(empty_string   )  }, { StringArray     , Value(empty_string_array   )  },
    { Token       , Value(empty_token    )  }, { TokenArray      , Value(empty_token_array    )  },
    { AssetPath   , Value(empty_assetpath)  }, { AssetPathArray  , Value(empty_assetpath_array)  },
    //--------------------------------------------------------------------------------------------
    { Bool        , Value(/*empty_bool*/ )  }, { BoolArray       , Value(/*empty_bool_array*/ )  },
    { Int         , Value(empty_int32    )  }, { IntArray        , Value(empty_int32_array    )  },
    { Int64       , Value(empty_int64    )  }, { Int64Array      , Value(empty_int64_array    )  },
    { UChar       , Value(empty_char     )  }, { UCharArray      , Value(empty_char_array     )  },
    { UInt        , Value(empty_int32    )  }, { UIntArray       , Value(empty_int32_array    )  },
    { UInt64      , Value(empty_uint64   )  }, { UInt64Array     , Value(empty_uint64_array   )  },
    { Half        , Value(empty_half     )  }, { HalfArray       , Value(empty_half_array     )  },
    { Float       , Value(empty_float    )  }, { FloatArray      , Value(empty_float_array    )  },
    { Double      , Value(empty_double   )  }, { DoubleArray     , Value(empty_double_array   )  },
    { TimeCode    , Value(/* ?????*/     )  }, { TimeCodeArray   , Value(/* ?????*/     )  },
    //--------------------------------------------------------------------------------------------
    { Int2        , Value(empty_vec2i    )  }, { Int2Array       , Value(empty_vec2i_array    )  },
    { Half2       , Value(empty_vec2h    )  }, { Half2Array      , Value(empty_vec2h_array    )  },
    { Float2      , Value(empty_vec2f    )  }, { Float2Array     , Value(empty_vec2f_array    )  },
    { Double2     , Value(empty_vec2d    )  }, { Double2Array    , Value(empty_vec2d_array    )  },
    //--------------------------------------------------------------------------------------------
    { Int3        , Value(empty_vec3i    )  }, { Int3Array       , Value(empty_vec3i_array    )  },
    { Half3       , Value(empty_vec3h    )  }, { Half3Array      , Value(empty_vec3h_array    )  },
    { Float3      , Value(empty_vec3f    )  }, { Float3Array     , Value(empty_vec3f_array    )  },
    { Double3     , Value(empty_vec3d    )  }, { Double3Array    , Value(empty_vec3d_array    )  },
    //--------------------------------------------------------------------------------------------
    { Int4        , Value(empty_vec4i    )  }, { Int4Array       , Value(empty_vec4i_array    )  },
    { Half4       , Value(empty_vec4h    )  }, { Half4Array      , Value(empty_vec4h_array    )  },
    { Float4      , Value(empty_vec4f    )  }, { Float4Array     , Value(empty_vec4f_array    )  },
    { Double4     , Value(empty_vec4d    )  }, { Double4Array    , Value(empty_vec4d_array    )  },
    //--------------------------------------------------------------------------------------------
    { Matrix2d    , Value(/* ?????*/     )  }, { Matrix2dArray   , Value(/* ?????*/           )  },
    { Matrix3d    , Value(/* ?????*/     )  }, { Matrix3dArray   , Value(/* ?????*/           )  },
    { Matrix4d    , Value(empty_mat4d    )  }, { Matrix4dArray   , Value(empty_mat4d_array    )  },
    //--------------------------------------------------------------------------------------------
    { Channel     , Value(empty_int32    )  }, { ChannelArray    , Value(empty_int32_array    )  },
    //--------------------------------------------------------------------------------------------
    { Color       , Value(empty_vec3i    )  }, { ColorArray      , Value(empty_vec3i_array    )  },
    { Color3h     , Value(empty_vec3h    )  }, { Color3hArray    , Value(empty_vec3h_array    )  },
    { Color3f     , Value(empty_vec3f    )  }, { Color3fArray    , Value(empty_vec3f_array    )  },
    { Color3d     , Value(empty_vec3d    )  }, { Color3dArray    , Value(empty_vec3d_array    )  },
    //--------------------------------------------------------------------------------------------
    { Normal      , Value(empty_vec3i    )  }, { NormalArray     , Value(empty_vec3i_array    )  },
    { Normal3h    , Value(empty_vec3h    )  }, { Normal3hArray   , Value(empty_vec3h_array    )  },
    { Normal3f    , Value(empty_vec3f    )  }, { Normal3fArray   , Value(empty_vec3f_array    )  },
    { Normal3d    , Value(empty_vec3d    )  }, { Normal3dArray   , Value(empty_vec3d_array    )  },
    //--------------------------------------------------------------------------------------------
    { Point       , Value(empty_vec3i    )  }, { PointArray      , Value(empty_vec3i_array    )  },
    { Point3h     , Value(empty_vec3h    )  }, { Point3hArray    , Value(empty_vec3h_array    )  },
    { Point3f     , Value(empty_vec3f    )  }, { Point3fArray    , Value(empty_vec3f_array    )  },
    { Point3d     , Value(empty_vec3d    )  }, { Point3dArray    , Value(empty_vec3d_array    )  },
    //--------------------------------------------------------------------------------------------
    { Vector      , Value(empty_vec3i    )  }, { VectorArray     , Value(empty_vec3i_array    )  },
    { Vector3h    , Value(empty_vec3h    )  }, { Vector3hArray   , Value(empty_vec3h_array    )  },
    { Vector3f    , Value(empty_vec3f    )  }, { Vector3fArray   , Value(empty_vec3f_array    )  },
    { Vector3d    , Value(empty_vec3d    )  }, { Vector3dArray   , Value(empty_vec3d_array    )  },
    //--------------------------------------------------------------------------------------------
    { Color4h     , Value(empty_vec4h    )  }, { Color4hArray    , Value(empty_vec4h_array    )  },
    { Color4f     , Value(empty_vec4f    )  }, { Color4fArray    , Value(empty_vec4f_array    )  },
    { Color4d     , Value(empty_vec4d    )  }, { Color4dArray    , Value(empty_vec4d_array    )  },
    //--------------------------------------------------------------------------------------------
    { TexCoord2h  , Value(empty_vec2h    )  }, { TexCoord2hArray , Value(empty_vec2h_array    )  },
    { TexCoord2f  , Value(empty_vec2f    )  }, { TexCoord2fArray , Value(empty_vec2f_array    )  },
    { TexCoord2d  , Value(empty_vec2d    )  }, { TexCoord2dArray , Value(empty_vec2d_array    )  },
    { TexCoord3h  , Value(empty_vec2h    )  }, { TexCoord3hArray , Value(empty_vec2h_array    )  },
    { TexCoord3f  , Value(empty_vec2f    )  }, { TexCoord3fArray , Value(empty_vec2f_array    )  },
    { TexCoord3d  , Value(empty_vec2d    )  }, { TexCoord3dArray , Value(empty_vec2d_array    )  },
    //--------------------------------------------------------------------------------------------
    { Matrix      , Value(empty_mat4d    )  }, { MatrixArray     , Value(empty_mat4d_array    )  },
    { Frame4d     , Value(/*????*/       )  }, { Frame4dArray    , Value(/*????*/             )  },
    //--------------------------------------------------------------------------------------------
    { Quath       , Value(empty_quath    )  }, { QuathArray      , Value(empty_quath_array    )  },
    { Quatf       , Value(empty_quatf    )  }, { QuatfArray      , Value(empty_quatf_array    )  },
    { Quatd       , Value(empty_quatd    )  }, { QuatdArray      , Value(empty_quatd_array    )  },
    //--------------------------------------------------------------------------------------------
    { Struct      , Value(/*????*/       )  }, { StructArray     , Value(/*????*/             )  },
    { Terminal    , Value(/*????*/       )  }, { TerminalArray   , Value(/*????*/             )  },
    { Vstruct     , Value(/*????*/       )  }, { VstructArray    , Value(/*????*/             )  },
    //---------------------------------------------------------------------------------------
    { InvalidType , Value()                 },
    //---------------------------------------------------------------------------------------
  };

  const std::map<Value::Type, Value>::const_iterator it = type_default_values_map.find(type);
  if (it != type_default_values_map.end()) {
    return it->second;
  }

  // Shouldn't happen... If it does then we're missing some types!
  std::cerr << "Value::getDefaultValueForType(): warning, unknown Value::Type index " << type;
  std::cerr << ", this is likely due to a coding error." << std::endl;

  return Value();
}


} // namespace usg
