// Copyright (c) 2022 The Foundry Visionmongers Ltd. All Rights Reserved.

#include "usg/geom/AssetPath.h"

#include "usg/private/MemoryPoolWrapperImpl.h"
#include "fdk/base/Hash.h"

#include "pxr/usd/sdf/assetPath.h"

namespace usg {
  namespace {
    MemoryPoolWrapperImpl<PXR_NS::SdfAssetPath, AssetPath, AssetPath::Handle>& wrapper()
    {
      static MemoryPoolWrapperImpl<PXR_NS::SdfAssetPath, AssetPath, AssetPath::Handle> sWrapper;
      return sWrapper;
    }

    const PXR_NS::SdfAssetPath& unwrap(const MemoryPoolWrapperImpl<PXR_NS::SdfAssetPath, AssetPath, AssetPath::Handle>::MemoryPoolHandle& h)
    {
      return wrapper().unwrap(h);
    }

    AssetPath wrap(const PXR_NS::SdfAssetPath& assetPath)
    {
      return wrapper().wrap(assetPath);
    }
  }

  AssetPath::AssetPath()
    : _handle(wrapper().allocate())
  {
  }

  AssetPath::AssetPath(const std::string& path)
    : _handle(wrapper().allocate(path))
  {

  }

  AssetPath::AssetPath(const std::string& path, const std::string& resolvedPath)
    : _handle(wrapper().allocate(path, resolvedPath))
  {
    
  }

  AssetPath::~AssetPath()
  {
    wrapper().free(_handle);
  }

  AssetPath::AssetPath(const AssetPath::Handle& handle)
    : _handle(wrapper().copyAllocate(handle))
  {

  }

  AssetPath::AssetPath(AssetPath&& rhs) noexcept
  {
    _handle = rhs._handle;
    rhs._handle = fdk::MemoryPool<PXR_NS::SdfAssetPath>::nullhandle;
  }

  AssetPath& AssetPath::operator=(AssetPath&& rhs) noexcept
  {
    if (this != &rhs) {
      _handle = rhs._handle;
      rhs._handle = fdk::MemoryPool<PXR_NS::SdfAssetPath>::nullhandle;
    }
    return *this;
  }

  AssetPath::AssetPath(const AssetPath& rhs)
    : _handle(wrapper().copyAllocate(rhs._handle))
  {

  }

  AssetPath& AssetPath::operator=(const AssetPath& rhs)
  {
    if (this != &rhs) {
      _handle = wrapper().copyAllocate(rhs._handle);
    }
    return *this;
  }

  bool AssetPath::operator==(const AssetPath &rhs) const
  {
    return unwrap(_handle) == unwrap(rhs._handle);
  }

  bool AssetPath::operator!=(const AssetPath &rhs) const
  {
    return unwrap(_handle) != unwrap(rhs._handle);
  }

  bool AssetPath::operator<(const AssetPath &rhs) const
  {
      return unwrap(_handle) < unwrap(rhs._handle);
  }

  const std::string& AssetPath::assetPath() const
  {
    return unwrap(_handle).GetAssetPath();
  }

  const std::string& AssetPath::resolvedPath() const
  {
    return unwrap(_handle).GetResolvedPath();
  }

  const std::string& AssetPath::asString() const
  {
    return assetPath();
  }

  AssetPath::Handle& AssetPath::getAssetPath(int usd_version)
  {
    return wrapper().getWrappedObject(_handle);
  }

  const AssetPath::Handle& AssetPath::getAssetPath(int usd_version) const
  {
    return wrapper().getWrappedObject(_handle);
  }

  uint64_t AssetPath::hash() const
  {
    return unwrap(_handle).GetHash();
  }

  void AssetPath::append(fdk::Hash& hash) const
  {
    const auto h = unwrap(_handle).GetHash();
    hash.append(&h, sizeof(h));
  }

  USG_API fdk::Hash& operator<<(fdk::Hash& hash, const AssetPath& path)
  {
    path.append(hash);
    return hash;
  }

  USG_API std::ostream& operator<<(std::ostream& stream, const AssetPath& path)
  {
    stream << path.assetPath();
    return stream;
  }
}
