// Copyright (c) 2022 The Foundry Visionmongers Ltd. All Rights Reserved.

#include "usg/geom/Attribute.h"

#include "usg/private/UsdImpl.h"
#include "usg/private/UsdAttrImpl.h"

namespace usg
{

extern std::string empty_string;
extern usg::Token  empty_token;
extern usg::Path   empty_geompath;


/*! Default ctor is invalid.
*/
Attribute::Attribute() :
  _pImpl(nullptr)
{
  //
}


Attribute::Attribute(const Attribute& b) :
  _pImpl((b._pImpl.get() != nullptr) ? new Attribute::Impl(*b._pImpl.get()) : nullptr)
{
  //
}
Attribute& Attribute::operator = (const Attribute& b)
{
  if (_pImpl.get() != b._pImpl.get()) {
    _pImpl.reset(new Attribute::Impl(*b._pImpl.get()));
  }
  return *this;
}


Attribute::Attribute(const Attribute::Impl& b) :
  _pImpl(new Attribute::Impl(b))
{
  //
}
Attribute& Attribute::operator = (const Attribute::Impl& b)
{
  if (_pImpl.get() != &b) {
    _pImpl.reset(new Attribute::Impl(b));
  }
  return *this;
}

Attribute::~Attribute()
{
  //
}

bool
Attribute::setFromImpl(Attribute::Impl&& b)
{
  _pImpl.reset(new Attribute::Impl(std::move(b)));
  return _pImpl->isValid();
}


bool
Attribute::isValid() const
{
  return (_pImpl && _pImpl->isValid());
}


std::ostream& operator << (std::ostream& o, const Attribute& attr)
{
  if (attr._pImpl) {
    fdk::TimeValueList times; attr.getTimeSamples(times);

    o << attr.getTypeName() << " " << attr.getName();
    if (times.size() == 0) {
      Value val; attr.getValue(val, fdk::defaultTimeValue());
      o << " = def(" << val << ")";
    }
    else if (times.size() == 1) {
      Value val; attr.getValue(val, *times.begin());
      o << " = ~0(" << val << ")";
    }
    else {
      o << " = ~(" << times.size() << " samples)";
    }
  }
  else {
    o << "<empty>";
  }

  return o;
}

const std::string& Attribute::getName()                const { return (_pImpl) ? _pImpl->getName() : empty_string; }
Path               Attribute::getPath()                const { return (_pImpl) ? _pImpl->getPath() : empty_geompath; }
Value::Type        Attribute::getType()                const { return (_pImpl) ? _pImpl->valuetype : Value::InvalidType; }
const std::string& Attribute::getTypeName()            const { return Value::getTypeName((_pImpl) ? _pImpl->valuetype : Value::InvalidType); }
size_t             Attribute::getNumDataElements()     const { return (_pImpl) ? Value::getNumDataElements(_pImpl->valuetype) : 0; }
size_t             Attribute::getBytesPerDataElement() const { return (_pImpl) ? Value::getBytesPerDataElement(_pImpl->valuetype) : 0; }
bool               Attribute::isDefined()              const { return (_pImpl) ? _pImpl->isDefined() : false; }
bool               Attribute::isAuthored()             const { return (_pImpl) ? _pImpl->isAuthored() : false; }

size_t Attribute::getSize(const fdk::TimeValue& time) const
{
  return (_pImpl) ? _pImpl->getSize(time) : 0;
}

size_t Attribute::getArraySize(const fdk::TimeValue& time) const
{
  return (_pImpl && _pImpl->isArrayType()) ? _pImpl->getSize(time) : 0;
}


void Attribute::clear() { if (_pImpl) _pImpl->clear(); }
void Attribute::clearDefault() { if (_pImpl) _pImpl->clearDefault(); }
void Attribute::clearAllTimeSamples() { if (_pImpl) _pImpl->clearAllTimeSamples(); }
void Attribute::clearAtTime(const fdk::TimeValue& time) { if (_pImpl) _pImpl->clearAtTime(time); }

void Attribute::block() { if (_pImpl) _pImpl->block(); }

#if 0
const std::type_info &
TfType::GetTypeid() const
{
    std::type_info const *typeInfo = _info->typeInfo;
    return typeInfo ? *typeInfo : typeid(void);
}
#endif


Attribute::AttrSpecHandle* Attribute::getSdfAttributeSpecHandle(int usd_version)
{
  return (_pImpl && usdVersionMatchesUsg(usd_version)) ?
            reinterpret_cast<Attribute::AttrSpecHandle*>(&_pImpl->sdfattr) :
            nullptr;
}
const Attribute::AttrSpecHandle* Attribute::getSdfAttributeSpecHandle(int usd_version) const
{
  return (_pImpl && usdVersionMatchesUsg(usd_version)) ?
            reinterpret_cast<const Attribute::AttrSpecHandle*>(&_pImpl->sdfattr) :
            nullptr;
}

Attribute::UsdAttrHandle* Attribute::getUsdAttribute(int usd_version)
{
  return (_pImpl && usdVersionMatchesUsg(usd_version)) ?
            reinterpret_cast<Attribute::UsdAttrHandle*>(&_pImpl->usdattr) :
            nullptr;
}
const Attribute::UsdAttrHandle* Attribute::getUsdAttribute(int usd_version) const
{
  return (_pImpl && usdVersionMatchesUsg(usd_version)) ?
            reinterpret_cast<const Attribute::UsdAttrHandle*>(&_pImpl->usdattr) :
            nullptr;
}


//----------------------------------------------------------------


void Attribute::getFieldNames(fdk::StringList& field_names) const { if (_pImpl) _pImpl->getFieldNames(field_names); }
bool Attribute::hasField(const Token& field_name) const { return (_pImpl) ? _pImpl->hasField(toTfToken(field_name)) : false; }
bool Attribute::deleteField(const Token& field_name) { return (_pImpl) ? _pImpl->deleteField(toTfToken(field_name)) : false; }

Value Attribute::getField(const Token& field_name) const
{
  return (_pImpl) ? Value(Value::Impl(_pImpl->getField(toTfToken(field_name)))) : Value();
}

bool Attribute::setField(const Token& field_name,
                         const Value& value)
{
  if (!_pImpl || !value._pImpl) {
    return false;
  }
  return _pImpl->setField(toTfToken(field_name), value._pImpl->vtvalue);
}

bool   Attribute::isAnimating() const { return (_pImpl) ? _pImpl->isAnimating() : false; }
size_t Attribute::getNumTimeSamples() const { return (_pImpl) ? _pImpl->getNumTimeSamples() : 0; }

size_t Attribute::getTimeSamples(fdk::TimeValueList& times) const
{
  if (!_pImpl) {
    times.clear();
    return 0;
  }
  return _pImpl->getTimeSamples(times);
}

fdk::TimeValueList Attribute::getTimeSamples() const
{
  fdk::TimeValueList times;
  getTimeSamples(times);
  return times;
}


void Attribute::setInterpolation(const Token& interpolation)
{
  if (_pImpl) {
    _pImpl->setField(PXR_NS::UsdGeomTokens->interpolation, PXR_NS::VtValue(toTfToken(interpolation)));
  }
}

Token Attribute::getInterpolation() const
{
  if (_pImpl) {
    const auto value = _pImpl->getField(PXR_NS::UsdGeomTokens->interpolation);
    if (!value.IsEmpty()) {
      return fromTfToken(value.Get<PXR_NS::TfToken>());
    }
  }
  return Token();
}


//----------------------------------------------------------------


bool Attribute::addConnectionTo(const Path& source) { return (_pImpl) ? _pImpl->addConnectionTo(source) : false; }
bool Attribute::removeConnectionTo(const Path& source) { return (_pImpl) ? _pImpl->removeConnectionTo(source) : false; }
void Attribute::clearConnections() { if (_pImpl) _pImpl->clearConnections(); }

bool
Attribute::setConnectionTo(const Path& source)
{
  if (_pImpl) {
    _pImpl->clearConnections();
    return _pImpl->addConnectionTo(source);
  }
  return false;
}

size_t Attribute::numConnections() const
{
// TODO ndk14: implement!!!
  return 0;
}

Path Attribute::getConnection(int i) const
{
// TODO ndk14: implement!!!
  return empty_string;
}


//----------------------------------------------------------------


// Default template assumes in-memory representation of the types match!
template <typename usgT, typename usdT>
inline void getValue(const usg::Value& usg_value,
                     usgT&             v,
                     const usgT&       d)
{
  if (usg_value.isValid()) {
    const PXR_NS::VtValue& vtvalue = usg_value.impl()->vtvalue;
    if (vtvalue.IsHolding<usdT>() && !vtvalue.IsEmpty()) {
      v = reinterpret_cast<const usgT&>(vtvalue.UncheckedGet<usdT>());
      return;
    }
  }
  v = d;
}
// Specialization for TfToken conversion to usg::Token:
inline void getTokenValue(const usg::Value&  usg_value,
                          usg::Token&        v,
                          const usg::Token&  d)
{
  if (usg_value.isValid()) {
    const PXR_NS::VtValue& vtvalue = usg_value.impl()->vtvalue;
    if (vtvalue.IsHolding<PXR_NS::TfToken>() && !vtvalue.IsEmpty()) {
      v = fromTfToken(vtvalue.UncheckedGet<PXR_NS::TfToken>());
      return;
    }
  }
  v = d;
}
// Specialization for SdfPath conversion to usg::Path:
inline void getPathValue(const usg::Value& usg_value,
                         usg::Path&        v,
                         const usg::Path&  d)
{
  if (usg_value.isValid()) {
    const PXR_NS::VtValue& vtvalue = usg_value.impl()->vtvalue;
    if (vtvalue.IsHolding<PXR_NS::SdfPath>() && !vtvalue.IsEmpty()) {
      v = fromSdfPath(vtvalue.UncheckedGet<PXR_NS::SdfPath>());
      return;
    }
  }
  v = d;
}
// Specialization for SdfAssetPath conversion to usg::AssetPath:
inline void getAssetPathValue(const usg::Value&     usg_value,
                              usg::AssetPath&       v,
                              const usg::AssetPath& d)
{
  if (usg_value.isValid()) {
    const PXR_NS::VtValue& vtvalue = usg_value.impl()->vtvalue;
    if (vtvalue.IsHolding<PXR_NS::SdfAssetPath>() && !vtvalue.IsEmpty()) {
      v = fromSdfAssetPath(vtvalue.UncheckedGet<PXR_NS::SdfAssetPath>());
      return;
    }
  }
  v = d;
}


// Default template assumes in-memory representation of the types match!
template <typename usgT, typename usdT>
inline void getVtArrayValue(const usg::Value&        usg_value,
                            std::vector<usgT>&       usg_array,
                            const std::vector<usgT>& def_array)
{
  if (usg_value.isValid()) {
    const PXR_NS::VtValue& vtvalue = usg_value.impl()->vtvalue;
    if (vtvalue.IsHolding<PXR_NS::VtArray<usdT> >() && !vtvalue.IsEmpty()) {
      const PXR_NS::VtArray<usdT> vtarray = vtvalue.UncheckedGet<PXR_NS::VtArray<usdT> >();
      usg_array.resize(vtarray.size());
      memcpy(usg_array.data(), vtarray.data(), sizeof(usgT)*usg_array.size());
      return;
    }
  }
  usg_array = def_array;
}

// string specialization to avoid memcpy-ing pointers to managed memory
template <>
inline void getVtArrayValue<std::string, std::string>(
        const usg::Value&               ndk_value,
        std::vector<std::string>&       ndk_array,
        const std::vector<std::string>& def_array)
{
  if (ndk_value.isValid())
  {
    const PXR_NS::VtValue& vtvalue = ndk_value.impl()->vtvalue;
    if (vtvalue.IsHolding<PXR_NS::VtArray<std::string> >() && !vtvalue.IsEmpty())
    {
      const PXR_NS::VtArray<std::string> vtarray = vtvalue.UncheckedGet<PXR_NS::VtArray<std::string> >();
      std::copy(vtarray.begin(), vtarray.end(), std::back_inserter(ndk_array));
      return;
    }
  }
  ndk_array = def_array;
}

// Specialization for TfToken conversion to usg::Token:
inline void getTokenArray(const usg::Value&              usg_value,
                          std::vector<usg::Token>&       usg_array,
                          const std::vector<usg::Token>& def_array)
{
  if (usg_value.isValid()) {
    const PXR_NS::VtValue& vtvalue = usg_value.impl()->vtvalue;
    if (vtvalue.IsHolding<PXR_NS::VtArray<PXR_NS::TfToken> >() && !vtvalue.IsEmpty()) {
      const PXR_NS::VtArray<PXR_NS::TfToken> vtarray = vtvalue.UncheckedGet<PXR_NS::VtArray<PXR_NS::TfToken> >();
      usg_array.resize(vtarray.size());
      for (size_t i=0; i < vtarray.size(); ++i) {
        usg_array[i] = fromTfToken(vtarray[i]);
      }
      return;
    }
  }
  usg_array = def_array;
}


template <> USG_API void Value::get(std::string   & v, const std::string   & d) const { getValue<std::string, std::string>(*this, v, d); }
                    void Value::get(std::string   & v, const char*           d) const { this->get(v, std::string(d)); }
template <> USG_API void Value::get(usg::Token    & v, const usg::Token    & d) const { getTokenValue(*this, v, d); }
template <> USG_API void Value::get(usg::Path     & v, const usg::Path     & d) const { getPathValue(*this, v, d); }
template <> USG_API void Value::get(usg::AssetPath& v, const usg::AssetPath& d) const { getAssetPathValue(*this, v, d); }
//
template <> USG_API void Value::get(bool       & v, const bool       & d) const { getValue<bool       , bool           >(*this, v, d); }
template <> USG_API void Value::get(char       & v, const char       & d) const { getValue<char       , char           >(*this, v, d); }
template <> USG_API void Value::get(fdk::half  & v, const fdk::half  & d) const { getValue<fdk::half  , PXR_NS::GfHalf >(*this, v, d); }
template <> USG_API void Value::get(float      & v, const float      & d) const { getValue<float      , float          >(*this, v, d); }
template <> USG_API void Value::get(double     & v, const double     & d) const { getValue<double     , double         >(*this, v, d); }
template <> USG_API void Value::get(int32_t    & v, const int32_t    & d) const { getValue<int32_t    , int32_t        >(*this, v, d); }
template <> USG_API void Value::get(int64_t    & v, const int64_t    & d) const { getValue<int64_t    , int64_t        >(*this, v, d); }
template <> USG_API void Value::get(uint32_t   & v, const uint32_t   & d) const { getValue<uint32_t   , uint32_t       >(*this, v, d); }
template <> USG_API void Value::get(uint64_t   & v, const uint64_t   & d) const { getValue<uint64_t   , uint64_t       >(*this, v, d); }
//
template <> USG_API void Value::get(fdk::Vec2h & v, const fdk::Vec2h & d) const { getValue<fdk::Vec2h , PXR_NS::GfVec2h   >(*this, v, d); }
template <> USG_API void Value::get(fdk::Vec2f & v, const fdk::Vec2f & d) const { getValue<fdk::Vec2f , PXR_NS::GfVec2f   >(*this, v, d); }
template <> USG_API void Value::get(fdk::Vec2d & v, const fdk::Vec2d & d) const { getValue<fdk::Vec2d , PXR_NS::GfVec2d   >(*this, v, d); }
template <> USG_API void Value::get(fdk::Vec2i & v, const fdk::Vec2i & d) const { getValue<fdk::Vec2i , PXR_NS::GfVec2i   >(*this, v, d); }
//
template <> USG_API void Value::get(fdk::Vec3h & v, const fdk::Vec3h & d) const { getValue<fdk::Vec3h , PXR_NS::GfVec3h   >(*this, v, d); }
template <> USG_API void Value::get(fdk::Vec3f & v, const fdk::Vec3f & d) const { getValue<fdk::Vec3f , PXR_NS::GfVec3f   >(*this, v, d); }
template <> USG_API void Value::get(fdk::Vec3d & v, const fdk::Vec3d & d) const { getValue<fdk::Vec3d , PXR_NS::GfVec3d   >(*this, v, d); }
template <> USG_API void Value::get(fdk::Vec3i & v, const fdk::Vec3i & d) const { getValue<fdk::Vec3i , PXR_NS::GfVec3i   >(*this, v, d); }
//
template <> USG_API void Value::get(fdk::Vec4h & v, const fdk::Vec4h & d) const { getValue<fdk::Vec4h , PXR_NS::GfVec4h   >(*this, v, d); }
template <> USG_API void Value::get(fdk::Vec4f & v, const fdk::Vec4f & d) const { getValue<fdk::Vec4f , PXR_NS::GfVec4f   >(*this, v, d); }
template <> USG_API void Value::get(fdk::Vec4d & v, const fdk::Vec4d & d) const { getValue<fdk::Vec4d , PXR_NS::GfVec4d   >(*this, v, d); }
template <> USG_API void Value::get(fdk::Vec4i & v, const fdk::Vec4i & d) const { getValue<fdk::Vec4i , PXR_NS::GfVec4i   >(*this, v, d); }
//
template <> USG_API void Value::get(fdk::Quath & v, const fdk::Quath & d) const { getValue<fdk::Quath , PXR_NS::GfQuath   >(*this, v, d); }
template <> USG_API void Value::get(fdk::Quatf & v, const fdk::Quatf & d) const { getValue<fdk::Quatf , PXR_NS::GfQuatf   >(*this, v, d); }
template <> USG_API void Value::get(fdk::Quatd & v, const fdk::Quatd & d) const { getValue<fdk::Quatd , PXR_NS::GfQuatd   >(*this, v, d); }
//
#if 0
// Disabling Mat4f version to avoid creating an invalid Value::Type. TODO: make this throw a static_assert()...?
template <> USG_API void Value::get(fdk::Mat4f & v, const fdk::Mat4f & d) const { getValue<fdk::Mat4f , PXR_NS::GfMatrix4f>(*this, v, d); }
#endif
template <> USG_API void Value::get(fdk::Mat4d & v, const fdk::Mat4d & d) const { getValue<fdk::Mat4d , PXR_NS::GfMatrix4d>(*this, v, d); }
//-------------------------------------------
template <> USG_API void Value::get(fdk::StringList& v, const fdk::StringList& d) const { getVtArrayValue<std::string, std::string>(*this, v, d); }
//
template <> USG_API void Value::get(usg::TokenArray & v, const usg::TokenArray & d) const { getTokenArray(*this, v, d); }
//
//template <> USG_API void Value::get(usg::BoolArray  & v, const usg::BoolArray  & d) const { getVtArrayValue<bool     , bool          >(*this, v, d); }
template <> USG_API void Value::get(usg::CharArray  & v, const usg::CharArray  & d) const { getVtArrayValue<char     , char          >(*this, v, d); }
template <> USG_API void Value::get(usg::HalfArray  & v, const usg::HalfArray  & d) const { getVtArrayValue<fdk::half, PXR_NS::GfHalf>(*this, v, d); }
template <> USG_API void Value::get(usg::FloatArray & v, const usg::FloatArray & d) const { getVtArrayValue<float    , float         >(*this, v, d); }
template <> USG_API void Value::get(usg::DoubleArray& v, const usg::DoubleArray& d) const { getVtArrayValue<double   , double        >(*this, v, d); }
template <> USG_API void Value::get(usg::Int32Array & v, const usg::Int32Array & d) const { getVtArrayValue<int32_t  , int32_t       >(*this, v, d); }
template <> USG_API void Value::get(usg::Int64Array & v, const usg::Int64Array & d) const { getVtArrayValue<int64_t  , int64_t       >(*this, v, d); }
template <> USG_API void Value::get(usg::Uint32Array& v, const usg::Uint32Array& d) const { getVtArrayValue<uint32_t , uint32_t      >(*this, v, d); }
template <> USG_API void Value::get(usg::Uint64Array& v, const usg::Uint64Array& d) const { getVtArrayValue<uint64_t , uint64_t      >(*this, v, d); }
//
template <> USG_API void Value::get(usg::Vec2hArray& v, const usg::Vec2hArray& d) const { getVtArrayValue<fdk::Vec2h, PXR_NS::GfVec2h>(*this, v, d); }
template <> USG_API void Value::get(usg::Vec2fArray& v, const usg::Vec2fArray& d) const { getVtArrayValue<fdk::Vec2f, PXR_NS::GfVec2f>(*this, v, d); }
template <> USG_API void Value::get(usg::Vec2dArray& v, const usg::Vec2dArray& d) const { getVtArrayValue<fdk::Vec2d, PXR_NS::GfVec2d>(*this, v, d); }
template <> USG_API void Value::get(usg::Vec2iArray& v, const usg::Vec2iArray& d) const { getVtArrayValue<fdk::Vec2i, PXR_NS::GfVec2i>(*this, v, d); }
//
template <> USG_API void Value::get(usg::Vec3hArray& v, const usg::Vec3hArray& d) const { getVtArrayValue<fdk::Vec3h, PXR_NS::GfVec3h>(*this, v, d); }
template <> USG_API void Value::get(usg::Vec3fArray& v, const usg::Vec3fArray& d) const { getVtArrayValue<fdk::Vec3f, PXR_NS::GfVec3f>(*this, v, d); }
template <> USG_API void Value::get(usg::Vec3dArray& v, const usg::Vec3dArray& d) const { getVtArrayValue<fdk::Vec3d, PXR_NS::GfVec3d>(*this, v, d); }
template <> USG_API void Value::get(usg::Vec3iArray& v, const usg::Vec3iArray& d) const { getVtArrayValue<fdk::Vec3i, PXR_NS::GfVec3i>(*this, v, d); }
//
template <> USG_API void Value::get(usg::Vec4hArray& v, const usg::Vec4hArray& d) const { getVtArrayValue<fdk::Vec4h, PXR_NS::GfVec4h>(*this, v, d); }
template <> USG_API void Value::get(usg::Vec4fArray& v, const usg::Vec4fArray& d) const { getVtArrayValue<fdk::Vec4f, PXR_NS::GfVec4f>(*this, v, d); }
template <> USG_API void Value::get(usg::Vec4dArray& v, const usg::Vec4dArray& d) const { getVtArrayValue<fdk::Vec4d, PXR_NS::GfVec4d>(*this, v, d); }
template <> USG_API void Value::get(usg::Vec4iArray& v, const usg::Vec4iArray& d) const { getVtArrayValue<fdk::Vec4i, PXR_NS::GfVec4i>(*this, v, d); }
//
template <> USG_API void Value::get(usg::QuathArray& v, const usg::QuathArray& d) const { getVtArrayValue<fdk::Quath, PXR_NS::GfQuath>(*this, v, d); }
template <> USG_API void Value::get(usg::QuatfArray& v, const usg::QuatfArray& d) const { getVtArrayValue<fdk::Quatf, PXR_NS::GfQuatf>(*this, v, d); }
template <> USG_API void Value::get(usg::QuatdArray& v, const usg::QuatdArray& d) const { getVtArrayValue<fdk::Quatd, PXR_NS::GfQuatd>(*this, v, d); }
//
template <> USG_API void Value::get(usg::Mat4dArray& v, const usg::Mat4dArray& d) const { getVtArrayValue<fdk::Mat4d , PXR_NS::GfMatrix4d>(*this, v, d); }
//--------------------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------------------


bool
Attribute::setValue(const usg::Value&     srcValue,
                    const fdk::TimeValue& time)
{
  if (_pImpl) {
    return _pImpl->setVtValue(srcValue._pImpl->vtvalue, time);
  }
  else {
    return false;
  }
}

//template <> bool Attribute::setValue<const char*>(const char*&       v) { return (_pImpl) ? _pImpl->set<const char*, PXR_NS::TfToken>(v, t) : false; }
template <> USG_API bool Attribute::setValue<usg::Token    >(const usg::Token    & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<PXR_NS::TfToken     , PXR_NS::TfToken     >(toTfToken(v), t) : false; }
template <> USG_API bool Attribute::setValue<usg::Path     >(const usg::Path     & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<PXR_NS::SdfPath     , PXR_NS::SdfPath     >(toSdfPath(v), t) : false; }
template <> USG_API bool Attribute::setValue<usg::AssetPath>(const usg::AssetPath& v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<PXR_NS::SdfAssetPath, PXR_NS::SdfAssetPath>(toSdfAssetPath(v), t) : false; }
template <> USG_API bool Attribute::setValue<std::string   >(const std::string   & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<std::string         , PXR_NS::TfToken     >(v, t) : false; }
//
template <> USG_API bool Attribute::setValue<bool       >(const bool       & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<bool      , bool           >(v, t) : false; }
template <> USG_API bool Attribute::setValue<char       >(const char       & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<char      , char           >(v, t) : false; }
template <> USG_API bool Attribute::setValue<unsigned char>(const unsigned char & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<unsigned char, unsigned char>(v, t) : false; }
template <> USG_API bool Attribute::setValue<fdk::half  >(const fdk::half  & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<fdk::half , PXR_NS::GfHalf >(v, t) : false; }
template <> USG_API bool Attribute::setValue<float      >(const float      & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<float     , float          >(v, t) : false; }
template <> USG_API bool Attribute::setValue<double     >(const double     & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<double    , double         >(v, t) : false; }
template <> USG_API bool Attribute::setValue<int32_t    >(const int32_t    & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<int32_t   , int32_t        >(v, t) : false; }
template <> USG_API bool Attribute::setValue<int64_t    >(const int64_t    & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<int64_t   , int64_t        >(v, t) : false; }
template <> USG_API bool Attribute::setValue<uint32_t   >(const uint32_t   & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<uint32_t  , uint32_t       >(v, t) : false; }
template <> USG_API bool Attribute::setValue<uint64_t   >(const uint64_t   & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<uint64_t  , uint64_t       >(v, t) : false; }
//
template <> USG_API bool Attribute::setValue<fdk::Vec2h>(const fdk::Vec2h  & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<fdk::Vec2h, PXR_NS::GfVec2h   >(v, t) : false; }
template <> USG_API bool Attribute::setValue<fdk::Vec2f>(const fdk::Vec2f  & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<fdk::Vec2f, PXR_NS::GfVec2f   >(v, t) : false; }
template <> USG_API bool Attribute::setValue<fdk::Vec2d>(const fdk::Vec2d  & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<fdk::Vec2d, PXR_NS::GfVec2d   >(v, t) : false; }
template <> USG_API bool Attribute::setValue<fdk::Vec2i>(const fdk::Vec2i  & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<fdk::Vec2i, PXR_NS::GfVec2i   >(v, t) : false; }
//
template <> USG_API bool Attribute::setValue<fdk::Vec3h>(const fdk::Vec3h  & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<fdk::Vec3h, PXR_NS::GfVec3h   >(v, t) : false; }
template <> USG_API bool Attribute::setValue<fdk::Vec3f>(const fdk::Vec3f  & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<fdk::Vec3f, PXR_NS::GfVec3f   >(v, t) : false; }
template <> USG_API bool Attribute::setValue<fdk::Vec3d>(const fdk::Vec3d  & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<fdk::Vec3d, PXR_NS::GfVec3d   >(v, t) : false; }
template <> USG_API bool Attribute::setValue<fdk::Vec3i>(const fdk::Vec3i  & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<fdk::Vec3i, PXR_NS::GfVec3i   >(v, t) : false; }
//
template <> USG_API bool Attribute::setValue<fdk::Vec4h>(const fdk::Vec4h  & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<fdk::Vec4h, PXR_NS::GfVec4h   >(v, t) : false; }
template <> USG_API bool Attribute::setValue<fdk::Vec4f>(const fdk::Vec4f  & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<fdk::Vec4f, PXR_NS::GfVec4f   >(v, t) : false; }
template <> USG_API bool Attribute::setValue<fdk::Vec4d>(const fdk::Vec4d  & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<fdk::Vec4d, PXR_NS::GfVec4d   >(v, t) : false; }
template <> USG_API bool Attribute::setValue<fdk::Vec4i>(const fdk::Vec4i  & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<fdk::Vec4i, PXR_NS::GfVec4i   >(v, t) : false; }
//
template <> USG_API bool Attribute::setValue<fdk::Quath>(const fdk::Quath  & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<fdk::Quath, PXR_NS::GfQuath   >(v, t) : false; }
template <> USG_API bool Attribute::setValue<fdk::Quatf>(const fdk::Quatf  & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<fdk::Quatf, PXR_NS::GfQuatf   >(v, t) : false; }
template <> USG_API bool Attribute::setValue<fdk::Quatd>(const fdk::Quatd  & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<fdk::Quatd, PXR_NS::GfQuatd   >(v, t) : false; }
//
template <> USG_API bool Attribute::setValue<fdk::Mat4d>(const fdk::Mat4d  & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->set<fdk::Mat4d, PXR_NS::GfMatrix4d>(v, t) : false; }
//
template <> USG_API bool Attribute::setValue<fdk::StringList>(const fdk::StringList& v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->setArray<std::string, std::string>(v, t) : false; }
//
template <> USG_API bool Attribute::setValue<usg::TokenArray >(const usg::TokenArray & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->setArray<usg::Token , PXR_NS::TfToken>(v, t) : false; }
template <> USG_API bool Attribute::setValue<usg::CharArray  >(const usg::CharArray  & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->setArray<char       , char           >(v, t) : false; }
template <> USG_API bool Attribute::setValue<usg::UCharArray >(const usg::UCharArray& v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->setArray<unsigned char, unsigned char>(v, t) : false; }
template <> USG_API bool Attribute::setValue<usg::HalfArray  >(const usg::HalfArray  & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->setArray<fdk::half  , PXR_NS::GfHalf >(v, t) : false; }
template <> USG_API bool Attribute::setValue<usg::FloatArray >(const usg::FloatArray & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->setArray<float      , float          >(v, t) : false; }
template <> USG_API bool Attribute::setValue<usg::DoubleArray>(const usg::DoubleArray& v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->setArray<double     , double         >(v, t) : false; }
template <> USG_API bool Attribute::setValue<usg::Int32Array >(const usg::Int32Array & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->setArray<int32_t    , int32_t        >(v, t) : false; }
template <> USG_API bool Attribute::setValue<usg::Int64Array >(const usg::Int64Array & v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->setArray<int64_t    , int64_t        >(v, t) : false; }
template <> USG_API bool Attribute::setValue<usg::Uint32Array>(const usg::Uint32Array& v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->setArray<uint32_t, uint32_t>(v, t) : false; }
template <> USG_API bool Attribute::setValue<usg::Uint64Array>(const usg::Uint64Array& v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->setArray<uint64_t, uint64_t>(v, t) : false; }
//
template <> USG_API bool Attribute::setValue<usg::Vec2hArray>(const usg::Vec2hArray& v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->setArray<fdk::Vec2h, PXR_NS::GfVec2h>(v, t) : false; }
template <> USG_API bool Attribute::setValue<usg::Vec2fArray>(const usg::Vec2fArray& v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->setArray<fdk::Vec2f, PXR_NS::GfVec2f>(v, t) : false; }
template <> USG_API bool Attribute::setValue<usg::Vec2dArray>(const usg::Vec2dArray& v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->setArray<fdk::Vec2d, PXR_NS::GfVec2d>(v, t) : false; }
template <> USG_API bool Attribute::setValue<usg::Vec2iArray>(const usg::Vec2iArray& v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->setArray<fdk::Vec2i, PXR_NS::GfVec2i>(v, t) : false; }
//
template <> USG_API bool Attribute::setValue<usg::Vec3hArray>(const usg::Vec3hArray& v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->setArray<fdk::Vec3h, PXR_NS::GfVec3h>(v, t) : false; }
template <> USG_API bool Attribute::setValue<usg::Vec3fArray>(const usg::Vec3fArray& v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->setArray<fdk::Vec3f, PXR_NS::GfVec3f>(v, t) : false; }
template <> USG_API bool Attribute::setValue<usg::Vec3dArray>(const usg::Vec3dArray& v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->setArray<fdk::Vec3d, PXR_NS::GfVec3d>(v, t) : false; }
template <> USG_API bool Attribute::setValue<usg::Vec3iArray>(const usg::Vec3iArray& v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->setArray<fdk::Vec3i, PXR_NS::GfVec3i>(v, t) : false; }
//
template <> USG_API bool Attribute::setValue<usg::Vec4hArray>(const usg::Vec4hArray& v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->setArray<fdk::Vec4h, PXR_NS::GfVec4h>(v, t) : false; }
template <> USG_API bool Attribute::setValue<usg::Vec4fArray>(const usg::Vec4fArray& v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->setArray<fdk::Vec4f, PXR_NS::GfVec4f>(v, t) : false; }
template <> USG_API bool Attribute::setValue<usg::Vec4dArray>(const usg::Vec4dArray& v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->setArray<fdk::Vec4d, PXR_NS::GfVec4d>(v, t) : false; }
template <> USG_API bool Attribute::setValue<usg::Vec4iArray>(const usg::Vec4iArray& v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->setArray<fdk::Vec4i, PXR_NS::GfVec4i>(v, t) : false; }
//
template <> USG_API bool Attribute::setValue<usg::QuathArray>(const usg::QuathArray& v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->setArray<fdk::Quath, PXR_NS::GfQuath>(v, t) : false; }
template <> USG_API bool Attribute::setValue<usg::QuatfArray>(const usg::QuatfArray& v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->setArray<fdk::Quatf, PXR_NS::GfQuatf>(v, t) : false; }
template <> USG_API bool Attribute::setValue<usg::QuatdArray>(const usg::QuatdArray& v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->setArray<fdk::Quatd, PXR_NS::GfQuatd>(v, t) : false; }
//
template <> USG_API bool Attribute::setValue<usg::Mat4dArray>(const usg::Mat4dArray& v, const fdk::TimeValue& t) { return (_pImpl) ? _pImpl->setArray<fdk::Mat4d, PXR_NS::GfMatrix4d>(v, t) : false; }


//--------------------------------------------------------------------------------------------------


size_t
Attribute::getValue(usg::Value&           dstValue,
                    const fdk::TimeValue& time) const
{
  if (_pImpl) {
    PXR_NS::VtValue vt_value;
    if (_pImpl->getVtValue(vt_value, time)) {
      if (dstValue.setFromImpl(Value::Impl(vt_value))) {
        return dstValue._pImpl->getSize();
      }
    }
  }
  return 0;
}

template <> USG_API size_t Attribute::getValue<usg::Token    >(usg::Token    & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<usg::Token    , PXR_NS::TfToken     >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<usg::Path     >(usg::Path     & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<usg::Path     , PXR_NS::SdfPath     >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<usg::AssetPath>(usg::AssetPath& v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<usg::AssetPath, PXR_NS::SdfAssetPath>(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<std::string   >(std::string   & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<std::string   , PXR_NS::TfToken     >(v, t) : 0; }
//
template <> USG_API size_t Attribute::getValue<bool       >(bool        & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<bool      , bool           >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<char       >(char        & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<char      , char           >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<unsigned char>(unsigned char & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<unsigned char, unsigned char>(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<fdk::half  >(fdk::half   & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<fdk::half , PXR_NS::GfHalf >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<float      >(float       & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<float     , float          >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<double     >(double      & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<double    , double         >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<int32_t    >(int32_t     & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<int32_t   , int32_t        >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<int64_t    >(int64_t     & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<int64_t   , int64_t        >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<uint32_t   >(uint32_t    & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<uint32_t  , uint32_t       >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<uint64_t   >(uint64_t    & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<uint64_t  , uint64_t       >(v, t) : 0; }
//
template <> USG_API size_t Attribute::getValue<fdk::Vec2h >(fdk::Vec2h  & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<fdk::Vec2h, PXR_NS::GfVec2h >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<fdk::Vec2f >(fdk::Vec2f  & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<fdk::Vec2f, PXR_NS::GfVec2f >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<fdk::Vec2d >(fdk::Vec2d  & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<fdk::Vec2d, PXR_NS::GfVec2d >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<fdk::Vec2i >(fdk::Vec2i  & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<fdk::Vec2i, PXR_NS::GfVec2i >(v, t) : 0; }
//
template <> USG_API size_t Attribute::getValue<fdk::Vec3h >(fdk::Vec3h  & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<fdk::Vec3h, PXR_NS::GfVec3h >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<fdk::Vec3f >(fdk::Vec3f  & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<fdk::Vec3f, PXR_NS::GfVec3f >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<fdk::Vec3d >(fdk::Vec3d  & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<fdk::Vec3d, PXR_NS::GfVec3d >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<fdk::Vec3i >(fdk::Vec3i  & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<fdk::Vec3i, PXR_NS::GfVec3i >(v, t) : 0; }
//
template <> USG_API size_t Attribute::getValue<fdk::Vec4h >(fdk::Vec4h  & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<fdk::Vec4h, PXR_NS::GfVec4h >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<fdk::Vec4f >(fdk::Vec4f  & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<fdk::Vec4f, PXR_NS::GfVec4f >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<fdk::Vec4d >(fdk::Vec4d  & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<fdk::Vec4d, PXR_NS::GfVec4d >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<fdk::Vec4i >(fdk::Vec4i  & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<fdk::Vec4i, PXR_NS::GfVec4i >(v, t) : 0; }
//
template <> USG_API size_t Attribute::getValue<fdk::Quath >(fdk::Quath  & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<fdk::Quath, PXR_NS::GfQuath >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<fdk::Quatf >(fdk::Quatf  & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<fdk::Quatf, PXR_NS::GfQuatf >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<fdk::Quatd >(fdk::Quatd  & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<fdk::Quatd, PXR_NS::GfQuatd >(v, t) : 0; }
//
#if 0
// Disabling Mat4f version to avoid creating an invalid Value::Type. TODO: make this throw a static_assert()...?
template <> USG_API size_t Attribute::getValue<fdk::Mat4f >(fdk::Mat4f  & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<fdk::Mat4f, PXR_NS::GfMatrix4f>(v, t) : 0; }
#endif
template <> USG_API size_t Attribute::getValue<fdk::Mat4d >(fdk::Mat4d  & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->get<fdk::Mat4d, PXR_NS::GfMatrix4d>(v, t) : 0; }
//-------------------------------------------
template <> USG_API size_t Attribute::getValue<fdk::StringList>(fdk::StringList& v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->getArray<std::string, std::string>(v, t) : 0; }
//
template <> USG_API size_t Attribute::getValue<usg::TokenArray >(usg::TokenArray & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->getArray<usg::Token, PXR_NS::TfToken>(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<usg::CharArray  >(usg::CharArray  & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->getArray<char      , char           >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<usg::UCharArray >(usg::UCharArray & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->getArray<unsigned char, unsigned char>(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<usg::HalfArray  >(usg::HalfArray  & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->getArray<fdk::half , PXR_NS::GfHalf >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<usg::FloatArray >(usg::FloatArray & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->getArray<float     , float          >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<usg::DoubleArray>(usg::DoubleArray& v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->getArray<double    , double         >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<usg::Int32Array >(usg::Int32Array & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->getArray<int32_t   , int32_t        >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<usg::Int64Array >(usg::Int64Array & v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->getArray<int64_t   , int64_t        >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<usg::Uint32Array>(usg::Uint32Array& v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->getArray<uint32_t  , uint32_t       >(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<usg::Uint64Array>(usg::Uint64Array& v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->getArray<uint64_t  , uint64_t       >(v, t) : 0; }
//
template <> USG_API size_t Attribute::getValue<usg::Vec2hArray>(usg::Vec2hArray& v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->getArray<fdk::Vec2h, PXR_NS::GfVec2h>(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<usg::Vec2fArray>(usg::Vec2fArray& v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->getArray<fdk::Vec2f, PXR_NS::GfVec2f>(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<usg::Vec2dArray>(usg::Vec2dArray& v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->getArray<fdk::Vec2d, PXR_NS::GfVec2d>(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<usg::Vec2iArray>(usg::Vec2iArray& v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->getArray<fdk::Vec2i, PXR_NS::GfVec2i>(v, t) : 0; }
//
template <> USG_API size_t Attribute::getValue<usg::Vec3hArray>(usg::Vec3hArray& v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->getArray<fdk::Vec3h, PXR_NS::GfVec3h>(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<usg::Vec3fArray>(usg::Vec3fArray& v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->getArray<fdk::Vec3f, PXR_NS::GfVec3f>(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<usg::Vec3dArray>(usg::Vec3dArray& v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->getArray<fdk::Vec3d, PXR_NS::GfVec3d>(v, t) : 0;}
template <> USG_API size_t Attribute::getValue<usg::Vec3iArray>(usg::Vec3iArray& v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->getArray<fdk::Vec3i, PXR_NS::GfVec3i>(v, t) : 0; }
//
template <> USG_API size_t Attribute::getValue<usg::Vec4hArray>(usg::Vec4hArray& v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->getArray<fdk::Vec4h, PXR_NS::GfVec4h>(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<usg::Vec4fArray>(usg::Vec4fArray& v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->getArray<fdk::Vec4f, PXR_NS::GfVec4f>(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<usg::Vec4dArray>(usg::Vec4dArray& v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->getArray<fdk::Vec4d, PXR_NS::GfVec4d>(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<usg::Vec4iArray>(usg::Vec4iArray& v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->getArray<fdk::Vec4i, PXR_NS::GfVec4i>(v, t) : 0; }
//
template <> USG_API size_t Attribute::getValue<usg::QuathArray>(usg::QuathArray& v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->getArray<fdk::Quath, PXR_NS::GfQuath>(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<usg::QuatfArray>(usg::QuatfArray& v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->getArray<fdk::Quatf, PXR_NS::GfQuatf>(v, t) : 0; }
template <> USG_API size_t Attribute::getValue<usg::QuatdArray>(usg::QuatdArray& v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->getArray<fdk::Quatd, PXR_NS::GfQuatd>(v, t) : 0; }
//
template <> USG_API size_t Attribute::getValue<usg::Mat4dArray>(usg::Mat4dArray& v, const fdk::TimeValue& t) const { return (_pImpl) ? _pImpl->getArray<fdk::Mat4d, PXR_NS::GfMatrix4d>(v, t) : 0; }


} // namespace usg
