// Copyright (c) 2022 The Foundry Visionmongers Ltd. All Rights Reserved.

#include "usg/geom/Layer.h"

#include "usg/private/UsdImpl.h"

#include "pxr/usd/sdf/layerUtils.h"

namespace usg {

Layer::Layer(const Layer::Impl& b)
  : _pImpl(new Layer::Impl(b))
{
}

Layer::~Layer()
{
}


LayerRef Layer::Create(const std::string& identifier, const fdk::KeyValueMapSorted& file_args)
{
  if (!identifier.empty()) {
    auto sdfLayer = PXR_NS::SdfLayer::CreateNew(identifier, file_args);
    assert(sdfLayer);
    sdfLayer->SetOwner(identifier);
    return std::make_shared<Layer>(Layer::Impl(sdfLayer));
  }
  return LayerRef();
}

LayerRef Layer::CreateAnonymous(const std::string& identifier, const fdk::KeyValueMapSorted& file_args)
{
  auto sdfLayer = PXR_NS::SdfLayer::CreateAnonymous(identifier, file_args);
  assert(sdfLayer);
  sdfLayer->SetOwner(identifier);
  return std::make_shared<Layer>(Layer::Impl(sdfLayer));
}

LayerRef Layer::Find(const std::string&            identifier,
                     const fdk::KeyValueMapSorted& file_args)
{
  if (identifier.empty()) {
    return LayerRef();
  }

  PXR_NS::SdfLayerRefPtr sdflayer = PXR_NS::SdfLayer::Find(identifier, file_args);
  if (!sdflayer) {
    return LayerRef();
  }

  return std::make_shared<Layer>(Layer::Impl(sdflayer));
}

LayerRef Layer::FindOrOpen(const std::string&            identifier,
                           const fdk::KeyValueMapSorted& file_args)
{
  if (identifier.empty()) {
    return LayerRef();
  }

  PXR_NS::SdfLayerRefPtr sdflayer = PXR_NS::SdfLayer::FindOrOpen(identifier, file_args);
  if (!sdflayer) {
    return LayerRef();
  }

  return std::make_shared<Layer>(Layer::Impl(sdflayer));
}

LayerRef Layer::FindRelativeToLayer(const LayerRef&               anchor,
                                    const std::string&            identifier,
                                    const fdk::KeyValueMapSorted& file_args)
{
  if (identifier.empty() || !anchor->isValid()) {
    return LayerRef();
  }

  PXR_NS::SdfLayerRefPtr sdflayer = PXR_NS::SdfLayer::FindRelativeToLayer(anchor->_pImpl->layer, identifier, file_args);
  if (!sdflayer) {
    return LayerRef();
  }

  return std::make_shared<Layer>(Layer::Impl(sdflayer));
}

LayerRef Layer::FindOrOpenRelativeToLayer(const LayerRef&               anchor,
                                          const std::string&            identifier,
                                          const fdk::KeyValueMapSorted& file_args)
{
  if (identifier.empty() || !anchor->isValid()) {
    return LayerRef();
  }

  PXR_NS::SdfLayerRefPtr sdflayer = PXR_NS::SdfLayer::FindOrOpenRelativeToLayer(anchor->_pImpl->layer, identifier, file_args);
  if (!sdflayer) {
    return LayerRef();
  }

  return std::make_shared<Layer>(Layer::Impl(sdflayer));
}

LayerRef Layer::Create(const Layer::Impl& handle)
{
  return std::make_shared<Layer>(handle);
}


bool
Layer::isValid() const
{
  return (_pImpl && _pImpl->layer);
}

Layer::Handle* Layer::getSdfLayerRefPtr(int usd_version)
{
  return (_pImpl && usdVersionMatchesUsg(usd_version)) ?
            reinterpret_cast<Layer::Handle*>(&_pImpl->layer) :
            nullptr;
}

const Layer::Handle* Layer::getSdfLayerRefPtr(int usd_version) const
{
  return (_pImpl && usdVersionMatchesUsg(usd_version)) ?
            reinterpret_cast<const Layer::Handle*>(&_pImpl->layer) :
            nullptr;
}

extern std::string empty_string;

const std::string& Layer::getIdentifier() const { return _pImpl ? _pImpl->layer->GetIdentifier() : empty_string; }

void Layer::setIdentifier(const std::string& identifier)
{
  if (_pImpl && _pImpl->layer) {
    _pImpl->layer->SetIdentifier(identifier);
  }
}


void
Layer::clear()
{
  if (!_pImpl || (_pImpl->layer && _pImpl->layer->IsEmpty()))
    return; // already empty

  _pImpl->layer->Clear();
}


bool
Layer::isEmpty() const
{
  return (!isValid() || _pImpl->layer->IsEmpty());
}


void
Layer::reload()
{
  if (isValid()) {
    _pImpl->layer->Reload();
  }
}


void
Layer::combineSubLayers(const LayerRef& b)
{
  if (!isValid() || !b || !b->isValid()) {
    return;
  }

  fdk::StringList addPaths; addPaths.reserve(50);

  // Note that SdfSubLayerProxy underlying type is a std::string. Look
  // at pxr/usd/sdf/proxyPolicies.h:SdfSubLayerTypePolicy.
  PXR_NS::SdfSubLayerProxy sublayers(b->_pImpl->layer->GetSubLayerPaths());
  const size_t nLayers = sublayers.size();
  for (size_t i = 0; i < nLayers; ++i) {
    const std::string& path = sublayers[i];
    if (std::find(addPaths.begin(), addPaths.end(), path) != addPaths.end()) {
      addPaths.insert(addPaths.end(), path);
    }
  }

  if (!addPaths.empty()) {
    _pImpl->layer->SetSubLayerPaths(addPaths);
  }
}


size_t
Layer::getSubLayers(fdk::StringList& sublayerIdentifiers) const
{
  sublayerIdentifiers.clear();
  if (!isValid()) {
    return 0;
  }

  // Note that SdfSubLayerProxy underlying type is a std::string. Look
  // at pxr/usd/sdf/proxyPolicies.h:SdfSubLayerTypePolicy.
  PXR_NS::SdfSubLayerProxy sublayers(_pImpl->layer->GetSubLayerPaths());
  const size_t nLayers = sublayers.size();
  sublayerIdentifiers.reserve(nLayers);
  for (size_t i = 0; i < nLayers; ++i) {
    sublayerIdentifiers.push_back(sublayers[i]);
  }
  return nLayers;
}


void
Layer::setSubLayers(const fdk::StringList& sublayerIdentifiers)
{
  if (sublayerIdentifiers.empty() || !isValid()) {
    return;
  }
  // Note that SdfSubLayerProxy underlying type is a std::string. Look
  // at pxr/usd/sdf/proxyPolicies.h:SdfSubLayerTypePolicy.
  _pImpl->layer->SetSubLayerPaths(sublayerIdentifiers);
}

void
Layer::setSubLayers(const LayerRef& b)
{
  if (!isValid() || !b || !b->isValid()) {
    return;
  }

  // Copy all the paths from the other layer:
  if (b->getNumSubLayers() > 0) {
    fdk::StringList sublayers;
    b->getSubLayers(sublayers);
    // Note that SdfSubLayerProxy underlying type is a std::string. Look
    // at pxr/usd/sdf/proxyPolicies.h:SdfSubLayerTypePolicy.
    _pImpl->layer->SetSubLayerPaths(sublayers);
  }
}


void
Layer::insertSubLayer(const std::string& sublayerIdentifier,
                      int                index)
{
  if (!sublayerIdentifier.empty() && isValid()) {
    _pImpl->layer->InsertSubLayerPath(sublayerIdentifier, index);
  }
}


void
Layer::insertSubLayer(const LayerRef& sublayer,
                      int          index)
{
  if (!sublayer || !sublayer->isValid()) {
    return;
  }
  insertSubLayer(sublayer->_pImpl->layer->GetIdentifier(), index);
}


size_t
Layer::getNumSubLayers() const
{
  return (isValid()) ? _pImpl->layer->GetNumSubLayerPaths() : 0;
}


void
Layer::changeBlockPush()
{
  if (_pImpl && _pImpl->layer) {
    const int cnt = _pImpl->blockCount++;
    // First push creates block:
    if (cnt == 0) {
      _pImpl->changeBlock = new PXR_NS::SdfChangeBlock();
    }
  }
}

void
Layer::changeBlockPop()
{
  if (_pImpl && _pImpl->layer) {
    const int cnt = --_pImpl->blockCount;
    // Last pop deletes block:
    if (cnt == 0) {
      delete _pImpl->changeBlock;
    }
  }
}


//--------------------------------------------------------------------------------------------------

size_t
Layer::getTimeSamples(std::set<fdk::TimeValue>& times) const
{
  if (isValid()) {
    times = _pImpl->layer->ListAllTimeSamples();
  }
  else {
    times.clear();
  }
  return times.size();
}


//--------------------------------------------------------------------------------------------------


Prim
Layer::definePrim(const Path&  path,
                  const Token& typeName)
{
  if (!isValid() || path.empty()) {
    return Prim();
  }

  PXR_NS::SdfPrimSpecHandle primspec = PXR_NS::SdfCreatePrimInLayer(_pImpl->layer, toSdfPath(path));
  if (!primspec) {
    return Prim();
  }

  primspec->SetSpecifier(PXR_NS::SdfSpecifierDef);
  if (!typeName.empty()) {
    primspec->SetTypeName(typeName.asString());
  }

  return Prim(Prim::Impl(primspec));
}


Prim
Layer::overridePrim(const Path&  path,
                    const Token& typeName)
{
  if (!isValid() || path.empty()) {
    return Prim();
  }

  PXR_NS::SdfPrimSpecHandle primspec = PXR_NS::SdfCreatePrimInLayer(_pImpl->layer, toSdfPath(path));
  if (!primspec) {
    return Prim();
  }

  primspec->SetSpecifier(PXR_NS::SdfSpecifierOver);
  primspec->SetTypeName(typeName.asString());

  return Prim(Prim::Impl(primspec));
}


Prim
Layer::overridePrim(const Prim& src_prim)
{
  if (!isValid() || !src_prim) {
    return Prim();
  }

  PXR_NS::SdfPrimSpecHandle primspec = PXR_NS::SdfCreatePrimInLayer(_pImpl->layer, toSdfPath(src_prim.getPath()));
  if (!primspec) {
    return Prim();
  }

  primspec->SetSpecifier(PXR_NS::SdfSpecifierOver);
  primspec->SetTypeName(src_prim.getTypeName().asString());

  return Prim(Prim::Impl(primspec));
}


Prim
Layer::getRootPrim() const
{
  return (isValid()) ? Prim(Prim::Impl(_pImpl->layer->GetPseudoRoot())) : Prim();
}

Prim
Layer::getPrimAtPath(const Path& path) const
{
  if (!isValid()) {
    return Prim();
  }

  PXR_NS::SdfPrimSpecHandle primspec = _pImpl->layer->GetPrimAtPath(toSdfPath(path));
  return (primspec) ? Prim(Prim::Impl(primspec)) : Prim();
}

static std::string scope_string("Scope");

Prim
Layer::defineScopeParents(const Path&  path,
                          const Token& last_prim_schema)
{
  if (path.empty() || path == Path::AbsoluteRootPath()) {
    return Prim();
  }

  Prim prim = getPrimAtPath(path);
  if (!prim) {
    definePrim(path, last_prim_schema);
  }

  Path parentPath(path.parent());
  for (; parentPath != Path::AbsoluteRootPath(); parentPath = parentPath.parent()) {
    Prim parentPrim = getPrimAtPath(parentPath);
    if (!parentPrim) {
      definePrim(parentPath, scope_string);
    }
  }

  return prim;
}


bool
Layer::hasDefaultPrim() const
{
  return (_pImpl && _pImpl->layer && _pImpl->layer->HasDefaultPrim());
}

Token
Layer::defaultPrim() const
{
  return (_pImpl && _pImpl->layer) ? fromTfToken(_pImpl->layer->GetDefaultPrim()) : Token();
}

void
Layer::setDefaultPrim(const Token& name)
{
  if (_pImpl && _pImpl->layer) {
    _pImpl->layer->SetDefaultPrim(toTfToken(name));
  }
}

void
Layer::clearDefaultPrim()
{
  if (_pImpl && _pImpl->layer)
    _pImpl->layer->ClearDefaultPrim();
}

void Layer::setCustomLayerData(const Token& key, const Value& value)
{
  if (isValid()) {
    auto data = _pImpl->layer->GetCustomLayerData();
    data[toTfToken(key)] = value.impl()->vtvalue;
    _pImpl->layer->SetCustomLayerData(data);
  }
}

Value Layer::getCustomLayerData(const Token& key) const
{
  if (isValid()) {
    auto data = _pImpl->layer->GetCustomLayerData();
    return Value(Value::Impl(data[toTfToken(key)]));
  }
  return Value();
}

void Layer::setModifiedPaths(const PathArray& paths)
{
  if (isValid()) {
    PXR_NS::VtStringArray sdfPaths;
    std::transform(paths.begin(), paths.end(),
      std::back_inserter(sdfPaths), [](const Path& path) { return path.asString(); });

    auto data = _pImpl->layer->GetCustomLayerData();
    data["usg:modified"] = sdfPaths;
    _pImpl->layer->SetCustomLayerData(data);
  }
}

PathArray Layer::getModifiedPaths() const
{
  PathArray paths;
  if (isValid()) {
    auto data = _pImpl->layer->GetCustomLayerData();
    const auto it = data.find("usg:modified");
    if (it != data.end()) {
      const auto& value = it->second;
      if (value.IsHolding<PXR_NS::VtStringArray>()) {
        const auto& sdfPaths = value.Get<PXR_NS::VtStringArray>();
        std::transform(sdfPaths.begin(), sdfPaths.end(),
          std::back_inserter(paths), [](const std::string& path) { return Path(path); });
      }
    }
  }
  return paths;
}

void Layer::setCreatedPaths(const PathArray& paths)
{
  if (isValid()) {
    PXR_NS::VtArray<PXR_NS::SdfPath> sdfPaths;
    std::transform(paths.begin(), paths.end(),
      std::back_inserter(sdfPaths), [](const Path& path) { return toSdfPath(path); });

    auto data = _pImpl->layer->GetCustomLayerData();
    data["usg:created"] = sdfPaths;
    _pImpl->layer->SetCustomLayerData(data);
  }
}

PathArray Layer::getCreatedPaths() const
{
  PathArray paths;
  if (isValid()) {
    auto data = _pImpl->layer->GetCustomLayerData();
    const auto it = data.find("usg:created");
    if (it != data.end()) {
      const auto& value = it->second;
      if (value.IsHolding<PXR_NS::SdfPathVector>()) {
        const auto& sdfPaths = value.Get<PXR_NS::SdfPathVector>();
        std::transform(sdfPaths.begin(), sdfPaths.end(),
          std::back_inserter(paths), [](const PXR_NS::SdfPath& path) { return fromSdfPath(path); });
      }
    }
  }
  return paths;
}

void Layer::setDefaultParentPath(const Path& path)
{
  if (isValid()) {
    auto data = _pImpl->layer->GetCustomLayerData();
    data["usg:parent"] = path.asString();
    _pImpl->layer->SetCustomLayerData(data);
  }
}

Path Layer::getDefaultParentPath() const
{
  if (isValid()) {
    auto data = _pImpl->layer->GetCustomLayerData();
    const auto it = data.find("usg:parent");
    if (it != data.end()) {
      const auto& value = it->second;
      if (value.IsHolding<std::string>()) {
        const auto& path = value.Get<std::string>();
        return Path(path);
      }
    }
  }
  return Path();
}

void Layer::setOwnerNodeName(const std::string& owner)
{
  if (isValid()) {
    auto data = _pImpl->layer->GetCustomLayerData();
    data["usg:owner"] = owner;
    _pImpl->layer->SetCustomLayerData(data);
  }
}

std::string Layer::getOwnerNodeName() const
{
  if (isValid()) {
    auto data = _pImpl->layer->GetCustomLayerData();
    const auto it = data.find("usg:owner");
    if (it != data.end()) {
      const auto& value = it->second;
      if (value.IsHolding<std::string>()) {
        return value.Get<std::string>();
      }
    }
  }
  return std::string();
}

size_t Layer::getCompositionAssetDependencies(fdk::StringList& assetIdentifiers) const
{
  assetIdentifiers.clear();
  if (!isValid()) {
    return 0;
  }

  std::set<std::string> assetSet = _pImpl->layer->GetCompositionAssetDependencies();
  assetIdentifiers.assign(assetSet.begin(), assetSet.end());
  return assetIdentifiers.size();
}

Token Layer::getUpAxis() const
{
  if (isValid()) {
    return fromTfToken(_pImpl->layer->GetPseudoRoot()->GetInfo(PXR_NS::UsdGeomTokens->upAxis).Get<PXR_NS::TfToken>());
  }
  return GeomTokens.Y;
}

//--------------------------------------------------------------------------------------------------

extern std::string g_preedit_cache_id;
extern std::string g_preedit_cache_hash;


void
Layer::setPreEditCacheMetadata(uint64_t cache_hash)
{
  if (!isValid())
    return; // don't crash...

  PXR_NS::VtDictionary dict;
  dict[g_preedit_cache_id  ] = PXR_NS::VtValue(getIdentifier());
  dict[g_preedit_cache_hash] = PXR_NS::VtValue(cache_hash);

  _pImpl->layer->SetCustomLayerData(dict);
}


bool
Layer::getPreEditCacheMetadata(std::string& cache_layer_id,
                               uint64_t&    cache_layer_hash)
{
  if (!isValid()) {
    return false; // don't crash...
  }

  const PXR_NS::VtDictionary dict = _pImpl->layer->GetCustomLayerData();
  return getPreEditCacheInfoFromDict(dict,
                                     cache_layer_id,
                                     cache_layer_hash);
}



size_t
Layer::getPrimPaths(PathArray& prim_paths) const
{
  prim_paths.clear();
  prim_paths.reserve(50);
  const PXR_NS::SdfLayerHandle& layer = _pImpl->layer;
  layer->Traverse(PXR_NS::SdfPath::AbsoluteRootPath(),
                  [&prim_paths, &layer](const PXR_NS::SdfPath& path)
                  {
                    if (path.IsPrimPath()) {
                      //std::cout << "  " << prim_paths.size() << ": layer prim '" << path << "'" << std::endl;
                      if (prim_paths.size() >= prim_paths.capacity())
                        prim_paths.reserve(prim_paths.capacity() * 2);
                      prim_paths.push_back(path.GetString());
                    }
                  });
  return prim_paths.size();
}


size_t
Layer::getAttrPaths(PathArray& attr_paths) const
{
  attr_paths.clear();
  attr_paths.reserve(50);
  const PXR_NS::SdfLayerHandle& layer = _pImpl->layer;
  layer->Traverse(PXR_NS::SdfPath::AbsoluteRootPath(),
                  [&attr_paths, &layer](const PXR_NS::SdfPath& path)
                  {
                    if (path.IsPrimPropertyPath()) {
                      //std::cout << "  " << attr_paths.size() << ": layer prop '" << path << "'" << std::endl;
                      if (attr_paths.size() >= attr_paths.capacity())
                        attr_paths.reserve(attr_paths.capacity() * 2);
                      attr_paths.push_back(path.GetString());
                    }
                  });
  return attr_paths.size();
}


bool Layer::isAnonymous() const
{
  return _pImpl->layer->IsAnonymous();
}

std::string Layer::getRealPath() const
{
  return _pImpl->layer->GetRealPath();
}

bool Layer::save(bool force)
{
  return _pImpl->layer->Save(force);
}

bool Layer::exportToFile(const std::string& filepath,
                         const std::string& comment,
                         const std::map<std::string, std::string>& fileFormatArgs) const
{
  return _pImpl && _pImpl->layer && _pImpl->layer->Export(filepath, comment, fileFormatArgs);
}

bool Layer::exportToString(std::string& text) const
{
  return _pImpl->layer->ExportToString(&text);
}

bool Layer::importFromString(const std::string& text) const
{
  return _pImpl->layer->ImportFromString(text);
}

void Layer::transferContent(const LayerRef& layer)
{
  _pImpl->layer->TransferContent(layer->_pImpl->layer);
}

void Layer::printContents(const char* prefix, std::ostream& o) const
{
  std::string s;
  _pImpl->layer->ExportToString(&s);
  if (prefix) {
    o << prefix << s;
  }
  else {
    o << s;
  }
}


struct SummaryStats
{
  size_t numSpecs = 0;
  size_t numPrimSpecs = 0;
  size_t numPropertySpecs = 0;
  size_t numFields = 0;
  size_t numSampleTimes = 0;
};

/*! Print out Layer components to a stream.
*/
std::ostream&
operator << (std::ostream& o, const Layer& geomLayer)
{
  const PXR_NS::SdfLayerHandle& layer = geomLayer._pImpl->layer;
  o << "Layer['" << layer->GetIdentifier() << "']";
  o << ", owner='" << layer->GetOwner() << "'";

  SummaryStats stats;
  layer->Traverse(PXR_NS::SdfPath::AbsoluteRootPath(),
                  [&stats, &layer](const PXR_NS::SdfPath& path)
                  {
                    //std::cout << "  '" << path << "'" << std::endl;
                    ++stats.numSpecs;
                    stats.numPrimSpecs     += path.IsPrimPath();
                    //stats.numPropertySpecs += path.IsPropertyPath();
                    //stats.numFields        += layer->ListFields(path).size();
                  });
  //stats.numSampleTimes = layer->ListAllTimeSamples().size();

  o << ", paths=" << stats.numSpecs;
  o << ", prims=" << stats.numPrimSpecs;

  PXR_NS::SdfSubLayerProxy sublayers(layer->GetSubLayerPaths());
  o << ", sublayers=" << sublayers.size();
  return o;
}


} // namespace usg

