// Copyright (c) 2022 The Foundry Visionmongers Ltd. All Rights Reserved.

#include "usg/geom/LightPrim.h"

#include "usg/lux/LuxTokens.h"
#include "usg/private/UsdImpl.h"

#if PXR_VERSION >= 2111
#include "pxr/usd/usdLux/lightAPI.h"
#else
#include "pxr/usd/usdLux/light.h"
#endif

//------------------------------------------------------------------------------------


const std::string gEnvDebugShaderBuilding    = ((std::getenv("USG_DEBUG_SHADER_BUILDING") != nullptr) ?
                                                 std::getenv("USG_DEBUG_SHADER_BUILDING") : "");
const std::string gEnvDebugShaderConnections = ((std::getenv("USG_DEBUG_SHADER_CONNECTIONS") != nullptr) ?
                                                 std::getenv("USG_DEBUG_SHADER_CONNECTIONS") : "");
const bool gDebugShaderBuilding    = (gEnvDebugShaderBuilding    == "1");
const bool gDebugShaderConnections = (gEnvDebugShaderConnections == "1");


//------------------------------------------------------------------------------------

using namespace usg::lux;

namespace usg
{

LightPrim::LightPrim() :
  XformablePrim()
{
  //
}

LightPrim::LightPrim(const Prim& prim) :
  XformablePrim(prim)
{
  //
}

LightPrim::LightPrim(const LightPrim& prim) :
  XformablePrim(prim)
{
  //
}

LightPrim::LightPrim(const Prim::Impl& handle) :
  XformablePrim(handle)
{
  //
}


/*!
*/
/*virtual*/
LightPrim::~LightPrim()
{
  //
}


bool
LightPrim::validate()
{
  if (!_pImpl.get())
    return false; // not valid...

  if (!isValidPrim(*this))
  {
    _pImpl.reset();
    return false;
  }
  return true; // still valid!
}


/*static*/ bool
LightPrim::isValidPrim(const Prim& prim)
{
  return prim.hasApi(LuxTokens.types.LightAPI)
      || prim.isA(LuxTokens.types.DomeLight)
      || prim.isA(LuxTokens.types.DistantLight)
      || prim.isA(LuxTokens.types.GeometryLight)
      || prim.isA(LuxTokens.types.CylinderLight)
      || prim.isA(LuxTokens.types.DiskLight)
      || prim.isA(LuxTokens.types.PortalLight)
      || prim.isA(LuxTokens.types.RectLight)
      || prim.isA(LuxTokens.types.SphereLight);
}

/*
UsdLuxLight:
  pxr/usd/usdLux/cylinderLight.h      class UsdLuxCylinderLight
  pxr/usd/usdLux/diskLight.h          class UsdLuxDiskLight
  pxr/usd/usdLux/distantLight.h:      class UsdLuxDistantLight
  pxr/usd/usdLux/domeLight.h:         class UsdLuxDomeLight
  pxr/usd/usdLux/geometryLight.h:     class UsdLuxGeometryLight
  pxr/usd/usdLux/pluginLight.h:       class UsdLuxPluginLight
  pxr/usd/usdLux/portalLight.h:       class UsdLuxPortalLight
  pxr/usd/usdLux/rectLight.h:         class UsdLuxRectLight
  pxr/usd/usdLux/sphereLight.h:       class UsdLuxSphereLight

UsdLuxLightFilter:
  pxr/usd/usdLux/pluginLightFilter.h: class UsdLuxPluginLightFilter
*/


//--------------------------------------------------------------------------------------------------


ShaderDesc*
LightPrim::createShaderDesc(const fdk::TimeValue& time) const
{
  if (!isValid()) {
    return nullptr;
  }

  ShaderDesc* shader_desc = new ShaderDesc(Schema::find(getTypeName().text()));
  if (gDebugShaderBuilding) {
    std::cout << "  LightPrim::createShaderDesc(" << this << "): schema=" << getTypeName() << std::endl;
  }
  if (!shader_desc) {
    return nullptr;
  }

  //shader_desc->print("    ", std::cout, true/*print_sdrnode_info*/);

  // Set the relative path of the shader within the parent group:
  //shader_desc->setMaterialRelativeName(makeRelativePath(parent_material.getPath(), getPath()).c_str());

  // Copy overrides and connections from the UsdAttributes:
  for (const auto attr : getAttributes()) {
    const Attribute prop = getAttr(attr);
    if (!prop.isValid()) {
      continue;
    }

    std::string prefix, name;
    fdk::splitPropertyName(prop.getName(), prefix, name);

    if (gDebugShaderBuilding) {
      std::cout << "    prop '" << name << "' = " << prop;
      std::cout << ", connected=" << prop.impl()->usdattr.HasAuthoredConnections();
    }

    if (prefix == "inputs") {
      if (gDebugShaderBuilding) {
        std::cout << " INPUT:" << std::endl;
      }

      if (prop.impl()->isUsd()) {
        if (prop.impl()->usdattr.HasAuthoredConnections()) {
          PXR_NS::SdfPathVector sources;
          prop.impl()->usdattr.GetConnections(&sources);
          if (sources.size() > 0) {
            if (gDebugShaderBuilding) {
              std::cout << "                connect='" << sources[0] << "'" << std::endl;
            }
            shader_desc->connectInput(name.c_str(), sources[0].GetString());
          }
        }
        else if (prop.impl()->usdattr.HasAuthoredValue()) {
          Value val;
          if (prop.getType() == Value::AssetPath) {
            PXR_NS::SdfAssetPath path; prop.impl()->usdattr.Get<PXR_NS::SdfAssetPath>(&path);
            val = Value(path.GetResolvedPath());
          }
          else {
            prop.getValue(val, time);
          }

          if (gDebugShaderBuilding) {
            std::cout << "                val=(" << val << ")" << std::endl;
          }
          shader_desc->overrideInput(name, val, time);
        }
      }
      else {
        // TODO: do we need an Sdf implementation here? Normally we're always
        // dealing with a UsdStage at this point...
      }
    }
    else if (prefix == "outputs") {
      if (gDebugShaderBuilding) {
        std::cout << " OUTPUT:" << std::endl;
      }

      if (prop.impl()->isUsd()) {

      }
      else {
        // TODO: do we need an Sdf implementation here? Normally we're always
        // dealing with a UsdStage at this point...
      }
    }
    else {
      // Local prop:
      if (gDebugShaderBuilding) {
        std::cout << " LOCAL:" << std::endl;
      }

      if (prop.impl()->isUsd()) {
        if (prop.impl()->usdattr.HasAuthoredValue()) {
          Value val;
          if (prop.getType() == Value::AssetPath) {
            PXR_NS::SdfAssetPath path; prop.impl()->usdattr.Get<PXR_NS::SdfAssetPath>(&path);
            val = Value(path.GetResolvedPath());
          }
          else {
            prop.getValue(val, time);
          }

          if (gDebugShaderBuilding) {
            std::cout << "                val=(" << val << ")" << std::endl;
          }
          shader_desc->overrideLocalProperty(prop.getName(), val, time);
        }
      }
      else {
        // TODO: do we need an Sdf implementation here? Normally we're always
        // dealing with a UsdStage at this point...
      }
    }
  }

  return shader_desc;
}

Relationship LightPrim::createFiltersRel() {
  return createRelationship(LuxTokens.filters);
}
Relationship LightPrim::getFiltersRel() const {
  return getRelationship(LuxTokens.filters);
}
void LightPrim::setFilters(const PathArray& value, fdk::TimeValue time) {
  createRelationship(LuxTokens.filters).setTargets(value);
}
PathArray LightPrim::getFilters(fdk::TimeValue time) const {
  return getRelationship(LuxTokens.filters).getTargets();
}

Attribute LightPrim::createColorAttr(const Value& defaultValue) {
  return createAttr(LuxTokens.inputsColor, Value::Color3f, false, TimeVariability::Varying, defaultValue);
}
Attribute LightPrim::getColorAttr() const {
  return getAttr(LuxTokens.inputsColor);
}
void LightPrim::setColor(const fdk::Vec3f& value, fdk::TimeValue time) {
  createColorAttr().setValue(value, time);
}
fdk::Vec3f LightPrim::getColor(fdk::TimeValue time) const {
  fdk::Vec3f v; getAttr(LuxTokens.inputsColor).getValue<fdk::Vec3f>(v, time); return v;
}

Attribute LightPrim::createColorTemperatureAttr(const Value& defaultValue) {
  return createAttr(LuxTokens.inputsColorTemperature, Value::Float, false, TimeVariability::Varying, defaultValue);
}
Attribute LightPrim::getColorTemperatureAttr() const {
  return getAttr(LuxTokens.inputsColorTemperature);
}
void LightPrim::setColorTemperature(float value, fdk::TimeValue time) {
  createColorTemperatureAttr().setValue(value, time);
}
float LightPrim::getColorTemperature(fdk::TimeValue time) const {
  float v; getAttr(LuxTokens.inputsColorTemperature).getValue<float>(v, time); return v;
}

Attribute LightPrim::createDiffuseAttr(const Value& defaultValue) {
  return createAttr(LuxTokens.inputsDiffuse, Value::Float, false, TimeVariability::Varying, defaultValue);
}
Attribute LightPrim::getDiffuseAttr() const {
  return getAttr(LuxTokens.inputsDiffuse);
}
void LightPrim::setDiffuse(float value, fdk::TimeValue time) {
  createDiffuseAttr().setValue(value, time);
}
float LightPrim::getDiffuse(fdk::TimeValue time) const {
  float v; getAttr(LuxTokens.inputsDiffuse).getValue<float>(v, time); return v;
}

Attribute LightPrim::createEnableColorTemperatureAttr(const Value& defaultValue) {
  return createAttr(LuxTokens.inputsEnableColorTemperature, Value::Bool, false, TimeVariability::Varying, defaultValue);
}
Attribute LightPrim::getEnableColorTemperatureAttr() const {
  return getAttr(LuxTokens.inputsEnableColorTemperature);
}
void LightPrim::setEnableColorTemperature(bool value, fdk::TimeValue time) {
  createEnableColorTemperatureAttr().setValue(value, time);
}
bool LightPrim::getEnableColorTemperature(fdk::TimeValue time) const {
  bool v; getAttr(LuxTokens.inputsEnableColorTemperature).getValue<bool>(v, time); return v;
}

Attribute LightPrim::createExposureAttr(const Value& defaultValue) {
  return createAttr(LuxTokens.inputsExposure, Value::Float, false, TimeVariability::Varying, defaultValue);
}
Attribute LightPrim::getExposureAttr() const {
  return getAttr(LuxTokens.inputsExposure);
}
void LightPrim::setExposure(float value, fdk::TimeValue time) {
  createExposureAttr().setValue(value, time);
}
float LightPrim::getExposure(fdk::TimeValue time) const {
  float v; getAttr(LuxTokens.inputsExposure).getValue<float>(v, time); return v;
}

Attribute LightPrim::createIntensityAttr(const Value& defaultValue) {
  return createAttr(LuxTokens.inputsIntensity, Value::Float, false, TimeVariability::Varying, defaultValue);
}
Attribute LightPrim::getIntensityAttr() const {
  return getAttr(LuxTokens.inputsIntensity);
}
void LightPrim::setIntensity(float value, fdk::TimeValue time) {
  createIntensityAttr().setValue(value, time);
}
float LightPrim::getIntensity(fdk::TimeValue time) const {
  float v; getAttr(LuxTokens.inputsIntensity).getValue<float>(v, time); return v;
}

Attribute LightPrim::createNormalizeAttr(const Value& defaultValue) {
  return createAttr(LuxTokens.inputsNormalize, Value::Bool, false, TimeVariability::Varying, defaultValue);
}
Attribute LightPrim::getNormalizeAttr() const {
  return getAttr(LuxTokens.inputsNormalize);
}
void LightPrim::setNormalize(bool value, fdk::TimeValue time) {
  createNormalizeAttr().setValue(value, time);
}
bool LightPrim::getNormalize(fdk::TimeValue time) const {
  bool v; getAttr(LuxTokens.inputsNormalize).getValue<bool>(v, time); return v;
}

Attribute LightPrim::createSpecularAttr(const Value& defaultValue) {
    return createAttr(LuxTokens.inputsSpecular, Value::Float, false, TimeVariability::Varying, defaultValue);
}
Attribute LightPrim::getSpecularAttr() const {
    return getAttr(LuxTokens.inputsSpecular);
}
void LightPrim::setSpecular(float value, fdk::TimeValue time) {
    createSpecularAttr().setValue(value, time);
}
float LightPrim::getSpecular(fdk::TimeValue time) const {
    float v; getAttr(LuxTokens.inputsSpecular).getValue<float>(v, time); return v;
}



#if 0
// TODO: Move these back to LightPrim header and implement the ones we want to keep

    //----------------------------------


//    UsdRelationship GetFiltersRel() const;

    /// See GetFiltersRel(), and also
    /// \ref Usd_Create_Or_Get_Property for when to use Get vs Create
//    UsdRelationship CreateFiltersRel() const;

#endif


} // namespace usg
