// Copyright (c) 2022 The Foundry Visionmongers Ltd. All Rights Reserved.

#include "usg/geom/MaterialPrim.h"
#include "usg/geom/Layer.h"
#include "usg/geom/Stage.h"

#include "usg/private/UsdImpl.h"

#include "pxr/usd/usdShade/material.h"

//------------------------------------------------------------------------------------


const std::string gEnvDebugShaderBuilding    = ((std::getenv("USG_DEBUG_SHADER_BUILDING") != nullptr) ?
                                                 std::getenv("USG_DEBUG_SHADER_BUILDING") : "");
const std::string gEnvDebugShaderConnections = ((std::getenv("USG_DEBUG_SHADER_CONNECTIONS") != nullptr) ?
                                                 std::getenv("USG_DEBUG_SHADER_CONNECTIONS") : "");
const bool gDebugShaderBuilding    = (gEnvDebugShaderBuilding    == "1");
const bool gDebugShaderConnections = (gEnvDebugShaderConnections == "1");


//------------------------------------------------------------------------------------


namespace usg
{


MaterialPrim::MaterialPrim() :
  Prim()
{
  //
}

MaterialPrim::MaterialPrim(const Prim& b) :
  Prim(b)
{
  validate();
}

MaterialPrim::MaterialPrim(const MaterialPrim& b) :
  Prim(b)
{
  //std::cout << "MaterialPrim::ctor-shaderprimcopy(" << this << ")" << std::endl;
}

MaterialPrim&
MaterialPrim::operator = (const MaterialPrim& b)
{
  if (_pImpl.get() != b._pImpl.get())
  {
    _pImpl.reset(new Prim::Impl(*b._pImpl.get()));
  }
  return *this;
}

MaterialPrim::MaterialPrim(const Prim::Impl& b) :
  Prim(b)
{
  validate();
}


/*!
*/
/*virtual*/
MaterialPrim::~MaterialPrim()
{
  //
}


bool
MaterialPrim::validate()
{
  if (!_pImpl.get())
    return false; // not valid...

  if (!isValidPrim(*this))
  {
    _pImpl.reset();
    return false;
  }
  return true; // still valid!
}


/*static*/ bool
MaterialPrim::isValidPrim(const Prim& prim)
{
  if (!prim)
    return false;
  else if (prim.impl()->sdfprim)
    return (prim.impl()->sdfprim->GetTypeName() == PXR_NS::TfToken("Material")); // Sdf Material schema
  else if (prim.impl()->usdprim)
    return prim.impl()->usdprim.IsA<PXR_NS::UsdShadeMaterial>(); // UsdShadeMaterial
  return false; // not a valid handle
}


//--------------------------------------------------------------------------------------------------


/*static*/ MaterialPrim
MaterialPrim::defineInLayer(const LayerRef& layer,
                            const Path& path)
{
  return (layer) ? MaterialPrim(layer->definePrim(path, "Material")) : MaterialPrim();
}


/*static*/ MaterialPrim
MaterialPrim::getInStage(const StageConstRef& stage,
                         const Path&  path)
{
  if (!stage || !stage->isValid() || path.empty())
    return MaterialPrim();

  PXR_NS::UsdPrim prim = stage->impl()->stage->GetPrimAtPath(toSdfPath(path));
  if (!prim.IsValid() || !prim.IsA<PXR_NS::UsdShadeMaterial>())
    return MaterialPrim(); // not aUsdShadeMaterial

  return MaterialPrim(Prim::Impl(prim));
}


//--------------------------------------------------------------------------------------------------


bool
MaterialPrim::hasSurfaceOutput(const Token& render_target) const
{
  if (!_pImpl || !_pImpl->usdprim)
  {
    return false;
  }
  const PXR_NS::UsdShadeOutput terminal = PXR_NS::UsdShadeMaterial(_pImpl->usdprim).GetSurfaceOutput(toTfToken(render_target));
  return (terminal && !terminal.IsDefined());
}

bool
MaterialPrim::hasDisplacementOutput(const Token& render_target) const
{
  if (!_pImpl || !_pImpl->usdprim)
  {
    return false;
  }
  const PXR_NS::UsdShadeOutput terminal = PXR_NS::UsdShadeMaterial(_pImpl->usdprim).GetDisplacementOutput(toTfToken(render_target));
  return (terminal && !terminal.IsDefined());
}

bool
MaterialPrim::hasVolumeOutput(const Token& render_target) const
{
  if (!_pImpl || !_pImpl->usdprim)
  {
    return false;
  }
  const PXR_NS::UsdShadeOutput terminal = PXR_NS::UsdShadeMaterial(_pImpl->usdprim).GetVolumeOutput(toTfToken(render_target));
  return (terminal && !terminal.IsDefined());
}


//--------------------------------------------------------------------------------------------------


struct _ShaderDescConn
{
  std::string input_name;
  ShaderDesc* source_shader;
  std::string source_output_name;

  _ShaderDescConn(const std::string& in,
                  ShaderDesc*        s,
                  const std::string& out) : input_name(in), source_shader(s), source_output_name(out) {}
};
using _ShaderDescConnArray = std::vector<_ShaderDescConn>;

struct _ShaderDescOverride
{
  std::string  input_name;
  PXR_NS::VtValue value;

  _ShaderDescOverride(const std::string&  in,
                      const PXR_NS::VtValue& val) : input_name(in), value(val) {}
};
using _ShaderDescOverrideArray = std::vector<_ShaderDescOverride>;
 
 
/*  Convert a single ShaderPrim and recurse up its input connections.

    Generated ShaderDescs are added to shader_desc_group_out and the source ShaderPrim
    name noted in the 'converted' set so that multiple connections to the same ShaderPrim
    are handled only once.
*/
ShaderDesc*
convertShaderPrim(const MaterialPrim&        parent_material,
                  PXR_NS::UsdPrim&           shader_prim,
                  const Path&                shader_relative_name,
                  std::set<PXR_NS::UsdPrim>& converted_shader_prims,
                  ShaderDescGroup&           shader_desc_group_out,
                  const fdk::TimeValue&      time)
{
  if (!shader_prim.IsValid())
    return nullptr;

  // Find all input attributes with valid connections and recurse up those:
  std::vector<PXR_NS::UsdAttribute> attrs = shader_prim.GetAttributes();

  _ShaderDescConnArray     shader_desc_connections; shader_desc_connections.reserve(20);
  _ShaderDescOverrideArray shader_desc_overrides; shader_desc_overrides.reserve(20);

  bool ok = true;
  PXR_NS::UsdShadeConnectableAPI source;
  PXR_NS::TfToken                sourceName;
  PXR_NS::UsdShadeAttributeType  sourceType;
  for (size_t attr_index=0; attr_index < attrs.size(); ++attr_index)
  {
    const PXR_NS::UsdAttribute& attr = attrs[attr_index];
    if (!PXR_NS::TfStringStartsWith(attr.GetName().GetString(), PXR_NS::UsdShadeTokens->inputs))
      continue;

    // Just the name without the leading 'inputs:':
    const std::string input_name = attr.GetBaseName().GetString();

    // Walk up connection if there is one:
    if (PXR_NS::UsdShadeConnectableAPI::GetConnectedSource(attr, 
                                                        &source,
                                                        &sourceName,
                                                        &sourceType))
    {
      PXR_NS::UsdPrim input_shader_prim = PXR_NS::UsdShadeShader(source).GetPrim();

      // Build the material-relative path which is used to retrieve ShaderDesc from the group:
      const auto input_shader_relative_name =
        fromSdfPath(input_shader_prim.GetPath()).makeRelativePath(parent_material.getPath());

      //std::cout << "    connect '" << input_name << "' to '" << input_shader_relative_name << ":" << sourceName << "'" << std::endl;

      ShaderDesc* input_shader_desc = nullptr;

      // Already converted?
      const std::set<PXR_NS::UsdPrim>::const_iterator it = converted_shader_prims.find(input_shader_prim);
      if (it != converted_shader_prims.end())
      {
        // converted already, retrieve it from the output group - needs the path of the
        // ShaderPrim relative to the MaterialPrim:
        input_shader_desc = shader_desc_group_out.getShaderNode(input_shader_relative_name.asString());
        //std::cout << "        prim already converted, retrieving '" << input_shader_relative_name << "'=" << input_shader_desc << std::endl;
      }
      else
      {
        // not converted yet:
        input_shader_desc = convertShaderPrim(parent_material,
                                              input_shader_prim,
                                              input_shader_relative_name,
                                              converted_shader_prims,
                                              shader_desc_group_out,
                                              time);
        converted_shader_prims.insert(input_shader_prim);
        //std::cout << "        NEW prim '" << input_shader_relative_name << "'=" << input_shader_desc << std::endl;
      }

      if (!input_shader_desc)
      {
std::cerr << "    ...failed to convert input shader prim '" << shader_prim.GetName() << "'" << std::endl;
        ok = false;
        // TODO: should we stop conversion...?
        //break;
      }

      // Success, make the connection - note this does not use the Connections list
      // from the Attribute class, it's only a single connection from the GetConnectedSource()
      // call above:
      // TODO: make sure we're not going to more trouble handling lists of connections per property for shaders!
      shader_desc_connections.push_back(_ShaderDescConn(input_name,
                                                        input_shader_desc,
                                                        sourceName.GetString()));
    }
    else
    {
      // No connection, get the override value:
      PXR_NS::VtValue val; attr.Get(&val, time);
      shader_desc_overrides.push_back(_ShaderDescOverride(input_name, val));
    }

  }

  if (!ok)
  {
    // TODO: create an error shader here?
    return nullptr;
  }

  if (gDebugShaderBuilding)
  {
    std::cout << "  MaterialPrim::convertShaderPrim('" << shader_relative_name << "')" << std::endl;
  }

  // Create a ShaderDesc from the UsdShader using its identifier (if it has one):
  const SchemaDescription* shader_schema = ShaderPrim::retrieveShaderSchema(Prim(Prim::Impl(shader_prim)));
  if (!shader_schema)
  {
    // TODO: return an error or warning that the info:id was not converted to a ShaderSchema ...?
    //
    // TODO: this likely should be a fatal error for the entire material...how to communicate
    // and show this? I think other systems put a bright Magenta color or stripey pattern
    // on objects that have material assignment or binding failures.
    //
    if (gDebugShaderBuilding)
    {
      std::cout << "    MaterialPrim::convertShaderPrim(" << shader_relative_name << "): error, shader schema id ";
      std::cout << " was not found in ShaderSchema registry, cannot create ShaderDesc..." << std::endl;
    }
    return nullptr; // shader doesn't have a schema to construct with
  }

  ShaderDesc* shader_desc = new ShaderDesc(shader_schema);
  if (gDebugShaderBuilding)
    std::cout << "    schema-id='" << shader_schema->registryIdentifier() << "'" << std::endl;

  // Set the relative path of the shader within the parent group:
  shader_desc->setMaterialRelativeName(shader_relative_name.asString());

  // Connect the prior noted connections:
  for (size_t i=0; i < shader_desc_connections.size(); ++i)
  {
    const _ShaderDescConn& conn = shader_desc_connections[i];
    if (shader_desc->connectInput(conn.input_name, conn.source_shader, conn.source_output_name))
    {
      if (gDebugShaderBuilding)
      {
        std::cout << "      connected input '" << conn.input_name << "' to ";
        std::cout << "'" << conn.source_shader->getMaterialRelativeName();
        std::cout << ".outputs:" << conn.source_output_name << "'" << std::endl;
      }
    }
  }

  // Assign any overrides:
  for (size_t i=0; i < shader_desc_overrides.size(); ++i)
  {
    const _ShaderDescOverride& over = shader_desc_overrides[i];
    if (gDebugShaderBuilding)
      std::cout << "      override input '" << over.input_name << "' = (" << Value(Value::Impl(over.value)) << ")" << std::endl;
    shader_desc->overrideInput(over.input_name, Value(Value::Impl(over.value)), time);
  }

  shader_desc_group_out.addShaderDesc(shader_desc);
  converted_shader_prims.insert(shader_prim);
  if (gDebugShaderBuilding)
    shader_desc->print("    ", std::cout);

  return shader_desc;
}


size_t
MaterialPrim::getShaderDescGroup(const Token&          render_target,
                                 const std::string&    terminal_type,
                                 const fdk::TimeValue& time,
                                 ShaderDescGroup&      shader_desc_group_out) const
{
  if (gDebugShaderBuilding)
  {
    std::cout << "===============================================================================================" << std::endl;
    std::cout << "MaterialPrim::getShaderDescGroup() '" << getPath() << "'";
    std::cout << " render_target='" << render_target << "'";
    std::cout << ", terminal_type='" << terminal_type << "', time=" << time << std::endl;
  }

  // Clear out and initialize the group:
  shader_desc_group_out.deleteShaders();
  shader_desc_group_out.material_base_name.clear();
  shader_desc_group_out.material_prim_path.clear();

  // Only support UsdShade Prims in this method:
  if (!_pImpl || !_pImpl->usdprim)
  {
    // TODO: print an invalid Material warning...?
    return 0;
  }

  shader_desc_group_out.material_base_name = getName();
  shader_desc_group_out.material_prim_path = getPath();

  PXR_NS::UsdPrim output_shader_prim;

  if (terminal_type == "surface")
  {
    // Get the surface output port - if it's not available we bail:
    const PXR_NS::UsdShadeOutput surface_terminal =
      PXR_NS::UsdShadeMaterial(_pImpl->usdprim).GetSurfaceOutput(toTfToken(render_target));
    if (!surface_terminal || !surface_terminal.IsDefined())
    {
      // Port not found, bail:
      if (gDebugShaderBuilding)
        std::cout << "  no surface port for render target '" << render_target << "' found on material, skipping." << std::endl;
      return 0;
    }

    if (gDebugShaderBuilding)
      std::cout << "  found surface port for render target '" << render_target << "' on material, extracting ShaderDescGroup..." << std::endl;

    //if (ignoreBaseMaterial && PXR_NS::UsdShadeConnectableAPI::IsSourceConnectionFromBaseMaterial(surface_terminal))
    //  return 0;

    // Get output port on output shader:
    PXR_NS::UsdShadeConnectableAPI source;
    PXR_NS::TfToken                sourceName;
    PXR_NS::UsdShadeAttributeType  sourceType;
    if (!PXR_NS::UsdShadeConnectableAPI::GetConnectedSource(surface_terminal, 
                                                         &source,
                                                         &sourceName,
                                                         &sourceType))
    {
      // TODO: print a shader construction warning...? 'Output does not exist, cannot connect Material'
      if (gDebugShaderBuilding)
        std::cout << "  error, output property '" << surface_terminal.GetFullName() << "' does not exist, cannot connect Material!" << std::endl;
      return 0;
    }

    output_shader_prim = PXR_NS::UsdShadeShader(source).GetPrim();
    if (gDebugShaderBuilding)
      std::cout << "  material output prim='" << output_shader_prim.GetName() << "', sourceConnectionName='" << sourceName << "'" << std::endl;
  }
  else if (terminal_type == "displacement")
  {
    // TODO: support displacement terminal!
  }
  else if (terminal_type == "volume")
  {
    // TODO: support volume terminal!
  }
  else
  {
    // TODO: support other terminals besides surface, displacement & volume...?
    // TODO: print an unsupported terminal type warning...?
    return 0;
  }

  if (!output_shader_prim.IsValid())
  {
    // TODO: print a failed to bind output prim connection?
    return 0;
  }

  const std::string& output_prim_name = output_shader_prim.GetName().GetString();

  shader_desc_group_out.shaders.reserve(20);

  // Build the material-relative path which is used to retrieve ShaderDesc from the group:
  const auto output_shader_relative_name =
    fromSdfPath(output_shader_prim.GetPath()).makeRelativePath(getPath());

  std::set<PXR_NS::UsdPrim> converted_shader_prims;
  shader_desc_group_out.output = convertShaderPrim(*this,
                                                   output_shader_prim,
                                                   output_shader_relative_name,
                                                   converted_shader_prims,
                                                   shader_desc_group_out,
                                                   time);

  if (!shader_desc_group_out.output)
  {
    // TODO: raise a conversion warning? Throw an error?
    shader_desc_group_out.deleteShaders();
    return 0;
  }

  return shader_desc_group_out.size();
}


#if 0

// TODO: Move these back to MaterialPrim header and implement the ones we want to keep

  public:

    //Attribute  getSurfaceAttr()      const;
    //Attribute  getDisplacementAttr() const;
    //Attribute  getVolumeAttr()       const;

    //UsdShadeOutput              getSurfaceOutput(const TfToken& renderTarget = UsdShadeTokens->universalRenderContext) const;
    //std::vector<UsdShadeOutput> getSurfaceOutputs() const;
    //UsdShadeShader              computeSurfaceSource(const TfToken& renderTarget = UsdShadeTokens->universalRenderContext,
    //                                                 TfToken* sourceName = nullptr, 
    //                                                 UsdShadeAttributeType* sourceType = nullptr) const;

    //UsdShadeOutput              getDisplacementOutput(const TfToken& renderTarget = UsdShadeTokens->universalRenderContext) const;
    //std::vector<UsdShadeOutput> getDisplacementOutputs() const;
    //UsdShadeShader              computeDisplacementSource(const TfToken& renderTarget = UsdShadeTokens->universalRenderContext,
    //                                                      TfToken* sourceName = nullptr, 
    //                                                      UsdShadeAttributeType* sourceType = nullptr) const;

    //UsdShadeOutput              getVolumeOutput(const TfToken& renderTarget = UsdShadeTokens->universalRenderContext) const;
    //std::vector<UsdShadeOutput> getVolumeOutputs() const;
    //UsdShadeShader              computeVolumeSource(const TfToken& renderTarget = UsdShadeTokens->universalRenderContext,
    //                                                TfToken* sourceName = nullptr, 
    //                                                UsdShadeAttributeType* sourceType = nullptr) const;


  public:
    //ShaderDescMap             shaders;      //!< SdfPath, ShaderDesc
    //ShaderConnectionMap       terminals;    //!< TfToken, ShaderConnection
    //std::vector<std::string>  primvars;     //!< TfToken

#endif

} // namespace usg
