// Copyright (c) 2022 The Foundry Visionmongers Ltd. All Rights Reserved.

#include "usg/geom/Path.h"

#include "usg/private/MemoryPoolWrapperImpl.h"
#include "usg/private/UsdImpl.h"
#include "fdk/base/Hash.h"

#include "pxr/usd/sdf/path.h"

#include <iostream>

namespace usg {

  // Pseudo-root, or '/':
  Path absolute_root_geompath = fromSdfPath(PXR_NS::SdfPath::AbsoluteRootPath());

  namespace {
    MemoryPoolWrapperImpl<PXR_NS::SdfPath, Path, Path::Handle>& wrapper()
    {
      static MemoryPoolWrapperImpl<PXR_NS::SdfPath, Path, Path::Handle> sWrapper;
      return sWrapper;
    }

    const PXR_NS::SdfPath& unwrap(const MemoryPoolWrapperImpl<PXR_NS::SdfPath, Path, Path::Handle>::MemoryPoolHandle& h)
    {
      return wrapper().unwrap(h);
    }

    Path wrap(const PXR_NS::SdfPath& sdfPath)
    {
      return wrapper().wrap(sdfPath);
    }
  }

  Path::Path()
    : _handle(wrapper().allocate())
  {
  }

  Path::Path(const char* text)
    : _handle(wrapper().allocate(text))
  {
  }

  Path::Path(const std::string& text)
    : _handle(wrapper().allocate(text.c_str()))
  {
  }

  Path::Path(const Path::Handle& handle)
    : _handle(wrapper().copyAllocate(handle))
  {
  }

  Path::~Path()
  {
    wrapper().free(_handle);
  }

  Path::Path(Path&& Path) noexcept
  {
    _handle = Path._handle;
    Path._handle = fdk::MemoryPool<PXR_NS::SdfPath>::nullhandle;
  }

  Path& Path::operator=(Path&& rhs) noexcept
  {
    if (this != &rhs) {
      _handle = rhs._handle;
      rhs._handle = fdk::MemoryPool<PXR_NS::SdfPath>::nullhandle;
    }
    return *this;
  }

  Path::Path(const Path& rhs)
    : _handle(wrapper().copyAllocate(rhs._handle))
  {
  }

  Path& Path::operator=(const Path& rhs)
  {
    if (this != &rhs) {
      _handle = wrapper().copyAllocate(rhs._handle);
    }
    return *this;
  }

  void Path::clear()
  {
    _handle = wrapper().allocate();
  }

  const std::string& Path::name() const
  {
    return unwrap(_handle).GetName();
  }

  Path Path::parent() const
  {
    return wrap(unwrap(_handle).GetParentPath());
  }

  Path Path::appendChild(const Token& suffix) const
  {
    return wrap(unwrap(_handle).AppendChild(reinterpret_cast<const PXR_NS::TfToken&>(suffix.getToken(PXR_VERSION))));
  }

  Path Path::appendChild(const char* suffix) const
  {
    return wrap(unwrap(_handle).AppendChild(PXR_NS::TfToken(suffix)));
  }

  Path Path::appendChild(const std::string& suffix) const
  {
    return wrap(unwrap(_handle).AppendChild(PXR_NS::TfToken(suffix)));
  }

  Path Path::appendPath(const Path& path) const
  {
    return wrap(unwrap(_handle).AppendPath(unwrap(path._handle)));
  }

  Path Path::appendProperty(const Token& property) const
  {
    return wrap(unwrap(_handle).AppendProperty(
        reinterpret_cast<const PXR_NS::TfToken&>(property.getToken(PXR_VERSION))));
  }

  Path Path::makeAbsolutePath(const Path& anchor) const
  {
    return wrap(unwrap(_handle).MakeAbsolutePath(unwrap(anchor._handle)));
  }

  Path Path::makeRelativePath(const Path& anchor) const
  {
    return wrap(unwrap(_handle).MakeRelativePath(unwrap(anchor._handle)));
  }

  std::string Path::asString() const
  {
    return unwrap(_handle).GetAsString();
  }

  bool Path::operator==(const Path& rhs) const
  {
    return unwrap(_handle) == unwrap(rhs._handle);
  }

  bool Path::operator!=(const Path& rhs) const
  {
    return !(unwrap(_handle) == unwrap(rhs._handle));
  }

  bool Path::operator<(const Path& rhs) const
  {
    return unwrap(_handle) < unwrap(rhs._handle);
  }

  bool Path::empty() const
  {
    return unwrap(_handle).IsEmpty();
  }

  bool Path::isAbsolutePath() const
  {
    return unwrap(_handle).IsAbsolutePath();
  }

  Path::Handle& Path::getPath(int usd_version)
  {
    return wrapper().getWrappedObject(_handle);
  }
 
  const Path::Handle& Path::getPath(int usd_version) const
  {
    return wrapper().getWrappedObject(_handle);
  }

  bool Path::isValidPath(std::string* errorMsg) const
  {
    return PXR_NS::SdfPath::IsValidPathString(unwrap(_handle).GetAsString(), errorMsg);
  }

  bool Path::IsValidPathString(const std::string& pathString, std::string* errorMsg)
  {
    return PXR_NS::SdfPath::IsValidPathString(pathString, errorMsg);
  }

  const Path& Path::AbsoluteRootPath()
  {
    static Path rootPath(wrap(PXR_NS::SdfPath::AbsoluteRootPath()));
    return rootPath;
  }

  uint64_t Path::hash() const
  {
    return unwrap(_handle).GetHash();
  }

  void Path::append(fdk::Hash& hash) const
  {
    const auto h = unwrap(_handle).GetHash();
    hash.append(&h, sizeof(h));
  }

  fdk::Hash& operator<<(fdk::Hash& hash, const Path& path)
  {
    path.append(hash);
    return hash;
  }

  std::ostream& operator<<(std::ostream& stream, const Path& path)
  {
    stream << path.asString();
    return stream;
  }

}
