// Copyright (c) 2022 The Foundry Visionmongers Ltd. All Rights Reserved.

#include "usg/private/PathExpression.h"
#include "usg/private/UsdImpl.h"

#include <cassert>

#include <pxr/usd/usd/collectionAPI.h>
#include <pxr/usd/usd/prim.h>
#include <pxr/usd/usd/primRange.h>

#include <algorithm>
#include <iostream>
#include <set>
#include <vector>

using namespace usg;

namespace {

  //! Class describing a scenegraph path pattern. A path is a vector of path elements. Each element
  //! is a constant string, a wildcard string or a globstar. The path "//*/foo" has four elements: "/", [globstar],
  //! "*" (wildcard) and "foo" (constant).
  struct PathPattern {
    struct Element {
      enum class Type {
        FixedText,          //<! Text with no wildcards
        TextWithWildcards,  //<! Text with *? wildcards
        Wildcard,           //<! A globstar wildcard: "a//b"
      };

      Element(const PXR_NS::TfToken& name, Type type)
        : _name(name)
        , _type(type)
      {
      }

      const char* name() const { return _name.GetText(); }
      const Type type() const { return _type; }

      PXR_NS::TfToken _name;
      Type _type;
    };

    void clear()
    {
      _elements.clear();
    }

    void addElement(const Element& element, bool hasCollection = false)
    {
      _elements.push_back(element);
      if (_prefixLength == 0 && (element.type() == Element::Type::Wildcard || element.type() == Element::Type::TextWithWildcards)) {
        _prefixLength = static_cast<int>(size());
      }
      _hasGlobStar |= element.type() == Element::Type::Wildcard;
      _hasWildcards |= element.type() == Element::Type::TextWithWildcards;
      _hasCollection |= hasCollection;
    }

    size_t size() const
    {
      return _elements.size();
    }

    const Element& operator[](int i) const
    {
      return _elements[i];
    }

    bool hasGlobStar() const { return _hasGlobStar; }
    bool hasWildcards() const { return _hasWildcards; }
    bool hasCollection() const { return _hasCollection; }
    bool isConstant() const { return !(_hasWildcards || _hasGlobStar); }
    int prefixLength() const { return _prefixLength; }

    PXR_NS::SdfPath asPath() const
    {
      PXR_NS::SdfPath path = PXR_NS::SdfPath::AbsoluteRootPath();
      for (auto it = _elements.cbegin() + 1; it != _elements.cend(); ++it) {
        path = path.AppendChild(it->_name);
      }
      return path;
    }

  private:
    bool _hasGlobStar = false;
    bool _hasWildcards = false;
    bool _hasCollection = false;
    int _prefixLength = 0; //!< The length of the fixed path prefix, up to the first wildcard
    std::vector<Element> _elements;
  };

  //! Wildcard string matching. Note this differs from the one in StringUtils.h which has different handling for ?/.
  bool wildcardMatch(const char *text, const char* pattern)
  {
    const char *textBackup = nullptr;
    const char *patternBackup = nullptr;
    while (*text) {
      if (*pattern == '*') {
        textBackup = text;
        patternBackup = ++pattern;
      }
      else if (*pattern == *text || *pattern == '?') {
        text++;
        pattern++;
      }
      else {
        if (patternBackup == nullptr) {
          return false;
        }
        text = ++textBackup;
        pattern = patternBackup;
      }
    }
    while (*pattern == '*') {
      pattern++;
    }
    return *pattern == '\0';
  }

  //! This is a wildcard matcher for path elements. It's basically the same as the one above but works with tokens
  //! instead of characters. We use this for globstars, and the one above for each path element.
  bool wildcardMatch(const PXR_NS::TfToken* text, int textSize, const PathPattern::Element* pattern, int patternSize)
  {
    const auto* textEnd = text + textSize;
    const auto* patternEnd = pattern + patternSize;
    const PXR_NS::TfToken *textBackup = nullptr;
    const PathPattern::Element *patternBackup = nullptr;
    while (text != textEnd) {
      const bool patternValid = (pattern < patternEnd); // Don't read off the end of the pattern (the text version of this function relies on a null terminator)
      if (patternValid && pattern->type() == PathPattern::Element::Type::Wildcard) {
        textBackup = text;
        patternBackup = ++pattern;
      }
      else if (patternValid && (pattern->type() == PathPattern::Element::Type::TextWithWildcards ?
                  wildcardMatch(text->GetText(), pattern->name()) :
                    *text == pattern->_name)) {
        text++;
        pattern++;
      }
      else {
        if (patternBackup == nullptr) {
          return false;
        }
        text = ++textBackup;
        pattern = patternBackup;
      }
    }
    while (pattern->type() == PathPattern::Element::Type::Wildcard) {
      pattern++;
    }
    return pattern == patternEnd;
  }


  //! The execution context passed when executing a compiled expression
  struct Context {
    PXR_NS::UsdPrim _currentPrim;  //!< The prim being matched
    std::vector<PXR_NS::TfToken> _currentPath; //!< The path to the prim being matched
    int _depth = 0; //!< The depth in the hierarchy (== _currentPath.size())

    struct State {
      bool _finished; //!< true if the node has finished
      bool _pruneChildren; //!< true if the node has finished
      bool _matchedPrefix; //!< true if the node has finished
    };
    std::vector<State> _state; //!< Per-ASTNode data
    std::vector<PXR_NS::SdfPathSet> _pathSets; //!< PathSets used in the expression
    std::vector<PXR_NS::UsdCollectionMembershipQuery> _collectionQueries; //!< UsdCollectionMembershipQuery used in the expression

    //! For debugging: Contains some stats about how many prims were tested and matches tried
    struct {
      int _matches = 0;
      int _prims = 0;
    } _stats;
  };

  //! AST node for the expression parser. These are stored in a flattened array with integer
  //! "pointers" to child nodes. This is to save space and make it easier to store compiled expressions.
  //! A node also has a "pointer" to an object. This is an index into one of the arrays of constants on the
  //! expression, such as _patterns or _tokens. This layout makes it easy to make per-node data in the
  //! execution Context instead of storing it in the nodes. This makes it possible to evaluate on multiple threads.
  class ASTNode {
  public:
    //! Op codes
    enum class Op {
      Null,
      MatchPath,
      MatchAll,
      MatchType,
      MatchDefaultPrim,
      MatchLastModified,
      MatchCollection,
      Union,
      Intersection,
      Difference,
      Filter,
    };

    ASTNode(Op op, int left = 0, int right = 0, int object = 0)
      : _op(op)
      , _left(left)
      , _right(right)
      , _object(object)
    {
    }

    Op _op; //! The op code for the node
    int _left = 0; //! Left descendent node
    int _right = 0; //! Right descendent node
    int _object = 0; //! < Index of a possible constant object used by the node
  };

}

namespace usg {

  class PathExpressionImpl {
  public:
    PathArray match(const PXR_NS::UsdStageRefPtr& stage) const;

    bool executeNode(Context& context, int node) const;

    bool isValid() const { return _errorMessage.size() == 0; }

    template<class...Args>
    int addNode(Args&&... args)
    {
      _astNodes.emplace_back(ASTNode(std::forward<Args>(args)...));
      return _astNodes.size() - 1;
    }

    int rootNode() const { return static_cast<int>(_astNodes.size()) - 1; }

    const ASTNode& node(int nodeIndex) const { return _astNodes[nodeIndex]; }
    const PathPattern& pattern(int nodeIndex) const { return _patterns[node(nodeIndex)._object]; }

    std::vector<ASTNode> _astNodes; //!< The compiled AST
    std::vector<PathPattern> _patterns; //!< Patterns used in the expression
    std::vector<PXR_NS::TfToken> _tokens; //!< Tokens used in the expression
    int _numPathSets = 0; //!< The number of pathSets used
    int _numQueries = 0; //!< The number of collection queries used
    std::string _errorMessage;
  };

  bool PathExpressionImpl::executeNode(Context& context, int node) const
  {
    const ASTNode& self = _astNodes[node];
    if (context._state[node]._finished) {
      return false;
    }
    bool matched = false;
    const auto left = self._left;
    const auto right = self._right;
    const auto leftState = context._state[self._left];
    const auto rightState = context._state[self._right];
    auto& state = context._state[node];
    state._pruneChildren = false;

    switch (self._op) {
    case ASTNode::Op::MatchAll:
      matched = true;
      break;
    case ASTNode::Op::MatchPath: {
        context._stats._matches++;
        auto& pattern = _patterns[self._object];
        matched = wildcardMatch(context._currentPath.data(), context._currentPath.size(), &pattern[0], pattern.size());
        if (matched) {
          if (pattern.isConstant()) {
            state._finished = true;
          }
          if (context._depth == pattern.size() && !pattern.hasGlobStar()) {
            state._pruneChildren = true;
          }
        }
        // Have we matched the constant prefix of the pattern?
        if (matched && context._depth == pattern.prefixLength()) {
          state._matchedPrefix = true;
        }
        // If we previously matched the prefix and have now gone up a level, we're done as nothing else can match
        if (state._matchedPrefix && context._depth < pattern.prefixLength()) {
          state._finished = true;
        }
      }
      break;
      case ASTNode::Op::MatchDefaultPrim:
      case ASTNode::Op::MatchLastModified: {
        context._stats._matches++;
        auto& pathSet = context._pathSets[self._object];
        matched = pathSet.find(context._currentPrim.GetPath()) != pathSet.end();
      }
      break;
    case ASTNode::Op::MatchCollection:
      context._stats._matches++;
      matched = context._collectionQueries[self._object].IsPathIncluded(context._currentPrim.GetPath());
      break;
    case ASTNode::Op::MatchType:
      matched = executeNode(context, left) && context._currentPrim.GetTypeName() == _tokens[self._object];
      break;
    case ASTNode::Op::Union:
      matched = executeNode(context, left) | executeNode(context, right);
      state._finished = leftState._finished & context._state[right]._finished;
      state._pruneChildren = leftState._pruneChildren & rightState._pruneChildren;
      break;
    case ASTNode::Op::Intersection:
      matched = executeNode(context, left) ? executeNode(context, right) : false;
      state._finished = leftState._finished | rightState._finished;
      state._pruneChildren = leftState._pruneChildren | rightState._pruneChildren;
      break;
    case ASTNode::Op::Difference:
      matched = executeNode(context, left) ? !executeNode(context, right) : false;
      state._finished = leftState._finished;
      state._pruneChildren = leftState._pruneChildren & rightState._pruneChildren;
      break;
    case ASTNode::Op::Null:
      break;
    default:
      assert(false);
      break;
    }
    return matched;
  }

  PathArray PathExpressionImpl::match(const PXR_NS::UsdStageRefPtr& stage) const
  {
    PathArray result;

    const auto numNodes = _astNodes.size();
    const auto rootNode = numNodes - 1;

    auto rootOp = _astNodes[rootNode]._op;

    // Optimization: If the root node is a simple fixed path or collection, we can skip the
    // iteration through the stage.
    // TODO: There's much more we could do here: e.g.
    // - Spot lists of fixed paths
    // - For Intersection and Difference, we can reduce the iteration to be over any collection on the left node
    // - Test the leaf node name first when matching globstars
    if (rootOp == ASTNode::Op::MatchPath && pattern(rootNode).isConstant()) {
      const auto path = pattern(rootNode).asPath();
      if (stage->GetPrimAtPath(path)) {
        result.push_back(usg::fromSdfPath(path));
      }
    }
    else if (rootOp == ASTNode::Op::MatchDefaultPrim) {
      if (stage->HasDefaultPrim()) {
        result.push_back(usg::fromSdfPath(stage->GetDefaultPrim().GetPath()));
      }
    }
    else if (rootOp == ASTNode::Op::MatchLastModified) {
      const auto rootLayer = stage->GetRootLayer();
      auto sublayerPaths = rootLayer->GetSubLayerPaths();
      if (!sublayerPaths.empty()) {
        auto layer = PXR_NS::SdfLayer::FindOrOpen(sublayerPaths.front());
        if (layer) {
          const auto data = layer->GetCustomLayerData();
          const auto it = data.find("usg:modified");
          if (it != data.end()) {
            const auto& value = it->second;
            if (value.IsHolding<PXR_NS::VtStringArray>()) {
              const auto& sdfPaths = value.Get<PXR_NS::VtStringArray>();
              for (const auto& path : sdfPaths) {
                result.push_back(usg::fromSdfPath(PXR_NS::SdfPath(path)));
              }
            }
          }
        }
      }
    }
    else {
      Context context;
      context._state.clear();
      context._state.resize(numNodes);

      if (isValid()) {
        // Initialize any path sets needed
        context._pathSets.resize(_numPathSets);
        context._collectionQueries.resize(_numQueries);
        for (auto& n : _astNodes) {
          switch (n._op) {
          case ASTNode::Op::MatchDefaultPrim:
            if (stage->HasDefaultPrim()) {
              context._pathSets[n._object].insert(stage->GetDefaultPrim().GetPath());
            }
            break;
          case ASTNode::Op::MatchLastModified:
            if (const auto rootLayer = stage->GetRootLayer()) {
              auto sublayerPaths = rootLayer->GetSubLayerPaths();
              if (!sublayerPaths.empty()) {
                auto layer = PXR_NS::SdfLayer::FindOrOpen(sublayerPaths.front());
                if (layer) {
                  const auto data = layer->GetCustomLayerData();
                  const auto it = data.find("usg:modified");
                  if (it != data.end()) {
                    const auto& value = it->second;
                    if (value.IsHolding<PXR_NS::VtStringArray>()) {
                      const auto& sdfPaths = value.Get<PXR_NS::VtStringArray>();
                      for (const auto& path : sdfPaths) {
                        context._pathSets[n._object].insert(PXR_NS::SdfPath(path));
                      }
                    }
                  }
                }
              }
            }
            break;
          case ASTNode::Op::MatchCollection: {
              auto pathNode = _astNodes[n._right];
              auto path = pattern(pathNode._object).asPath();
              auto parentPath = path.GetParentPath();
              auto collectionName = path.GetNameToken();
              if (const auto prim = stage->GetPrimAtPath(parentPath)) {
                PXR_NS::UsdCollectionAPI usdCollectionAPI(prim, collectionName);
                context._collectionQueries[n._object] = usdCollectionAPI.ComputeMembershipQuery();
              }
            }
            break;
          default:
            break;
          }
        }

        // Now iterate the stage, matching as we go
        auto range = PXR_NS::UsdPrimRange::AllPrimsPreAndPostVisit(stage->GetPseudoRoot());
        for (auto it = range.begin(); it != range.end(); ++it) {
          if (it.IsPostVisit()) {
            context._depth--;
            context._currentPath.pop_back();
          }
          else {
            context._depth++;
            context._currentPrim = *it;
            context._currentPath.push_back(it->GetName());
            context._stats._prims++;
            if (context._depth != 1) { // Skip the pseudo-root itself
              if (executeNode(context, rootNode)) {
                result.push_back(usg::fromSdfPath(it->GetPath()));
              }
              if (context._state[rootNode]._pruneChildren) {
                it.PruneChildren();
              }
            }
          }
        }
      }
#if FN_MATCH_STATS
      std::cout << "Prims tested: " << context._stats._prims << "  Matches: " << context._stats._matches << std::endl;
#endif
    }

    return result;
  }
}

namespace {

  //! The parser for path expressions.
  class PathExpressionParser {
  public:
    PathExpressionParser(const std::string s, usg::PathExpressionImpl& expression);

  private:
    /// Tokens for the expression parser
    enum class Token {
      EndOfFile,
      Identifier,
      MagicIdentifier,
      Path,
      MatchAllPath,
      OpenParen,
      CloseParen,
      OpenBrace,
      CloseBrace,
      Union,
      Difference,
      Intersection,
      Unexpected,
    };

    int isExpression();
    int isPath();
    int isPathExpression();

    void nextToken();

    bool error(const char* s);
    bool error(const std::string& s);

    usg::PathExpressionImpl& _expression;

    std::istream* _stream;
    char _c;
    Token _token;
    PathPattern _path;
    std::string _string;
    bool _debug = true;
  };


  PathExpressionParser::PathExpressionParser(const std::string in, usg::PathExpressionImpl& expression)
    : _expression(expression)
  {
    std::istringstream stream(in);
    _stream = &stream;
    _c = ' ';
    _token = Token::EndOfFile;
    nextToken();
    _expression.addNode(ASTNode::Op::Null); // Add the null ASTNode

    // Empty expression is valid and matches nothing
    if (_token != Token::EndOfFile) {
      auto r = isExpression();
      if (r && _token != Token::EndOfFile) {
        _expression._errorMessage = "Extra characters after expression";
      }
    }
  }

  void PathExpressionParser::nextToken()
  {
    while (!_stream->eof() && isspace(_c)) {
      _c = _stream->get();
    }

    if (_stream->eof()) {
      _token = Token::EndOfFile;
      return;
    }

    if (_c == '/') {
      std::string path;
      while (!_stream->eof() && (isalnum(_c) || _c == '_' || _c == '/' || _c == '*' || _c == '?' || _c == '$')) {
        path += _c;
        _c = _stream->get();
      }

      if (path == "//*") {
        _token = Token::MatchAllPath;
        return;
      }

      _path.clear();
      _token = Token::Path;
      bool hasGlobStar = false;
      bool lastWasSlash = false;
      bool hasWildcards = false;
      bool hasCollection = false;
      std::string name;
      for (auto c : path) {
        if (c == '/') {
          if (_path.hasCollection()) {
            error("Collection name must be last element in path");
            _token = Token::Unexpected;
            return;
          }
          if (lastWasSlash) {
            hasGlobStar = true;
          }
          if (_path.size() == 0 && name.empty()) {
            name = "/";
          }
          _path.addElement(PathPattern::Element(PXR_NS::TfToken(name),
            hasGlobStar ? PathPattern::Element::Type::Wildcard :
            (hasWildcards ? PathPattern::Element::Type::TextWithWildcards :
            PathPattern::Element::Type::FixedText)), hasCollection);
          lastWasSlash = true;
          hasWildcards = false;
          hasGlobStar = false;
          name.clear();
        }
        else if (c == '$') {
          if (lastWasSlash) {
            hasCollection = true;
          }
          else {
            error("Illegal character in path: $");
            _token = Token::Unexpected;
            return;
          }
        }
        else {
          lastWasSlash = false;
          if (c == '*' || c == '?') {
            hasWildcards = true;
          }
          name += c;
        }
      }
      if (!name.empty()) {
        _path.addElement(PathPattern::Element(PXR_NS::TfToken(name),
            lastWasSlash ? PathPattern::Element::Type::Wildcard :
            (hasWildcards ? PathPattern::Element::Type::TextWithWildcards :
            PathPattern::Element::Type::FixedText)), hasCollection);
      }
      return;
    }

    if (_c == '{') {
      _c = _stream->get();
      _string.clear();
      while (!_stream->eof() && isalpha(_c)) {
        _string += _c;
        _c = _stream->get();
      }
      if (_c != '}') {
        error(R"(Expected "}" after "{".)");
        _token = Token::Unexpected;
        return;
      }
      _c = _stream->get();
      _token = Token::MagicIdentifier;
      return;
    }

    if (isalpha(_c) || _c == '_') {
      _string.clear();
      while (!_stream->eof() && (isalnum(_c) || _c == '_')) {
        _string += _c;
        _c = _stream->get();
      }
      _token = Token::Identifier;
      return;
    }

    char lastC = _c;
    _token = Token::Unexpected;
    _c = _stream->get();

    switch (lastC) {
    case '+': _token = Token::Union; break;
    case '-': _token = Token::Difference; break;
    case '^': _token = Token::Intersection; break;
    case '(': _token = Token::OpenParen; break;
    case ')': _token = Token::CloseParen; break;
    case '{': _token = Token::OpenBrace; break;
    case '}': _token = Token::CloseBrace; break;
    default: break;
    }
  }

  bool PathExpressionParser::error(const char* s)
  {
    if (_expression._errorMessage.empty()) {
      _expression._errorMessage = s;
    }
    return false;
  }

  bool PathExpressionParser::error(const std::string& s)
  {
    if (_expression._errorMessage.empty()) {
      _expression._errorMessage = s;
    }
    return false;
  }

  int PathExpressionParser::isPath()
  {
    if (_token == Token::MatchAllPath) {
      auto n = _expression.addNode(ASTNode::Op::MatchAll);
      nextToken();
      return n;
    }
    if (_token == Token::Path) {
      if (_path.hasCollection()) {
        if (!_path.isConstant()) {
          nextToken();
          return error("Collection paths can't contain wildcards");
        }
        else {
          nextToken();
          auto n = _expression.addNode(ASTNode::Op::MatchPath, 0, 0, _expression._patterns.size());
          _expression._patterns.push_back(_path);
          return _expression.addNode(ASTNode::Op::MatchCollection, 0, n, _expression._numQueries++);
        }
      }

      auto n = _expression.addNode(ASTNode::Op::MatchPath, 0, 0, _expression._patterns.size());
      _expression._patterns.push_back(_path);
      nextToken();
      return n;
    }
    return 0;
  }

  int PathExpressionParser::isPathExpression()
  {
    if (_token == Token::OpenParen) {
      nextToken();
      const auto node = isExpression();
      if (_token == Token::CloseParen) {
        nextToken();
        return node;
      }
      else {
        return error("')' expected");
      }
    }

    if (_token == Token::MagicIdentifier) {
      if (_string == "default") {
        nextToken();
        return _expression.addNode(ASTNode::Op::MatchDefaultPrim, 0, 0, _expression._numPathSets++);
      }
      else if (_string == "lastmodified") {
        nextToken();
        return _expression.addNode(ASTNode::Op::MatchLastModified, 0, 0, _expression._numPathSets++);
      }
      error("Bad path replacement: " + _string);
      return 0;
    }

    if (auto node = isPath()) {
      if (_token == Token::OpenBrace) {
        nextToken();
        // TODO: Handle full filter expressions here some day
        if (_token == Token::Identifier) {
          node = _expression.addNode(ASTNode::Op::MatchType, node, 0, _expression._tokens.size());
          _expression._tokens.push_back(PXR_NS::TfToken(_string));
          nextToken();
        }
        if (_token == Token::CloseBrace) {
          nextToken();
        }
      }
      return node;
    }

    return 0;
  }

  int PathExpressionParser::isExpression()
  {
    auto n = isPathExpression();
    if (n) {
      while (true) {
        if (_token == Token::Union || _token == Token::Difference || _token == Token::Intersection) {
          const auto t = _token;
          nextToken();
          if (auto right = isPathExpression()) {
            if (t == Token::Union) {
              n = _expression.addNode(ASTNode::Op::Union, n, right);
            }
            else if (t == Token::Intersection) {
              n = _expression.addNode(ASTNode::Op::Intersection, n, right);
            }
            else {
              n = _expression.addNode(ASTNode::Op::Difference, n, right);
            }
          }
          else {
            error("Path expression expected");
            return 0;
          }
        }
        else if (auto right = isPathExpression()) {
          n = _expression.addNode(ASTNode::Op::Union, n, right);
        }
        else {
          return n;
        }
      }
    }
    error("Path expression expected");
    return 0;
  }

}


PathExpression::PathExpression(const std::string s)
  : _pImpl(new PathExpressionImpl())
{
  PathExpressionParser parser(s, *_pImpl);
}

PathExpression::~PathExpression()
{
  // Required for std::unique_ptr
}

bool PathExpression::isValid() const
{
  return _pImpl->isValid();
}

std::string PathExpression::errorMessage() const
{
  return _pImpl->_errorMessage;
}

usg::PathArray PathExpression::match(const usg::StageRef& stage) const
{
  return _pImpl->match(stage->impl()->stage);
}
