// Copyright (c) 2022 The Foundry Visionmongers Ltd. All Rights Reserved.

#include "usg/geom/Prim.h"
#include "usg/geom/Layer.h"
#include "usg/geom/GeomTokens.h"

#include "usg/private/UsdImpl.h"

#include "usg/geom/ShaderDesc.h"

//------------------------------------------------------------------------------------


const std::string gEnvDebugUsgMaterials = ((std::getenv("USG_DEBUG_USG_MATERIALS") != nullptr) ?
                                            std::getenv("USG_DEBUG_USG_MATERIALS") : "");
const bool gDebugUsgMaterials = (gEnvDebugUsgMaterials == "1");


//-----------------------------------------------------------------------------------

namespace usg
{

extern std::string empty_string;
extern Path    empty_geompath;


Prim::Prim() :
  _pImpl(nullptr)
{
  //
}

Prim::Prim(const Prim& b) :
  _pImpl(b._pImpl ? new Prim::Impl(*b._pImpl) : nullptr)
{
  //
}

Prim&
Prim::operator = (const Prim& b)
{
  if (_pImpl != b._pImpl) {
    if (b._pImpl) {
      _pImpl.reset(new Prim::Impl(*b._pImpl));
    }
    else {
      _pImpl.reset();
    }
  }
  return *this;
}

Prim::Prim(const Prim::Impl& b) :
  _pImpl(new Prim::Impl(b))
{
  //
}

/*virtual*/
Prim::~Prim()
{
  //
}


bool
Prim::isValid() const
{
  return (_pImpl && _pImpl->isValid());
}

bool Prim::validate()
{
  return isValid();
}


Prim::SpecPrimHandle* Prim::getSdfPrimSpec(int usd_version)
{
  return (_pImpl && usdVersionMatchesUsg(usd_version)) ?
            reinterpret_cast<Prim::SpecPrimHandle*>(&_pImpl->sdfprim) :
            nullptr;
}
const Prim::SpecPrimHandle* Prim::getSdfPrimSpec(int usd_version) const
{
  return (_pImpl && usdVersionMatchesUsg(usd_version)) ?
            reinterpret_cast<const Prim::SpecPrimHandle*>(&_pImpl->sdfprim) :
            nullptr;
}

Prim::UsdPrimHandle* Prim::getUsdPrim(int usd_version)
{
  return (_pImpl && usdVersionMatchesUsg(usd_version)) ?
            reinterpret_cast<Prim::UsdPrimHandle*>(&_pImpl->usdprim) :
            nullptr;
}
const Prim::UsdPrimHandle* Prim::getUsdPrim(int usd_version) const
{
  return (_pImpl && usdVersionMatchesUsg(usd_version)) ?
            reinterpret_cast<const Prim::UsdPrimHandle*>(&_pImpl->usdprim) :
            nullptr;
}


bool
Prim::operator == (const Prim& b) const
{
  return (_pImpl && b._pImpl && *_pImpl == *b._pImpl);
}

bool
Prim::operator < (const Prim& b) const
{
  return (_pImpl && b._pImpl && *_pImpl < *b._pImpl);
}

bool Prim::isA(const Token& typeName) const
{
  if (_pImpl) {
    const auto tfType = PXR_NS::UsdSchemaRegistry::GetTypeFromName(toTfToken(typeName));
    if (_pImpl->sdfprim) {
      const auto thisType = PXR_NS::UsdSchemaRegistry::GetTypeFromName(_pImpl->sdfprim->GetTypeName());
      return thisType.IsA(tfType);
    }
    else if (_pImpl->usdprim) {
      return _pImpl->usdprim.IsA(tfType);
    }
  }
  return false;
}

bool Prim::hasApi(const Token& typeName) const
{
  if (_pImpl) {
    const auto tfType = PXR_NS::UsdSchemaRegistry::GetTypeFromName(toTfToken(typeName));
    if (tfType) {
      if (_pImpl->sdfprim) {
        return true;
      }
      else if (_pImpl->usdprim) {
        return _pImpl->usdprim.HasAPI(tfType);
      }
    }
  }
  return false;
}

bool Prim::hasApi(const Token& typeName, const Token& instanceName) const
{
  if (_pImpl) {
    const auto tfType = PXR_NS::UsdSchemaRegistry::GetTypeFromName(toTfToken(typeName));
    if (tfType) {
      if (_pImpl->sdfprim) {
        return true;
      }
      else if (_pImpl->usdprim) {
        return _pImpl->usdprim.HasAPI(tfType, toTfToken(instanceName));
      }
    }
  }
  return false;
}

/*! Print out base Prim info to a stream.
*/
std::ostream&
operator << (std::ostream& o, const Prim& prim)
{
  o << "Prim('" << prim.getName() << "')[" << prim.getTypeName();
  if (!prim.isValid()) {
    o << "<invalid>";
  }
  else if (prim.impl()->isSdfPrim()) {
    o << "<SdfPrimSpec>";
  }
  else if (prim.impl()->isUsdPrim()) {
    o << "<UsdPrim>";
  }
  else {
    o << "<error>";
  }
  o << "]";
  return o;
}


/*! Print out base Prim info and attributes to a stream.
*/
/*virtual*/ void
Prim::print(const char*   prefix,
            std::ostream& o) const
{
  o << prefix << "Prim('" << getName() << "')[" << getTypeName();
  if (!isValid()) {
    o << "<invalid>]" << std::endl;
    return;
  }
  else if (_pImpl->isSdfPrim()) {
    o << "<SdfPrimSpec>]" << std::endl;
    return;
  }
  else if (!_pImpl->isUsdPrim()) {
    o << "<error>]" << std::endl;
    return;
  }
  o << "<UsdPrim>";
  o << " active=" << _pImpl->usdprim.IsActive();
  o << ", loaded=" << _pImpl->usdprim.IsLoaded();
  o << ", defined=" << _pImpl->usdprim.IsDefined();
  o << ", abstract=" << _pImpl->usdprim.IsAbstract();
  o << ", model=" << _pImpl->usdprim.IsModel();
  o << ", group=" << _pImpl->usdprim.IsGroup();
  o << " ]" << std::endl;

  for (const auto attr : getAttributes()) {
    const Attribute prop = getAttr(attr);
    if (prop.isAuthored())
      o << prefix << "  " << prop << std::endl;
  }
}


/*!
*/
const std::string& Prim::getName()       const { return (_pImpl) ? _pImpl->getName()  : empty_string; }
Path           Prim::getPath()       const { return (_pImpl) ? _pImpl->getPath()  : empty_geompath;   }
const Token        Prim::getTypeName()   const { return (_pImpl) ? _pImpl->getTypeName() : Token(); }
Prim               Prim::getParent()     const { return (_pImpl && _pImpl->usdprim) ? Prim(Prim::Impl(_pImpl->usdprim.GetParent())) : Prim(); }

extern Path absolute_root_geompath; // pseudo-root, or '/'

bool
Prim::isPseudoRoot() const
{
  if (_pImpl) {
    return (_pImpl->getPath() == absolute_root_geompath);
  }
  return false;
}

void Prim::setActive(bool active) { if (_pImpl) _pImpl->setActive(active); }
bool Prim::isActive() const { return (_pImpl) ? _pImpl->isActive() : false; }

bool Prim::isDefined() const { return (_pImpl) ? _pImpl->isDefined() : false; }

bool Prim::hasPayload() const { return (_pImpl) ? _pImpl->hasPayload() : false; }
bool Prim::isLoaded() const { return (_pImpl) ? _pImpl->isLoaded() : false; }
void Prim::load() const { if (_pImpl) _pImpl->load(); }
void Prim::unload() const { if (_pImpl) _pImpl->unload(); }

void Prim::setVisible(bool visible, fdk::TimeValue time)
{
  auto visibilityAttr = createIntrinsicAttr(GeomTokens.visibility, Value::Token);
  visibilityAttr.setValue(visible ? GeomTokens.inherited : GeomTokens.invisible, time);
}

bool Prim::isVisible(fdk::TimeValue time) const
{
  if (auto visibilityAttr = getAttr(GeomTokens.visibility)) {
    Token result;
    visibilityAttr.getValue(result, time);
    return result == GeomTokens.inherited;
  }
  return true;
}

void Prim::setKind(const Token& kind)
{
  if (_pImpl) {
    _pImpl->setKind(kind);
  }
}

Token Prim::getKind() const
{
  return (_pImpl) ? _pImpl->getKind() : Token();
}

void Prim::setInstanceable(bool instanceable)
{
  if (_pImpl) {
    _pImpl->setInstanceable(instanceable);
  }
}

bool Prim::isInstanceable() const
{
  return (_pImpl) ? _pImpl->getInstanceable() : false;
}

bool Prim::addAppliedSchema(const Token& schemaName)
{
  return (_pImpl) ? _pImpl->addAppliedSchema(schemaName) : false;
}

void Prim::setCreator(const Token& creator)
{
  setCustomData("creator_app", Value(creator));
}

Token Prim::creatorApp() const
{
  if (auto creator = getCustomData("creator_app")) {
    return creator.getToken();
  }
  return Token{};
}

bool Prim::isCreator(const Token& creator) const
{
  return creatorApp() == creator;
}

void Prim::setCustomData(const std::string& key, const Value& value)
{
  if (_pImpl) {
    _pImpl->setCustomData(key, value);
  }
}

bool Prim::getCustomData(const std::string& key, Value& value) const
{
  if (_pImpl) {
    return _pImpl->getCustomData(key, value);
  }
  return false;
}

Value Prim::getCustomData(const std::string& key) const
{
  if (_pImpl) {
    return _pImpl->getCustomData(key);
  }
  return {};
}

bool Prim::hasCustomData(const std::string& key) const
{
  if (_pImpl) {
    return _pImpl->hasCustomData(key);
  }
  return false;
}

void Prim::removeCustomData(const std::string& key)
{
  if (_pImpl) {
    _pImpl->removeCustomData(key);
  }
}


//--------------------------------------------------------------------------------------------------


/*static*/ Prim
Prim::defineInLayer(const LayerRef& layer,
                    const Path&     path,
                    const Token&    typeName)
{
  return (layer && layer->isValid()) ? layer->definePrim(path, typeName) : Prim();
}

/*static*/ Prim
Prim::overrideInLayer(const LayerRef& layer,
                      const Path&     path,
                      const Token&    typeName)
{
  return (layer && layer->isValid()) ? layer->overridePrim(path, typeName) : Prim();
}

/*static*/ void
Prim::createMissingParents(const StageRef& referenceStage,
                           const LayerRef& layer,
                           const Path&     path,
                           const Token&    parentPrimType)
{
  for (auto p = path.parent(); !p.empty(); p = p.parent()) {
    if (!referenceStage->getPrimAtPath(p)) {
      layer->definePrim(p, parentPrimType);
    }
    else {
      break;
    }
  }
}


//--------------------------------------------------------------------------------------------------


Attribute
Prim::getAttr(const Token& name) const
{
  Attribute attr;
  this->getAttr(name, attr);
  return attr;
}
bool
Prim::getAttr(const Token& name,
              Attribute&    attr) const
{
  return (_pImpl) ? _pImpl->getAttr(name, attr) : false;
}


/*! Create and attach an intrinsic(non-custom) attribute to this Prim.

    An 'intrinsic' attribute is considered part of the primitive's schema, or fundamental
    design, such as positions on a PointBased Prim or the face vertex list on a Mesh Prim.

    'Custom' attributes are arbitrary data attached to a Prim such as uvs, normals or
    IDs.
*/
Attribute
Prim::createIntrinsicAttr(const Token&    name,
                          Value::Type     type,
                          TimeVariability variability_hint)
{
  if (!_pImpl) {
    return Attribute();
  }
  return _pImpl->createAttr(name, type, variability_hint, false/*is_custom*/);
}


/*! Create and attach a custom(arbitrary) attribute to this Prim.
  
    A 'custom' attribute is arbitrary data attached to a Prim such as uvs, normals or
    IDs.

*/
Attribute
Prim::createCustomAttr(const Token&    name,
                       Value::Type     type,
                       TimeVariability variability_hint)
{
  if (!_pImpl) {
    return Attribute();
  }
  return _pImpl->createAttr(name, type, variability_hint, true/*is_custom*/);
}

Attribute Prim::createAttr(const Token&    name,
                           Value::Type     type,
                           bool            is_custom,
                           TimeVariability variability_hint,
                           const Value&    defaultValue)
{
  if (_pImpl) {
    auto attr = _pImpl->createAttr(name, type, variability_hint, is_custom);
    if (defaultValue) {
      attr.setValue(defaultValue);
    }
    return attr;
  }
  return Attribute();
}

//--------------------------------------------------------------------------------------------------


Relationship Prim::createRelationship(const Token& name, bool custom)
{
  if (!_pImpl) {
    return Relationship();
  }
  return _pImpl->createRelationship(toTfToken(name), empty_geompath, custom);
}

Relationship
Prim::getRelationship(const Token& name) const
{
  Relationship rel;
  if (_pImpl) {
    _pImpl->getRelationship(toTfToken(name), rel);
  }
  return rel;
}

bool
Prim::getRelationship(const Token&  name, Relationship& rel) const
{
  return (_pImpl) ? _pImpl->getRelationship(toTfToken(name), rel) : false;
}


//--------------------------------------------------------------------------------------------------


std::vector<Token> Prim::getAttributes() const
{
  std::vector<Token> attributes;
  if (!_pImpl || !_pImpl->usdprim) {
    return attributes;
  }

  //std::vector<UsdProperty> properties = GetPropertiesInNamespace()
#if 1
  std::vector<PXR_NS::UsdAttribute> attrs = _pImpl->usdprim.GetPrim().GetAttributes();
  attributes.reserve(attrs.size());
  for (const auto attr : attrs) {
    attributes.push_back(attr.GetName().GetString());
  }

#else
  prim.GetProperties([&](const PXR_NS::TfToken &propname)
                        { return (UT_String(propname.GetText()).multiMatch(*compiled_pattern_ptr) != 0); }
                     );
  attributes.reserve(properties.size());
  for (auto && property : properties)
  {
    if (property)
      myPrivate->myExpandedPathSet.sdfPathSet().insert(property.GetPath());
  }
#endif

  return attributes;
}

Attribute Prim::copyAttr(const Prim& fromPrim, const Token& name, fdk::TimeValue time)
{
  Attribute fromAttr = fromPrim.getAttr(name);
  if (!fromAttr.isValid()) {
    return {};
  }

  Value value;
  if (!fromAttr.getValue(value, time)) {
    return {};
  }

  Attribute toAttr = createCustomAttr(name, fromAttr.getType(), TimeVariability::Varying);
  if (!toAttr.isValid()) {
    return {};
  }

  toAttr.setValue(value, time);

  return toAttr;
}

size_t
Prim::getRelationships(std::vector<const char*>& relationships) const
{
  relationships.clear();
  if (!_pImpl || !_pImpl->usdprim) {
    return 0;
  }

  std::vector<PXR_NS::UsdRelationship> rels = _pImpl->usdprim.GetPrim().GetRelationships();
  relationships.reserve(rels.size());
  for (size_t i=0; i < rels.size(); ++i) {
    relationships.push_back(rels[i].GetName().GetString().c_str());
  }

  return relationships.size();
}

void Prim::addReference(const std::string& identifier, const Path& path, ListPosition position)
{
  if (_pImpl) {
    _pImpl->addReference(identifier, path, position);
  }
}

void Prim::addPayload(const std::string& identifier, const Path& path, ListPosition position)
{
  if (_pImpl) {
    _pImpl->addPayload(identifier, path, position);
  }
}

void Prim::addVariantSet(const std::string& variantSetName, ListPosition position)
{
  if (_pImpl) {
    _pImpl->addVariantSet(variantSetName, position);
  }
}

void Prim::addVariant(const std::string& variantSetName, const std::string& variantName)
{
  if (_pImpl) {
    _pImpl->addVariant(variantSetName, variantName);
  }
}

void Prim::setVariantSelection(const std::string& variantSetName, const std::string& selection)
{
  if (_pImpl) {
    _pImpl->setVariantSelection(variantSetName, selection);
  }
}

std::vector<std::string> Prim::getVariantSets() const
{
  return _pImpl ? _pImpl->getVariantSets() : std::vector<std::string>();
}

std::vector<std::string> Prim::getVariants(const std::string& variantSetName) const
{
  return _pImpl ? _pImpl->getVariants(variantSetName) : std::vector<std::string>();
}

/*virtual*/ bool
Prim::applyAttributeOverrides(const ShaderDesc& schema_desc,
                              bool              skip_connected)
{
  if (!isValid()) {
    return false;
  }
  if (gDebugUsgMaterials) {
    std::cout << "        " << getName() << "::applyAttributeOverrides():" << std::endl;
  }

  bool ok = true;
  if (_pImpl->isEditable()) {
    //---------------------------------------------------
    // Update overrides mode:

    // TODO: make this be non-input specific!
    const ShaderConnectionMap& connections = schema_desc.getInputConnections();

    // Update each property, checking if there's an override for it or not:
    fdk::StringPtrList prop_names;
    schema_desc.getInputNames(prop_names);
    for (const std::string* prop_name : prop_names) {
      // Skip any with connections:
      if (!prop_name || (skip_connected && connections.find(*prop_name) != connections.end())) {
        continue;
      }
      if (gDebugUsgMaterials) {
        std::cout << "          update property '" << *prop_name << "'" << std::endl;
      }

      // Get attribute to edit - if it doesn't exist on the Prim yet then it's not yet defined:
      Attribute attr; getAttr(*prop_name, attr);

      // TODO: make this be non-input specific!
      const ShaderProperty* override_schema_prop = schema_desc.getInputOverride(*prop_name);
      if (override_schema_prop) {
        if (gDebugUsgMaterials) {
          std::cout << "          update property '" << *override_schema_prop << "'" << std::endl;
        }
        // Update the Attribute with the override property value:
        if (!attr) {
          attr = createIntrinsicAttr(override_schema_prop->getFullName(),
                                     override_schema_prop->getValueType());
        }
        assert(attr.isValid());
        attr.setValue(override_schema_prop->getValue(),
                      override_schema_prop->getValueTime());
        if (gDebugUsgMaterials) {
          std::cout << " SET OVER   = " << attr << std::endl;
        }

        continue;
      }

      // TODO: make this be non-input specific!
      const ShaderProperty schema_prop = schema_desc.getInput(*prop_name);
      if (schema_prop && attr) {
        // Clear the Attribute to the default property value. Since we can't really delete
        // the defined attr we'll set its value to the property's:
        attr.setValue(schema_prop.getValue(),
                      schema_prop.getValueTime());
        if (gDebugUsgMaterials) {
          std::cout << " CLEAR ATTR = " << attr << std::endl;
        }

        continue;
      }

    }

  }
  else {
    //---------------------------------------------------
    // Define overrides mode:

    const ShaderPropertyMap& overrides = schema_desc.getInputOverrides();
    if (overrides.size() == 0) {
      return true; // no overrides
    }

    // TODO: make this be non-input specific!
    const ShaderConnectionMap& connections = schema_desc.getInputConnections();

    // Write out input overrides, skipping the ones that have defined connections:
    for (ShaderPropertyMap::const_iterator it=overrides.begin(); it != overrides.end(); ++it) {
      // Skip any with connections:
      if (skip_connected && connections.find(it->first) != connections.end()) {
        continue;
      }

      // Sanity check override property pointer:
      if (!it->second || !it->second->isValid()) {
        // TODO should log a schema translation error!
        ok = false;
        continue; // invalid property, shouldn't happen but don't crash...
      }

      const ShaderProperty& override_schema_prop = *it->second;
      if (gDebugUsgMaterials) {
        std::cout << "          override input '" << it->first << "'" << std::endl;
      }

      if (Attribute attr = createIntrinsicAttr(override_schema_prop.getFullName(),
                                              override_schema_prop.getValueType())) {
        if (gDebugUsgMaterials) {
          std::cout << "            ...with prop " << override_schema_prop;
          if (!fdk::isDefaultTimeValue(override_schema_prop.getValueTime())) {
            std::cout << " at time " << override_schema_prop.getValueTime();
          }
          else {
            std::cout << " at <default> time";
          }
          std::cout << std::endl;
        }
        attr.setValue(override_schema_prop.getValue(),
                      override_schema_prop.getValueTime());
      }
      else {
        // TODO should log a schema translation error!
        ok = false;
        continue; // failed to create the attribute...
      }

    }

  }

  return ok;
}


//--------------------------------------------------------------------------------------------------

size_t
Prim::getChildren(std::vector<Prim>& children, bool activeOnly) const
{
  children.clear();
  if (_pImpl) {
    if (_pImpl->sdfprim) {
      for (auto n : _pImpl->sdfprim->GetNameChildren()) {
        if (!activeOnly || n->GetActive()) {
          children.push_back(Prim(Prim::Impl(n)));
        }
      }
    }
    else if (_pImpl->usdprim) {
      PXR_NS::Usd_PrimFlagsConjunction prim_flags;
      prim_flags &= !PXR_NS::UsdPrimIsAbstract;
      if (activeOnly) {
        prim_flags = prim_flags && PXR_NS::UsdPrimIsActive && PXR_NS::UsdPrimIsLoaded;
      }

      const PXR_NS::UsdPrim::SiblingRange child_range = _pImpl->usdprim.GetFilteredChildren(PXR_NS::UsdTraverseInstanceProxies(prim_flags));

      children.reserve(50);
      for (PXR_NS::UsdPrim::SiblingRange::iterator it=child_range.begin(); it != child_range.end(); ++it)
      {
        children.push_back(Prim(Prim::Impl(*it)));
      }
    }
  }

  return children.size();
}

LayerRef Prim::getDefiningLayer() const
{
  if (_pImpl && _pImpl->usdprim) {
    for (auto& spec : _pImpl->usdprim.GetPrimStack()) {
      if (spec->GetSpecifier() == PXR_NS::SdfSpecifierDef) {
        return std::make_shared<Layer>(Layer::Impl(spec->GetLayer()));
      }
    }
  }
  return LayerRef();
}


//-----------------------------------------------------------------------------------------

template <typename usgT, typename usdT>
inline Attribute setPrimAttr(Prim*                 prim,
                             const Token&          name,
                             const usgT&           v,
                             Value::Type           value_type,
                             const fdk::TimeValue& time)
{
  if (prim->impl()) {
    // Get or create attribute:
    Attribute attr = prim->impl()->createAttr(name, value_type, TimeVariability::Uniform, true/*is_custom*/);
    if (attr.impl()) {
      attr.impl()->set<usgT, usdT>(v, time);
      return attr;
    }
  }
  return Attribute();
}

template <typename usgT, typename usdT>
inline Attribute setPrimArrayAttr(Prim*                    prim,
                                  const Token&             name,
                                  const std::vector<usgT>& array,
                                  Value::Type              value_type,
                                  const fdk::TimeValue&    time)
{
  if (prim->impl()) {
    // Get or create attribute:
    Attribute attr = prim->impl()->createAttr(name, value_type, TimeVariability::Uniform, true/*is_custom*/);
    if (attr.impl()) {
      attr.impl()->setArray<usgT, usdT>(array, time);
      return attr;
    }
  }
  return Attribute();
}

template <> USG_API Attribute Prim::setAttr(const Token& n, const std::string& v, const fdk::TimeValue& t) { return setPrimAttr<std::string      , std::string      >(this, n, v                , Value::String,    t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const Token      & v, const fdk::TimeValue& t) { return setPrimAttr<PXR_NS::TfToken     , PXR_NS::TfToken     >(this, n, toTfToken(v)     , Value::Token,     t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const Path       & v, const fdk::TimeValue& t) { return setPrimAttr<PXR_NS::SdfPath     , PXR_NS::SdfPath     >(this, n, toSdfPath(v)     , Value::Path,  t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const AssetPath  & v, const fdk::TimeValue& t) { return setPrimAttr<PXR_NS::SdfAssetPath, PXR_NS::SdfAssetPath>(this, n, toSdfAssetPath(v), Value::AssetPath, t); }
//
template <> USG_API Attribute Prim::setAttr(const Token& n, const bool      & v, const fdk::TimeValue& t) { return setPrimAttr<bool      , bool          >(this, n, v, Value::Bool   , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const char      & v, const fdk::TimeValue& t) { return setPrimAttr<char      , char          >(this, n, v, Value::UChar  , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const fdk::half & v, const fdk::TimeValue& t) { return setPrimAttr<fdk::half , PXR_NS::GfHalf>(this, n, v, Value::Half   , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const float     & v, const fdk::TimeValue& t) { return setPrimAttr<float     , float         >(this, n, v, Value::Float  , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const double    & v, const fdk::TimeValue& t) { return setPrimAttr<double    , double        >(this, n, v, Value::Double , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const int32_t   & v, const fdk::TimeValue& t) { return setPrimAttr<int32_t   , int32_t       >(this, n, v, Value::Int    , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const int64_t   & v, const fdk::TimeValue& t) { return setPrimAttr<int64_t   , int64_t       >(this, n, v, Value::Int64  , t); }
//
template <> USG_API Attribute Prim::setAttr(const Token& n, const fdk::Vec2h& v, const fdk::TimeValue& t) { return setPrimAttr<fdk::Vec2h, PXR_NS::GfVec2h>(this, n, v, Value::Half2  , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const fdk::Vec2f& v, const fdk::TimeValue& t) { return setPrimAttr<fdk::Vec2f, PXR_NS::GfVec2f>(this, n, v, Value::Float2 , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const fdk::Vec2d& v, const fdk::TimeValue& t) { return setPrimAttr<fdk::Vec2d, PXR_NS::GfVec2d>(this, n, v, Value::Double2, t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const fdk::Vec2i& v, const fdk::TimeValue& t) { return setPrimAttr<fdk::Vec2i, PXR_NS::GfVec2i>(this, n, v, Value::Int2   , t); }
//
template <> USG_API Attribute Prim::setAttr(const Token& n, const fdk::Vec3h& v, const fdk::TimeValue& t) { return setPrimAttr<fdk::Vec3h, PXR_NS::GfVec3h>(this, n, v, Value::Half3  , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const fdk::Vec3f& v, const fdk::TimeValue& t) { return setPrimAttr<fdk::Vec3f, PXR_NS::GfVec3f>(this, n, v, Value::Float3 , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const fdk::Vec3d& v, const fdk::TimeValue& t) { return setPrimAttr<fdk::Vec3d, PXR_NS::GfVec3d>(this, n, v, Value::Double3, t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const fdk::Vec3i& v, const fdk::TimeValue& t) { return setPrimAttr<fdk::Vec3i, PXR_NS::GfVec3i>(this, n, v, Value::Int3   , t); }
//
template <> USG_API Attribute Prim::setAttr(const Token& n, const fdk::Vec4h& v, const fdk::TimeValue& t) { return setPrimAttr<fdk::Vec4h, PXR_NS::GfVec4h>(this, n, v, Value::Half4  , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const fdk::Vec4f& v, const fdk::TimeValue& t) { return setPrimAttr<fdk::Vec4f, PXR_NS::GfVec4f>(this, n, v, Value::Float4 , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const fdk::Vec4d& v, const fdk::TimeValue& t) { return setPrimAttr<fdk::Vec4d, PXR_NS::GfVec4d>(this, n, v, Value::Double4, t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const fdk::Vec4i& v, const fdk::TimeValue& t) { return setPrimAttr<fdk::Vec4i, PXR_NS::GfVec4i>(this, n, v, Value::Int4   , t); }
//
template <> USG_API Attribute Prim::setAttr(const Token& n, const fdk::Quath& v, const fdk::TimeValue& t) { return setPrimAttr<fdk::Quath, PXR_NS::GfQuath>(this, n, v, Value::Quath  , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const fdk::Quatf& v, const fdk::TimeValue& t) { return setPrimAttr<fdk::Quatf, PXR_NS::GfQuatf>(this, n, v, Value::Quatf  , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const fdk::Quatd& v, const fdk::TimeValue& t) { return setPrimAttr<fdk::Quatd, PXR_NS::GfQuatd>(this, n, v, Value::Quatd  , t); }
//
#if 0
// Disabling Mat4f version to avoid creating an invalid Value::Type. TODO: make this throw a static_assert()...?
template <> USG_API Attribute Prim::setAttr(const Token& n, const fdk::Mat4f& v, const fdk::TimeValue& t) { return setPrimAttr<fdk::Mat4f, PXR_NS::GfMatrix4f>(this, n, v, Value::Matrix4f, t); }
#endif
template <> USG_API Attribute Prim::setAttr(const Token& n, const fdk::Mat4d& v, const fdk::TimeValue& t) { return setPrimAttr<fdk::Mat4d, PXR_NS::GfMatrix4d>(this, n, v, Value::Matrix4d, t); }
//-------------------------------------------
template <> USG_API Attribute Prim::setAttr(const Token& n, const fdk::StringList& v, const fdk::TimeValue& t) { return setPrimArrayAttr<std::string, std::string >(this, n, v, Value::StringArray , t); }
//template <> USG_API void Prim::setAttr(const Token& n, const TokenArray&  v, const fdk::TimeValue& t) { return setPrimArrayAttr<std::string, PXR_NS::TfToken>(this, n, v, Value::TokenArray , t); }
//
// TODO: std::vector<bool> does not support the data() method at getting at the array contents! wth? Is it because it's a bitmask implementation specialization?
//template <> USG_API void Prim::setAttr(const Token& n, const BoolArray  & v, const fdk::TimeValue& t) { return setPrimArrayAttr<bool     , bool        >(this, n, v, Value::BoolArray   , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const CharArray  & v, const fdk::TimeValue& t) { return setPrimArrayAttr<char      , char           >(this, n, v, Value::UCharArray  , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const HalfArray  & v, const fdk::TimeValue& t) { return setPrimArrayAttr<fdk::half , PXR_NS::GfHalf >(this, n, v, Value::HalfArray   , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const FloatArray & v, const fdk::TimeValue& t) { return setPrimArrayAttr<float     , float          >(this, n, v, Value::FloatArray  , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const DoubleArray& v, const fdk::TimeValue& t) { return setPrimArrayAttr<double    , double         >(this, n, v, Value::DoubleArray , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const Int32Array & v, const fdk::TimeValue& t) { return setPrimArrayAttr<int32_t   , int32_t        >(this, n, v, Value::IntArray    , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const Int64Array & v, const fdk::TimeValue& t) { return setPrimArrayAttr<int64_t   , int64_t        >(this, n, v, Value::Int64Array  , t); }
//
template <> USG_API Attribute Prim::setAttr(const Token& n, const Vec2hArray & v, const fdk::TimeValue& t) { return setPrimArrayAttr<fdk::Vec2h, PXR_NS::GfVec2h>(this, n, v, Value::Half2Array  , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const Vec2fArray & v, const fdk::TimeValue& t) { return setPrimArrayAttr<fdk::Vec2f, PXR_NS::GfVec2f>(this, n, v, Value::Float2Array , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const Vec2dArray & v, const fdk::TimeValue& t) { return setPrimArrayAttr<fdk::Vec2d, PXR_NS::GfVec2d>(this, n, v, Value::Double2Array, t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const Vec2iArray & v, const fdk::TimeValue& t) { return setPrimArrayAttr<fdk::Vec2i, PXR_NS::GfVec2i>(this, n, v, Value::Int2Array   , t); }
//
template <> USG_API Attribute Prim::setAttr(const Token& n, const Vec3hArray & v, const fdk::TimeValue& t) { return setPrimArrayAttr<fdk::Vec3h, PXR_NS::GfVec3h>(this, n, v, Value::Half3Array  , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const Vec3fArray & v, const fdk::TimeValue& t) { return setPrimArrayAttr<fdk::Vec3f, PXR_NS::GfVec3f>(this, n, v, Value::Float3Array , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const Vec3dArray & v, const fdk::TimeValue& t) { return setPrimArrayAttr<fdk::Vec3d, PXR_NS::GfVec3d>(this, n, v, Value::Double3Array, t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const Vec3iArray & v, const fdk::TimeValue& t) { return setPrimArrayAttr<fdk::Vec3i, PXR_NS::GfVec3i>(this, n, v, Value::Int3Array   , t); }
//
template <> USG_API Attribute Prim::setAttr(const Token& n, const Vec4hArray & v, const fdk::TimeValue& t) { return setPrimArrayAttr<fdk::Vec4h, PXR_NS::GfVec4h>(this, n, v, Value::Half4Array  , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const Vec4fArray & v, const fdk::TimeValue& t) { return setPrimArrayAttr<fdk::Vec4f, PXR_NS::GfVec4f>(this, n, v, Value::Float4Array , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const Vec4dArray & v, const fdk::TimeValue& t) { return setPrimArrayAttr<fdk::Vec4d, PXR_NS::GfVec4d>(this, n, v, Value::Double4Array, t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const Vec4iArray & v, const fdk::TimeValue& t) { return setPrimArrayAttr<fdk::Vec4i, PXR_NS::GfVec4i>(this, n, v, Value::Int4Array   , t); }
//
template <> USG_API Attribute Prim::setAttr(const Token& n, const QuathArray & v, const fdk::TimeValue& t) { return setPrimArrayAttr<fdk::Quath, PXR_NS::GfQuath>(this, n, v, Value::QuathArray  , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const QuatfArray & v, const fdk::TimeValue& t) { return setPrimArrayAttr<fdk::Quatf, PXR_NS::GfQuatf>(this, n, v, Value::QuatfArray  , t); }
template <> USG_API Attribute Prim::setAttr(const Token& n, const QuatdArray & v, const fdk::TimeValue& t) { return setPrimArrayAttr<fdk::Quatd, PXR_NS::GfQuatd>(this, n, v, Value::QuatdArray  , t); }
//
#if 0
// Disabling Mat4f version to avoid creating an invalid Value::Type. TODO: make this throw a static_assert()...?
template <> USG_API Attribute Prim::setAttr(const Token& n, const Mat4fArray & v, const fdk::TimeValue& t) { return setPrimArrayAttr<fdk::Mat4f, PXR_NS::GfMatrix4f>(this, n, v, Value::Matrix4fArray, t); }
#endif
template <> USG_API Attribute Prim::setAttr(const Token& n, const Mat4dArray & v, const fdk::TimeValue& t) { return setPrimArrayAttr<fdk::Mat4d, PXR_NS::GfMatrix4d>(this, n, v, Value::Matrix4dArray, t); }


} // namespace usg
