// Copyright (c) 2022 The Foundry Visionmongers Ltd. All Rights Reserved.

#include "usg/geom/Schema.h"
#include "usg/geom/ShaderSchema.h"

#include "usg/private/UsdImpl.h"

#include "fdk/base/PathUtils.h"

#include "pxr/usd/ndr/parserPlugin.h"
#include "pxr/usd/sdr/registry.h"
#include "pxr/usd/sdr/shaderNode.h"
#include "pxr/base/tf/stringUtils.h" // for TfStringGetBeforeSuffix
#include "pxr/base/plug/registry.h" // for RegisterPlugins
#include "pxr/base/arch/systemInfo.h" // for ArchGetExecutablePath

#include "pxr/usd/usdShade/shader.h"          // for _GetSdrMetadata args
#include "pxr/usd/usdShade/shaderDefUtils.h"

#include <mutex> // for std::mutex
#include <unordered_map>
#include <ostream>


//------------------------------------------------------------------------------------


const std::string gEnvDebugShaderPlugins     = ((std::getenv("USG_DEBUG_SHADER_PLUGINS") != nullptr) ?
                                                 std::getenv("USG_DEBUG_SHADER_PLUGINS") : "");
const std::string gEnvDebugShaderBuilding    = ((std::getenv("USG_DEBUG_SHADER_BUILDING") != nullptr) ?
                                                 std::getenv("USG_DEBUG_SHADER_BUILDING") : "");
const std::string gEnvDebugShaderConnections = ((std::getenv("USG_DEBUG_SHADER_CONNECTIONS") != nullptr) ?
                                                 std::getenv("USG_DEBUG_SHADER_CONNECTIONS") : "");
const bool gDebugShaderPlugins     = (gEnvDebugShaderPlugins     == "1");
const bool gDebugShaderBuilding    = (gEnvDebugShaderBuilding    == "1");
const bool gDebugShaderConnections = (gEnvDebugShaderConnections == "1");


//------------------------------------------------------------------------------------


namespace usg
{

static std::mutex g_map_lock;

//------------------------------------------------------------------------------------

// Forward DDImage plugin methods to FnUsdAbstraction. The runtime connection is done in
// DDImage::Plugins.cpp:

static const char* plugin_methods_not_connected = "usg-plugin-methods-not-connected";

static const char* dummy_plugin_find(const char*, const char* const*) { return plugin_methods_not_connected; }
static const char* dummy_plugin_load(const char*, const char*) { return plugin_methods_not_connected; }
static const char* dummy_plugin_error() { return plugin_methods_not_connected; }
static const char* dummy_plugin_name() { return plugin_methods_not_connected; }

USG_API const char* (*ddimage_plugin_find)(const char* name, const char* const* extensions) = dummy_plugin_find;
USG_API const char* (*ddimage_plugin_load)(const char* name, const char* altname) = dummy_plugin_load;
USG_API const char* (*ddimage_plugin_error)() = dummy_plugin_error;
USG_API const char* (*ddimage_plugin_name)() = dummy_plugin_name;


//------------------------------------------------------------------------------------
//------------------------------------------------------------------------------------

static const PXR_NS::TfToken      usgBuiltinSchemaType(toTfToken(GeomTokens.usgSchema));
static const PXR_NS::TfToken      usgBuiltinShaderType(toTfToken(GeomTokens.usgShader));
static const PXR_NS::TfToken      glslfxSourceType(toTfToken(GeomTokens.glslfxSourceType));
static const PXR_NS::NdrTokenVec  usgBuiltinDiscoveryTypes = { usgBuiltinSchemaType /*, glslfxSourceType*/ };


#if PXR_VERSION < 2102
#else
// Copied from usd/usdShade/shaderDefParser.cpp, it's still not in UsdShadeShaderDefUtils...
static PXR_NS::NdrTokenMap
_GetSdrMetadata(const PXR_NS::UsdShadeShader& shaderDef,
                const PXR_NS::NdrTokenMap&    discoveryResultMetadata) 
{
  // XXX Currently, this parser does not support 'vstruct' parsing, but if
  //     we decide to support 'vstruct' type in the future, we would need to
  //     identify 'vstruct' types in this function by examining the metadata.

  PXR_NS::NdrTokenMap metadataOut = discoveryResultMetadata;

  auto shaderDefMetadata = shaderDef.GetSdrMetadata();
  metadataOut.insert(shaderDefMetadata.begin(), shaderDefMetadata.end());

  //metadataOut[PXR_NS::SdrNodeMetadata->Primvars] =
  //  PXR_NS::UsdShadeShaderDefUtils::GetPrimvarNamesMetadataString(metadataOut, shaderDef);

  return metadataOut;
}
#endif


/*! The SchemaBuiltinParser is hooked into by Usd looking at the asset path
    extension (yikes...)

    TODO: currently there's no obvious way to directly add a NdrNode definition to
    the SdrRegistry, and we must use an NdrParserPlugin as the hook to add a schema
    dynamically when the Nuke SchemaDescription gets loaded.

*/
class SchemaBuiltinParser : public PXR_NS::NdrParserPlugin
{
  // This stage holds the schema prototype Prims that allows us to use the Schema name as
  // an identifier vs. the complex ones returned from the Asset or SourceCode schema
  // registration methods.
  PXR_NS::UsdStageRefPtr schemadefs_stage;


public:
  SchemaBuiltinParser() : PXR_NS::NdrParserPlugin()
  {
    schemadefs_stage = PXR_NS::UsdStage::CreateInMemory("__usgBuiltinNodeDefs.usda");
  }
 

  /*! Returns the types of schemas that this plugin can parse.

      "Type" here is the discovery type (in the case of files, this will
      probably be the file extension, but in other systems will be data that
      can be determined during discovery). This type should only be used to
      match up a `NdrNodeDiscoveryResult` to its parser plugin; this value is
      not exposed in the schema's API.
  */
  /*virtual*/
  const PXR_NS::NdrTokenVec& GetDiscoveryTypes() const override { return usgBuiltinDiscoveryTypes; }


  /*! Returns the source type that this parser operates on.
      
      A source type is the most general type for a schema. The parser plugin is
      responsible for parsing all discovery results that have the types
      declared under `GetDiscoveryTypes()`, and those types are collectively
      identified as one "source type".
  */
  /*virtual*/
  const PXR_NS::TfToken& GetSourceType() const override { return usgBuiltinSchemaType; }


  /*! Takes the specified `NdrNodeDiscoveryResult` instance, which was a
      result of the discovery process, and generates a new `NdrNode`.

      The schema's name, source type, and family must match.
  */
  /*virtual*/
  PXR_NS::NdrNodeUniquePtr Parse(const PXR_NS::NdrNodeDiscoveryResult& discoveryResult) override
  {
    // std::cout << "  ----------------------------------------------------------------" << std::endl;
    // std::cout << "  SchemaBuiltinParser::Parse(" << this << ") discoveryResult:" << std::endl;
    // std::cout << "      identifier '" << discoveryResult.identifier << "'" << std::endl;
    // std::cout << "         version '" << discoveryResult.version.GetString() << "'" << std::endl;
    // std::cout << "            name '" << discoveryResult.name << "'" << std::endl;
    // std::cout << "          family '" << discoveryResult.family << "'" << std::endl;
    // std::cout << "   discoveryType '" << discoveryResult.discoveryType << "'" << std::endl;
    // std::cout << "      sourceType '" << discoveryResult.sourceType << "'" << std::endl;
    // std::cout << "             uri '" << discoveryResult.uri << "'" << std::endl;
    // std::cout << "     resolvedUri '" << discoveryResult.resolvedUri << "'" << std::endl;
    // std::cout << "      sourceCode '" << discoveryResult.sourceCode << "'" << std::endl;
    // std::cout << "  ----------------------------------------------------------------" << std::endl;

    const std::string schema_name = discoveryResult.identifier.GetString();
    // std::cout << "SchemaDescription() -> SchemaBuiltinParser::Parse()";
    // std::cout << " schema_name='" << schema_name << "'";
    // std::cout << ",  sourceType='" << discoveryResult.sourceType << "'" << std::endl;

    Schema* base_schema = Schema::create(schema_name.c_str());
    if (!base_schema)
    {
      // TODO: handle this error state somehow? We don't have an Op to declare the error on...

      // Report the error as if it was coming from the Schema plugin code:
      std::cerr << "Schema::create(): error, cannot allocate new node of type '" << schema_name << "'";
      std::cerr << " - this is likely due to a coding error.";

      return PXR_NS::NdrNodeUniquePtr();
    }
    // {
    //   std::cout << "  properties:" << std::endl;
    //   const PXR_NS::NdrPropertyUniquePtrVec& properties = base_schema->_pImpl->properties;
    //   for (size_t i=0; i < properties.size(); ++i)
    //   {
    //     ShaderPropertyHandle::printNdrProperty("    ", std::cout, *properties[i]);
    //     std::cout << std::endl;
    //   }
    // }

    PXR_NS::NdrNode* ndrnode = nullptr;

    // Always handle the base schema cleanly:
    if (discoveryResult.sourceType == usgBuiltinSchemaType)
    {
      // We still need to process the properties for a base schema type since Ndr still chokes
      // on certain types, like Asset, so adding a bit of metadata to these properties usually
      // gets it through the NdrRegistry_ValidateProperty() method:



      // Build a new NdrNode from the properties, metadata, etc retrieved
      // from the ShaderSchemaHandle, with no filtering:
      ndrnode = new PXR_NS::NdrNode(discoveryResult.identifier,
                                    discoveryResult.version,
                                    discoveryResult.name,
                                    discoveryResult.family,
                                    discoveryResult.discoveryType,
                                    discoveryResult.sourceType,
                                    discoveryResult.uri,
                                    discoveryResult.resolvedUri,
                                    std::move(base_schema->_pImpl->properties),
                                    std::move(base_schema->_pImpl->metadata),
                                    std::string()/*sourceCode*/);
    }
    else
    {
      // Special-case the schema types we want to filter or process the properties for.

      // See if this schema is a shader def vs a prim def:
      ShaderSchema* shader_schema = base_schema->asShaderSchema();
      if (shader_schema)
      {
        // Shaders are handled special because USD and SdrShaderNode has specific requirements
        // for its properties that don't work for general conversion to UsdPrims. The best example
        // of this is Asset properties get converted to String in the SdrShaderNode (with extra
        // metadata describing the string's 'asset-ness'. Naive converting of those properties to
        // UsdAttributes does not recover the Asset type.

        // Construct a UsdShadeShader so we can pass it to the various UsdShade utilities
        // that expect one. This should be similar to what usd/usdShade/shaderDefParser.cpp:Parse()
        // does, but that method relies on there being an external file to load into a stage:
        PXR_NS::SdfPath schema_path = PXR_NS::SdfPath(fdk::buildPath(fdk::buildPath("/shaders", discoveryResult.sourceType.GetString()), schema_name));

        assert(schemadefs_stage);
        PXR_NS::UsdPrim schema_prim = schemadefs_stage->GetPrimAtPath(schema_path);
        if (!schema_prim)
        {
          schema_prim = schemadefs_stage->DefinePrim(schema_path, PXR_NS::TfToken("Shader"));
          assert(schema_prim);

          PXR_NS::UsdShadeShader shader_prim(schema_prim);
          assert(shader_prim);

          // Assign 'info:id' attr:
          {
            PXR_NS::UsdAttribute attr = schema_prim.CreateAttribute(PXR_NS::UsdShadeTokens->infoId,
                                                                    PXR_NS::SdfValueTypeNames->Token,
                                                                    false/*custom*/);
            attr.Set(PXR_NS::TfToken(discoveryResult.identifier));
          }
#if 1
          // ***** NOTE ******
          // If 'implemetationSource' attr is not set to 'id' then Hydra will throw an error in its material
          // network parser saying there's no valid 'info:id' - despite there being a valid info:id assigned...
          // *****************
          {
            PXR_NS::UsdAttribute attr = schema_prim.CreateAttribute(PXR_NS::UsdShadeTokens->infoImplementationSource,
                                                                    PXR_NS::SdfValueTypeNames->Token,
                                                                    false/*custom*/);
            attr.Set(PXR_NS::TfToken("id"));
          }
#else
#if 1
          shaderprim.setImplementationSourceAttr("sourceAsset");
          shaderprim.setSourceAssetAttr("./Project2DTexture.glslfx", "glslfx");
#else
          shaderprim.setImplementationSourceAttr("sourceCode");
          shaderprim.setSourceCodeAttr(source_shader_desc.getSourceCode(), "glslfx");
#endif
#endif

          // Copy metadata fields to shader Prim as 'sdrMetadata':
          shader_prim.SetSdrMetadata(base_schema->_pImpl->metadata);

          // Copy Sdr properties to shader Prim:
          const PXR_NS::NdrPropertyUniquePtrVec& properties = base_schema->_pImpl->properties;
          for (size_t i=0; i < properties.size(); ++i)
          {
            const PXR_NS::NdrPropertyUniquePtr& prop = properties[i];
            // {
            //   const PXR_NS::VtValue& defaultValue = prop->GetDefaultValue();
            //   const PXR_NS::NdrSdfTypeIndicator ndrSdfTypeIndicator = prop->GetTypeAsSdfType();
            //   const PXR_NS::SdfValueTypeName sdfType = ndrSdfTypeIndicator.first;
            //   std::cout << "********************************* sdrproperty '" << prop->GetName() << "'";
            //   std::cout << ", GetType='" << prop->GetType() << "'";
            //   std::cout << ", GetArraySize=" << prop->GetArraySize();
            //   std::cout << ", def-vtvalue-GetTypeName=" << defaultValue.GetTypeName();
            //   std::cout << ", def-vtvalue-GetType='" << defaultValue.GetType() << "'";
            //   //std::cout << ", SdrShaderProperty::GetTypeAsSdfType()='" << sdfType.GetType() << "'";
            //   std::cout << std::endl;
            //   std::cout << "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << std::endl;
            // }

            // Note - this must be the full input name, even if it's a local property so that
            // USD recognizes it as an input:
            std::string prop_full_name = (prop->IsOutput()) ? "outputs:" : "inputs:";
            prop_full_name += prop->GetName().GetString();

            // Determine the proper Sdf type for this property.
            //
            // Get the Value::Type so we can explicitly handle SdfValueTypeName conversion.
            // If we only use SdrShaderProperty::GetTypeAsSdfType() it will always translate and may
            // do the incorrect thing if the expected metadata is not present on the properties.
            // 
            const Value::Type usg_type = getUsgValueType(prop->GetType());
            assert(usg_type != Value::Type::InvalidType); // just in case...the properties should already have been validated/cleaned

            // Get the default result from the SdrProperty translator:
#if PXR_VERSION >= 2111
            const PXR_NS::NdrSdfTypeIndicator ndrSdfTypeIndicator = prop->GetTypeAsSdfType();
            PXR_NS::SdfValueTypeName sdfType = ndrSdfTypeIndicator.first;
#else
            const PXR_NS::SdfTypeIndicator sdfTypeIndicator = prop->GetTypeAsSdfType();
            PXR_NS::SdfValueTypeName sdfType = sdfTypeIndicator.first;
#endif
            PXR_NS::NdrTokenMap      sdrMetadata; // additional tags to add to property

            if (!prop->IsOutput() && discoveryResult.sourceType == glslfxSourceType)
            {
              // Skip unsupported glslfx input types:
              if (usg_type == Value::Type::String    || usg_type == Value::Type::StringArray ||
                  usg_type == Value::Type::Token     || usg_type == Value::Type::TokenArray  ||
                  usg_type == Value::Type::AssetPath || usg_type == Value::Type::AssetPathArray)
              {
                continue; // skip string types, hdSt seems to choke on them...
              }
            }

            PXR_NS::UsdAttribute attr = schema_prim.CreateAttribute(PXR_NS::TfToken(prop_full_name),
                                                                 sdfType/*SdfValueTypeName*/,
                                                                 false/*custom*/);
            if (!prop->IsOutput())
              attr.Set(prop->GetDefaultValue());

            // Copy supported SdrProperty metadata fields:
            const PXR_NS::NdrTokenMap& metadata = prop->GetMetadata();
            for (PXR_NS::NdrTokenMap::const_iterator it=metadata.begin(); it != metadata.end(); ++it)
            {
              // Skip some of the tags that get created by SdrShaderProperty ctor as
              // we don't ned them:
              static PXR_NS::NdrStringSet skip_keys({ PXR_NS::SdrPropertyMetadata->Widget });
              if (skip_keys.find(it->first.GetString()) != skip_keys.end())
                continue;
#if 1
              attr.SetMetadataByDictKey(PXR_NS::UsdShadeTokens->sdrMetadata,
                                        it->first,
                                        PXR_NS::VtValue(PXR_NS::TfToken(it->second)));
#else
              if (PXR_NS::SdfSchema::GetInstance().IsRegistered(it->first))
                attr.SetMetadata(it->first, PXR_NS::VtValue(PXR_NS::TfToken(it->second)));
#endif
            }
            // Add any additional Sdr tags:
            for (PXR_NS::NdrTokenMap::const_iterator it=sdrMetadata.begin(); it != sdrMetadata.end(); ++it)
            {
              attr.SetMetadataByDictKey(PXR_NS::UsdShadeTokens->sdrMetadata,
                                        it->first,
                                        PXR_NS::VtValue(PXR_NS::TfToken(it->second)));
            }
          }


          if (0)//(gDebugShaderPlugins)
          {
            // Adding a SdrShaderProperty to a SdrShaderNode can result in a translation!!!
            // As a sanity test retrieve what we just added and check:
            std::cout << "==============================================================================" << std::endl;
            std::string str;
            schemadefs_stage->ExportToString(&str, false/*addSourceFileComment*/);
            std::cout << str;
          }
        }

        // We get the Sdr properties by using usd/usdShade/UsdShadeShaderDefUtils.cpp:GetShaderProperties()
        // which is what usd/usdShade/shaderDefParser.cpp:Parse() does:
        PXR_NS::UsdShadeShader shader_prim(schema_prim);
        assert(shader_prim);

  // {
  //   PXR_NS::NdrTokenMap metadata = _GetSdrMetadata(shader_prim, discoveryResult.metadata);
  //   std::cout << "               **********  metadata[";
  //   for (auto it=metadata.begin(); it != metadata.end(); ++it)
  //     std::cout << " '" << it->first << "'='" << it->second << "'";
  //   std::cout << " ]" << std::endl;
  // }

        // Build a new SdrShaderNode from the properties, metadata, etc retrieved
        // from the ShaderSchemaHandle:
        ndrnode = new PXR_NS::SdrShaderNode(discoveryResult.identifier,
                                            discoveryResult.version,
                                            discoveryResult.name,
                                            discoveryResult.family,
                                            discoveryResult.discoveryType,
                                            discoveryResult.sourceType,
                                            discoveryResult.uri,
                                            discoveryResult.resolvedUri,
                                            PXR_NS::UsdShadeShaderDefUtils::GetShaderProperties(PXR_NS::UsdShadeConnectableAPI(shader_prim)),
                                            _GetSdrMetadata(shader_prim, discoveryResult.metadata),
                                            discoveryResult.sourceCode);
      } // ShaderSchema
      else
      {
        // TODO: handle other base types here?
      }

    }
    assert(ndrnode);

    delete base_schema; // no longer needed as we've moved out the contents of the ShaderDescHandle

    return PXR_NS::NdrNodeUniquePtr(ndrnode);

  } // Parse()

}; // SchemaBuiltinParser


//------------------------------------------------------------------------------------
//------------------------------------------------------------------------------------


// This map stores the Description structures for every loaded Schema plugin or
// built-in class:
using SchemaDescriptionMap = std::unordered_map<std::string, const SchemaDescription*>;
static SchemaDescriptionMap g_schema_dso_map;


/*! This ctor should only get called once from the Parse() method for each
    subclass that wants to register itself.
*/
Schema::Schema(const SchemaDescription&   pluginDescription,
               const SchemaPropertyArray& properties,
               const fdk::KeyValueMap&    metadata) :
  _pImpl(nullptr)
{
  PXR_NS::NdrPropertyUniquePtrVec sdrproperties;
  ShaderPropertyHandle::usgToSdrPropertyArray(properties, sdrproperties);

  PXR_NS::NdrTokenMap sdrmetadata;
  ShaderPropertyHandle::usgToSdrMetadata(metadata, sdrmetadata);
  {
    // TODO: unsure if this is needed:
    sdrmetadata.insert({ PXR_NS::SdrPropertyMetadata->ImplementationName,
                         pluginDescription.schemaName()
                       });
  }

  _pImpl.reset(new ShaderSchemaHandle(std::move(sdrproperties),
                                      std::move(sdrmetadata)));
  assert(_pImpl);
}


/*virtual*/
Schema::~Schema()
{
  //
}


/*! Create a Schema instance based on the type name ('Project2DTexture', 'PreviewSurface', etc.)
    Calling code takes ownership of returned pointer.
*/
/*static*/
Schema*
Schema::create(const char* schema_name)
{
  if (!schema_name || !schema_name[0])
    return nullptr;
  //std::cerr << "Schema::create('" << schema_name << "')" << std::endl;

  // Get the description by name:
  const SchemaDescription* desc = find(schema_name);
  if (!desc)
    return nullptr; // can't find schema by name...

  // Allocate a new one and return it:
  Schema* dso = desc->builder_method();
  if (!dso)
  {
    std::cerr << "Schema::create(): error, cannot allocate new node of type '" << schema_name << "'" << std::endl;
    return nullptr;
  }
  //std::cerr << "loaded description '" << desc->schemaName() << "', dso=" << dso << std::endl;

  return dso;
}


//------------------------------------------------------------------------------------
//------------------------------------------------------------------------------------


/*! Constructor sets name and label to same value.
*/
SchemaDescription::SchemaDescription(const char*   schema_name,
                                     PluginBuilder builder) :
  _schema_name(schema_name),
  builder_method(builder),
  _registry_node(nullptr)
{
  _plugin_path = ddimage_plugin_name();

  if (gDebugShaderPlugins)
  {
    std::cout << "SchemaDescription::ctor(" << schema_name << ")"<< " pluginpath='" << _plugin_path << "'" << std::endl;
  }

  // Register the Description:
  {
    const std::string dso_name(_schema_name);

    std::lock_guard<std::mutex> guard(g_map_lock); // just in case...
    if (g_schema_dso_map.find(dso_name) == g_schema_dso_map.end())
    {
      g_schema_dso_map[dso_name] = this;
    }
    else
    {
      // TODO: already exists! This should either throw an error or print a warning...
    }
  }

  // Create a temp Schema to retrieve source types:
  if (!builder_method)
  {
    std::cerr << "Schema::create(): error, plugin '" << _schema_name << "' provided a <null> builder ";
    std::cerr << "method - this is likely due to a coding error.";
    return;
  }
  Schema* schema = builder_method();
  if (!schema)
  {
    std::cerr << "Schema::create(): error, builder method failed to allocate a schema '" << _schema_name << "'";
    std::cerr << " object - this is likely due to a coding error.";
    return;
  }

  // Base type name:
  _schema_base_type = schema->baseTypeName();

  // Is this a shader?
  ShaderSchema* shader_schema = schema->asShaderSchema();
  if (shader_schema)
  {
    // Build the list of source types to build for this schema:
    schema->getSourceTypes(_source_types);

    // And always add the UsgShader source type too:
    _source_types.insert(GeomTokens.usgShader);
  }
  // {
  //   std::cout << "  base type='" << _schema_base_type << "'";
  //   std::cout << ", source types[";
  //   for (const std::string& sourceType : _source_types)
  //   {
  //     std::cout << " '" << sourceType << "'";
  //   }
  //   std::cout << " ]" << std::endl;
  // }

  // Update compiled string:
  //  const char* compiled;   // Date and DD_IMAGE_VERSION_LONG this was compiled for
  //Description::compiled = __DATE__ " for Usg-" kDDImageVersion;

  // The NdrRegistry where all the NdrNode definitions are stored and searched in USD:
  PXR_NS::SdrRegistry& sdrreg = PXR_NS::SdrRegistry::GetInstance();

  // The base schema identifer name, without any source-type extension:
  const PXR_NS::NdrIdentifier ndr_identifier(_schema_name);

  // Always add the base UsgSchema type first if there is one so that the base prim definition will
  // exist prior to adding any additional types:
  {
    // Always use usgBuiltinSchemaType as discoveryType so ShaderSchemaBuiltinParser
    // can control the parsing of the schema properties:
    PXR_NS::NdrNodeDiscoveryResult base_dr(ndr_identifier,                // identifier
                                           PXR_NS::NdrVersion(),          // version
                                           _schema_name,                  // name          (string)
                                           PXR_NS::TfToken(_schema_name), // family        (token)
                                           usgBuiltinSchemaType,          // discoveryType (token)
                                           usgBuiltinSchemaType,          // sourceType    (token)
                                           "",                            // uri           (string)
                                           "",                            // resolvedUri   (string)
                                           "",                            // sourceCode    (string)
                                           PXR_NS::NdrTokenMap(),         // metadata      (token)
                                           "",                            // blindData     (string)
                                           PXR_NS::TfToken()              // subIdentifier (token)
                                          );
    // Move the discovery result to the registry *before* asking for Node:
#if PXR_SDR_DISCOVER_PATCHED || PXR_VERSION >= 2208
    sdrreg.AddDiscoveryResult(base_dr);
#endif

    // Cause the SchemaBuiltinParser::Parse() to get called which attempts to load
    // the Schema plugin matching the DiscoveryResult's identifier name.
    // Once the Schema is successfully instantiated we can call it to acquire its
    // glsl implementation source, properties, etc:
    PXR_NS::NdrNodeConstPtr base_ndrnode = sdrreg.GetNodeByIdentifier(ndr_identifier,
                                                                      PXR_NS::NdrTokenVec({usgBuiltinSchemaType}));
    if (base_ndrnode)
    {
      // Store a new handle in this description:
      _registry_node = new ShaderDescHandle(base_ndrnode);
      assert(_registry_node);
      assert(_registry_node->isValid());
      if (gDebugShaderPlugins)
      {
        std::cout << "  successfully registered base node '" << ndr_identifier << "'";
        std::cout << ", id='" << _registry_node->ndrnode->GetIdentifier() << "'" << std::endl;
        ShaderSchemaHandle::printNdrNode("    |", std::cout, *base_ndrnode);
      }
    }
    else
    {
      std::cerr << "Schema::create(): error, failed to register node asset '" << _schema_name << "'";
      std::cerr << " - this is likely due to a coding error.";
    }
  }

  // Add any additional schema types:
  for (const Token& sourceType : _source_types)
  {
    if (sourceType == GeomTokens.usgSchema)
      continue; // skip base type just in case...

    const PXR_NS::TfToken ndr_sourceType = toTfToken(sourceType);

    // Handle ShaderSchemas:
    if (shader_schema)
    {
      // Don't pass any source code for the base shader definition:
      const std::string& sourceCode = (sourceType != GeomTokens.usgShader) ? shader_schema->sourceCode(sourceType) : std::string();

      // Always use usgBuiltinSchemaType as discoveryType so ShaderSchemaBuiltinParser
      // can control the parsing of the schema properties:
      PXR_NS::NdrNodeDiscoveryResult shader_dr(ndr_identifier,                // identifier
                                               PXR_NS::NdrVersion(),          // version
                                               _schema_name,                  // name          (string)
                                               PXR_NS::TfToken(_schema_name), // family        (token)
                                               usgBuiltinSchemaType,          // discoveryType (token)
                                               ndr_sourceType,                // sourceType    (token)
                                               "",                            // uri           (string) - leave empty!!!
                                               "",                            // resolvedUri   (string) - leave empty!!!
                                               sourceCode,                    // sourceCode    (string)
                                               PXR_NS::NdrTokenMap(),         // metadata      (tokenmap)
                                               "",                            // blindData     (string)
                                               PXR_NS::TfToken()              // subIdentifier (token)
                                              );
      // Move the discovery result to the registry *before* asking for Node:
#if PXR_SDR_DISCOVER_PATCHED || PXR_VERSION >= 2208
      sdrreg.AddDiscoveryResult(shader_dr);
#endif

      PXR_NS::SdrShaderNodeConstPtr shader_sdrnode = sdrreg.GetShaderNodeByIdentifier(ndr_identifier,
                                                                                      PXR_NS::NdrTokenVec({ndr_sourceType}));
      if (shader_sdrnode)
      {
        if (gDebugShaderPlugins)
        {
          std::cout << "  successfully registered companion shader asset node '" << _schema_name << "'";
          std::cout << ", sourceType='" << ndr_sourceType << "'";
          std::cout << ", id='" << shader_sdrnode->GetIdentifier() << "'" << std::endl;
          ShaderSchemaHandle::printSdrShaderNode("      |", std::cout, *shader_sdrnode);
        }
      }
    }
    else
    {
      // TODO: handle other Schema base types here? Cameras? Lights?
    }
  }

  delete schema;

}


//----------------------------------------------------------------


const SchemaDescHandle*
SchemaDescription::registryNode() const { return _registry_node; }

const std::string&
SchemaDescription::registryIdentifier() const
{
  return (_registry_node) ? _registry_node->ndrnode->GetIdentifier().GetString() : empty_string;
}

/*static*/ const SchemaDescription*
SchemaDescription::getRegistryIdentifier(const std::string& identifier)
{
  std::lock_guard<std::mutex> guard(g_map_lock); // just in case...

  SchemaDescriptionMap::const_iterator it = g_schema_dso_map.find(identifier);
  return (it != g_schema_dso_map.end()) ? it->second : nullptr;
}


/*!
*/
/*static*/
const SchemaDescription*
SchemaDescription::find(const char* schema_name)
{
  if (!schema_name || !schema_name[0])
    return nullptr;  // just in case...

  const std::string dso_name(schema_name);
  //std::cout << "SchemaDescription::find('" << dso_name << "')" << std::endl;

  // Search for existing dso using the base schemaName() name
  {
    std::lock_guard<std::mutex> guard(g_map_lock); // just in case...
    SchemaDescriptionMap::const_iterator it = g_schema_dso_map.find(dso_name);
    if (it != g_schema_dso_map.end())
    {
      //std::cout << "  found existing Description=" << it->second << std::endl;
      return it->second;
    }
  }

  // Not found, prepend pluginPrefix() to name and search the plugin paths for
  // the plugin dso file (ie slrProject.so, slrDisplacement.tcl, etc)
  std::string plugin_name(pluginPrefix());
  plugin_name += schema_name;
  plugin_name += pluginSuffix();
  //std::cout << "  plugin_name='" << plugin_name << "'" << std::endl;

  // Use the stock DDImage plugin load method, which supports .tcl redirectors.

  // TODO: plugin_load() says that it returns nullptr if a plugin is not
  //  loaded but that does not appear to be the case. It returns the path
  //  to the plugin it *attempted* to load, but only by checking plugin_error()
  //  can we tell if dlopen() failed and what was returned in dlerror()...
  const char* plugin_path = ddimage_plugin_load(plugin_name.c_str(), nullptr/*altname*/);
  if (plugin_path == plugin_methods_not_connected)
  {
    std::cerr << "SchemaDescription::find('" << plugin_name << "') ";
    std::cerr << "fatal error: Usg plugin system is disconnected, no plugins will load!" << std::endl;
    return nullptr;  // yipe, can't load any plugins!
  }
  if (!plugin_path || !plugin_path[0])
  {
    std::cerr << "SchemaDescription::find('" << plugin_name << "') ";
    std::cerr << "error, plugin not found." << std::endl;
    return nullptr;  // plugin not found!
  }
  // Was there a dlerror() on load?
  if (ddimage_plugin_error())
  {
    std::cerr << "SchemaDescription::find('" << plugin_name << "') ";
    std::cerr << "error, plugin not loaded, dlopen error '" << ddimage_plugin_error() << "'" << std::endl;
    return nullptr;  // plugin not found!
  }

  // Plugin found and loaded, return the pointer that was added to the map:
  const SchemaDescription* dso_desc = nullptr;
  {
    std::lock_guard<std::mutex> guard(g_map_lock); // just in case...
    SchemaDescriptionMap::const_iterator it = g_schema_dso_map.find(dso_name);
    if (it != g_schema_dso_map.end())
      dso_desc = it->second;
  }
  if (!dso_desc)
  {
      // Error - the plugin should have been found! If not then the plugin
      // likely does not have defined SchemaDescriptions matching 'plugin_name':
      std::cerr << "SchemaDescription::find('" << dso_name << "') ";
      std::cerr << "error, plugin did not define a SchemaDescription matching ";
      std::cerr << "the plugin name - this is likely a coding error.";
      if (ddimage_plugin_error())
          std::cerr << " '" << ddimage_plugin_error() << "'";
      std::cerr << std::endl;
      return nullptr;  // plugin not found!
  }

  return dso_desc;
}

//------------------------------------------------------------------------------------
//------------------------------------------------------------------------------------

/*static*/ void
ShaderSchemaHandle::printNdrNode(const char*         prefix,
                                 std::ostream&       o,
                                 const PXR_NS::NdrNode& ndrnode)
{
  o << prefix << "----------------------------------------------------------------" << std::endl;
  o << prefix << "NdrNode():" << std::endl;
  o << prefix << "         identifier '" << ndrnode.GetIdentifier() << "'" << std::endl;
  o << prefix << "            version '" << ndrnode.GetVersion().GetString() << "'" << std::endl;
  o << prefix << "               name '" << ndrnode.GetName() << "'" << std::endl;
  o << prefix << "             family '" << ndrnode.GetFamily() << "'" << std::endl;
  o << prefix << "            context '" << ndrnode.GetContext() << "'" << std::endl;
  o << prefix << "         sourceType '" << ndrnode.GetSourceType() << "'" << std::endl;
  o << prefix << "      definitionUri '" << ndrnode.GetResolvedDefinitionURI() << "'" << std::endl;
  o << prefix << "  implementationUri '" << ndrnode.GetResolvedImplementationURI() << "'" << std::endl;
  //o << prefix << "         sourceCode '" << ndrnode.GetSourceCode() << "'" << std::endl;
  //o << prefix << "         infoString '" << ndrnode.GetInfoString() << "'" << std::endl;
  o << prefix << "    ----------------------------------------------" << std::endl;

  const PXR_NS::NdrTokenVec& input_names = ndrnode.GetInputNames();
  if (input_names.size() > 0)
  {
    o << prefix << "    ---------- inputs -----------" << std::endl;
    for (size_t i=0; i < input_names.size(); ++i)
    {
      const PXR_NS::NdrPropertyConstPtr input = ndrnode.GetInput(input_names[i]);
      if (input)
      {
        o << prefix;
        ShaderPropertyHandle::printNdrProperty("    ", o, *input);
        o << std::endl;
      }
    }
  }

  const PXR_NS::NdrTokenVec& output_names = ndrnode.GetOutputNames();
  if (output_names.size() > 0)
  {
    o << prefix << "    ---------- outputs -----------" << std::endl;
    for (size_t i=0; i < output_names.size(); ++i)
    {
      const PXR_NS::NdrPropertyConstPtr output = ndrnode.GetOutput(output_names[i]);
      if (output)
      {
        o << prefix;
        ShaderPropertyHandle::printNdrProperty("    ", o, *output);
        o << ")" << std::endl;
      }
    }
  }

  o << prefix << "----------------------------------------------------------------" << std::endl;

}

/*static*/ void
ShaderSchemaHandle::printSdrShaderNode(const char*               prefix,
                                       std::ostream&             o,
                                       const PXR_NS::SdrShaderNode& sdrnode)
{
  o << prefix << "----------------------------------------------------------------" << std::endl;
  o << prefix << "SdrShaderNode():" << std::endl;
  o << prefix << "         identifier '" << sdrnode.GetIdentifier() << "'" << std::endl;
  o << prefix << "            version '" << sdrnode.GetVersion().GetString() << "'" << std::endl;
  o << prefix << "               name '" << sdrnode.GetName() << "'" << std::endl;
  o << prefix << "             family '" << sdrnode.GetFamily() << "'" << std::endl;
  o << prefix << "            context '" << sdrnode.GetContext() << "'" << std::endl;
  o << prefix << "         sourceType '" << sdrnode.GetSourceType() << "'" << std::endl;
  o << prefix << "      definitionUri '" << sdrnode.GetResolvedDefinitionURI() << "'" << std::endl;
  o << prefix << "  implementationUri '" << sdrnode.GetResolvedImplementationURI() << "'" << std::endl;
  o << prefix << "         sourceCode '" << sdrnode.GetSourceCode() << "'" << std::endl;
  //o << prefix << "         infoString '" << sdrnode.GetInfoString() << "'" << std::endl;
  o << prefix << "    ----------------------------------------------" << std::endl;
  o << prefix << " implementationName '" << sdrnode.GetImplementationName() << "'" << std::endl;
  o << prefix << "            [label] '" << sdrnode.GetLabel() << "'" << std::endl;
  o << prefix << "         [category] '" << sdrnode.GetCategory() << "'" << std::endl;
  o << prefix << "             [role] '" << sdrnode.GetRole() << "'" << std::endl;
  o << prefix << "             [help] '" << sdrnode.GetHelp() << "'" << std::endl;

  const PXR_NS::NdrTokenVec& input_names = sdrnode.GetInputNames();
  if (input_names.size() > 0)
  {
    o << prefix << "    ---------- inputs -----------" << std::endl;
    for (size_t i=0; i < input_names.size(); ++i)
    {
      const PXR_NS::SdrShaderPropertyConstPtr input = sdrnode.GetShaderInput(input_names[i]);
      if (input)
      {
        o << prefix;
        ShaderPropertyHandle::printSdrProperty("    ", o, *input);
        o << std::endl;
      }
    }
  }

  const PXR_NS::NdrTokenVec& output_names = sdrnode.GetOutputNames();
  if (output_names.size() > 0)
  {
    o << prefix << "    ---------- outputs -----------" << std::endl;
    for (size_t i=0; i < output_names.size(); ++i)
    {
      const PXR_NS::SdrShaderPropertyConstPtr output = sdrnode.GetShaderOutput(output_names[i]);
      if (output)
      {
        o << prefix;
        ShaderPropertyHandle::printSdrProperty("    ", o, *output);
        o << ")" << std::endl;
      }
    }
  }

  o << prefix << "----------------------------------------------------------------" << std::endl;

}



} // namespace usg


PXR_NAMESPACE_OPEN_SCOPE

  // Register the Usd plugin outside usg namespace:
  NDR_REGISTER_PARSER_PLUGIN(usg::SchemaBuiltinParser)

  // Add the dedicated source path to the FnUsdAbstraction's plugInfo.json file:
  // Note that we only need to do this once at runtime.
  struct RegisterFnUsdAbstractionResources
  {
    RegisterFnUsdAbstractionResources()
    {
      const std::string nukeInstallDir = TfGetPathName(ArchGetExecutablePath());
      const std::string pluginDir      = TfStringCatPaths(nukeInstallDir, "FnUsdAbstraction/resources");
      PlugRegistry::GetInstance().RegisterPlugins(pluginDir);
    }
  } register_usg_lib;

PXR_NAMESPACE_CLOSE_SCOPE
