// Copyright (c) 2022 The Foundry Visionmongers Ltd. All Rights Reserved.

#include "usg/geom/SchemaProperty.h"
#include "usg/base/Token.h"

#include "usg/private/UsdImpl.h"

#include "pxr/usd/sdr/shaderProperty.h"
#include "pxr/usd/sdr/shaderMetadataHelpers.h"

#include "pxr/base/gf/matrix4f.h"
#include "pxr/base/tf/stringUtils.h" // for TfStringGetBeforeSuffix

#include <unordered_map>

//------------------------------------------------------------------------------------


const std::string gEnvDebugShaderBuilding    = ((std::getenv("USG_DEBUG_SHADER_BUILDING") != nullptr) ?
                                                 std::getenv("USG_DEBUG_SHADER_BUILDING") : "");
const std::string gEnvDebugShaderConnections = ((std::getenv("USG_DEBUG_SHADER_CONNECTIONS") != nullptr) ?
                                                 std::getenv("USG_DEBUG_SHADER_CONNECTIONS") : "");
const std::string gEnvDebugShaderProperties  = ((std::getenv("USG_DEBUG_SHADER_PROPERTIES") != nullptr) ?
                                                 std::getenv("USG_DEBUG_SHADER_PROPERTIES") : "");
const bool gDebugShaderBuilding    = (gEnvDebugShaderBuilding    == "1");
const bool gDebugShaderConnections = (gEnvDebugShaderConnections == "1");
const bool gDebugShaderProperties  = (gEnvDebugShaderProperties  == "1");


//------------------------------------------------------------------------------------


namespace usg
{


inline PXR_NS::GfVec3f
fdkVec3hAsGfVec3f(const fdk::Vec3h& v)
{
  const PXR_NS::GfVec3h& pv = reinterpret_cast<const PXR_NS::GfVec3h&>(v);
  return PXR_NS::GfVec3f(float(pv[0]), float(pv[1]), float(pv[2]));
}

inline PXR_NS::GfVec3f
fdkVec3dAsGfVec3f(const fdk::Vec3d& v)
{
  const PXR_NS::GfVec3d& pv = reinterpret_cast<const PXR_NS::GfVec3d&>(v);
  return PXR_NS::GfVec3f(float(pv[0]), float(pv[1]), float(pv[2]));
}


//------------------------------------------------------------------------------------

/* Usd supported type mappings defined in sdr/shaderProperty.cpp:

  // This only establishes EXACT mappings. If a mapping is not included here,
  // a one-to-one mapping isn't possible.
  const TokenToSdfTypeMap& GetTokenTypeToSdfType()
  {
      static const TokenToSdfTypeMap tokenTypeToSdfType  = {
          {SdrPropertyTypes->Int,     SdfValueTypeNames->Int},
          {SdrPropertyTypes->String,  SdfValueTypeNames->String},
          {SdrPropertyTypes->Float,   SdfValueTypeNames->Float},
          {SdrPropertyTypes->Color,   SdfValueTypeNames->Color3f},
          {SdrPropertyTypes->Point,   SdfValueTypeNames->Point3f},
          {SdrPropertyTypes->Normal,  SdfValueTypeNames->Normal3f},
          {SdrPropertyTypes->Vector,  SdfValueTypeNames->Vector3f},
          {SdrPropertyTypes->Matrix,  SdfValueTypeNames->Matrix4d}
      };
      return tokenTypeToSdfType;
  }

  // The array equivalent of the above map
  const TokenToSdfTypeMap& GetTokenTypeToSdfArrayType()
  { 
      static const TokenToSdfTypeMap tokenTypeToSdfArrayType  = {
          {SdrPropertyTypes->Int,     SdfValueTypeNames->IntArray},
          {SdrPropertyTypes->String,  SdfValueTypeNames->StringArray},
          {SdrPropertyTypes->Float,   SdfValueTypeNames->FloatArray},
          {SdrPropertyTypes->Color,   SdfValueTypeNames->Color3fArray},
          {SdrPropertyTypes->Point,   SdfValueTypeNames->Point3fArray},
          {SdrPropertyTypes->Normal,  SdfValueTypeNames->Normal3fArray},
          {SdrPropertyTypes->Vector,  SdfValueTypeNames->Vector3fArray},
          {SdrPropertyTypes->Matrix,  SdfValueTypeNames->Matrix4dArray}
      };
      return tokenTypeToSdfArrayType;
  }
*/


/*!
*/


/*  Value::Type map to convert commonly-used types to the correct Sdr type to avoid
    common propery declaration mistakes.

    ***** IMPORTANT NOTE ******
    'elements' cannot be copied directly to a SdrShaderProperty's arraySize as any non-zero value will
    trigger a conversion of the default type to a VtArray causing lots of errors...!

    When passing 'elements' to an SdrShaderProperty ctor use a if/then like this:
      arraySize = (params.elements > 1) ? params.elements : 0;


  See Pixar docs re. Sdr/Ndr types:
    https://graphics.pixar.com/usd/release/api/sdr_page_front.html
  "Sdr defines a limited set of types:
    Int
    String
    Float
    Color
    Point
    Normal
    Vector
    Matrix
    Struct
    Terminal
    VStruct
    Unknown
  "
*/
using ValueTypeToSdrTypeMap = std::unordered_map<Value::Type, ShaderPropertyHandle::SdrTypeParams>;
static const ValueTypeToSdrTypeMap conform_value_type_map =
{
  //    input type             sdr_type   elements is_array        input type                    sdr_type  elements is_array
  { usg::Value::Half,     {usg::Value::Float,  0,  false} }, { usg::Value::HalfArray,     {usg::Value::Float,  0,  true} },
  { usg::Value::Double,   {usg::Value::Float,  0,  false} }, { usg::Value::DoubleArray,   {usg::Value::Float,  0,  true} },
  { usg::Value::TimeCode, {usg::Value::Float,  0,  false} }, { usg::Value::TimeCodeArray, {usg::Value::Float,  0,  true} },

  { usg::Value::Bool,     {usg::Value::Int,    0,  false} }, { usg::Value::BoolArray,     {usg::Value::Int,    0,  true} },
  { usg::Value::Int,      {usg::Value::Int,    0,  false} }, { usg::Value::IntArray,      {usg::Value::Int,    0,  true} },
  { usg::Value::Int64,    {usg::Value::Int,    0,  false} }, { usg::Value::Int64Array,    {usg::Value::Int,    0,  true} },
  { usg::Value::UChar,    {usg::Value::Int,    0,  false} }, { usg::Value::UCharArray,    {usg::Value::Int,    0,  true} },
  { usg::Value::UInt,     {usg::Value::Int,    0,  false} }, { usg::Value::UIntArray,     {usg::Value::Int,    0,  true} },
  { usg::Value::UInt64,   {usg::Value::Int,    0,  false} }, { usg::Value::UInt64Array,   {usg::Value::Int,    0,  true} },

  { usg::Value::Int2,     {usg::Value::Int,    2,  false} }, // convert to int2 array
  { usg::Value::Half2,    {usg::Value::Float,  2,  false} }, // convert to float2 array
  { usg::Value::Float2,   {usg::Value::Float,  2,  false} }, // convert to float2 array
  { usg::Value::Double2,  {usg::Value::Float,  2,  false} }, // convert to float2 array

  { usg::Value::Int3,     {usg::Value::Int,    3,  false} }, // convert to int3 array
  { usg::Value::Half3,    {usg::Value::Float,  3,  false} }, // convert to float3 array
  { usg::Value::Float3,   {usg::Value::Float,  3,  false} }, // convert to float3 array
  { usg::Value::Double3,  {usg::Value::Float,  3,  false} }, // convert to float3 array

  { usg::Value::Int4,     {usg::Value::Int,    4,  false} }, // convert to int4 array
  { usg::Value::Half4,    {usg::Value::Float,  4,  false} }, // convert to float4 array
  { usg::Value::Float4,   {usg::Value::Float,  4,  false} }, // convert to float4 array
  { usg::Value::Double4,  {usg::Value::Float,  4,  false} }, // convert to float4 array

  { usg::Value::Channel,  {usg::Value::Int,    0,  false} }, { usg::Value::ChannelArray,  {usg::Value::Int,    0,  true} },

  { usg::Value::Color3h,  {usg::Value::Color,  0,  false} }, { usg::Value::Color3hArray,  {usg::Value::Color,  0,  true} },
  { usg::Value::Color3f,  {usg::Value::Color,  0,  false} }, { usg::Value::Color3fArray,  {usg::Value::Color,  0,  true} },
  { usg::Value::Color3d,  {usg::Value::Color,  0,  false} }, { usg::Value::Color3dArray,  {usg::Value::Color,  0,  true} },

  { usg::Value::Color4h,  {usg::Value::Float,  4,  false} }, // convert to float4 array
  { usg::Value::Color4f,  {usg::Value::Float,  4,  false} }, // convert to float4 array
  { usg::Value::Color4d,  {usg::Value::Float,  4,  false} }, // convert to float4 array

  { usg::Value::Point3h,  {usg::Value::Point,  0,  false} }, { usg::Value::Point3hArray,  {usg::Value::Point,  0,  true} },
  { usg::Value::Point3f,  {usg::Value::Point,  0,  false} }, { usg::Value::Point3fArray,  {usg::Value::Point,  0,  true} },
  { usg::Value::Point3d,  {usg::Value::Point,  0,  false} }, { usg::Value::Point3dArray,  {usg::Value::Point,  0,  true} },

  { usg::Value::Normal3h, {usg::Value::Normal, 0,  false} }, { usg::Value::Normal3hArray, {usg::Value::Normal, 0,  true} },
  { usg::Value::Normal3f, {usg::Value::Normal, 0,  false} }, { usg::Value::Normal3fArray, {usg::Value::Normal, 0,  true} },
  { usg::Value::Normal3d, {usg::Value::Normal, 0,  false} }, { usg::Value::Normal3dArray, {usg::Value::Normal, 0,  true} },

  { usg::Value::Vector3h, {usg::Value::Vector, 0,  false} }, { usg::Value::Vector3hArray, {usg::Value::Vector, 0,  true} },
  { usg::Value::Vector3f, {usg::Value::Vector, 0,  false} }, { usg::Value::Vector3fArray, {usg::Value::Vector, 0,  true} },
  { usg::Value::Vector3d, {usg::Value::Vector, 0,  false} }, { usg::Value::Vector3dArray, {usg::Value::Vector, 0,  true} },

  { usg::Value::Matrix4d, {usg::Value::Matrix, 0,  false} }, { usg::Value::Matrix4dArray, {usg::Value::Matrix, 0,  true} },

};


/*! Convert commonly-used Value::Types to the correct Sdr type to avoid common propery declaration mistakes.

    This is a convenience function as the schema property declarations should be restricted
    to the allowable Sdr types but it's far too easy to make a mistake and Usd doesn't provide
    a useful error message to identify the problem at runtime...
*/
/*static*/ Value::Type
ShaderPropertyHandle::conformUsgValueTypeToSdrValueType(Value::Type usg_type)
{

  const ValueTypeToSdrTypeMap::const_iterator it = conform_value_type_map.find(usg_type);
  return (it != conform_value_type_map.end()) ? it->second.sdr_type : usg_type;
}

template <class usdT>
inline PXR_NS::VtValue
buildMinVtArray(const usdT& val)
{
  return PXR_NS::VtValue( PXR_NS::VtArray<usdT>({val}) );
}



/*! Make sure the VtValue matches the formal SdrShader property types, translating any
    that we can to avoid common propery declaration mistakes.

    This is a convenience function as the schema property declarations should be restricted
    to the allowable Sdr types but it's far too easy to make a mistake and Usd doesn't provide
    a useful error message to identify the problem at runtime...

    Also arrays are tricky to handle as USD requires the default value passed to the property
    ctor to ** match ** the declared type but if the VtArray for the default value is empty
    then USD says 'oh, you really wanted an Int property, not an IntArray property'...
    So to avoid those situations we fill in a minimum-sized VtArray with zeroes for the default.
    ...sigh...
*/
/*static*/ PXR_NS::VtValue
ShaderPropertyHandle::conformUsgValueToVtValue(const usg::Value&    usg_val_in,
                                               const SdrTypeParams& out_type_params)
{
  const usg::Value::Type val_in_type       = usg_val_in.getType();
  const size_t           val_in_array_size = usg_val_in.getArraySize();
  // std::cout << "      conformUsgValueToVtValue()";
  // std::cout << " sdr_type='" << Value::getTypeName(out_type_params.sdr_type);
  // std::cout << "', elements=" << out_type_params.elements;
  // std::cout << ", val.type='" << Value::getTypeName(val_in_type) << "'";
  // std::cout << ", val.array_size=" << val_in_array_size;
  // std::cout << std::endl;

  //----------------------------------------------------------------------------------------
  if      (out_type_params.sdr_type == usg::Value::AssetPath)
  {
    if (out_type_params.is_array)
      return (val_in_array_size > 0) ? usg_val_in.impl()->vtvalue : buildMinVtArray(PXR_NS::SdfAssetPath()); // handle zero-sized source arrays

    if      (val_in_type == usg::Value::String   ) return PXR_NS::VtValue(PXR_NS::SdfAssetPath(usg_val_in.get<std::string>()));
    else if (val_in_type == usg::Value::Token    ) return PXR_NS::VtValue(PXR_NS::SdfAssetPath(usg_val_in.getToken().asString()));
    else if (val_in_type == usg::Value::Path     ) return PXR_NS::VtValue(PXR_NS::SdfAssetPath(usg_val_in.getPath().asString()));
    else if (val_in_type == usg::Value::AssetPath) return PXR_NS::VtValue(toSdfAssetPath(usg_val_in.getAssetPath()));
    // Default to empty SdfAssetPath if input type is unrecognized as a string type:
    return PXR_NS::VtValue(PXR_NS::SdfAssetPath());
  }
  //----------------------------------------------------------------------------------------
  else if (out_type_params.sdr_type == usg::Value::Path)
  {
    if (out_type_params.is_array)
      return (val_in_array_size > 0) ? usg_val_in.impl()->vtvalue : buildMinVtArray(PXR_NS::SdfPath()); // handle zero-sized source arrays

    // Be careful to not return an SdfPath with '' passed into it, it will cause a Tf error:
    if (val_in_type == usg::Value::Path) return PXR_NS::VtValue(toSdfPath(usg_val_in.getPath()));
    std::string s;
    if      (val_in_type == usg::Value::String   ) s = usg_val_in.get<std::string>();
    else if (val_in_type == usg::Value::Token    ) s = usg_val_in.getToken().asString();
    else if (val_in_type == usg::Value::AssetPath) s = usg_val_in.getAssetPath().asString();
    return PXR_NS::VtValue((!s.empty()) ? PXR_NS::SdfPath(s) : PXR_NS::SdfPath());
  }
  //----------------------------------------------------------------------------------------
  else if (out_type_params.sdr_type == usg::Value::Token)
  {
    if (out_type_params.is_array)
      return (val_in_array_size > 0) ? usg_val_in.impl()->vtvalue : buildMinVtArray(PXR_NS::TfToken("")); // handle zero-sized source arrays

    if      (val_in_type == usg::Value::String   ) return PXR_NS::VtValue(PXR_NS::TfToken(usg_val_in.get<std::string>()));
    else if (val_in_type == usg::Value::Token    ) return PXR_NS::VtValue(toTfToken(usg_val_in.getToken()));
    else if (val_in_type == usg::Value::Path     ) return PXR_NS::VtValue(PXR_NS::TfToken(usg_val_in.getPath().asString()));
    else if (val_in_type == usg::Value::AssetPath) return PXR_NS::VtValue(PXR_NS::TfToken(usg_val_in.getAssetPath().asString()));
    // Default to empty TfToken if input type is unrecognized as a string type:
    return PXR_NS::VtValue(PXR_NS::TfToken());
  }
  //----------------------------------------------------------------------------------------
  else if (out_type_params.sdr_type == usg::Value::String)
  {
// TODO: make this convert when we actually have implemented Token and Path types
    if (out_type_params.is_array)
      return (val_in_array_size > 0) ? usg_val_in.impl()->vtvalue : buildMinVtArray(std::string("")); // handle zero-sized source arrays

    if      (val_in_type == usg::Value::String   ) return PXR_NS::VtValue(usg_val_in.get<std::string>());
    else if (val_in_type == usg::Value::Token    ) return PXR_NS::VtValue(usg_val_in.getToken().asString());
    else if (val_in_type == usg::Value::Path     ) return PXR_NS::VtValue(usg_val_in.getPath().asString());
    else if (val_in_type == usg::Value::AssetPath) return PXR_NS::VtValue(usg_val_in.getAssetPath().asString());
    // Default to empty string if input type is unrecognized as a string type:
    return PXR_NS::VtValue(std::string());
  }
  //----------------------------------------------------------------------------------------
  else if (out_type_params.sdr_type == usg::Value::Int)
  {
    if (out_type_params.is_array)
      return (val_in_array_size > 0) ? usg_val_in.impl()->vtvalue : buildMinVtArray<int32_t>(0); // handle zero-sized source arrays

    if      (val_in_type == usg::Value::Int  ) return usg_val_in.impl()->vtvalue;
    else if (val_in_type == usg::Value::Int64) return PXR_NS::VtValue(int32_t(usg_val_in.get<int64_t>()));
    // These aren't needed until usg::Value supports these types in the Value::ctors:
    //else if (val_in_type == usg::Value::UChar ) return PXR_NS::VtValue(int32_t(usg_val_in.get<uint8_t>()));
    //else if (val_in_type == usg::Value::UInt  ) return PXR_NS::VtValue(int32_t(usg_val_in.get<uint32_t>()));
    //else if (val_in_type == usg::Value::UInt64) return PXR_NS::VtValue(int32_t(usg_val_in.get<uint64_t>()));
    //else if (val_in_type == usg::Value::Bool  ) return PXR_NS::VtValue(int32_t(usg_val_in.get<int32_t>()));
  }
  //----------------------------------------------------------------------------------------
  else if (out_type_params.sdr_type == usg::Value::Float)
  {
// TODO: I think we also need to handle conversion of multi-element arrays in here!
    if (out_type_params.is_array)
      return (val_in_array_size > 0) ? usg_val_in.impl()->vtvalue : buildMinVtArray<float>(0.0f); // handle zero-sized source arrays

    if (out_type_params.elements == 1)
    {
      if      (val_in_type == usg::Value::Float ) return usg_val_in.impl()->vtvalue;
      else if (val_in_type == usg::Value::Double) return PXR_NS::VtValue(float(usg_val_in.get<double>()));
      else if (val_in_type == usg::Value::Half  ) return PXR_NS::VtValue(float(reinterpret_cast<const PXR_NS::GfHalf&>(usg_val_in.get<fdk::half>())));
    }
    else if (out_type_params.elements == 2)
    {
      PXR_NS::GfVec2f vector;
      if      (val_in_type == usg::Value::Float2)
      {
        const fdk::Vec2f& pv = usg_val_in.get<fdk::Vec2f>();
        vector = reinterpret_cast<const PXR_NS::GfVec2f&>(pv);
      }
      else if (val_in_type == usg::Value::Double2)
      {
        const fdk::Vec2f pv = usg_val_in.get<fdk::Vec2d>();
        vector = reinterpret_cast<const PXR_NS::GfVec2f&>(pv);
      }
      else if (val_in_type == usg::Value::Half2) 
      {
        const fdk::Vec2f pv = usg_val_in.get<fdk::Vec2h>();
        vector = reinterpret_cast<const PXR_NS::GfVec2f&>(pv);
      }
      return PXR_NS::VtValue(vector);
    }
    else if (out_type_params.elements == 3)
    {
      PXR_NS::GfVec3f vector;
      if      (val_in_type == usg::Value::Float3)
      {
        const fdk::Vec3f& pv = usg_val_in.get<fdk::Vec3f>();
        vector = reinterpret_cast<const PXR_NS::GfVec3f&>(pv);
      }
      else if (val_in_type == usg::Value::Double3)
      {
        const fdk::Vec3f pv(usg_val_in.get<fdk::Vec3d>());
        vector = PXR_NS::GfVec3f(pv.x, pv.y, pv.z);
      }
      else if (val_in_type == usg::Value::Half3) 
      {
        const fdk::Vec3f pv = usg_val_in.get<fdk::Vec3h>();
        vector = PXR_NS::GfVec3f(pv.x, pv.y, pv.z);
      }
      return PXR_NS::VtValue(vector);
    }
    else if (out_type_params.elements == 4)
    {
      PXR_NS::GfVec4f vector;
      if      (val_in_type == usg::Value::Float4)
      {
        const fdk::Vec4f& pv = usg_val_in.get<fdk::Vec4f>();
        vector = reinterpret_cast<const PXR_NS::GfVec4f&>(pv);
      }
      else if (val_in_type == usg::Value::Double4)
      {
        const fdk::Vec4f pv = usg_val_in.get<fdk::Vec4d>();
        vector = reinterpret_cast<const PXR_NS::GfVec4f&>(pv);
      }
      else if (val_in_type == usg::Value::Half4) 
      {
        const fdk::Vec4f pv = usg_val_in.get<fdk::Vec4h>();
        vector = reinterpret_cast<const PXR_NS::GfVec4f&>(pv);
      }
      return PXR_NS::VtValue(vector);
    }
  }
  //----------------------------------------------------------------------------------------
  else if (out_type_params.sdr_type == usg::Value::Color)
  {
    if      (val_in_type == usg::Value::Float3 ) return usg_val_in.impl()->vtvalue;
    else if (val_in_type == usg::Value::Double3) return PXR_NS::VtValue(fdkVec3dAsGfVec3f(usg_val_in.get<fdk::Vec3d>()));
    else if (val_in_type == usg::Value::Half3  ) return PXR_NS::VtValue(fdkVec3hAsGfVec3f(usg_val_in.get<fdk::Vec3h>()));
  }
  //----------------------------------------------------------------------------------------
  else if (out_type_params.sdr_type == usg::Value::Point)
  {
    if      (val_in_type == usg::Value::Float3 ) return usg_val_in.impl()->vtvalue;
    else if (val_in_type == usg::Value::Double3) return PXR_NS::VtValue(fdkVec3dAsGfVec3f(usg_val_in.get<fdk::Vec3d>()));
    else if (val_in_type == usg::Value::Half3  ) return PXR_NS::VtValue(fdkVec3hAsGfVec3f(usg_val_in.get<fdk::Vec3h>()));
  }
  //----------------------------------------------------------------------------------------
  else if (out_type_params.sdr_type == usg::Value::Normal)
  {
    if      (val_in_type == usg::Value::Float3 ) return usg_val_in.impl()->vtvalue;
    else if (val_in_type == usg::Value::Double3) return PXR_NS::VtValue(fdkVec3dAsGfVec3f(usg_val_in.get<fdk::Vec3d>()));
    else if (val_in_type == usg::Value::Half3  ) return PXR_NS::VtValue(fdkVec3hAsGfVec3f(usg_val_in.get<fdk::Vec3h>()));
  }
  //----------------------------------------------------------------------------------------
  else if (out_type_params.sdr_type == usg::Value::Vector)
  {
    if      (val_in_type == usg::Value::Float3 ) return usg_val_in.impl()->vtvalue;
    else if (val_in_type == usg::Value::Double3) return PXR_NS::VtValue(fdkVec3dAsGfVec3f(usg_val_in.get<fdk::Vec3d>()));
    else if (val_in_type == usg::Value::Half3  ) return PXR_NS::VtValue(fdkVec3hAsGfVec3f(usg_val_in.get<fdk::Vec3h>()));
  }
  //----------------------------------------------------------------------------------------
  else if (out_type_params.sdr_type == usg::Value::Matrix)
  {
    if (val_in_type == usg::Value::Matrix4d) return usg_val_in.impl()->vtvalue;
  }

  // No filter applied
  return usg_val_in.impl()->vtvalue;
}

/*static*/ PXR_NS::VtValue
ShaderPropertyHandle::conformUsgValueToVtValue(const Value& val,
                                               Value::Type  usgtype_hint)
{
  const ValueTypeToSdrTypeMap::const_iterator it = conform_value_type_map.find(usgtype_hint);
  if (it != conform_value_type_map.end())
    return conformUsgValueToVtValue(val, it->second);

  return conformUsgValueToVtValue(val/*usg_val*/,
                                  ShaderPropertyHandle::SdrTypeParams(usgtype_hint, 0/*elements*/, false/*is_array*/));
}


void
ShaderPropertyHandle::usgToSdrMetadata(const fdk::KeyValueMap& usgmetadata,
                                       PXR_NS::NdrTokenMap&    sdrmetadata)
{
  sdrmetadata.clear();
  for (auto it=usgmetadata.cbegin(); it != usgmetadata.cend(); ++it)
  {
    sdrmetadata[PXR_NS::TfToken(it->first)] = it->second;
  }
}


void
ShaderPropertyHandle::usgToSdrPropertyArray(const SchemaPropertyArray&       usgproperties,
                                            PXR_NS::NdrPropertyUniquePtrVec& sdrproperties)
{
  sdrproperties.clear();
  sdrproperties.reserve(usgproperties.size());
  for (size_t i=0; i < usgproperties.size(); ++i)
  {
    const SchemaProperty& prop = usgproperties[i];
    if (!prop.isValid())
      continue; // just in case...

    const PXR_NS::SdrShaderProperty* src_sdrprop = prop.impl()->getProperty();
    assert(src_sdrprop);
    const PXR_NS::NdrTokenMap& src_metadata = src_sdrprop->GetMetadata();

    PXR_NS::NdrTokenMap out_metadata = src_metadata;

    // Note: much of the SdrProperty conversion logic resides in usdShade/shderDefUtils.cpp
    // in the GetShaderProperties() method which converts a UsdShadeShader to a series of
    // SdrProperties.
    // We don't want to replicate that code but this routine is a good place to do
    // any specific pre-massaging required before the SdrShaderProperty ctor does its
    // own thing. For example the ctor tests for 'connectable' and 'role' tags to set
    // specific aspects of the resulting property.
#if 0
    // Create special metadata tags if they don't exist on the incoming metadata:
    if (!src_sdrprop->IsOutput())
    {
      // Input:
      // Check for the existence of the 'connectable' token and if its value is not 'connectable':
      //if (src_metadata.count(PXR_NS::SdrPropertyMetadata->Connectable))
      //{
      //  if (PXR_NS::IsTruthy(PXR_NS::SdrPropertyMetadata->Connectable, src_metadata))
      //    out_metadata.insert({PXR_NS::SdrPropertyMetadata->Connectable, ""});
      //}
      out_metadata.insert({PXR_NS::SdrPropertyMetadata->Connectable, "1"});
    }
    else
    {
      // Output:
    }
#endif

    // Support LocalProp category by saving it in prop's metadata tag.
    // We check for its existence when building a ShaderPropertyHandle from an
    // existing SdrShaderProperty:
    if (prop.impl()->isLocal())
      insertUsgLocalTaginMetadata(out_metadata);


    // Unfortunately we can't simply copy the underlying PXR_NS::SdrShaderProperty inside
    // the SchemaProperty since there's no copy ctor...:
    PXR_NS::SdrShaderProperty* sdrprop =
      new PXR_NS::SdrShaderProperty(src_sdrprop->GetName(),
                                 src_sdrprop->GetType(),
                                 src_sdrprop->GetDefaultValue(),
                                 src_sdrprop->IsOutput(),
                                 src_sdrprop->GetArraySize(),
                                 out_metadata,
                                 src_sdrprop->GetHints(),
                                 src_sdrprop->GetOptions());
    assert(sdrprop);
    // {
    //   const PXR_NS::VtValue& defaultValue = sdrprop->GetDefaultValue();
    //   const PXR_NS::NdrSdfTypeIndicator ndrSdfTypeIndicator = sdrprop->GetTypeAsSdfType();
    //   const PXR_NS::SdfValueTypeName sdfType = ndrSdfTypeIndicator.first;
    //   std::cout << "********************************* created sdrproperty '" << sdrprop->GetName() << "'";
    //   std::cout << ", type-string='" << sdrprop->GetType() << "'";
    //   std::cout << ", array-size=" << sdrprop->GetArraySize();
    //   std::cout << ", def-vtvalue-type-name=" << defaultValue.GetTypeName();
    //   std::cout << ", def-vtvalue-type='" << defaultValue.GetType() << "'";
    //   std::cout << ", DERIVED-SDF-TYPE='" << sdfType.GetType() << "'";
    //   std::cout << std::endl;
    //   std::cout << "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << std::endl;
    // }

    sdrproperties.emplace_back(PXR_NS::SdrShaderPropertyUniquePtr(sdrprop));
  }
}


/*static*/ PXR_NS::TfToken
ShaderPropertyHandle::conformPropNameToCategory(const char* prop_name,
                                                int         usg_prop_category)
{
  // Removes any leading 'inputs:' or 'outputs:' prefix from the name. This can
  // commonly happen when the property names are coming from a Prim schema and
  // not a shader.
  // 'usg_prop_category' can be provided to reinforce this, but at the moment is
  // unused.
  if (!prop_name || !prop_name[0]) {
    return PXR_NS::TfToken();
  }

  if (strncmp(prop_name, "inputs:", 7) == 0) {
    return PXR_NS::TfToken(prop_name + 7);
  }
  else if (strncmp(prop_name, "outputs:", 8) == 0) {
    return PXR_NS::TfToken(prop_name + 8);
  }

  return PXR_NS::TfToken(prop_name);
}


/*static*/ ShaderPropertyHandle::SdrTypeDescription
ShaderPropertyHandle::convertToNdrValueAndType(const char*             debug_name,
                                               int                     usg_prop_category,
                                               Value::Type             usg_type,
                                               const Value&            usg_value,
                                               size_t                  array_size,
                                               const fdk::KeyValueMap& usg_metadata,
                                               const fdk::KeyValueMap& usg_hints,
                                               const fdk::StringList&  usg_options)
{
  const usg::ValueTypeInfo& usg_type_info = getValueTypeInfo(usg_type);

  SdrTypeDescription out;
  if (gDebugShaderProperties)
  {
    std::cout << "  '" << ((debug_name != nullptr) ? debug_name : "<>") << "' Usg-IN: ";
    std::cout << " usg_type-string='" << Value::getTypeName(usg_type) << "', usg_type=" << usg_type;
    std::cout << ", usg_type-is_array=" << usg_type_info.is_array;
    std::cout << ", array-size=" << array_size << ", usg_value-array-size=" << usg_value.getArraySize();
    std::cout << ", usg_value'" << usg_value.getTypeName() << "'=(" << usg_value << ")" << std::endl;
  }

  // Default the Sdr prop type to the declared one passed in:
  SdrTypeParams sdr_typeparams(usg_type, 0/*elements*/, usg_type_info.is_array/*is_array*/);

  // Do we need to convert the declared prop type to a reasonable one for Sdr use?
  {
    const ValueTypeToSdrTypeMap::const_iterator it = conform_value_type_map.find(usg_type);
    if (it != conform_value_type_map.end())
    {
      sdr_typeparams = it->second;
    }
  }

  //----------------------------------------------------------------------
  //                 ***** IMPORTANT NOTE ******
  //    sdr_typeparams.elements cannot be copied directly to the
  //    SdrShaderProperty's arraySize as any non-zero value will trigger
  //    a conversion of the default type to a VtArray causing lots of
  //    errors...!
  //----------------------------------------------------------------------
  out.usg_type        = usg_type; // remember the original type
  out.sdr_typename    = getUsgTypeNameAsToken(sdr_typeparams.sdr_type);
  out.default_vtvalue = conformUsgValueToVtValue(usg_value, sdr_typeparams);
  if (usg_type_info.is_array)
  {
    // conformUsgValueToVtValue() should have handled empty default value arrays!
    out.array_size = out.default_vtvalue.GetArraySize();
  }
  else
  {
    // conformUsgValueToVtValue() should have expanded multi-element float types to float[] arrays:
    out.array_size = (sdr_typeparams.elements > 1) ? sdr_typeparams.elements : 0; // << IMPORTANT!!!!
  }
  //----------------------------------------------------------------------

  // Copy metadata tags, and filter / add tags to support the data type:
  for (auto it=usg_metadata.cbegin(); it != usg_metadata.cend(); ++it)
  {
    // TODO: any filtering to do here? We are stripping the PXR_NS::SdrPropertyMetadata->Widget
    //  tag during Schema building, not sure if we want to do so here.
    //if (it->first == "foo")
    //
    out.metadata[PXR_NS::TfToken(it->first)] = it->second;
  }

  // If Asset or AssetArray type we need to add a tag to indicate this:
  if (usg_type == Value::Type::AssetPath || usg_type == Value::Type::AssetPathArray)
  {
    // Add the '__SDR__isAssetIdentifier' tag:
    out.metadata[PXR_NS::SdrPropertyMetadata->IsAssetIdentifier] = "1";
  }

  out.options.reserve(usg_options.size());
  for (const auto& option : usg_options) {
    const PXR_NS::TfToken t(option);
    out.options.emplace_back(t, t);
  }

  //----------------------------------------------------------------------

  usgToSdrMetadata(usg_hints,    out.hints);
  // TODO: convert this list to Tokens:  //out.options

  //----------------------------------------------------------------------

  // Support LocalProp category by saving it in prop's metadata tag.
  // We check for its existence when building a ShaderPropertyHandle from an
  // existing SdrShaderProperty:
  if (usg_prop_category == usg::LocalProp)
    insertUsgLocalTaginMetadata(out.metadata);

  if (gDebugShaderProperties)
  {
    std::cout << "    \\----------> Sdr-OUT: ";
    std::cout << " sdr_type-string='"    << out.sdr_typename << "'";
    std::cout << ", array-size="         << out.array_size;
    std::cout << ", vtvalue-type-name='" << out.default_vtvalue.GetTypeName() << "'";
    std::cout << ", vtvalue-type='"      << out.default_vtvalue.GetType() << "'";
    std::cout << ", vtvalue=["           << out.default_vtvalue << "]";
    if (out.metadata.size() > 0)
    {
      std::cout << ", metadata[";
      for (auto it=out.metadata.begin(); it != out.metadata.end(); ++it)
        std::cout << " '" << it->first << "'='" << it->second << "'";
      std::cout << " ]";
    }
    std::cout << std::endl;
  }

  return out;
}


//------------------------------------------------------------------------------------
//------------------------------------------------------------------------------------


//
// All SchemaProperty constructors must allocate a valid pImpl!
//

SchemaProperty::SchemaProperty() :
  _pImpl(new ShaderPropertyHandle())
{
  assert(_pImpl);
}

/*!
*/
SchemaProperty::SchemaProperty(int          category,
                               Value::Type  type,
                               const char*  name,
                               const Value& value,
                               size_t       arraySize) :
  _pImpl(new ShaderPropertyHandle(ShaderPropertyHandle::conformPropNameToCategory(name, category),
                                  category,
                                  ShaderPropertyHandle::convertToNdrValueAndType(name/*debug-use*/,
                                                                                 category,
                                                                                 type,
                                                                                 value,
                                                                                 arraySize,
                                                                                 fdk::KeyValueMap()/*metadata*/,
                                                                                 fdk::KeyValueMap()/*options*/,
                                                                                 fdk::StringList()/*hints*/)))
{
  assert(_pImpl);
}

/*!
*/
SchemaProperty::SchemaProperty(int                     category,
                               Value::Type             type,
                               const char*             name,
                               const Value&            value,
                               const fdk::KeyValueMap& metadata) :
  _pImpl(new ShaderPropertyHandle(ShaderPropertyHandle::conformPropNameToCategory(name, category),
                                  category,
                                  ShaderPropertyHandle::convertToNdrValueAndType(name/*debug-use*/,
                                                                                 category,
                                                                                 type,
                                                                                 value,
                                                                                 0/*arraySize*/,
                                                                                 metadata,
                                                                                 fdk::KeyValueMap()/*options*/,
                                                                                 fdk::StringList()/*hints*/)))
{
  assert(_pImpl);
}

/*!
*/
SchemaProperty::SchemaProperty(int                     category,
                               Value::Type             type,
                               const char*             name,
                               const fdk::KeyValueMap& metadata) :
  _pImpl(new ShaderPropertyHandle(ShaderPropertyHandle::conformPropNameToCategory(name, category),
                                  category,
                                  ShaderPropertyHandle::convertToNdrValueAndType(name/*debug-use*/,
                                                                                 category,
                                                                                 type,
                                                                                 Value::getDefaultValueForType(type),
                                                                                 0/*arraySize*/,
                                                                                 metadata,
                                                                                 fdk::KeyValueMap()/*options*/,
                                                                                 fdk::StringList()/*hints*/)))
{
  assert(_pImpl);
}

/*!
*/
SchemaProperty::SchemaProperty(int                     category,
                               Value::Type             type,
                               const char*             name,
                               const Value&            value,
                               size_t                  arraySize,
                               const fdk::KeyValueMap& metadata,
                               const fdk::KeyValueMap& hints,
                               const fdk::StringList&  options) :
  _pImpl(new ShaderPropertyHandle(ShaderPropertyHandle::conformPropNameToCategory(name, category),
                                  category,
                                  ShaderPropertyHandle::convertToNdrValueAndType(name/*debug-use*/,
                                                                                 category,
                                                                                 type,
                                                                                 value,
                                                                                 arraySize,
                                                                                 metadata,
                                                                                 hints,
                                                                                 options)))
{
  assert(_pImpl);
}


SchemaProperty::SchemaProperty(const SchemaProperty& b) :
  _pImpl(new ShaderPropertyHandle(*b._pImpl))
{
  assert(_pImpl);
}
SchemaProperty& SchemaProperty::operator = (const SchemaProperty& b)
{
  if (this != &b)
    _pImpl.reset(new ShaderPropertyHandle(*b._pImpl));
  assert(_pImpl);
  return *this;
}

SchemaProperty::SchemaProperty(const ShaderPropertyHandle& handle) :
  _pImpl(new ShaderPropertyHandle(handle))
{
  assert(_pImpl);
}


bool SchemaProperty::isValid() const
{
  return (_pImpl->getProperty() != nullptr);
}

/*virtual*/
SchemaProperty::~SchemaProperty()
{
  //
}


const std::string& SchemaProperty::getName()         const { return _pImpl->getProperty()->GetName(); }
const std::string& SchemaProperty::getType()         const { return _pImpl->getProperty()->GetType(); }
Value::Type        SchemaProperty::getValueType()    const { return _pImpl->valuetype; }
bool               SchemaProperty::isOutput()        const { return _pImpl->getProperty()->IsOutput(); }
bool               SchemaProperty::isArray()         const { return _pImpl->getProperty()->IsArray(); }
bool               SchemaProperty::isDynamicArray()  const { return _pImpl->getProperty()->IsDynamicArray(); }
int                SchemaProperty::getArraySize()    const { return _pImpl->getProperty()->GetArraySize(); }

bool               SchemaProperty::isReference()     const { return (_pImpl->sdrrefproperty != nullptr); }
bool               SchemaProperty::isOverriden()     const { return (_pImpl->sdrovrproperty != nullptr); }

std::string
SchemaProperty::getFullName() const
{
  const PXR_NS::SdrShaderProperty* sdrprop = _pImpl->getProperty();
  std::string full_name = (_pImpl->isLocal()) ? "" : ((sdrprop->IsOutput()) ? "outputs:" : "inputs:");
  full_name += sdrprop->GetName().GetString();

  return full_name;
}


const Value
SchemaProperty::getValue() const
{
  return Value(Value::Impl(_pImpl->getProperty()->GetDefaultValue()));
}


void            SchemaProperty::setValueTime(const fdk::TimeValue& time) { _pImpl->override_time = time; }
fdk::TimeValue  SchemaProperty::getValueTime() const { return _pImpl->override_time; }


//!
fdk::KeyValueMap
SchemaProperty::getMetadata() const
{
  const PXR_NS::NdrTokenMap& ndrMetadata = _pImpl->getProperty()->GetMetadata();
  fdk::KeyValueMap metadata;
  metadata.reserve(ndrMetadata.size());
  for (const auto& m : ndrMetadata) {
    metadata.emplace(m.first.GetText(), m.second);
  }
  return metadata;
}


bool SchemaProperty::isConnectable() const { return _pImpl->getProperty()->IsConnectable(); }
bool SchemaProperty::canConnectTo(const SchemaProperty& output) const
{
  assert(output._pImpl->getProperty());
  return _pImpl->getProperty()->CanConnectTo(*output._pImpl->getProperty());
}



/*! Print the name, type and contents of knob to stream.
*/
void
SchemaProperty::print(std::ostream& o) const
{
  if (!isValid()) {
    o << "<invalid>";
  }
  else {
    o << getType();
    if (isArray()) {
      o << "[]";
    }
    o << " " << getFullName() << " = (" << getValue() << ")";
    o << "[";
    const PXR_NS::NdrTokenMap& metadata = _pImpl->getProperty()->GetMetadata();
    for (auto it=metadata.begin(); it != metadata.end(); ++it) {
      o << " '" << it->first << "'='" << it->second << "'";
    }
    o << " ]";
  }
}


/*static*/ void
ShaderPropertyHandle::printNdrProperty(const char* prefix, std::ostream& o, const PXR_NS::NdrProperty& prop)
{
  if (prefix) {
    o << prefix;
  }
  o << prop.GetType();
  if (prop.IsArray()) {
    o << "[" << prop.GetArraySize() << "]";
  }
  o << " " << (prop.IsOutput() ? "outputs:" : "inputs:") << prop.GetName();
  o << " = (" << prop.GetDefaultValue() << ")";
  o << " - [";
  const PXR_NS::NdrTokenMap& metadata = prop.GetMetadata();
  for (auto it=metadata.begin(); it != metadata.end(); ++it) {
    o << " '" << it->first << "'='" << it->second << "'";
  }
  o << " ]";
#if 0
  prop.GetHints(),
  prop.GetOptions());
#endif
}


} // namespace usg
