// Copyright (c) 2022 The Foundry Visionmongers Ltd. All Rights Reserved.

#include "usg/geom/ShaderDesc.h"
#include "usg/geom/GeomTokens.h"

#include "usg/private/UsdImpl.h"

#include "fdk/base/PathUtils.h"

#include "pxr/usd/sdr/shaderNode.h"

#include "pxr/usd/usdShade/material.h"

#include "pxr/base/tf/stringUtils.h" // for TfStringGetBeforeSuffix

//------------------------------------------------------------------------------------


const std::string gEnvDebugShaderBuilding    = ((std::getenv("USG_DEBUG_SHADER_BUILDING") != nullptr) ?
                                                 std::getenv("USG_DEBUG_SHADER_BUILDING") : "");
const std::string gEnvDebugShaderConnections = ((std::getenv("USG_DEBUG_SHADER_CONNECTIONS") != nullptr) ?
                                                 std::getenv("USG_DEBUG_SHADER_CONNECTIONS") : "");
const bool gDebugShaderBuilding    = (gEnvDebugShaderBuilding    == "1");
const bool gDebugShaderConnections = (gEnvDebugShaderConnections == "1");


//-----------------------------------------------------------------------------------


namespace usg
{


ShaderDescGroup::ShaderDescGroup(const std::string& _target_renderer) :
  target_renderer(_target_renderer),
  output(nullptr),
  displacement_output(nullptr)
{
  //
}

ShaderDescGroup::~ShaderDescGroup()
{
  for (ShaderDescPtrMap::const_iterator it=shaders.begin(); it !=shaders.end(); ++it) {
    if (it->second) {
      delete it->second;
    }
  }
}

void
ShaderDescGroup::deleteShaders()
{
  for (ShaderDescPtrMap::const_iterator it=shaders.begin(); it !=shaders.end(); ++it) {
    if (it->second) {
      delete it->second;
    }
  }
  shaders.clear();
  output = nullptr;
  topology_hash.reset();
  property_hash.reset();
}

ShaderDesc*
ShaderDescGroup::getShaderNode(const std::string& shader_node_name) const
{
  const ShaderDescPtrMap::const_iterator it = shaders.find(shader_node_name);
  return (it != shaders.end()) ? it->second : nullptr;
}

namespace
{
  void topologyHashVisitor(fdk::Hash& hash, const ShaderDescGroup* group, const ShaderDesc* desc)
  {
    hash.append(desc->schemaName());
    for (const auto& inputs : desc->getInputConnections()) {
      for (const auto& i : *inputs.second) {
        topologyHashVisitor(hash, group, group->getShaderNode(i.upstreamNodePath.asString()));
      }
    }
  }
}

fdk::Hash ShaderDescGroup::getTopologyHash()
{
  if (topology_hash.isDefault() && output) {
    topologyHashVisitor(topology_hash, this, output);
  }
  return topology_hash;
}

bool
ShaderDescGroup::addShaderDesc(ShaderDesc* shader)
{
  if (!shader) {
    return false;
  }
  return addShaderDesc(shader, shader->getMaterialRelativeName());
}

bool
ShaderDescGroup::addShaderDesc(ShaderDesc*        shader,
                               const std::string& prim_name)
{
  if (!shader || prim_name.empty() || shaders.find(prim_name) != shaders.end()) {
    return false;
  }

  shader->setMaterialRelativeName(prim_name);
  shaders[shader->getMaterialRelativeName()] = shader;
  return true;
}


//------------------------------------------------------------------------------------
//------------------------------------------------------------------------------------


ShaderDesc::ShaderDesc() :
  _pImpl(nullptr)
{
  //
}

//!
ShaderDesc::ShaderDesc(ShaderDesc&& b) :
  _pImpl(std::move(b._pImpl))
{
  //
}

/*! Construct from a ShaderSchema registry description.

    Direct copying ShaderDesc is disabled, use the ShaderDesc::createFromSchema()
    method instead to create additional copies.
*/
ShaderDesc::ShaderDesc(const SchemaDescription* schema) :
  _pImpl((schema && schema->registryNode()) ?
            new ShaderDescHandle(schema->registryNode()->ndrnode) :
            nullptr)
{
  //
}


ShaderDesc::~ShaderDesc()
{
  //
}

/*!
*/
bool
ShaderDesc::isValid() const
{
  return (_pImpl && _pImpl->isValid());
}

/*static*/ ShaderDesc*
ShaderDesc::createFromSchema(const char* schema_name)
{
  return createFromSchema(schema_name, std::string());
}

/*static*/ ShaderDesc*
ShaderDesc::createFromSchema(const char*        schema_name,
                             const std::string& prim_name)
{
  const SchemaDescription* schema = SchemaDescription::find(schema_name);
  if (!schema || !schema->registryNode()) {
    std::cerr << "ShaderDesc::createFromSchema(): error, cannot find shader schema for type name '" << schema_name << "'" << std::endl;
    return nullptr; // can't find shader class by name...
  }

  ShaderDesc* desc = new ShaderDesc(schema);
  if (desc) {
    desc->setMaterialRelativeName(prim_name);
  }
  return desc;
}


//!
/*static*/ ShaderDesc
ShaderDesc::createFromIdentifier(const std::string& identifier)
{
  const SchemaDescription* schema = SchemaDescription::getRegistryIdentifier(identifier);
  if (!schema || !schema->registryNode()) {
    std::cerr << "ShaderDesc::createFromIdentifier(): error, cannot find shader schema for identifier '" << identifier << "'" << std::endl;
    return ShaderDesc(); // can't find shader class by name...
  }

  return ShaderDesc(schema);
}


const std::string&
ShaderDesc::schemaName() const
{
  return (isValid()) ? _pImpl->class_name : empty_string;
}


const std::string&
ShaderDesc::getIdentifier() const
{ 
  return (isValid()) ? _pImpl->ndrnode->GetIdentifier().GetString() : empty_string;
}


/*static*/ std::string
ShaderDesc::buildName(const std::string& target_renderer,
                      const std::string& prim_name)
{
  return (!target_renderer.empty()) ? target_renderer + "_" + prim_name : prim_name;
}

void
ShaderDesc::setMaterialRelativeName(const std::string& prim_name)
{
  if (!isValid()) {
    return;
  }
  _pImpl->rel_prim_path = prim_name;
}


const std::string&
ShaderDesc::getMaterialRelativeName() const
{
  return (isValid()) ? _pImpl->rel_prim_path : empty_string;
}


//--------------------------------------------

const std::string&
ShaderDesc::getSourceType() const
{
  return (isValid()) ? _pImpl->ndrnode->GetSourceType().GetString() : empty_string;
}

const std::string&
ShaderDesc::getSourceCode() const
{
  return (isValid()) ? _pImpl->ndrnode->GetSourceCode() : empty_string;
}

const std::string&
ShaderDesc::getResolvedDefinitionURI() const
{
  return (isValid()) ? _pImpl->ndrnode->GetResolvedDefinitionURI() : empty_string;
}

const std::string&
ShaderDesc::getResolvedImplementationURI() const
{
  return (isValid()) ? _pImpl->ndrnode->GetResolvedImplementationURI() : empty_string;
}


//----------------------------------------------------------

static ShaderPropertyMap empty_override_map;


size_t
ShaderDesc::getLocalPropertyNames(fdk::StringPtrList& names) const
{
  names.clear();
  if (!isValid()) {
    return 0;
  }

  // Filter out just the local props:
  const PXR_NS::NdrTokenVec& prop_names = _pImpl->ndrnode->GetInputNames();
  names.reserve(prop_names.size());
  for (const PXR_NS::TfToken& name : prop_names) {
    if (ShaderPropertyHandle::isPropertyLocal(_pImpl->ndrnode->GetInput(name)))
      names.push_back(&name.GetString());
  }

  return names.size();
}

bool
ShaderDesc::hasLocalProperty(const std::string& prop_name) const
{
  return hasInput(prop_name);
}

ShaderProperty
ShaderDesc::getLocalProperty(const std::string& prop_name) const
{
  return std::move(getInput(prop_name));
}

ShaderProperty*
ShaderDesc::overrideLocalProperty(const std::string&    prop_name,
                                  const Value&          override_value,
                                  const fdk::TimeValue& override_time)
{
  return overrideInput(prop_name, override_value, override_time);
}


//--------------------------------------------


size_t
ShaderDesc::getInputNames(fdk::StringPtrList& names) const
{
  names.clear();
  if (!isValid()) {
    return 0;
  }

  const PXR_NS::NdrTokenVec& prop_names = _pImpl->ndrnode->GetInputNames();
  names.reserve(prop_names.size());
  for (size_t i=0; i < prop_names.size(); ++i) {
    const PXR_NS::TfToken& name = prop_names[i];
    if (!name.IsEmpty()) {
      names.push_back(&name.GetString());
    }
  }

  return names.size();
}

bool
ShaderDesc::hasInput(const std::string& input_name) const
{
  if (!isValid() || input_name.empty()) {
    return false;
  }
  // TODO: Use immortal tag on TfToken ctor...?
  return (_pImpl->ndrnode->GetInput(PXR_NS::TfToken(input_name)));
}

ShaderProperty
ShaderDesc::getInput(const std::string& input_name) const
{
  if (!isValid() || input_name.empty()) {
    return ShaderProperty();
  }
  // TODO: Use immortal tag on TfToken ctor...?
  PXR_NS::NdrPropertyConstPtr sdrprop = _pImpl->ndrnode->GetInput(PXR_NS::TfToken(input_name));
  return (sdrprop) ?
            ShaderProperty(ShaderPropertyHandle(dynamic_cast<const PXR_NS::SdrShaderPropertyConstPtr>(sdrprop))) :
            ShaderProperty();
}


usg::Value
ShaderDesc::getInputDefaultValue(const std::string& input_name) const
{
  if (!isValid() || input_name.empty()) {
    return Value();
  }
  // TODO: Use immortal tag on TfToken ctor...?
  PXR_NS::NdrPropertyConstPtr sdrprop = _pImpl->ndrnode->GetInput(PXR_NS::TfToken(input_name));
  return (sdrprop) ? Value(Value::Impl(sdrprop->GetDefaultValue())) : Value();
}


usg::Value
ShaderDesc::getInputValue(const std::string& input_name) const
{
  if (!isValid() || input_name.empty()) {
    return Value();
  }

  // Check for an override first:
  ShaderPropertyMap::iterator it = _pImpl->input_overrides.find(input_name);
  if (it != _pImpl->input_overrides.end()) {
    const PXR_NS::SdrShaderProperty* over = it->second->impl()->getProperty();
    if (over) {
      return Value(Value::Impl(over->GetDefaultValue()));
    }
  }

  // No override return the default value:
  PXR_NS::NdrPropertyConstPtr sdrprop = _pImpl->ndrnode->GetInput(PXR_NS::TfToken(input_name));
  return (sdrprop) ? Value(Value::Impl(sdrprop->GetDefaultValue())) : Value();
}


int
findPropMatch(const std::string&                  name,
              const PXR_NS::NdrPropertyUniquePtrVec& props)
{
  const size_t nProps = props.size();
  for (size_t i=0; i < nProps; ++i) {
    const PXR_NS::NdrPropertyUniquePtr& prop = props[i];
    if (prop.get() && prop.get()->GetName() == name) {
      return i;
    }
  }
  return -1;
}


//----------------------------------------------------------


ShaderProperty*
ShaderDesc::overrideInput(const std::string&    input_name,
                          const Value&          override_value,
                          const fdk::TimeValue& override_time)
{
  if (!isValid() || input_name.empty()) {
    return nullptr;
  }

  const PXR_NS::SdrShaderPropertyConstPtr sdrprop =
    dynamic_cast<const PXR_NS::SdrShaderPropertyConstPtr>(_pImpl->ndrnode->GetInput(PXR_NS::TfToken(input_name)));
  if (!sdrprop) {
    if (getMaterialRelativeName().empty()) {
      std::cerr << getIdentifier();
    }
    else {
      std::cerr << getMaterialRelativeName();
    }
    std::cerr << "::overrideInput('" << input_name << "')";
    std::cerr << " error - unable to retrieve property" << std::endl;
    return nullptr;
  }
  if (gDebugShaderConnections) {
    std::cout << "  override input prop '" << input_name << "', SdrType=" << sdrprop->GetType();
    std::cout << ", override_value[" << override_value << "]";
    std::cout << ", override_time=" << override_time;
    std::cout << std::endl;
  }

  // Delete any existing override:
  ShaderPropertyMap::iterator it = _pImpl->input_overrides.find(input_name);
  if (it != _pImpl->input_overrides.end()) {
    delete it->second;
    it->second = nullptr;
  }

  // Let the value go through the conformer first before checking compatibility:
  const Value::Type prop_type = getUsgValueType(sdrprop->GetType());
  const PXR_NS::VtValue conformed_override_value = ShaderPropertyHandle::conformUsgValueToVtValue(override_value, prop_type);

  // Check the base types match:
  const Value::Type prop_base_type = Value::getBaseType(prop_type);
  const Value::Type over_base_type = Value::getBaseType(getUsgValueType(conformed_override_value.GetType()));

  PXR_NS::TfToken over_type = sdrprop->GetType();

  // If mismatch can we translate?
  bool translate_mismatch = (over_base_type != prop_base_type);
  if (translate_mismatch) {
    // TODO: handle string, token and asset here manually - this should be done!

    if (prop_base_type == Value::Type::Bool   ||
        prop_base_type == Value::Type::Int    ||
        prop_base_type == Value::Type::Int64  ||
        prop_base_type == Value::Type::UInt   ||
        prop_base_type == Value::Type::UInt64) {
      // Special case matching integer types to uchar & bool:
      if (over_base_type == Value::Type::Bool) {
        over_type = PXR_NS::SdfValueTypeNames->Bool.GetAsToken();
        translate_mismatch = false;
      }
      else if (over_base_type == Value::Type::UChar) {
        over_type = PXR_NS::SdfValueTypeNames->UChar.GetAsToken();
        translate_mismatch = false;
      }
    }
    else {
      // Special case vector types
      switch (over_base_type) {
        default:
          break;
        case Value::Type::Float2:
          over_type = PXR_NS::SdfValueTypeNames->Float2.GetAsToken();
          translate_mismatch = false;
          break;
        case Value::Type::Float3:
          over_type = PXR_NS::SdfValueTypeNames->Float3.GetAsToken();
          translate_mismatch = false;
          break;
        case Value::Type::Float4:
          over_type = PXR_NS::SdfValueTypeNames->Float4.GetAsToken();
          translate_mismatch = false;
          break;
        case Value::Type::Int4:
          over_type = PXR_NS::SdfValueTypeNames->Int4.GetAsToken();
          translate_mismatch = false;
          break;
      }
    }
  }

  if (translate_mismatch) {
    if (getMaterialRelativeName().empty()) {
      std::cerr << getIdentifier();
    }
    else {
      std::cerr << getMaterialRelativeName();
    }
    std::cerr << "::overrideInput('" << input_name << "')";
    std::cerr << " error - type mismatch: input prop type='" << Value::getTypeName(prop_base_type) << "'";
    std::cerr << " and override prop type='" << Value::getTypeName(over_base_type) << "'";
    std::cerr << ", ignoring override" << std::endl;
    return nullptr;
  }

  // Create override property with an override value and an override time:
  ShaderProperty* over =
    new ShaderProperty(ShaderPropertyHandle(sdrprop->GetName(),
                                            over_type,
                                            conformed_override_value,
                                            sdrprop->IsOutput(),
                                            sdrprop->IsArray(),
                                            sdrprop->GetMetadata(),
                                            sdrprop->GetHints(),
                                            sdrprop->GetOptions()));
  over->setValueTime(override_time);

  _pImpl->input_overrides[input_name] = over;
  if (gDebugShaderConnections) {
    std::cout << "    ...with prop " << *over << std::endl;
  }
  return over;
}


bool
ShaderDesc::overrideInputs(const ShaderPropertyMap& overrides)
{
  if (!isValid()) {
    return false;
  }

  bool ok = true;
  for (ShaderPropertyMap::const_iterator it=overrides.begin(); it != overrides.end(); ++it) {
    if (!it->second) {
      continue;
    }
    if (!overrideInput(it->first, it->second->getValue(), it->second->getValueTime())) {
      ok = false;
    }
  }
  return ok;
}


bool
ShaderDesc::clearInputOverride(const std::string& input_name)
{
  if (!isValid() || input_name.empty()) {
    return false;
  }

  ShaderPropertyMap::iterator it = _pImpl->input_overrides.find(input_name);
  if (it == _pImpl->input_overrides.end()) {
    return false;
  }

  delete it->second;
  _pImpl->input_overrides.erase(it);

  return true;
}


ShaderProperty*
ShaderDesc::getInputOverride(const std::string& input_name) const
{
  if (!isValid() || input_name.empty()) {
    return nullptr;
  }

  ShaderPropertyMap::iterator it = _pImpl->input_overrides.find(input_name);
  return (it != _pImpl->input_overrides.end()) ? it->second : nullptr;
}


const ShaderPropertyMap&
ShaderDesc::getInputOverrides() const
{
  return (isValid()) ? _pImpl->input_overrides : empty_override_map;
}


//--------------------------------------------------------------

static ShaderConnectionMap empty_connection_map;


ShaderConnectionArray* 
ShaderDesc::connectInput(const std::string& input_name,
                         const ShaderDesc*  source_shader,
                         const std::string& source_shader_output_name)
{
  if (gDebugShaderConnections) {
    std::cout << "ShaderDesc('" << getIdentifier() << "')::connectInput() input_name='" << input_name << "'";
    std::cout << ", source_shader=" << source_shader;
    std::cout << ", output_name='" << source_shader_output_name << "'";
  }

  if (!isValid() || input_name.empty()) {
    if (gDebugShaderConnections) {
      std::cout << ": cannot connect, this ShaderDesc is invalid" << std::endl;
    }
    return nullptr;
  }

  ShaderConnectionMap::iterator it = _pImpl->input_connections.find(input_name);
  if (it == _pImpl->input_connections.end()) {
    if (!source_shader || !source_shader->isValid() || source_shader_output_name.empty()) {
      if (gDebugShaderConnections) {
        std::cout << ": cannot connect to null or invalid source shader" << std::endl;
      }
      return nullptr; // don't bother creating a connection
    }

    // Create the connections list:
    _pImpl->input_connections[input_name] = new ShaderConnectionArray();
    it = _pImpl->input_connections.find(input_name);
  }
  else {
    // Input connection already exists:
    if (!source_shader || !source_shader->isValid() || source_shader_output_name.empty()) {
      if (gDebugShaderConnections) {
        std::cout << ": cannot connect to null or invalid source shader" << std::endl;
      }
      // If no source disconnect the input by deleting its connection list:
      delete it->second;
      _pImpl->input_connections.erase(it);
      return nullptr;
    }
  }
  assert(it != _pImpl->input_connections.end());
  assert(it->second != nullptr);
  assert(source_shader);

  const ShaderProperty source_output = source_shader->getOutput(source_shader_output_name);
  if (!source_output) {
    if (gDebugShaderConnections) {
      std::cout << ": cannot connect, source shader '" << source_shader->schemaName() << "' does not have output ";
      std::cout << "port '" << source_shader_output_name << "'" << std::endl;
    }
    // If no source output prop disconnect the input by deleting its connection list:
    delete it->second;
    _pImpl->input_connections.erase(it);
    return nullptr;
  }

  // Add the connection:
  const std::string output_name = "outputs:" + source_shader_output_name;
  it->second->resize(1);
  (*it->second)[0] = ShaderConnection(source_shader->getMaterialRelativeName(), output_name);

  if (gDebugShaderConnections) {
    std::cout << ": connected to source shader '" << source_shader->getMaterialRelativeName() << "' output " << source_output << std::endl;
  }
  return it->second;
}
ShaderConnectionArray* 
ShaderDesc::connectInput(const ShaderProperty& input_property,
                         const ShaderDesc*     source_shader,
                         const std::string&    source_shader_output_name)
{
  return connectInput(input_property.getName(), source_shader, source_shader_output_name);
}


ShaderConnectionArray* 
ShaderDesc::connectInput(const std::string& input_name,
                         const std::string& connection_path)
{
  if (gDebugShaderConnections) {
    std::cout << "ShaderDesc('" << getIdentifier() << "')::connectInput() input_name='" << input_name << "'";
    std::cout << ", connection_path='" << connection_path << "'";
  }

  if (!isValid() || input_name.empty()) {
    if (gDebugShaderConnections) {
      std::cout << ": cannot connect, this ShaderDesc is invalid" << std::endl;
    }
    return nullptr;
  }

  ShaderConnectionMap::iterator it = _pImpl->input_connections.find(input_name);
  if (it == _pImpl->input_connections.end()) {
    if (connection_path.empty()) {
      if (gDebugShaderConnections) {
        std::cout << ": cannot connect to empty connection path" << std::endl;
      }
      return nullptr; // don't bother creating a connection
    }

    // Create the connections list:
    _pImpl->input_connections[input_name] = new ShaderConnectionArray();
    it = _pImpl->input_connections.find(input_name);
  }
  else {
    // Input connection already exists:
    if (connection_path.empty()) {
      if (gDebugShaderConnections) {
        std::cout << ": cannot connect to empty connection path" << std::endl;
      }
      // If no source disconnect the input by deleting its connection list:
      delete it->second;
      _pImpl->input_connections.erase(it);
      return nullptr;
    }
  }
  assert(it != _pImpl->input_connections.end());
  assert(it->second != nullptr);

  // Add the connection:
  std::string source_shader_path, source_shader_output_name;
  fdk::splitPropertyPath(connection_path, source_shader_path, source_shader_output_name);

  it->second->resize(1);
  (*it->second)[0] = ShaderConnection(source_shader_path, source_shader_output_name);

  if (gDebugShaderConnections) {
    std::cout << ": connected to source shader '" << source_shader_path << "' output " << source_shader_output_name << std::endl;
  }
  return it->second;
}
ShaderConnectionArray* 
ShaderDesc::connectInput(const ShaderProperty& input_property,
                         const std::string&    connection_path)
{
  return connectInput(input_property.getName(), connection_path);
}


bool
ShaderDesc::connectInputs(const ShaderConnectionMap& connections)
{
  bool ok = true;
  for (ShaderConnectionMap::const_iterator it=connections.begin(); it != connections.end(); ++it) {
    const std::string&           input_name      = it->first;
    const ShaderConnectionArray* connection_list = it->second;
//    if (!connectInput(input_name, connection.upstreamNodePath, connection.upstreamOutputName)) {
//      ok = false;
//    }
  }
  return ok;
}


const ShaderConnectionMap&
ShaderDesc::getInputConnections() const
{
  return (isValid()) ? _pImpl->input_connections : empty_connection_map;
}


const ShaderConnectionArray*
ShaderDesc::getInputConnection(const std::string& input_name) const
{
  if (!isValid()) {
    return nullptr;
  }
  const ShaderConnectionMap::const_iterator it = _pImpl->input_connections.find(input_name);
  return (it != _pImpl->input_connections.end()) ? it->second : nullptr;
}


//--------------------------------------------


size_t
ShaderDesc::getOutputNames(fdk::StringPtrList& names) const
{
  names.clear();
  if (!isValid()) {
    return 0;
  }

  const PXR_NS::NdrTokenVec& prop_names = _pImpl->ndrnode->GetOutputNames();
  names.reserve(prop_names.size());
  for (size_t i=0; i < prop_names.size(); ++i) {
    const PXR_NS::TfToken& name = prop_names[i];
    if (!name.IsEmpty()) {
      names.push_back(&name.GetString());
    }
  }

  return names.size();
}

bool
ShaderDesc::hasOutput(const std::string& output_name) const
{
  if (!isValid() || output_name.empty()) {
    return false;
  }
  // TODO: Use immortal tag on TfToken ctor...?
  return (_pImpl->ndrnode->GetOutput(PXR_NS::TfToken(output_name)));
}

ShaderProperty
ShaderDesc::getOutput(const std::string& output_name) const
{
  if (!isValid() || output_name.empty()) {
    return ShaderProperty();
  }
  // TODO: Use immortal tag on TfToken ctor...?
  PXR_NS::NdrPropertyConstPtr sdrprop = _pImpl->ndrnode->GetOutput(PXR_NS::TfToken(output_name));
  return (sdrprop) ?
            ShaderProperty(ShaderPropertyHandle(dynamic_cast<const PXR_NS::SdrShaderPropertyConstPtr>(sdrprop))) :
            ShaderProperty();
}


//--------------------------------------------


/*! Print input and output knob values to stream.
*/
void
ShaderDesc::print(const char*   prefix,
                  std::ostream& o,
                  bool          print_contents) const
{
  if (!isValid()) {
    o << prefix << "<empty ShaderDesc>" << std::endl;
    return;
  }

  o << prefix << schemaName();
  if (1) {
    o << " (" << this << ")";
  }
  if (!getMaterialRelativeName().empty()) {
    o << ": group-path='" << getMaterialRelativeName() << "'";
  }
  o << std::endl;

  if (print_contents && _pImpl->ndrnode) {
    ShaderSchemaHandle::printNdrNode(prefix, o, *_pImpl->ndrnode);
  }
}


#if 0
// TODO: Not yet implemented methods from Sdr/Ndr node api
// TODO: Move these back to ShaderDesc header and implement the ones we want to keep

    //! Return the version of the node
    NdrVersion GetVersion() const;

    //! Gets the name of the node.
    const std::string& GetName() const;

    /*! Gets the name of the family that the node belongs to. An empty token
        will be returned if the node does not belong to a family.
    */
    const TfToken& GetFamily() const;

    /*! Gets the context of the node.
        
        The context is the context that the node declares itself as having (or,
        if a particular node does not declare a context, it will be assigned a
        default context by the parser).
    
        As a concrete example from the `Sdr` library, a shader with a specific
        source type may perform different duties vs. another shader with the
        same source type. For example, one shader with a source type of
        `SdrArgsParser::SourceType` may declare itself as having a context of
        'pattern', while another shader of the same source type may say it is
        used for lighting, and thus has a context of 'light'.
    */
    const TfToken& GetContext() const;

    /*! Gets the type of source that this node originated from.
        
        Note that this is distinct from `GetContext()`, which is the type that
        the node declares itself as having.
    
        As a concrete example from the `Sdr` library, several shader parsers
        exist and operate on different types of shaders. In this scenario, each
        distinct type of shader (OSL, Args, etc) is considered a different
        _source_, even though they are all shaders. In addition, the shaders
        under each source type may declare themselves as having a specific
        context (shaders can serve different roles). See `GetContext()` for
        more information on this.
    */
    const TfToken& GetSourceType() const;

    /*! Gets the URI to the resource that provided this node's
        definition. Could be a path to a file, or some other resource
        identifier. This URI should be fully resolved.
    
        \sa NdrNode::GetResolvedImplementationURI()
    */
    const std::string& GetResolvedDefinitionURI() const;

    /*! Gets the URI to the resource that provides this node's
        implementation. Could be a path to a file, or some other resource
        identifier. This URI should be fully resolved.
    
        \sa NdrNode::GetResolvedDefinitionURI()
    */
    const std::string& GetResolvedImplementationURI() const;

    /*! Returns  the source code for this node. This will be empty for most 
        nodes. It will be non-empty only for the nodes that are constructed 
        using \ref NdrRegistry::GetNodeFromSourceCode(), in which case, the 
        source code has not been parsed (or even compiled) yet. 
    
        An unparsed node with non-empty source-code but no properties is 
        considered to be invalid. Once the node is parsed and the relevant 
        properties and metadata are extracted from the source code, the node 
        becomes valid.
    
        \sa NdrNode::IsValid
    */
    const std::string &GetSourceCode() const;

    /*! Whether or not this node is valid. A node that is valid indicates that
        the parser plugin was able to successfully parse the contents of this
        node.
    
        Note that if a node is not valid, some data like its name, URI, source 
        code etc. could still be available (data that was obtained during the 
        discovery process). However, other data that must be gathered from the 
        parsing process will NOT be available (eg, inputs and outputs).
    */
    virtual bool IsValid() const;

    /*! Gets a string with basic information about this node. Helpful for
        things like adding this node to a log.
    */
    virtual std::string GetInfoString() const;


  public:
    //------------------------------------------------------
    // Inputs and Outputs
    // An input or output is also generically referred to
    // as a "property".
    //------------------------------------------------------

    //! Get an ordered list of all the input names on this node.
    const NdrTokenVec& GetInputNames() const;

    //! Get an ordered list of all the output names on this node.
    const NdrTokenVec& GetOutputNames() const;

    /*! Get an input property by name. `nullptr` is returned if an input with
        the given name does not exist.
    */
    NdrPropertyConstPtr GetInput(const TfToken& inputName) const;

    /*! Get an output property by name. `nullptr` is returned if an output with
        the given name does not exist.
    */
    NdrPropertyConstPtr GetOutput(const TfToken& outputName) const;


  public:
    //------------------------------------------------------
    // Metadata
    //------------------------------------------------------

    /*! All metadata that came from the parse process. Specialized nodes may
        isolate values in the metadata (with possible manipulations and/or
        additional parsing) and expose those values in their API.
    */
    const NdrTokenMap& GetMetadata() const;


    /*! The label assigned to this node, if any. Distinct from the name
        returned from `getName()`. In the context of a UI, the label value
        might be used as the display name for the node instead of the name.
    */
    const TfToken& getLabel() const;

    /*! The category assigned to this node, if any. Distinct from the family
        returned from `getFamily()`.
    */
    const TfToken& getCategory() const;

    /*! Returns the role of this node. This is used to annotate the role that 
        the shader node plays inside a shader network. We can tag certain
        shaders to indicate their role within a shading network. We currently
        tag primvar reading nodes, texture reading nodes and nodes that access
        volume fields (like extinction or scattering). This is done to identify
        resources used by a shading network.
    */
    std::string getRole() const;

    //! The help message assigned to this node, if any.
    std::string getHelp() const;

    //! The departments this node is associated with, if any.
    const NdrTokenVec& getDepartments() const;

    /*! Gets the pages on which the node's properties reside (an aggregate of
        the unique `SdrShaderProperty::getPage()` values for all of the node's
        properties). Nodes themselves do not reside on pages. In an example
        scenario, properties might be divided into two pages, 'Simple' and
        'Advanced'.
    */
    const NdrTokenVec& getPages() const;

    /*! The list of primvars this node knows it requires / uses.
        For example, a shader node may require the 'normals' primvar to function
        correctly. Additional, user specified primvars may have been authored on
        the node. These can be queried via `GetAdditionalPrimvarProperties()`.
        Together, `GetPrimvars()` and `GetAdditionalPrimvarProperties()`,
        provide the complete list of primvar requirements for the node.
    */
    const NdrTokenVec& getPrimvars() const;

    /*! The list of string input properties whose values provide the names of
        additional primvars consumed by this node. For example, this may return
        a token named `varname`. This indicates that the client should query the
        value of a (presumed to be string-valued) input attribute named varname
        from its scene description to determine the name of a primvar the 
        node will consume. See `getPrimvars()` for additional information.
    */
    const NdrTokenVec& getAdditionalPrimvarProperties() const;

    /*! Returns the implementation name of this node.  The name of the node
        is how to refer to the node in shader networks.  The label is how to
        present this node to users.  The implementation name is the name of
        the function (or something) this node represents in the
        implementation.  Any client using the implementation \b must call
        this method to get the correct name;  using \c getName() is not
        correct.
    */
    std::string getImplementationName() const;
#endif


} // namespace usg
