// Copyright (c) 2022 The Foundry Visionmongers Ltd. All Rights Reserved.

#include "usg/geom/ShaderPrim.h"
#include "usg/geom/ShaderSchema.h"
#include "usg/geom/MaterialPrim.h"
#include "usg/geom/Layer.h"
#include "usg/geom/Stage.h"

#include "usg/private/UsdImpl.h"

#include "fdk/base/PathUtils.h"

#include "pxr/usd/usdShade/shader.h"

//------------------------------------------------------------------------------------


const std::string gEnvDebugShaderBuilding    = ((std::getenv("USG_DEBUG_SHADER_BUILDING") != nullptr) ?
                                                 std::getenv("USG_DEBUG_SHADER_BUILDING") : "");
const std::string gEnvDebugShaderConnections = ((std::getenv("USG_DEBUG_SHADER_CONNECTIONS") != nullptr) ?
                                                 std::getenv("USG_DEBUG_SHADER_CONNECTIONS") : "");
const bool gDebugShaderBuilding    = (gEnvDebugShaderBuilding    == "1");
const bool gDebugShaderConnections = (gEnvDebugShaderConnections == "1");


//-----------------------------------------------------------------------------------


namespace usg
{

extern std::string empty_string;


ShaderPrim::ShaderPrim() :
  Prim(),
  _schema(nullptr)
{
  //
}

ShaderPrim::ShaderPrim(const Prim& b) :
  Prim(b),
  _schema(nullptr)
{
  // Verify as Usd Shader:
  if (validate())
    _schema = retrieveShaderSchema(*this);
}

ShaderPrim::ShaderPrim(const ShaderPrim& b) :
  Prim(b),
  _schema(b._schema)
{
  //std::cout << "ShaderPrim::ctor-shaderprimcopy(" << this << ") schema=" << _schema << ", " << *this << ", b=" << b << std::endl;
}

ShaderPrim&
ShaderPrim::operator = (const ShaderPrim& b)
{
  if (_pImpl.get() != b._pImpl.get())
  {
    _pImpl.reset(new Prim::Impl(*b._pImpl.get()));
    _schema = b._schema;
  }
  //std::cout << "ShaderPrim:: = shaderprimcopy(" << this << ") schema=" << _schema << ", " << *this << std::endl;
  return *this;
}

ShaderPrim::ShaderPrim(const Prim::Impl& b) :
  Prim(b),
  _schema(nullptr)
{

  // Verify as Usd Shader:
  if (validate())
    _schema = retrieveShaderSchema(*this);
}


/*!
*/
/*virtual*/
ShaderPrim::~ShaderPrim()
{
  //
}


bool
ShaderPrim::validate()
{
  if (!_pImpl.get())
    return false; // not valid...

  if (!isValidPrim(*this))
  {
    _pImpl.reset();
    return false;
  }
  return true; // still valid!
}


/*static*/ bool
ShaderPrim::isValidPrim(const Prim& prim)
{
  if (!prim)
    return false;
  else if (prim.impl()->sdfprim)
    return (prim.impl()->sdfprim->GetTypeName() == PXR_NS::TfToken("Shader")); // Sdf Shader schema
  else if (prim.impl()->usdprim)
    return prim.impl()->usdprim.IsA<PXR_NS::UsdShadeShader>(); // UsdShadeShader
  return false; // not a valid handle
}


//--------------------------------------------------------------------------------------------------


/*static*/ ShaderPrim
ShaderPrim::defineInLayer(const ShaderDesc& source_shader_desc,
                          const LayerRef&   layer,
                          const Path&       material_path)
{
  if (!layer || !source_shader_desc.isValid())
    return ShaderPrim();

  //source_shader_desc.print("        ", std::cout);

  ShaderPrim shaderprim(layer->definePrim(material_path.appendPath(
                                                    source_shader_desc.getMaterialRelativeName()), "Shader"));
  if (shaderprim.isValid())
  {
    shaderprim.setIdentifierAttr(source_shader_desc.getIdentifier());
    shaderprim.connectInputs(source_shader_desc, material_path);
    shaderprim.applyAttributeOverrides(source_shader_desc, true/*skip_connected_inputs*/);
    shaderprim.defineOutputs(source_shader_desc, material_path);

#if 1
    // ***** NOTE ******
    // If 'implemetationSource' attr is not set to 'id' then Hydra will throw an error in its material
    // network parser saying there's no valid 'info:id' - despite there being a valid info:id assigned...
    // *****************
    shaderprim.setImplementationSourceAttr("id");
#else
#  if 1
    shaderprim.setImplementationSourceAttr("sourceAsset");
    shaderprim.setSourceAssetAttr("./Project2DTexture.glslfx", "glslfx");
#  else
    shaderprim.setImplementationSourceAttr("sourceCode");
    shaderprim.setSourceCodeAttr(source_shader_desc.getSourceCode(), "glslfx");
#  endif
#endif
  }

  return shaderprim;
}


/*static*/ ShaderPrim
ShaderPrim::updateInLayer(const ShaderDesc& source_shader_desc,
                          const LayerRef&   layer,
                          const Path&       material_path)
{
  if (!layer || !layer->isValid() || !source_shader_desc.isValid())
    return ShaderPrim();

  ShaderPrim shaderprim(defineInLayer(layer, material_path.appendPath(source_shader_desc.getMaterialRelativeName())));
  if (shaderprim.isValid())
  {
    shaderprim.applyAttributeOverrides(source_shader_desc, true/*skip_connected*/);
  }

  return shaderprim;
}


/*static*/ ShaderPrim
ShaderPrim::updateInStage(const ShaderDesc& source_shader_desc,
                          const StageRef&   stage,
                          const Path&       material_path)
{
  if (!stage || !stage->isValid() || !source_shader_desc.isValid())
    return ShaderPrim();

  ShaderPrim shaderprim(getInStage(stage, material_path.appendPath(source_shader_desc.getMaterialRelativeName())));
  if (shaderprim.isValid())
  {
    shaderprim.applyAttributeOverrides(source_shader_desc, true/*skip_connected*/);
  }

  return shaderprim;
}


/*static*/ ShaderPrim
ShaderPrim::defineInLayer(const LayerRef& layer,
                          const Path&     path)
{
  return (layer) ? ShaderPrim(layer->definePrim(path, "Shader")) : ShaderPrim();
}


/*static*/ ShaderPrim
ShaderPrim::getInStage(const StageRef& stage,
                       const Path&     path)
{
  if (!stage || !stage->isValid() || path.empty())
    return ShaderPrim();

  PXR_NS::UsdPrim prim = stage->impl()->stage->GetPrimAtPath(toSdfPath(path));
  if (!prim.IsValid() || !prim.IsA<PXR_NS::UsdShadeShader>())
    return ShaderPrim(); // not a UsdShadeShader!

  return ShaderPrim(Prim::Impl(prim));
}


//--------------------------------------------------------------------------------------------------


bool
ShaderPrim::setIdentifierAttr(const std::string& identifier)
{
  if (!isValid())
    return false;

  if (Attribute id = createIntrinsicAttr(PXR_NS::UsdShadeTokens->infoId.GetString(), Value::Token))
  {
    id.setValue(identifier);
    return true;
  }
  return false;
}
std::string
ShaderPrim::getIdentifier() const
{
  std::string info_id;
  if (isValid())
    getAttr(PXR_NS::UsdShadeTokens->infoId.GetString()).getValue(info_id);
  return info_id;
}

//-------------------------------

bool
ShaderPrim::setImplementationSourceAttr(const std::string& implementation_source)
{
  if (!isValid())
    return false;

  if (Attribute attr = createIntrinsicAttr(PXR_NS::UsdShadeTokens->infoImplementationSource.GetString(), Value::Token))
  {
    attr.setValue(implementation_source);
    return true;
  }
  return false;
}
std::string
ShaderPrim::getImplementationSource() const
{
  std::string implementation_source;
  if (isValid())
    getAttr(PXR_NS::UsdShadeTokens->infoImplementationSource.GetString()).getValue(implementation_source);
  return implementation_source;
}

//-------------------------------

// This is same logic that's *hidden* in pxr/usd/usdShade/nodeDefAPI.cpp...
std::string
getSourceAssetAttrName(const std::string& source_namespace)
{
  if (source_namespace.empty())
    return std::string("info:sourceAsset");//PXR_NS::UsdShadeTokens->infoSourceAsset.GetString();
  return std::string("info:") + source_namespace + ":" + PXR_NS::UsdShadeTokens->sourceAsset.GetString();
}

bool
ShaderPrim::setSourceAssetAttr(const Path&    source_assert_path,
                               const std::string& source_namespace)
{
  if (!isValid())
    return false;

  if (Attribute attr = createIntrinsicAttr(getSourceAssetAttrName(source_namespace), Value::AssetPath))
  {
    attr.setValue(source_assert_path);
    return true;
  }
  return false;
}
std::string
ShaderPrim::getSourceAsset(const std::string& source_namespace) const
{
  std::string source_assert_path;
  if (isValid())
    getAttr(getSourceAssetAttrName(source_namespace)).getValue(source_assert_path);
  return source_assert_path;
}

//-------------------------------

// This is same logic that's *hidden* in pxr/usd/usdShade/nodeDefAPI.cpp...
std::string
getSourceCodeAttrName(const std::string& source_namespace)
{
  if (source_namespace.empty())
    return std::string("info:sourceCode");//PXR_NS::UsdShadeTokens->infoSourceCode.GetString();
  return std::string("info:") + source_namespace + ":" + PXR_NS::UsdShadeTokens->sourceCode.GetString();
}

bool
ShaderPrim::setSourceCodeAttr(const std::string& source_code,
                              const std::string& source_namespace)
{
  if (!isValid())
    return false;

  if (Attribute attr = createIntrinsicAttr(getSourceCodeAttrName(source_namespace), Value::String))
  {
    attr.setValue(source_code);
    return true;
  }
  return false;
}
std::string
ShaderPrim::getSourceCode(const std::string& source_namespace) const
{
  std::string source_code;
  if (isValid())
    getAttr(getSourceCodeAttrName(source_namespace)).getValue(source_code);
  return source_code;
}


//--------------------------------------------------------------------------------------------------


/*static*/
const SchemaDescription*
ShaderPrim::retrieveShaderSchema(const Prim& prim)
{
  if (!prim.isValid())
    return nullptr;

  // TODO: We should cast the prim to a ShaderPrim instead of replicating this getAttr code:
  std::string info_id; prim.getAttr(PXR_NS::UsdShadeTokens->infoId.GetString()).getValue(info_id);
  //const std::string info_id = shaderprim.getIdentifier();

  //std::cout << "ShaderPrim::retrieveShaderSchema(" << prim.getName() << ") info:id='" << info_id << "'" << std::endl;
  if (info_id.empty())
    return nullptr;

  // Schema identifier already in registry?
  const SchemaDescription* schema = SchemaDescription::getRegistryIdentifier(info_id);
  if (schema)
    return schema;

  // Not in registry yet, try to load identifier as schema name:
  schema = SchemaDescription::find(info_id.c_str());
  if (!schema || !schema->registryNode())
  {
    std::cerr << "ShaderPrim::retrieveShaderSchema(): error, cannot find shader schema for type name '" << info_id << "'" << std::endl;
    return nullptr; // can't find shader class by name...
  }

  return schema;
}


//--------------------------------------------------------------------------------------------------


bool
ShaderPrim::connectInputs(const ShaderDesc& source_shader_desc,
                          const Path&   material_path)
{
  if (!isValid())
    return false;

  const ShaderConnectionMap& connections = source_shader_desc.getInputConnections();
  if (connections.size() == 0)
    return true; // nothing to connect

  // Attach shader's input connections:
  bool ok = true;
  for (ShaderConnectionMap::const_iterator it=connections.begin(); it != connections.end(); ++it)
  {
    if (!it->second || it->second->size() == 0)
      return false; // shouldn't happen...
    const std::string&           connect_prop = it->first;
    const ShaderConnectionArray& connect_list = *it->second;
    if (gDebugShaderConnections)
    {
      std::cout << "  connect input prop '" << connect_prop << "'" << std::endl;
    }

    ShaderProperty input_prop = source_shader_desc.getInput(connect_prop.c_str());
    if (!input_prop.isValid())
    {
      // TODO should log a shader translation error!
      ok = false;
      continue; // connection is for an invalid or non-existant property...
    }

    if (Attribute attr = createIntrinsicAttr(input_prop.getFullName(),
                                            input_prop.getValueType()))
    {
      attr.clearConnections();
      for (size_t j=0; j < connect_list.size(); ++j)
      {
        const ShaderConnection& connection = connect_list[j];
        const auto output_prop_path = material_path.appendPath(
                                                               connection.upstreamNodePath).appendProperty(
                                                               connection.upstreamOutputName);
        if (gDebugShaderConnections)
        {
          std::cout << "    ...to output prop '" << output_prop_path << "'" << std::endl;
        }
        attr.addConnectionTo(output_prop_path);

      }
    }
    else
    {
      // TODO should log a shader translation error!
      ok = false;
      continue; // failed to create the attribute...
    }

  }

  return ok;
}


bool
ShaderPrim::defineOutputs(const ShaderDesc& source_shader_desc,
                          const Path&       material_path)
{
  if (!isValid())
    return false;

//TODO: we should only define output properties if they are overriden
// or actively connected to another ShaderPrim. I'm not sure if the UsdShadeConnectableAPI
// should be used for this...
// For the moment we just define all outputs so that things work.

  fdk::StringPtrList output_names;
  const size_t nShaderOutputs = source_shader_desc.getOutputNames(output_names);

  bool ok = true;
  for (size_t output=0; output < nShaderOutputs; ++output)
  {
    const ShaderProperty output_prop = source_shader_desc.getOutput(*output_names[output]);
    if (!output_prop.isValid())
    {
      // TODO should log a shader translation error!
      ok = false;
      continue; // invalid or non-existant property...
    }

    if (Attribute attr = createIntrinsicAttr(output_prop.getFullName(),
                                            output_prop.getValueType()))
    {

    }
    else
    {
      // TODO should log a shader translation error!
      ok = false;
      continue; // failed to create the attribute...
    }

  }

  return ok;
}


//--------------------------------------------------------------------------------------------------


ShaderDesc*
ShaderPrim::createShaderDesc(const MaterialPrim& parent_material,
                             bool                debug_info) const
{
  if (!isValid())
    return nullptr;

  if (!_schema)
  {
    // TODO: return an error or warning that the info:id was not converted to a ShaderSchema ...?
    if (debug_info)
    {
      std::cout << "    ShaderPrim::createShaderDesc(" << this << "): error, info:id='" << getIdentifier() << "'";
      std::cout << " was not found in ShaderSchema registry, cannot create ShaderDesc..." << std::endl;
    }
    return nullptr; // shader doesn't have a schema to construct with
  }

  ShaderDesc* shader_desc = new ShaderDesc(_schema);
  if (debug_info)
  {
    std::cout << "    ShaderPrim::createShaderDesc(" << this << "): schema=" << _schema->registryIdentifier() << std::endl;
  }

  // Set the relative path of the shader within the parent group:
  shader_desc->setMaterialRelativeName(parent_material.getPath().makeRelativePath(getPath()).asString());

  // Copy overrides and connections from the UsdAttributes:
  for (const auto attr : getAttributes()) {
    const Attribute prop = getAttr(attr);
    std::string prefix, name;
    fdk::splitPropertyName(prop.getName(), prefix, name);
    if (!prop.isValid() || prefix == "info")
      continue;
    if (debug_info)
    {
      std::cout << "                prop '" << name << "' = " << prop;
      std::cout << ", connected=" << prop.impl()->usdattr.HasAuthoredConnections();
      std::cout << std::endl;
    }

    if (prefix == "inputs")
    {
      if (prop.impl()->isUsd())
      {
        if (prop.impl()->usdattr.HasAuthoredConnections())
        {
          PXR_NS::SdfPathVector sources;
          prop.impl()->usdattr.GetConnections(&sources);
          if (sources.size() > 0)
          {
            if (debug_info)
            {
              std::cout << "                  connect='" << sources[0] << "'" << std::endl;
            }
            shader_desc->connectInput(name.c_str(), sources[0].GetString());
          }
        }
        else
        {
          Value val;
          if (prop.getType() == Value::AssetPath)
          {
            PXR_NS::SdfAssetPath path; prop.impl()->usdattr.Get<PXR_NS::SdfAssetPath>(&path);
            val = Value(path.GetResolvedPath());
          }
          else
          {
            prop.getValue(val);
          }

          if (debug_info)
          {
            std::cout << "                  val=(" << val << ")" << std::endl;
          }
          shader_desc->overrideInput(name, val);
        }
      }
      else
      {
        // TODO: do we need an Sdf implementation here? Normally we're always
        // dealing with a UsdStage at this point...
      }
    }

    if (prefix == "outputs")
    {
      if (prop.impl()->isUsd())
      {

      }
      else
      {
        // TODO: do we need an Sdf implementation here? Normally we're always
        // dealing with a UsdStage at this point...
      }
    }
  }

  return shader_desc;
}


} // namespace usg
