// Copyright (c) 2022 The Foundry Visionmongers Ltd. All Rights Reserved.

#include "usg/geom/Stage.h"
#include "usg/geom/Prim.h"
#include "usg/geom/GprimPrim.h"
#include "usg/geom/MeshPrim.h"

#include "usg/private/UsdImpl.h"
#include "fdk/base/Hash.h"
#include "fdk/base/private/TimeTracer.h"

#include "pxr/usd/usdGeom/bboxCache.h"
#include "pxr/usd/usdGeom/metrics.h"

#include <mutex>
#include <stack>

namespace usg {

Stage::Stage(const Stage::Impl& b)
  : _pImpl(new Stage::Impl(b))
{
}

Stage::~Stage()
{
}


bool
Stage::isValid() const
{
  return (_pImpl.get() != nullptr && _pImpl->stage);
}


Stage::Handle* Stage::Stage::getUsdStageRefPtr(int usd_version)
{
  return (_pImpl && usdVersionMatchesUsg(usd_version)) ?
            reinterpret_cast<Stage::Handle*>(&_pImpl->stage) :
            nullptr;
}

const Stage::Handle* Stage::getUsdStageRefPtr(int usd_version) const
{
  return (_pImpl && usdVersionMatchesUsg(usd_version)) ?
            reinterpret_cast<const Stage::Handle*>(&_pImpl->stage) :
            nullptr;
}


//----------------------------------------------------------------------------------


/*static*/ StageRef
Stage::Create(const LayerRef&  sessionLayer,
              const fdk::Hash* composeState,
              bool             loadAll)
{
  //std::cout << "  Stage::Create(" << this << ") sessionLayer=" << sessionLayer->getIdentifier() << std::endl;
  if (!sessionLayer || !sessionLayer->isValid()) {
    return StageRef();
  }

  const PXR_NS::UsdStage::InitialLoadSet loadSetPxr = loadAll ? PXR_NS::UsdStage::LoadAll : PXR_NS::UsdStage::LoadNone;
  StageRef stage(std::make_shared<Stage>(Stage::Impl(PXR_NS::UsdStage::Open(sessionLayer->impl()->layer, loadSetPxr))));

  if (composeState) {
    stage->_pImpl->compose_state = *composeState;
  }
  return stage;
}


/*static*/ StageRef
Stage::Create(const LayerRef&  rootLayer,
              const LayerRef&  sessionLayer,
              const fdk::Hash* composeState,
              const fdk::Hash* sessionState,
              bool             loadAll)
{
  //std::cout << "  Stage::Create(" << this << ") sessionLayer=" << sessionLayer->getIdentifier() << std::endl;
  if (!rootLayer || !rootLayer->isValid()) {
    return StageRef();
  }

  const PXR_NS::UsdStage::InitialLoadSet loadSetPxr = loadAll ? PXR_NS::UsdStage::LoadAll : PXR_NS::UsdStage::LoadNone;
  StageRef stage;

  if (sessionLayer  && sessionLayer->isValid()) {
    stage = std::make_shared<Stage>(Stage::Impl(PXR_NS::UsdStage::Open(rootLayer->impl()->layer, sessionLayer->impl()->layer, loadSetPxr)));
  }
  else {
    stage = std::make_shared<Stage>(Stage::Impl(PXR_NS::UsdStage::Open(rootLayer->impl()->layer, loadSetPxr)));
  }

  if (composeState) {
    stage->_pImpl->compose_state = *composeState;
  }
  if (sessionState) {
    stage->_pImpl->session_state = *sessionState;
  }
  return stage;
}


/*static*/ StageRef
Stage::CreateInMemory()
{
  auto sessionLayer = PXR_NS::SdfLayer::CreateAnonymous();
  return StageRef(std::make_shared<Stage>(Stage::Impl(PXR_NS::UsdStage::CreateInMemory(std::string(), sessionLayer))));
}


void      Stage::setComposeState(const fdk::Hash& compose_state) { if (_pImpl) { _pImpl->compose_state = compose_state; } }
fdk::Hash Stage::composeState() const { return (_pImpl) ? _pImpl->compose_state : fdk::Hash(); }

void      Stage::setProcessState(const fdk::Hash& process_state) { if (_pImpl) { _pImpl->process_state = process_state; } }
fdk::Hash Stage::processState() const { return (_pImpl) ? _pImpl->process_state : fdk::Hash(); }

void      Stage::setEditState(const fdk::Hash& edit_state) { if (_pImpl) { _pImpl->edit_state = edit_state; } }
fdk::Hash Stage::editState() const { return (_pImpl) ? _pImpl->edit_state : fdk::Hash(); }

void      Stage::setSessionState(const fdk::Hash& session_state) { if (_pImpl) { _pImpl->session_state = session_state; } }
fdk::Hash Stage::sessionState() const { return (_pImpl) ? _pImpl->session_state : fdk::Hash(); }

//-------------------------------------------------------

//-------------------------------------------------------

void
Stage::setEditTarget(const LayerRef& edit_layer)
{
  if (!edit_layer || !isValid())
    return;

  PXR_NS::UsdEditTarget target = _pImpl->stage->GetEditTargetForLocalLayer(edit_layer->impl()->layer);

  _pImpl->stage->SetEditTarget(target);
}

LayerRef Stage::getRootLayer() const
{
  return Layer::Create(Layer::Impl(_pImpl->stage->GetRootLayer()));
}

Path
Stage::getDefaultParentPath() const
{
  const auto rootLayer = getRootLayer();
  fdk::StringList identifiers;
  rootLayer->getSubLayers(identifiers);

  for (const auto& identifier : identifiers) {
    auto layer = Layer::FindOrOpen(identifier);
    const auto path = layer->getDefaultParentPath();
    if (!path.empty()) {
      return path;
    }
    break;
  }
  return Path();
}

Prim
Stage::getRootPrim() const
{
  return (isValid()) ? Prim(Prim::Impl(_pImpl->stage->GetPseudoRoot())) : Prim();
}

Prim
Stage::getPrimAtPath(const Path& path) const
{
  return (isValid() && !path.empty()) ? Prim(Prim::Impl(_pImpl->stage->GetPrimAtPath(toSdfPath(path)))) : Prim();
}


/*! Yuck! For debugging only!
*/
size_t
Stage::getPrimPaths(PathArray& prim_paths) const
{
  prim_paths.clear();
  if (isValid()) {
    prim_paths.reserve(50);
    for (const PXR_NS::UsdPrim& prim : PXR_NS::UsdPrimRange::Stage(_pImpl->stage)) {
      if (prim_paths.size() >= prim_paths.capacity()) {
        prim_paths.reserve(prim_paths.capacity() * 2);
      }
      prim_paths.push_back(prim.GetPath().GetString());
    }
  }

  return prim_paths.size();
}

template<class T>
PathArray
Stage::findPrimsOfType(const PathArray& primPaths,
                       bool             matchDescendents) const
{
  PathArray result;
  if (!matchDescendents) {
    // Only match the prims in the paths list, don't walk down the hierarchy:
    result.reserve(primPaths.size());
    for (const auto& path : primPaths) {
      if (const auto prim = getPrimAtPath(path)) {
        if (prim.isA<T>()) {
          result.push_back(prim.getPath());
        }
      }
    }
  }
  else {
    // Walk down the hierarchy matching all prims of primType:
    std::stack<Prim> stack;
    std::vector<Prim> children;
    std::set<Path> matched;

    // TODO ndk14: this could probably be made faster by using the USD api
    // to traverse the stage:
    for (const auto& path : primPaths) {
      if (const auto rootPrim = getPrimAtPath(path)) {
        stack.push(rootPrim);
        while (!stack.empty()) {
          const auto prim = stack.top();
          stack.pop();
          if (prim.isA<T>()) {
            matched.insert(prim.getPath());
          }
          prim.getChildren(children);
          for (const auto p : children) {
            stack.push(p);
          }
        }
      }
    }

    result.reserve(matched.size());
    for (const auto& path : matched) {
      result.push_back(path);
    }
  }
  return result;
}

// Explicit instantiations for our builtin types
template USG_API PathArray Stage::findPrimsOfType<BoundablePrim>(const PathArray&, bool) const;
template USG_API PathArray Stage::findPrimsOfType<ImageablePrim>(const PathArray&, bool) const;
template USG_API PathArray Stage::findPrimsOfType<GeometryPrim>(const PathArray&, bool) const;
template USG_API PathArray Stage::findPrimsOfType<XformablePrim>(const PathArray&, bool) const;
template USG_API PathArray Stage::findPrimsOfType<PointBasedPrim>(const PathArray&, bool) const;
template USG_API PathArray Stage::findPrimsOfType<MeshPrim>(const PathArray&, bool) const;

template<class T>
std::vector<T> Stage::findDescendentPrims(const PathArray& primPaths) const
{
  std::vector<T> result;
  std::stack<Prim> stack;
  std::vector<Prim> children;

  // TODO ndk14: this could probably be made faster by using the USD api
  // to traverse the stage:
  for (const auto& path : primPaths) {
    if (const auto rootPrim = getPrimAtPath(path)) {
      stack.push(rootPrim);
      while (!stack.empty()) {
        const auto prim = stack.top();
        stack.pop();
        if (prim.isA<T>()) {
          result.push_back(prim);
        }
        prim.getChildren(children);
        for (const auto p : children) {
          stack.push(p);
        }
      }
    }
  }
  return result;
}

// Explicit instantiations for our builtin types
template USG_API std::vector<BoundablePrim> Stage::findDescendentPrims<BoundablePrim>(const PathArray&) const;
template USG_API std::vector<ImageablePrim> Stage::findDescendentPrims<ImageablePrim>(const PathArray&) const;
template USG_API std::vector<GeometryPrim> Stage::findDescendentPrims<GeometryPrim>(const PathArray&) const;
template USG_API std::vector<XformablePrim> Stage::findDescendentPrims<XformablePrim>(const PathArray&) const;
template USG_API std::vector<PointBasedPrim> Stage::findDescendentPrims<PointBasedPrim>(const PathArray&) const;
template USG_API std::vector<MeshPrim> Stage::findDescendentPrims<MeshPrim>(const PathArray&) const;

//----------------------------------------------------------------------------------


//----------------------------------------------------------------------------------

fdk::Box3d
Stage::getWorldBbox(const fdk::TimeValue& time) const
{
  mFnTimeTracedFunction();
  if (!isValid()) {
    return fdk::Box3d();
  }

  fdk::Hash worldBoundsHash = processState();
  worldBoundsHash.append(time);

  auto& worldBoundsCache = _pImpl->world_bounds;
  auto cacheIter = worldBoundsCache.find(worldBoundsHash);
  if (cacheIter == worldBoundsCache.end()) {
    // Purposes:
    //   default, render, proxy, guide
    cacheIter = worldBoundsCache.emplace(worldBoundsHash,
                                         PXR_NS::UsdGeomBBoxCache(time,
                                                                  PXR_NS::TfTokenVector{PXR_NS::UsdGeomTokens->default_,
                                                                                        PXR_NS::UsdGeomTokens->proxy,
                                                                                        PXR_NS::UsdGeomTokens->render},
                                                                  true/*useExtentsHint*/,
                                                                  false/*ignoreVisibility*/)).first;
  }

  const PXR_NS::GfBBox3d world_bounds = cacheIter->second.ComputeWorldBound(_pImpl->stage->GetPseudoRoot());

  // TODO: enable the infinite bbox test:
  if (world_bounds.GetRange().IsEmpty() /*|| self._isInfiniteBBox(world_bounds)*/) {
    return fdk::Box3d(); // empty
  }

  const PXR_NS::GfRange3d bbox = world_bounds.ComputeAlignedRange();

  return reinterpret_cast<const fdk::Box3d&>(bbox);
}

void Stage::setMetersPerUnit(double metersPerUnit)
{
  if (isValid()) { PXR_NS::UsdGeomSetStageMetersPerUnit(_pImpl->stage, metersPerUnit); }
}

bool Stage::hasMetersPerUnit() const
{
  return (isValid()) ? PXR_NS::UsdGeomStageHasAuthoredMetersPerUnit(_pImpl->stage) : false;
}

double
Stage::getMetersPerUnit() const
{
  if (isValid()) {
    return PXR_NS::UsdGeomGetStageMetersPerUnit(_pImpl->stage);
  }
  // Replicate the centimeters fallback in UsdGeomGetStageMetersPerUnit():
  return PXR_NS::UsdGeomLinearUnits::centimeters;
}


//----------------------------------------------------------------------------------


std::string   g_preedit_cache_id = "GeomOp:preedit_cache:id";
std::string g_preedit_cache_hash = "GeomOp:preedit_cache:hash";

#if 0
void
Stage::updatePreEditCache(Layer*   cache_layer,
                          uint64_t cache_hash)
{
  assert(_pImpl);
  assert(_pImpl->stage);

  // Allow the cache id to be an empty string:
  PXR_NS::TfToken cache_layer_id;
  if (cache_layer && cache_layer->isValid())
    cache_layer_id = PXR_NS::TfToken(cache_layer->getIdentifier());

  PXR_NS::VtDictionary dict;
  dict[g_preedit_cache_id  ] = PXR_NS::VtValue(cache_layer_id);
  dict[g_preedit_cache_hash] = PXR_NS::VtValue(cache_hash);

  _pImpl->stage->SetMetadata(PXR_NS::SdfFieldKeys->CustomLayerData, dict);

  if (cache_layer && cache_layer->isValid())
    cache_layer->impl()->layer->SetCustomLayerData(dict);
}
#endif
void
Stage::setPreEditCacheMetadata(const std::string& cache_layer_id,
                               uint64_t           cache_hash)
{
  assert(_pImpl);
  assert(_pImpl->stage);

  PXR_NS::VtDictionary dict;
  dict[g_preedit_cache_id  ] = PXR_NS::VtValue(cache_layer_id);
  dict[g_preedit_cache_hash] = PXR_NS::VtValue(cache_hash);

  _pImpl->stage->SetMetadata(PXR_NS::SdfFieldKeys->CustomLayerData, dict);
}


/*! Retrieve the current cache layer metadata values, returning true on success.

    This metdata dictionary is carrying the cache information set by the GeomOp
    management code.
*/
bool
Stage::getPreEditCacheMetadata(std::string& cache_layer_id,
                               uint64_t&    cache_layer_hash)
{
  assert(_pImpl);
  assert(_pImpl->stage);

  // Get the metdata dictionary carrying the cache information set by the GeomOp
  // management code:
  PXR_NS::VtDictionary dict;
  if (!_pImpl->stage->GetMetadata(PXR_NS::SdfFieldKeys->CustomLayerData, &dict))
    return false; // don't crash...

  return getPreEditCacheInfoFromDict(dict,
                                     cache_layer_id,
                                     cache_layer_hash);
}


void
Stage::retrievePreEditCacheProperties(const LayerRef&       cache_layer,
                                      const fdk::TimeValue& time)
{
  assert(_pImpl);
  assert(_pImpl->stage);

  if (!cache_layer || !cache_layer->isValid())
    return;

  const PXR_NS::SdfLayerHandle& layer = cache_layer->impl()->layer;

  // Go through all cache Layer attributes copying their values to the stage:
  std::vector<PXR_NS::SdfPath> attr_paths;
  attr_paths.reserve(50);
  {
    layer->Traverse(PXR_NS::SdfPath::AbsoluteRootPath(),
                    [&attr_paths, &layer](const PXR_NS::SdfPath& path)
                    {
                      if (path.IsPrimPropertyPath())
                      {
                        if (attr_paths.size() >= attr_paths.capacity())
                          attr_paths.reserve(attr_paths.capacity() * 2);
                        attr_paths.push_back(path);
                      }
                    });
  }
  const size_t nCacheAttrs = attr_paths.size();

  for (size_t i=0; i < nCacheAttrs; ++i)
  {
    const PXR_NS::SdfPath attrpath = attr_paths[i];
    //std::cout << "    " << i << ": '" << attrpath << "'";

    PXR_NS::SdfAttributeSpecHandle attrspec = layer->GetAttributeAtPath(attrpath);
    if (!attrspec)
    {
      //std::cout << " - attrspec not found, skipping!" << std::endl;
      continue;
    }
    //std::cout << " - aquired attrspec, copying value to stage" << std::endl;

    // Copy attrspec's default value to UsdAttribute in stage:
    const PXR_NS::VtValue cached_val = attrspec->GetField(PXR_NS::SdfFieldKeys->Default);
    if (!cached_val.IsEmpty())
    {
      PXR_NS::UsdAttribute usdattr = _pImpl->stage->GetAttributeAtPath(attrpath);
      if (usdattr && usdattr.IsValid())
      {
        usdattr.Set(cached_val);
      }
    }
  }

}


bool Stage::exportToFile(const std::string& filepath,
                         bool addComment,
                         const std::map<std::string, std::string>& fileFormatArgs)
{
  return _pImpl && _pImpl->stage && _pImpl->stage->Export(filepath, addComment, fileFormatArgs);
}

bool Stage::exportToString(std::string& text, bool addComment)
{
  if (_pImpl && _pImpl->stage) {
    _pImpl->stage->ExportToString(&text, addComment);
    return true;
  }
  text.clear();
  return false;
}

//----------------------------------------------------------------------------------

void Stage::setPopulationMask(const PathArray& primPaths)
{
  std::vector<PXR_NS::SdfPath> sdfPaths;
  std::transform(primPaths.begin(), primPaths.end(), std::back_inserter(sdfPaths),
      [](const Path& path) {
        return toSdfPath(path);
      }
  );
  PXR_NS::UsdStagePopulationMask mask(sdfPaths.begin(), sdfPaths.end());
  _pImpl->stage->SetPopulationMask(mask);
}

LayerRef Stage::getSessionLayer()
{
  if (_sessionLayer.get()) {
    return _sessionLayer;
  }
  return _sessionLayer = Layer::Create(Layer::Impl(_pImpl->stage->GetSessionLayer()));
}

void Stage::setStartTimeCode(fdk::TimeValue time)
{
  _pImpl->stage->SetStartTimeCode(time);
}

fdk::TimeValue Stage::getStartTimeCode()
{
  return _pImpl->stage->GetStartTimeCode();
}

void Stage::setEndTimeCode(fdk::TimeValue time)
{
  _pImpl->stage->SetEndTimeCode(time);
}

fdk::TimeValue Stage::getEndTimeCode()
{
  return _pImpl->stage->GetEndTimeCode();
}

void   Stage::setTimeCodesPerSecond(double tps) { _pImpl->stage->SetTimeCodesPerSecond(tps); }
double Stage::getTimeCodesPerSecond() const     { return _pImpl->stage->GetTimeCodesPerSecond(); }

void   Stage::setFramesPerSecond(double fps) { _pImpl->stage->SetFramesPerSecond(fps); }
double Stage::getFramesPerSecond() const     { return _pImpl->stage->GetFramesPerSecond(); }

//----------------------------------------------------------------------------------

void
Stage::printReport(std::ostream& o,
                   bool          print_contents) const
{
  if (!_pImpl)
  {
    o << "  stage: <null pImpl>" << std::endl;
    return;
  }
  else if (!_pImpl->stage)
  {
    o << "  stage: <null UsdStageRefPtr>" << std::endl;
    return;
  }

  o << "  stage: compose_state=0x" << std::hex << _pImpl->compose_state;
  o << ", process_state=0x" << _pImpl->process_state << std::dec;
  o << ", session_state=0x" << _pImpl->session_state << std::dec;
  o << std::endl;
  PXR_NS::SdfLayerHandleVector layers = _pImpl->stage->GetLayerStack(false/*includeSessionLayers*/);
  o << "    layers: " << layers.size() << std::endl;
  for (size_t i=0; i < layers.size(); ++i)
  {
    o << "      " << i;
    if (i == 0) {
      o << "(strongest) ";
    }
    else if (i == layers.size()-1) {
      o << "( weakest ) ";
    }
    else {
      o << "            ";
    }
    o << layers[i]->GetIdentifier() << std::endl;
  }
#if 0
  const PXR_NS::UsdEditTarget& target = _pImpl->stage->GetEditTarget();
  if (target.IsValid())
    o << "      edit-target layer: " << target.GetLayer()->GetIdentifier() << std::endl;
#endif

  if (print_contents)
  {
#if 1
    std::string str;
    _pImpl->stage->ExportToString(&str, false/*addSourceFileComment*/);
    o << str;
#else
    for (size_t i=0; i < layers.size(); ++i)
    {
      std::string str;
      layers[i]->ExportToString(&str);
      o << i << ":" << std::endl;
      o << str;
    }
#endif
  }

}

class StageNotificationsImpl : public PXR_NS::TfWeakBase {
public:
  class ChangedObjectsImpl : public StageNotifications::ChangedObjects {
  public:
    ChangedObjectsImpl(const PXR_NS::UsdNotice::ObjectsChanged& changedObjects)
      : _changedObjects(changedObjects)
    {
    }

    PathArray getChangedPaths() const override
    {
      auto changedPaths = _changedObjects.GetResyncedPaths();
      PathArray result;
      result.reserve(changedPaths.size());
      for (auto path : changedPaths) {
        result.push_back(fromSdfPath(path));
      }
      return result;
    }

    PathArray getChangedInfoPaths() const override
    {
      auto changedPaths = _changedObjects.GetChangedInfoOnlyPaths();
      PathArray result;
      result.reserve(changedPaths.size());
      for (auto path : changedPaths) {
        result.push_back(fromSdfPath(path));
      }
      return result;
    }

    TokenArray getChangedFields(const Path& path) const override
    {
      auto changedFields = _changedObjects.GetChangedFields(toSdfPath(path));
      TokenArray result;
      result.reserve(changedFields.size());
      for (auto field : changedFields) {
        result.push_back(fromTfToken(field));
      }
      return result;
    }

  private:
    const PXR_NS::UsdNotice::ObjectsChanged& _changedObjects;
  };

  StageNotificationsImpl(StageNotifications* owner, const StageRef& stage)
  {
    auto usdStage = stage->impl()->stage;
    _owner = owner;
    PXR_NS::TfNotice::Register(PXR_NS::TfCreateWeakPtr(this), &StageNotificationsImpl::handleObjectsChanged, usdStage);
    PXR_NS::TfNotice::Register(PXR_NS::TfCreateWeakPtr(this), &StageNotificationsImpl::handleStageContentsChanged, usdStage);
    PXR_NS::TfNotice::Register(PXR_NS::TfCreateWeakPtr(this), &StageNotificationsImpl::handleLayerMutingChanged, usdStage);
    PXR_NS::TfNotice::Register(PXR_NS::TfCreateWeakPtr(this), &StageNotificationsImpl::handleStageEditTargetChanged, usdStage);
  }

  void handleObjectsChanged(const PXR_NS::UsdNotice::ObjectsChanged& notice, const PXR_NS::UsdStageWeakPtr &sender)
  {
    ChangedObjectsImpl changedObjects(notice);
    _owner->handleObjectsChanged(changedObjects);
  }

  void handleStageContentsChanged(const PXR_NS::UsdNotice::StageContentsChanged& notice, const PXR_NS::UsdStageWeakPtr &sender)
  {
    _owner->handleContentsChanged();
  }

  void handleLayerMutingChanged(const PXR_NS::UsdNotice::LayerMutingChanged& notice, const PXR_NS::UsdStageWeakPtr &sender)
  {
    _owner->handleLayerMutingChanged();
  }

  void handleStageEditTargetChanged(const PXR_NS::UsdNotice::StageEditTargetChanged& notice, const PXR_NS::UsdStageWeakPtr &sender)
  {
    _owner->handleEditTargetChanged();
  }

private:
  friend class StageNotifications;
  StageNotifications* _owner;
};


StageNotifications::StageNotifications(const StageRef& stage)
  : _pImpl(new StageNotificationsImpl(this, stage))
{
}

StageNotifications::~StageNotifications()
{
  // Needed for std::unique_ptr
}

StageNotifications::ChangedObjects::~ChangedObjects()
{
}

} // namespace usg
