// Copyright (c) 2022 The Foundry Visionmongers Ltd. All Rights Reserved.

#include "usg/geom/Subdivide.h"
#include "usg/geom/GeomTokens.h"
#include "usg/base/ArgSet.h"

// For OpenSubdiv - note this should be in a plugin:
#include <opensubdiv/far/topologyDescriptor.h>
#include <opensubdiv/far/primvarRefiner.h>

namespace usg {

  //TODO: This should probably go into a new file...
  //--------------------------------------------------------------------------------------------------
  //--------------------------------------------------------------------------------------------------


  // Wrap fdk::Vec classes for OpenSubdiv use.
  // See opensubdiv/far/primvarRefiner.h and opensubdiv/hbr/fvarData.h:
  // <snip>
  //    Interpolation methods template both the source and destination
  //    data buffer classes. Client-code is expected to provide interfaces
  //    that implement the functions specific to its primitive variable
  //    data layout. Template APIs must implement the following:
  //
  //    class MySource {
  //        MySource & operator[](int index);
  //    };
  //
  //    class MyDestination {
  //        void Clear();
  //        void AddWithWeight(MySource const & value, float weight);
  //        void AddWithWeight(MyDestination const & value, float weight);
  //    };
  // </snip>
  //
  // fdk::Vec classes already have [] defined so we just need to expose
  // setToZero() and implement AddWithWeight().

  class OsdFloat
  {
    public:
      void Clear() { val = 0.0f; }
      void AddWithWeight(OsdFloat const& b, float weight) { val += b.val*weight; }
    protected:
      float val;
  };

  class OsdVec2f : public fdk::Vec2f
  {
    public:
      void Clear() { setToZero(); }
      void AddWithWeight(OsdVec2f const& b, float weight) { *this += b*weight; }
  };

  class OsdVec3f : public fdk::Vec3f
  {
    public:
      void Clear() { setToZero(); }
      void AddWithWeight(OsdVec3f const& b, float weight) { *this += b*weight; }
  };

  class OsdVec4f : public fdk::Vec4f
  {
    public:
      void Clear() { setToZero(); }
      void AddWithWeight(OsdVec4f const& b, float weight) { *this += b*weight; }
  };


  //----------------------------------------------------------------------------------


  /*! Return the appropriate refiner object for the given arguments.

  */
  OpenSubdiv::Far::TopologyRefiner*
  getRefiner(const ArgSet&                              exec_args,
             int32_t                                    nRefinementLevels,
             const OpenSubdiv::Far::TopologyDescriptor& desc)
  {
    const Token scheme(exec_args.getString(GeomTokens.subdivisionScheme, GeomTokens.catmullClark.asString()/*default*/));
    OpenSubdiv::Sdc::SchemeType type;
    if      (scheme == GeomTokens.catmullClark) { type = OpenSubdiv::Sdc::SCHEME_CATMARK;  }
    else if (scheme == GeomTokens.loop        ) { type = OpenSubdiv::Sdc::SCHEME_LOOP;     }
    else if (scheme == GeomTokens.bilinear    ) { type = OpenSubdiv::Sdc::SCHEME_BILINEAR; }
    else {
      // TODO: throw unrecognized-scheme warning
      type = OpenSubdiv::Sdc::SCHEME_CATMARK;
    }

  // TODO: loop mode crashes for some reason...force to catclark instead...
  if (type == OpenSubdiv::Sdc::SCHEME_LOOP) {
    type = OpenSubdiv::Sdc::SCHEME_CATMARK;
  }

    // TODO: these are set to common defaults, also check primvars copied in from file meshes:
    OpenSubdiv::Sdc::Options options;
    options.SetVtxBoundaryInterpolation(OpenSubdiv::Sdc::Options::VTX_BOUNDARY_EDGE_AND_CORNER);
    options.SetFVarLinearInterpolation(OpenSubdiv::Sdc::Options::FVAR_LINEAR_CORNERS_ONLY);
    options.SetCreasingMethod(OpenSubdiv::Sdc::Options::CREASE_UNIFORM);
    options.SetTriangleSubdivision(OpenSubdiv::Sdc::Options::TRI_SUB_CATMARK);

    // Create a FarTopologyRefiner from the descriptor:
    OpenSubdiv::Far::TopologyRefiner* refiner =
      OpenSubdiv::Far::TopologyRefinerFactory<OpenSubdiv::Far::TopologyDescriptor>::Create(desc,
        OpenSubdiv::Far::TopologyRefinerFactory<OpenSubdiv::Far::TopologyDescriptor>::Options(type, options));
    assert(refiner);

    // Declare how we're doing refinement to the desired target level:
    // TODO: support other refinement types like adaptive?
    if (1) {
      // TODO: these are set to arbitrary defaults, also check primvars copied in from file meshes:
      OpenSubdiv::Far::TopologyRefiner::UniformOptions refine_options(nRefinementLevels);
      refine_options.orderVerticesFromFacesFirst = false;
      refine_options.fullTopologyInLastLevel     = false;

      refiner->RefineUniform(refine_options);
    }
    else {
      // TODO: these are set to arbitrary defaults, also check primvars copied in from file meshes:
      OpenSubdiv::Far::TopologyRefiner::AdaptiveOptions refine_options(nRefinementLevels);
      refine_options.secondaryLevel              = 15;
      refine_options.useSingleCreasePatch        = false;
      refine_options.useInfSharpPatch            = false;
      refine_options.considerFVarChannels        = false;
      refine_options.orderVerticesFromFacesFirst = false;

      refiner->RefineAdaptive(refine_options);
    }

    return refiner;
  }


  void
  subdivideUniform(const ArgSet&          exec_args,
                   size_t                 nFaces,
                   size_t                 nVerts,
                   size_t                 nPoints,
                   int32_t                current_subd_level,
                   int32_t                target_subd_level,
                   MeshTessellateContext& tess_ctx)
  {
    if (!tess_ctx.faceVertexCounts          ||
        !tess_ctx.vert_position_indices     ||
        tess_ctx.position_lists.size() == 0 ||
        !tess_ctx.position_lists[0]) {
      return; // don't crash...
    }

    const int32_t nSrcFaces  = nFaces;
    const int32_t nSrcVerts  = nVerts;
    const int32_t nSrcPoints = nPoints;
    if (nSrcFaces == 0 || nSrcVerts == 0 || nSrcPoints == 0) {
      return; // don't crash...
    }

    if (current_subd_level < 0) {
      current_subd_level = 0;
    }
    const int nRefinementLevels = (target_subd_level - current_subd_level);
    if (nRefinementLevels <= 0) {
      return; // no need to further subdivide
    }

    // Attribs are already expanded to vertex rate so the primvar index arrays
    // will all point at the same indice array:
    std::vector<int32_t> fvarIndices(nSrcVerts);
    for (int32_t i=0; i < nSrcVerts; ++i) {
      fvarIndices[i] = i;
    }

    // Get number of Fvars and set up channel list:
    const int32_t nFvarChans = (int32_t)(tess_ctx.vert_float_attribs.size() +
                                         tess_ctx.vert_vec2_attribs.size()  +
                                         tess_ctx.vert_vec3_attribs.size()  +
                                         tess_ctx.vert_vec4_attribs.size());

    std::vector<OpenSubdiv::Far::TopologyDescriptor::FVarChannel> primvar_channels;
    primvar_channels.reserve(nFvarChans);
    for (size_t i=0; i < nFvarChans; ++i) {
      primvar_channels[i].numValues    = nSrcVerts;
      primvar_channels[i].valueIndices = fvarIndices.data();
    }

    OpenSubdiv::Far::TopologyDescriptor desc;
    desc.numVertices        = nSrcPoints; // points count, not verts!
    desc.numFaces           = nSrcFaces;
    desc.numVertsPerFace    = reinterpret_cast<const int32_t*>(tess_ctx.faceVertexCounts->data());
    desc.vertIndicesPerFace = reinterpret_cast<const int32_t*>(tess_ctx.vert_position_indices->data()); // at per-vert rate, not per-face!!!!
    //
    desc.numFVarChannels    = nFvarChans;
    desc.fvarChannels       = primvar_channels.data();

    // Create a FarTopologyRefiner from the descriptor:
    OpenSubdiv::Far::TopologyRefiner* refiner = getRefiner(exec_args, nRefinementLevels, desc);

    OpenSubdiv::Far::PrimvarRefiner primvarRefiner(*refiner);
    //const size_t nTotalRefinedFaces  = refiner->GetNumFacesTotal();
    const size_t nTotalRefinedVerts  = (nFvarChans > 0) ?
                                            refiner->GetNumFVarValuesTotal(0/*fvarChannelIdx*/) : 0;
    const size_t nTotalRefinedPoints = refiner->GetNumVerticesTotal();
    //std::cout << "    nTotalRefinedFaces=" << refiner->GetNumFacesTotal();
    //std::cout << ", nTotalRefinedPoints=" << nTotalRefinedPoints;
    //std::cout << ", nTotalRefinedVerts=" << nTotalRefinedVerts;
    //std::cout << std::endl;

    // Source pointer lists to swap during refinement loop:
    // TODO: make a struct to hold these...
    std::vector<OsdVec3f*> srcPositions(tess_ctx.position_lists.size());
    std::vector<OsdFloat*> srcFloatAttribs(tess_ctx.vert_float_attribs.size());
    std::vector<OsdVec2f*> srcVec2Attribs(tess_ctx.vert_vec2_attribs.size());
    std::vector<OsdVec3f*> srcVec3Attribs(tess_ctx.vert_vec3_attribs.size());
    std::vector<OsdVec4f*> srcVec4Attribs(tess_ctx.vert_vec4_attribs.size());
    // Destination buffers to accommodate the new verts/points being added:
    std::vector<Vec3fArray> refined_position_lists(tess_ctx.position_lists.size());       // point rate
    std::vector<FloatArray> refined_vert_float_attribs(tess_ctx.vert_float_attribs.size()); // face-vertex rate
    std::vector<Vec2fArray> refined_vert_vec2_attribs(tess_ctx.vert_vec2_attribs.size()); // face-vertex rate
    std::vector<Vec3fArray> refined_vert_vec3_attribs(tess_ctx.vert_vec3_attribs.size()); // face-vertex rate
    std::vector<Vec4fArray> refined_vert_vec4_attribs(tess_ctx.vert_vec4_attribs.size()); // face-vertex rate

    //------------------------------------------
    // Initialize positions buffers:
    for (size_t i=0; i < refined_position_lists.size(); ++i) {
      assert(tess_ctx.position_lists[i] != nullptr);
      assert(tess_ctx.position_lists[i]->size() == nSrcPoints);
      srcPositions[i] = reinterpret_cast<OsdVec3f*>(tess_ctx.position_lists[i]->data());
      refined_position_lists[i].resize(nTotalRefinedPoints);
    }
    if (nFvarChans > 0) {
      //------------------------------------------
      // Initialize fvars buffers:
      for (size_t i=0; i < refined_vert_float_attribs.size(); ++i) {
        assert(tess_ctx.vert_float_attribs[i] != nullptr);
        assert(tess_ctx.vert_float_attribs[i]->size() == nSrcVerts);
        srcFloatAttribs[i] = reinterpret_cast<OsdFloat*>(tess_ctx.vert_float_attribs[i]->data());
        refined_vert_float_attribs[i].resize(nTotalRefinedVerts);
      }
      for (size_t i=0; i < refined_vert_vec2_attribs.size(); ++i) {
        assert(tess_ctx.vert_vec2_attribs[i] != nullptr);
        assert(tess_ctx.vert_vec2_attribs[i]->size() == nSrcVerts);
        srcVec2Attribs[i] = reinterpret_cast<OsdVec2f*>(tess_ctx.vert_vec2_attribs[i]->data());
        refined_vert_vec2_attribs[i].resize(nTotalRefinedVerts);
      }
      for (size_t i=0; i < refined_vert_vec3_attribs.size(); ++i) {
        assert(tess_ctx.vert_vec3_attribs[i] != nullptr);
        assert(tess_ctx.vert_vec3_attribs[i]->size() == nSrcVerts);
        srcVec3Attribs[i] = reinterpret_cast<OsdVec3f*>(tess_ctx.vert_vec3_attribs[i]->data());
        refined_vert_vec3_attribs[i].resize(nTotalRefinedVerts);
      }
      for (size_t i=0; i < refined_vert_vec4_attribs.size(); ++i) {
        assert(tess_ctx.vert_vec4_attribs[i] != nullptr);
        assert(tess_ctx.vert_vec4_attribs[i]->size() == nSrcVerts);
        srcVec4Attribs[i] = reinterpret_cast<OsdVec4f*>(tess_ctx.vert_vec4_attribs[i]->data());
        refined_vert_vec4_attribs[i].resize(nTotalRefinedVerts);
      }
    }

    // Refine mesh attributes to desired level:
    for (int srcLevel=0; srcLevel < nRefinementLevels; ++srcLevel) {
      const int dstLevel = (srcLevel + 1);
      //std::cout << "       srcLevel " << srcLevel << " dstLevel " << dstLevel << std::endl;

      //------------------------------------------
      // Positions:
      for (size_t i=0; i < refined_position_lists.size(); ++i) {
        OsdVec3f* dst = (srcLevel == 0) ?
                          reinterpret_cast<OsdVec3f*>(refined_position_lists[i].data()) :
                          srcPositions[i];
        dst += refiner->GetLevel(srcLevel).GetNumVertices();
        primvarRefiner.Interpolate(dstLevel, srcPositions[i], dst);
        srcPositions[i] = dst;
      }

      if (nFvarChans > 0)
      {
        //------------------------------------------
        // FVars:
        uint32_t fvar_chan = 0;
        for (size_t i=0; i < refined_vert_float_attribs.size(); ++i, ++fvar_chan) {
          const size_t nSrcFVars = refiner->GetLevel(srcLevel).GetNumFVarValues(fvar_chan);
          OsdFloat* src = srcFloatAttribs[i];
          OsdFloat* dst = ((srcLevel == 0) ?
                            reinterpret_cast<OsdFloat*>(refined_vert_float_attribs[i].data()) :
                            src) + nSrcFVars;
          primvarRefiner.InterpolateFaceVarying(dstLevel, src, dst, fvar_chan);
          srcFloatAttribs[i] = dst;
        }
        for (size_t i=0; i < refined_vert_vec2_attribs.size(); ++i, ++fvar_chan) {
          const size_t nSrcFVars = refiner->GetLevel(srcLevel).GetNumFVarValues(fvar_chan);
          OsdVec2f* src = srcVec2Attribs[i];
          OsdVec2f* dst = ((srcLevel == 0) ?
                            reinterpret_cast<OsdVec2f*>(refined_vert_vec2_attribs[i].data()) :
                            src) + nSrcFVars;
          primvarRefiner.InterpolateFaceVarying(dstLevel, src, dst, fvar_chan);
          srcVec2Attribs[i] = dst;
        }
        for (size_t i=0; i < refined_vert_vec3_attribs.size(); ++i, ++fvar_chan) {
          const size_t nSrcFVars = refiner->GetLevel(srcLevel).GetNumFVarValues(fvar_chan);
          OsdVec3f* src = srcVec3Attribs[i];
          OsdVec3f* dst = ((srcLevel == 0) ?
                            reinterpret_cast<OsdVec3f*>(refined_vert_vec3_attribs[i].data()) :
                            src) + nSrcFVars;
          primvarRefiner.InterpolateFaceVarying(dstLevel, src, dst, fvar_chan);
          srcVec3Attribs[i] = dst;
        }
        for (size_t i=0; i < refined_vert_vec4_attribs.size(); ++i, ++fvar_chan) {
          const size_t nSrcFVars = refiner->GetLevel(srcLevel).GetNumFVarValues(fvar_chan);
          OsdVec4f* src = srcVec4Attribs[i];
          OsdVec4f* dst = ((srcLevel == 0) ?
                            reinterpret_cast<OsdVec4f*>(refined_vert_vec4_attribs[i].data()) :
                            src) + nSrcFVars;
          primvarRefiner.InterpolateFaceVarying(dstLevel, src, dst, fvar_chan);
          srcVec4Attribs[i] = dst;
        }
      }
    }

    // Copy refined point/vert data.
    // We need to expand out the point/vert indices back to the flattened point-rate/vertex-rate:
    const OpenSubdiv::Far::TopologyLevel& refLastLevel = refiner->GetLevel(nRefinementLevels);

    const int32_t nLevelFaces  = refLastLevel.GetNumFaces();
    const int32_t nLevelVerts  = nLevelFaces*4; // always quads
    const int32_t nLevelPoints = refLastLevel.GetNumVertices();
    const int32_t nLevelFVars  = (nFvarChans > 0) ?
                                  refLastLevel.GetNumFVarValues(0/*fvarChannelIdx*/) : 0;

    // Gets the starting point & facevert index in the highest refinement level:
    const int32_t levelPointsStart = (refiner->GetNumVerticesTotal() - nLevelPoints);
    const int32_t levelFVarsStart  = (nFvarChans > 0) ?
                                        (refiner->GetNumFVarValuesTotal(0/*fvarChannelIdx*/) - nLevelFVars) : 0;
    //std::cout << "    final level " << current_subd_level+nRefinementLevels << ": nLevelFaces=" << nLevelFaces;
    //std::cout << ", nLevelPoints=" << nLevelPoints;
    //std::cout << ", nLevelFVars=" << nLevelFVars;
    //std::cout << ", levelPointsStart=" << levelPointsStart;
    //std::cout << ", levelFVarsStart=" << levelFVarsStart;
    //std::cout << std::endl;

    // Point data is copied straight over:
    assert((levelPointsStart + nLevelPoints) <= nTotalRefinedPoints);

    for (size_t i=0; i < tess_ctx.position_lists.size(); ++i) {
      tess_ctx.position_lists[i]->resize(nLevelPoints);
      memcpy(tess_ctx.position_lists[i]->data(), refined_position_lists[i].data()+levelPointsStart, sizeof(fdk::Vec3f)*nLevelPoints);
    }

    if (nFvarChans > 0) {
      for (size_t i=0; i < tess_ctx.vert_float_attribs.size(); ++i) {
        tess_ctx.vert_float_attribs[i]->resize(nLevelVerts);
      }
      for (size_t i=0; i < tess_ctx.vert_vec2_attribs.size(); ++i) {
        tess_ctx.vert_vec2_attribs[i]->resize(nLevelVerts);
      }
      for (size_t i=0; i < tess_ctx.vert_vec3_attribs.size(); ++i) {
        tess_ctx.vert_vec3_attribs[i]->resize(nLevelVerts);
      }
      for (size_t i=0; i < tess_ctx.vert_vec4_attribs.size(); ++i) {
        tess_ctx.vert_vec4_attribs[i]->resize(nLevelVerts);
      }
    }

    //

    tess_ctx.all_quads = false;
    tess_ctx.all_tris  = false;

    uint32_t nFaceVerts;
    switch (refiner->GetSchemeType()) {
      default:
      case OpenSubdiv::Sdc::SCHEME_CATMARK:  nFaceVerts = 4; tess_ctx.all_quads = true; break;
      case OpenSubdiv::Sdc::SCHEME_LOOP:     nFaceVerts = 3; tess_ctx.all_tris  = true; break;
      case OpenSubdiv::Sdc::SCHEME_BILINEAR: nFaceVerts = 4; tess_ctx.all_quads = true; break;
    }

    // Face-vert attrib data needs to be de-referenced via face indices:
    tess_ctx.faceVertexCounts->resize(nLevelFaces);
    tess_ctx.vert_position_indices->resize(nLevelVerts);

    uint32_t face_vert_start = 0; // global vert count
    for (int32_t f=0; f < nLevelFaces; ++f) {
      const OpenSubdiv::Far::ConstIndexArray facevert_position_indices = refLastLevel.GetFaceVertices(f);

      assert(facevert_position_indices.size() == nFaceVerts); // all refined faces should be same
      (*tess_ctx.faceVertexCounts)[f] = nFaceVerts; // refined face vert count

      for (uint32_t v=0; v < nFaceVerts; ++v) {
        (*tess_ctx.vert_position_indices)[face_vert_start+v] = facevert_position_indices[v];
      }

      // Flatten-copy the primvar values back the vert-rate buffers:
      if (nFvarChans > 0) {
        //------------------------------------------
        // FVars:
        uint32_t fvar_chan = 0;
        for (size_t i=0; i < tess_ctx.vert_float_attribs.size(); ++i, ++fvar_chan) {
          const OpenSubdiv::Far::ConstIndexArray face_fvar_indices = refLastLevel.GetFaceFVarValues(f, fvar_chan);
          //
          const float* src = refined_vert_float_attribs[i].data()   + levelFVarsStart;
          float*       dst = tess_ctx.vert_float_attribs[i]->data() + face_vert_start;
          for (int v=0; v < nFaceVerts; ++v) {
            *dst++ = src[face_fvar_indices[v]];
          }
        }
        for (size_t i=0; i < tess_ctx.vert_vec2_attribs.size(); ++i, ++fvar_chan) {
          const OpenSubdiv::Far::ConstIndexArray face_fvar_indices = refLastLevel.GetFaceFVarValues(f, fvar_chan);
          //
          const fdk::Vec2f* src = refined_vert_vec2_attribs[i].data()   + levelFVarsStart;
          fdk::Vec2f*       dst = tess_ctx.vert_vec2_attribs[i]->data() + face_vert_start;
          for (int v=0; v < nFaceVerts; ++v) {
            *dst++ = src[face_fvar_indices[v]];
          }
        }
        for (size_t i=0; i < tess_ctx.vert_vec3_attribs.size(); ++i, ++fvar_chan) {
          const OpenSubdiv::Far::ConstIndexArray face_fvar_indices = refLastLevel.GetFaceFVarValues(f, fvar_chan);
          //
          const fdk::Vec3f* src = refined_vert_vec3_attribs[i].data()   + levelFVarsStart;
          fdk::Vec3f*       dst = tess_ctx.vert_vec3_attribs[i]->data() + face_vert_start;
          for (int v=0; v < nFaceVerts; ++v) {
            *dst++ = src[face_fvar_indices[v]];
          }
        }
        for (size_t i=0; i < tess_ctx.vert_vec4_attribs.size(); ++i, ++fvar_chan) {
          const OpenSubdiv::Far::ConstIndexArray face_fvar_indices = refLastLevel.GetFaceFVarValues(f, fvar_chan);
          //
          const fdk::Vec4f* src = refined_vert_vec4_attribs[i].data()   + levelFVarsStart;
          fdk::Vec4f*       dst = tess_ctx.vert_vec4_attribs[i]->data() + face_vert_start;
          for (int v=0; v < nFaceVerts; ++v) {
            *dst++ = src[face_fvar_indices[v]];
          }
        }
      }

      face_vert_start += nFaceVerts;
    }

  }
}
