// Copyright (c) 2021 The Foundry Visionmongers Ltd. All Rights Reserved.

#include "usg/imaging/ImageInterface.h"

#include "pxr/imaging/hio/image.h"
#include "pxr/usd/ar/resolver.h"
#include <pxr/base/gf/vec3f.h>

#include <mutex>

//----------------------------------------------------------------------------------

const std::string gEnvDebugTextureHandling = ((std::getenv("USG_DEBUG_TEXTURE_HANDLING") != nullptr) ?
                                               std::getenv("USG_DEBUG_TEXTURE_HANDLING") : "");
const bool gDebugTextureHandling = (gEnvDebugTextureHandling == "1");

//----------------------------------------------------------------------------------

namespace usg
{

//----------------------------------------------------------------------------------

// This map stores the Description structures for every loaded ImageInterface plugin or
// built-in class:
using ImageInterfaceMap = std::unordered_map<std::string, const ImageInterface::Description*>;
static ImageInterfaceMap& getImageInterfaceMap()
{
  static ImageInterfaceMap g_interface_map;
  return g_interface_map;
}
static std::mutex g_map_lock;


ImageInterface::Description::Description(const char*      identifier,
                                         InterfaceBuilder builder) :
  image_identifier(identifier),
  interface_builder(builder)
{
  if (0/*gDebugTextureHandling*/) {
    std::cout << "ImageInterface::Description::ctor(" << this << ") image_identifier='" << image_identifier << "'" << std::endl;
  }

  // Register the Description:
  if (image_identifier && image_identifier[0]) {
    const std::string dso_name(image_identifier);

    std::lock_guard<std::mutex> guard(g_map_lock); // just in case...
    auto& interface_map = getImageInterfaceMap();
    if (interface_map.find(dso_name) == interface_map.end()) {
      interface_map[dso_name] = this;
    }
    else {
      // TODO: already exists! This should either throw an error or print a warning...
    }
  }
  else {
    // TODO: invalid identifier! This should either throw an error or print a warning...
  }
}

//----------------------------------------------------------------------------------


/*! Default ctor leaves an invalid ImageInterface.
*/
ImageInterface::ImageInterface(const std::string& nodepath) :
  _nodepath(nodepath)
{
  //
}


/*!
*/
/*virtual*/
ImageInterface::~ImageInterface()
{
  //
}


/*static*/
const ImageInterface::Description*
ImageInterface::find(const char* identifier)
{
  if (!identifier || !identifier[0]) {
    return nullptr;  // just in case...
  }

  const std::string dso_name(identifier);
  if (0/*gDebugTextureHandling*/) {
    std::cout << "ImageInterface::Description::find('" << dso_name << "')" << std::endl;
  }

  std::lock_guard<std::mutex> guard(g_map_lock); // just in case...
  auto& interface_map = getImageInterfaceMap();
  ImageInterfaceMap::const_iterator it = interface_map.find(dso_name);

  return (it != interface_map.end()) ? it->second : nullptr;
}


} // namespace usg


//----------------------------------------------------------------------------------
//----------------------------------------------------------------------------------


PXR_NAMESPACE_OPEN_SCOPE

//
// TODO: Note that HioImage subclass plugins MUST be in the pxr namespace otherwise
// the TfType plugin system fails to recognize the type...!
//

/*!
*/
class HioUsgImageInterface : public HioImage
{
public:
  usg::ImageInterface* _interface;    //!<
  std::string          _filename;     //!< Set during _OpenForReading()
  int32_t              _nchans;       //!< Number of channels packed (interleaved) in target buffer
  int32_t              _bpchan;       //!< Target buffer data type is 1,2, or 4 bytes per chan
  bool                 _signed;       //!< Target buffer is signed type
  bool                 _srgb;         //!< Target buffer should be sRGB biased (only for 1-byte buf)

public:
  HioUsgImageInterface() :
    HioImage(),
    _interface(nullptr),
    _nchans(0),
    _bpchan(0),
    _signed(false),
    _srgb(false)
  {
    if (gDebugTextureHandling) {
      std::cout << " HioUsgImageInterface::ctor(" << this << ")" << std::endl;
    }
  }

  ~HioUsgImageInterface()
  {
    if (_interface) {
      delete _interface;
    }
    if (gDebugTextureHandling) {
      std::cout << "~HioUsgImageInterface::dtor(" << this << ")" << std::endl;
    }
  }


  const std::string&  GetFilename() const override { return _filename; }
  int                 GetWidth()    const override { return (_interface) ? _interface->width()  : 0; }
  int                 GetHeight()   const override { return (_interface) ? _interface->height() : 0; }

  //----------------------------------------------------------

  //! Extract the extension from the _filename var. Ripped from pxr/imaging/hio/stbImage.cpp.
  std::string  getFilenameExtension() const
  {
    std::string fileExtension = ArGetResolver().GetExtension(_filename);
    //convert to lowercase
    transform(fileExtension.begin(), 
              fileExtension.end(), 
              fileExtension.begin(), ::tolower);
    return fileExtension;
  }

  //!
  bool  _OpenForReading(const std::string& filename,
                        int                subimage,
                        int                mip,
                        SourceColorSpace   sourceColorSpace,
                        bool               suppressErrors) override
  {
    if (gDebugTextureHandling) {
      std::cout << "  HioUsgImageInterface::_OpenForReading(" << this << ") file='" << filename << "'";
      std::cout << ", subimage=" << subimage << ", mip=" << mip;
      std::cout << ", sourceColorSpace=" << sourceColorSpace;
      std::cout << std::endl;
    }

    if (_interface) {
      delete _interface;
      _interface = nullptr;
    }
    _filename = filename;
    _bpchan = 0;
    _signed = false;
    _srgb   = false;
    _nchans = 0;

    // Find a matching ImageInterface subclass with this same extension:
    const std::string fileExtension = getFilenameExtension();
    const usg::ImageInterface::Description* desc = usg::ImageInterface::find(fileExtension.c_str());
    if (!desc) {
      std::cerr << "ImageInterface::create(): error, builder method failed to find an interface ";
      std::cerr << "for file type '" << fileExtension << "'  - this is likely due to a coding error.";
      return false;
    }

    _interface = desc->interface_builder(_filename);
    if (!_interface) {
      std::cerr << "ImageInterface::create(): error, builder method failed to allocate interface ";
      std::cerr << "for asset '" << filename << "'  - this is likely due to a coding error.";
      return false;
    }

    // ImageInterface ctor should have initialized and set up the interface instance,
    // returning true for isValid():
    if (!_interface->isValid()) {
      // TODO: error? warning?
      return false;
    }

    _nchans = std::min(4, _interface->numChannels()); // clamp to max
    if (_nchans == 0) {
      return false; // don't bother reading a zero-channel image
    }

    /*  There's only a few base HioImage data types that need real support (ignoring the compressed
        codecs which we don't know how to write) and the rest are just nChan(2,3,4) variants and
        signed/unsigned flavors of: char, short, int, half, float, double.

        Note there's only one non-linear one supported - sRGB - so we don't need to go crazy
        supporting 'n' variations of destination colorspaces, only 1.

        So, we only need to pass bytes-per-channel, signed/unsigned, sRGB-ness, and channel count
        to the texture Iop converter.

        Here's the HioImage base formats:
          HioFormatUNorm8,     // UNorm8 - a 1-byte value representing a float between 0 and 1.
          HioFormatSNorm8,     // SNorm8 - a 1-byte value representing a float between -1 and 1.
          HioFormatUNorm8srgb, // UNorm8 SRGB - a 1-byte value representing a float between 0 and 1.
          //
          HioFormatUInt16,     // UInt16 - a 2-byte unsigned short integer. 
          HioFormatInt16,      // Int16 - a 2-byte signed short integer. 
          //
          HioFormatUInt32,     // UInt32 - a 4-byte unsigned integer. 
          HioFormatInt32,      // Int32 - a 4-byte signed integer.
          //
          HioFormatFloat16,    // Float16 - a 2-byte IEEE half-precision float.
          HioFormatFloat32,    // Float32 - a 4-byte IEEE float.
          HioFormatDouble64,   // Double64 - a 8-byte IEEE double.
    */

    // TODO: Since HioImage currently does not appear to have a way of defining the type of
    // texture buffer except via the filename, this severely limits us since Nuke always
    // has float32 source data...

    // TODO: For now always output sRGB(Monitor LUT actually) 8-bit textures, which matches
    // legacy system but means we're more limited in our glslfx shaders...alternative is to
    // always use half-float buffers!

    if (1/*_interface->targetBufferType() == blah*/) {
      _bpchan = 1;      // 8-bit (no 10 or 12 bit support I suppose...)
      _signed = false;  // 0..255 = 0..1
      _srgb   = true;   // Apply monitor-correction curve
    }

    // sRGB always 8-bit unsigned:
    if (_srgb) {
      _bpchan = 1;
      _signed = false;
    }

    return true;
  }

  //----------------------------------------------------------

  HioFormat  GetFormat() const override
  {
    if (_nchans > 0) {
      if (_srgb) {
        // sRGB always 8-bit:
        switch (_nchans) {
          case 1: return HioFormatUNorm8srgb;
          case 2: return HioFormatUNorm8Vec2srgb;
          case 3: return HioFormatUNorm8Vec3srgb;
          case 4: return HioFormatUNorm8Vec4srgb;
        }
      }

      // Data or color values interpreted linearly (ie linear-light or raw data):
      switch (_nchans)
      {
        case 1: {
          switch (_bpchan) {
            case 1: return (_signed) ? HioFormatSNorm8 : HioFormatUNorm8;
            case 2: return HioFormatFloat16;
            case 4: return HioFormatFloat32;
            default: return HioFormatInvalid;
          }
        }
        case 2: {
          switch (_bpchan) {
            case 1: return (_signed) ? HioFormatSNorm8Vec2 : HioFormatUNorm8Vec2;
            case 2: return HioFormatFloat16Vec2;
            case 4: return HioFormatFloat32Vec2;
            default: return HioFormatInvalid;
          }
        }
        case 3: {
          switch (_bpchan) {
            case 1: return (_signed) ? HioFormatSNorm8Vec3 : HioFormatUNorm8Vec3;
            case 2: return HioFormatFloat16Vec3;
            case 4: return HioFormatFloat32Vec3;
            default: return HioFormatInvalid;
          }
        }
        case 4: {
          switch (_bpchan) {
            case 1: return (_signed) ? HioFormatSNorm8Vec4 : HioFormatUNorm8Vec4;
            case 2: return HioFormatFloat16Vec4;
            case 4: return HioFormatFloat32Vec4;
            default: return HioFormatInvalid;
          }
        }
      }
    }
    return HioFormatInvalid;
  }

  int  GetBytesPerPixel() const override { return (_bpchan * _nchans); }

  int  GetNumMipLevels() const override
  {
    // TODO: should we support mip levels in texture objects...? Probably...
    return 1;
  }

  bool  IsColorSpaceSRGB() const override { return _srgb; }

  //----------------------------------------------------------

  bool  GetMetadata(const TfToken& key,
                    VtValue*       value) const override
  {
    //! TODO: should we do something here...? We could provide Storm with information
    // from Nuke via this mechanism. What I have no idea atm...
    return false;
  }

  bool  GetSamplerMetadata(HioAddressDimension pname,
                           HioAddressMode*     param) const override
  {
    // Set the edge handling mode to the one that Nuke usually uses.
    // TODO: expose this somehow...? Not sure if any of the other
    // modes make much sense in normal Nuke context.
    /*
      enum HioAddressMode
      {
        HioAddressModeClampToEdge = 0,
        HioAddressModeMirrorClampToEdge,
        HioAddressModeRepeat,
        HioAddressModeMirrorRepeat,
        HioAddressModeClampToBorderColor
      };
    */
    if (param) {
      switch (pname) {
        case HioAddressDimensionU: { *param = HioAddressModeClampToEdge; return true; }
        case HioAddressDimensionV: { *param = HioAddressModeClampToEdge; return true; }
        default: return false;
      }
    }
    return false;
  }

  //----------------------------------------------------------

  //!
  bool  Read(const StorageSpec& storage) override { return ReadCropped(0, 0, 0, 0, storage); }
  //!
  bool  ReadCropped(int const          cropTop,
                    int const          cropBottom,
                    int const          cropLeft,
                    int const          cropRight,
                    const StorageSpec& storage) override
  {
    bool success = false;
    // Read should not have been called if we don't have a valid interface...!
    if (_interface) {
      const int32_t width  = _interface->width();
      const int32_t height = _interface->height();
      if (gDebugTextureHandling) {
        std::cout << "    HioUsgImageInterface::ReadCropped(" << this << ")";
        std::cout << " [" << width << " " << height << "] nChans=" << _nchans;
        std::cout << ", storage[" << storage.width << " " << storage.height << "] depth=" << storage.depth;
        std::cout << ", flipped=" << storage.flipped << ", data=" << storage.data;
        std::cout << ", format=" << storage.format;
#if 1
        std::cout << ", buffersize=" << HioGetDataSize(storage.format, GfVec3i(width, height, 1/*mip_levels?*/));
#else
        size_t blockWidth, blockHeight;
        const size_t bytesPerPixel = HioGetDataSizeOfFormat(storage.format, &blockWidth, &blockHeight);
        size_t numPixels = ((width  + (blockWidth  - 1)) / blockWidth ) *
                           ((height + (blockHeight - 1)) / blockHeight);
        std::cout << ", bytesPerPixel=" << bytesPerPixel;
        std::cout << ", block[" << blockWidth << "x" << blockHeight << "]";
        std::cout << ", numPixels=" << numPixels;
        std::cout << ", buffersize=" <<  numPixels * bytesPerPixel * std::max(1, 1/*mip_levels?*/);
#endif
        std::cout << std::endl;
      }

      // TODO: support the crop options - I think it's a crop *into* the target buffer,
      // not a crop outof the source buffer.


      if (width > 0 && height > 0 && _nchans > 0) {
        // Verify the data allocation is the expected size:
        //const size_t storageBufSize = HioGetDataSize(storage.format, GfVec3i(width, height, nChans));

        success = _interface->read(fdk::Box2i(0, 0, width, height),
                                   _nchans, _bpchan, _signed, _srgb,
                                   storage.data);
      }
    }

    if (!success) {
      // TODO: Produce a checkerboard pattern for debugging purposes
      unsigned char* p = reinterpret_cast<unsigned char*>(storage.data);
      for (int y=0; y < storage.height; ++y) {
        const int yc = (y/10) & 1;
        for (int x=0; x < storage.width; ++x) {
          int xc = (x/10) & 1;
          p[0] = static_cast<unsigned char>(0.18 * 0xff);//(xc & yc) ? 0xff : 0;
          p[1] = static_cast<unsigned char>(0.18 * 0xff);//(xc ^ yc) ? 0xff : 0;
          p[2] = static_cast<unsigned char>(0.18 * 0xff);//(xc | yc) ? 0xff : 0;
          p[3] = 0xff;
          p += _nchans;
        }
      }
    }

    return true;
  }


  //----------------------------------------------------------

  //!
  bool  _OpenForWriting(const std::string& filename) override { return false; }
  //!
  bool  Write(const StorageSpec&  storage,
              const VtDictionary& metadata) override { return false; }

};


TF_REGISTRY_FUNCTION(TfType)
{
  TfType t = TfType::Define<HioUsgImageInterface, TfType::Bases<HioImage>>();
  t.SetFactory< HioImageFactory<HioUsgImageInterface> >();
}

PXR_NAMESPACE_CLOSE_SCOPE

