// Copyright (c) 2021 The Foundry Visionmongers Ltd. All Rights Reserved.

#include "usg/imaging/RenderInterface.h"

#include "usg/private/UsdHdImpl.h"

#include <algorithm>

namespace usg
{

//----------------------------------------------------------------------------------

static std::vector<RenderInterface::RenderPluginDesc>  g_pluginIds;
static fdk::StringList g_pluginKnobStrings;


static void
_scanRendererPlugins()
{
  static std::once_flag pluginInitFlag;

  std::call_once(pluginInitFlag, [] {
    PXR_NS::HfPluginDescVector plugins;
    PXR_NS::HdRendererPluginRegistry::GetInstance().GetPluginDescs(&plugins);

    g_pluginIds.reserve(plugins.size());
    g_pluginKnobStrings.reserve(plugins.size());
    std::ostringstream buf;

    for (const PXR_NS::HfPluginDesc& pluginDesc : plugins) {
      g_pluginIds.push_back({fromTfToken(pluginDesc.id), pluginDesc.displayName, pluginDesc.priority});
      buf << pluginDesc.id.GetString() << '\t' << pluginDesc.displayName;
      g_pluginKnobStrings.push_back(buf.str());
      buf.str(std::string());
      buf.clear();
    }
  });
}


//----------------------------------------------------------------------------------


/*static*/ RenderInterface::DataFormat RenderInterface::FormatUNorm8       = PXR_NS::HdFormatUNorm8;
/*static*/ RenderInterface::DataFormat RenderInterface::FormatUNorm8Vec2   = PXR_NS::HdFormatUNorm8Vec2;
/*static*/ RenderInterface::DataFormat RenderInterface::FormatUNorm8Vec3   = PXR_NS::HdFormatUNorm8Vec3;
/*static*/ RenderInterface::DataFormat RenderInterface::FormatUNorm8Vec4   = PXR_NS::HdFormatUNorm8Vec4;
//
/*static*/ RenderInterface::DataFormat RenderInterface::FormatSNorm8       = PXR_NS::HdFormatSNorm8;
/*static*/ RenderInterface::DataFormat RenderInterface::FormatSNorm8Vec2   = PXR_NS::HdFormatSNorm8Vec2;
/*static*/ RenderInterface::DataFormat RenderInterface::FormatSNorm8Vec3   = PXR_NS::HdFormatSNorm8Vec3;
/*static*/ RenderInterface::DataFormat RenderInterface::FormatSNorm8Vec4   = PXR_NS::HdFormatSNorm8Vec4;
//
/*static*/ RenderInterface::DataFormat RenderInterface::FormatFloat16      = PXR_NS::HdFormatFloat16;
/*static*/ RenderInterface::DataFormat RenderInterface::FormatFloat16Vec2  = PXR_NS::HdFormatFloat16Vec2;
/*static*/ RenderInterface::DataFormat RenderInterface::FormatFloat16Vec3  = PXR_NS::HdFormatFloat16Vec3;
/*static*/ RenderInterface::DataFormat RenderInterface::FormatFloat16Vec4  = PXR_NS::HdFormatFloat16Vec4;
//
/*static*/ RenderInterface::DataFormat RenderInterface::FormatFloat32      = PXR_NS::HdFormatFloat32;
/*static*/ RenderInterface::DataFormat RenderInterface::FormatFloat32Vec2  = PXR_NS::HdFormatFloat32Vec2;
/*static*/ RenderInterface::DataFormat RenderInterface::FormatFloat32Vec3  = PXR_NS::HdFormatFloat32Vec3;
/*static*/ RenderInterface::DataFormat RenderInterface::FormatFloat32Vec4  = PXR_NS::HdFormatFloat32Vec4;
//
/*static*/ RenderInterface::DataFormat RenderInterface::FormatUInt16       = PXR_NS::HdFormatUInt16;
/*static*/ RenderInterface::DataFormat RenderInterface::FormatUInt16Vec2   = PXR_NS::HdFormatUInt16Vec2;
/*static*/ RenderInterface::DataFormat RenderInterface::FormatUInt16Vec3   = PXR_NS::HdFormatUInt16Vec3;
/*static*/ RenderInterface::DataFormat RenderInterface::FormatUInt16Vec4   = PXR_NS::HdFormatUInt16Vec4;
//
/*static*/ RenderInterface::DataFormat RenderInterface::FormatInt32        = PXR_NS::HdFormatInt32;
/*static*/ RenderInterface::DataFormat RenderInterface::FormatInt32Vec2    = PXR_NS::HdFormatInt32Vec2;
/*static*/ RenderInterface::DataFormat RenderInterface::FormatInt32Vec3    = PXR_NS::HdFormatInt32Vec3;
/*static*/ RenderInterface::DataFormat RenderInterface::FormatInt32Vec4    = PXR_NS::HdFormatInt32Vec4;
/*static*/ RenderInterface::DataFormat RenderInterface::FormatFloat32UInt8 = PXR_NS::HdFormatFloat32UInt8;
//
/*static*/ RenderInterface::DataFormat RenderInterface::FormatInvalid      = PXR_NS::HdFormatInvalid;
/*static*/ RenderInterface::DataFormat RenderInterface::FormatCount        = PXR_NS::HdFormatCount;


RenderInterface::OutputLayer::OutputLayer() :
  _pImpl(nullptr)
{
  //
}

RenderInterface::OutputLayer::OutputLayer(const OutputLayer& b) :
  _pImpl((b._pImpl.get() != nullptr) ? new RenderOutputLayerHandle(*b._pImpl.get()) : nullptr)
{
  //
}
RenderInterface::OutputLayer& RenderInterface::OutputLayer::operator = (const OutputLayer& b)
{
  if (_pImpl.get() != b._pImpl.get())
    _pImpl.reset(new RenderOutputLayerHandle(*b._pImpl.get()));
  return *this;
}


RenderInterface::OutputLayer::OutputLayer(const RenderOutputLayerHandle& b) :
  _pImpl(new RenderOutputLayerHandle(b))
{
  //
}

/*virtual*/
RenderInterface::OutputLayer::~OutputLayer()
{
  //
}

bool
RenderInterface::OutputLayer::isValid() const
{
  return (_pImpl && _pImpl->buffer);
}

void  RenderInterface::OutputLayer::resolve() { if (_pImpl && _pImpl->buffer) _pImpl->buffer->Resolve(); }
void* RenderInterface::OutputLayer::map()     { return (_pImpl && _pImpl->buffer) ? _pImpl->buffer->Map() : nullptr; }
void  RenderInterface::OutputLayer::unmap()   { if (_pImpl && _pImpl->buffer) _pImpl->buffer->Unmap(); }

bool        RenderInterface::OutputLayer::isMapped()    const { return _pImpl->buffer->IsMapped(); }
bool        RenderInterface::OutputLayer::isConverged() const { return _pImpl->buffer->IsConverged(); }
int32_t     RenderInterface::OutputLayer::width()  const { return _pImpl->buffer->GetWidth(); }
int32_t     RenderInterface::OutputLayer::height() const { return _pImpl->buffer->GetHeight(); }
int32_t     RenderInterface::OutputLayer::nChans() const { return _pImpl->nchans; }
int32_t     RenderInterface::OutputLayer::bytesPerChan() const { return _pImpl->bpchan; }
bool        RenderInterface::OutputLayer::isSignedData() const { return _pImpl->is_signed; }
bool        RenderInterface::OutputLayer::isFloatData()  const { return _pImpl->is_float; }

RenderInterface::DataFormat
RenderInterface::OutputLayer::dataFormat() const
{
  return static_cast<DataFormat>(_pImpl->buffer->GetFormat());
}

//----------------------------------------------------------------------------------


RenderInterface::RenderInterface() :
  _pImpl(new RenderInterfaceHandle()),
  _rerender(true),
  //
  _progressiveRendering(true),
  _showViewerGeometry(true),
  _enableLighting(true),
  _enableShadows(false),
  _enableSceneMaterials(true),
  _displayRender(true),
  _displayProxy(false),
  _displayGuides(false),
  _pointSize(3.0f),
  _refineLevel(0),
  //
  _flatShading(false),
  _cullStyle(static_cast<int>(PXR_NS::HdCullStyleNothing)),
  _textureSize(512),
  //
  _displayColor(0.18f),
  _ambientColor(0.2f),
  _wireframeColor(0.0f, 1.0f, 0.0f, 0.5f),
  _pointColor(1.0f, 0.0f, 0.0f, 1.0f)
{
  _scanRendererPlugins();

  if (!g_pluginIds.empty()) {
    _rendererId = g_pluginIds[0].id.asString();
  }
}


/*virtual*/
RenderInterface::~RenderInterface()
{
  //
}

bool
RenderInterface::isValid() const
{
  return (_pImpl != nullptr);
}

void RenderInterface::setRenderSetting(const Token& key, const Value& v)
{
  if (isRenderSetting(key)) {
    _pImpl->renderDelegate->SetRenderSetting(toTfToken(key), v.impl()->vtvalue);
  }
}

bool RenderInterface::isRenderSetting(const Token& key) const
{
  for (const auto& setting : _rendererSettings) {
    if (setting.key == key) {
      return true;
    }
  }
  return false;
}

const RenderInterface::RenderSettingsDescriptor& RenderInterface::getRenderSetting(const Token& key) const
{
  for (const auto& setting : _rendererSettings) {
    if (setting.key == key) {
      return setting;
    }
  }
  static const RenderSettingsDescriptor empty;
  return empty;
}

void
RenderInterface::initRenderer(const ArgSet& args)
{
  _rendererArgs = args;
  const bool initialized = _initRenderer(Token(args.getString("RendererId")));
}


/*static*/ const std::vector<RenderInterface::RenderPluginDesc>&
RenderInterface::rendererPlugins()
{
  if (g_pluginIds.empty()) {
    _scanRendererPlugins();
  }
  return g_pluginIds;
}

//! Checks whether th renderer plugin exists.
/*static*/ bool RenderInterface::rendererPluginExists(Token id, RenderPluginDesc* desc)
{
  if (g_pluginIds.empty()) {
    _scanRendererPlugins();
  }

  for (const auto& plugin : g_pluginIds) {
    if (plugin.id == id) {
      if (desc) {
        *desc = plugin;
      }
      return true;
    }
  }
  
  return false;
}


bool
RenderInterface::rendererInitialized() const
{
  return (_pImpl) ? _pImpl->initialized : false;
}


//--------------------------------------------------------------------

void
RenderInterface::setCameraFraming(const fdk::Box2i& displayWindow,
                                  const fdk::Box2i& dataWindow,
                                  float             pixelAspectRatio)
{
  if (!rendererInitialized()) {
    return;
  }
  _pImpl->taskController->SetFraming(PXR_NS::CameraUtilFraming(PXR_NS::GfRange2f(PXR_NS::GfVec2f(displayWindow.min.x,
                                                                                                 displayWindow.min.y),
                                                                                 PXR_NS::GfVec2f(displayWindow.max.x,
                                                                                                 displayWindow.max.y)),
                                                                                 PXR_NS::GfRect2i(PXR_NS::GfVec2i(dataWindow.x(), dataWindow.y()),
                                                                                 dataWindow.w(), dataWindow.h()),
                                                               pixelAspectRatio));
}


void
RenderInterface::setFreeCameraMatrices(const fdk::Mat4d& viewMatrix,
                                       const fdk::Mat4d& projectionMatrix)
{
  if (!rendererInitialized()) {
    return;
  }

  _pImpl->taskController->SetFreeCameraMatrices(reinterpret_cast<const PXR_NS::GfMatrix4d&>(viewMatrix),
                                                reinterpret_cast<const PXR_NS::GfMatrix4d&>(projectionMatrix));
}


void
RenderInterface::setOutputLayerSize(int32_t width,
                                    int32_t height)
{
  if (!rendererInitialized()) {
    return;
  }
  _pImpl->taskController->SetRenderBufferSize(PXR_NS::GfVec2i(width, height));
}

void RenderInterface::setSelected(const usg::PathSet& newSelection)
{
  std::set_difference(_selectedPrims.begin(), _selectedPrims.end(),
                      newSelection.begin(), newSelection.end(),
                      std::inserter(_deselectedPrims, _deselectedPrims.end()));
  _selectedPrims = newSelection;
}


//--------------------------------------------------------------------


void
RenderInterface::executeTasks(bool fullyConverge)
{
  if (!rendererInitialized()) {
    return;
  }

  PXR_NS::HdTaskSharedPtrVector tasks = _pImpl->taskController->GetRenderingTasks();
  if (fullyConverge) {
    do {
      _pImpl->engine->Execute(_pImpl->renderIndex, &tasks);
    }
    while (!_pImpl->taskController->IsConverged());
  }
  else {
    _pImpl->engine->Execute(_pImpl->renderIndex, &tasks);
  }
}


bool
RenderInterface::isConverged() const
{
  if (!rendererInitialized()) {
    return false;
  }
  return _pImpl->taskController->IsConverged();
}


void
RenderInterface::setOutputLayers(const TokenArray& names)
{
  if (!rendererInitialized()) {
    return;
  }

  PXR_NS::TfTokenVector tnames; tnames.reserve(names.size());
  for (const Token& name : names) {
    tnames.push_back(toTfToken(name));
  }
  _pImpl->taskController->SetRenderOutputs(tnames);
}


RenderInterface::OutputLayer
RenderInterface::getOutputLayer(const Token& name) const
{
  if (!rendererInitialized()) {
    return OutputLayer();
  }

  PXR_NS::HdRenderBuffer* buffer = _pImpl->taskController->GetRenderOutput(toTfToken(name));
  if (buffer) {
    return OutputLayer(RenderOutputLayerHandle(buffer));
  }
  return OutputLayer();
}


} //namespace usg

