// Copyright (c) 2021 The Foundry Visionmongers Ltd. All Rights Reserved.

#include "usg/imaging/StormRenderInterface.h"

#include "usg/geom/LightPrim.h"
#include "usg/geom/MaterialPrim.h"
#include "usg/geom/MeshPrim.h"
#include "usg/geom/ShaderDesc.h"
#include "usg/geom/Stage.h"
#include "usg/lux/ShapingAPI.h"
#include "usg/lux/LuxTokens.h"
#include "usg/lux/ShadowAPI.h"

#include "usg/private/UsdHdImpl.h"

#include "fdk/base/private/TimeTracer.h"
#include "fdk/math/Quat.h"

#include "pxr/usdImaging/usdImaging/version.h"
#if USD_IMAGING_API_VERSION >= 13
# include "pxr/imaging/hgi/hgi.h"
# include "pxr/imaging/hgi/tokens.h"
#endif

#include "pxr/imaging/glf/glContext.h"
#include "pxr/imaging/glf/contextCaps.h"

#include "pxr/base/gf/frustum.h"
#include "pxr/imaging/hd/light.h"
#include "pxr/imaging/hd/material.h"
#include "pxr/imaging/hd/rprimCollection.h"
#include "pxr/imaging/hd/tokens.h"
#include "pxr/imaging/hdSt/hgiConversions.h"
#include "pxr/imaging/hdSt/renderDelegate.h"
#include "pxr/imaging/hdx/shadowMatrixComputation.h"
#include "pxr/imaging/hdx/simpleLightTask.h"
#include "pxr/imaging/hio/glslfx.h"
#include "pxr/usd/sdr/registry.h"
#include "pxr/usd/sdr/shaderProperty.h"
#include "pxr/usd/usdLux/distantLight.h"
#include "pxr/usdImaging/usdImaging/lightAdapter.h"


#include <stack>

namespace {

// Shadows are implemented by telling the render delegate we have one extra simpleLight for each light on the scene.
// This is done by inserting a Sprim with path '{DelegateID}{light path}{kShadowsPath}'. Whenever the render delegate
// calls for paths ending with kShadowsPath we can extract the correct light path and return the correct attributes.
constexpr const char* kShadowsPath = "__UPE_shadows_workaround__";

PXR_NS::GfVec3f ConvertLightDirection(const usg::LightPrim& light, const fdk::Vec3f &direction) {
  const usg::XformablePrim xform(light);
  auto matrix = xform.getConcatenatedXform();
  const fdk::Vec3f rotate = matrix.getRotations(fdk::RotateOrder::ZXY);
  fdk::Mat4f rotationMatrix(1);
  rotationMatrix.rotateInOrder(fdk::RotateOrder::ZXY, rotate);
  fdk::Vec3f vec(rotationMatrix.vecTransform(direction));
  return {vec[0], vec[1], vec[2]};
}

}
namespace usg
{

class ShadowMatrix : public PXR_NS::HdxShadowMatrixComputation
{
public:
    ShadowMatrix(bool ortho, const PXR_NS::GfVec3d& pos, const PXR_NS::GfVec3f& rot)
    {
        PXR_NS::GfFrustum frustum;
        frustum.SetProjectionType(ortho ? PXR_NS::GfFrustum::Orthographic : PXR_NS::GfFrustum::Perspective);
        frustum.SetWindow(PXR_NS::GfRange2d(PXR_NS::GfVec2d(-10, -10), PXR_NS::GfVec2d(10, 10)));
        frustum.SetPosition(pos);
        frustum.SetViewDistance(1);
        frustum.SetRotation(PXR_NS::GfRotation({0,0,1}, {rot[0], rot[1], rot[2]}));

        if (ortho) {
            frustum.SetNearFar(PXR_NS::GfRange1d(-100, 100));
        }
        else {
            frustum.SetNearFar(PXR_NS::GfRange1d(0.1, 100));
        }

        _shadowMatrix =
            frustum.ComputeViewMatrix() * frustum.ComputeProjectionMatrix();
    }

    std::vector<PXR_NS::GfMatrix4d> Compute(
        const PXR_NS::GfVec4f &viewport, PXR_NS::CameraUtilConformWindowPolicy policy) override
    {
        return std::vector<PXR_NS::GfMatrix4d>(1, _shadowMatrix);
    }

    std::vector<PXR_NS::GfMatrix4d> Compute(
        const PXR_NS::CameraUtilFraming& framing, PXR_NS::CameraUtilConformWindowPolicy policy) override
    {
        return std::vector<PXR_NS::GfMatrix4d>(1, _shadowMatrix);
    }
private:
    PXR_NS::GfMatrix4d _shadowMatrix;
};

// We don't put this in the public API as it's a workround for a bug on macOS and may disappear if we find an alternative.
USG_API void Init()
{
  PXR_NS::SdrRegistry::GetInstance();
}

//------------------------------------------------------------------------------

const std::string gEnvDebugHydra                = ((std::getenv("USG_DEBUG_HYDRA") != nullptr) ?
                                                    std::getenv("USG_DEBUG_HYDRA") : "");
const std::string gEnvDebugGlslfxShaderBuilding = ((std::getenv("USG_DEBUG_GLSLFX_SHADER_BUILDING") != nullptr) ?
                                                    std::getenv("USG_DEBUG_GLSLFX_SHADER_BUILDING") : "");
const bool gDebugHydra                = (gEnvDebugHydra == "1");
const bool gDebugGlslfxShaderBuilding = (gEnvDebugGlslfxShaderBuilding == "1");

//------------------------------------------------------------------------------


class GlslfxGenerator
{
  public:

    GlslfxGenerator(usg::ShaderDescGroup& shaderGroup, const PXR_NS::SdfPath& materialId)
      : _shaderGroup{shaderGroup}, _materialId{materialId}
    {
      std::unordered_set<const usg::ShaderDesc*> visitedShaders;
      _sortTopologically(shaderGroup.output, visitedShaders);

      fdk::Hash hash = shaderGroup.getTopologyHash();
      hash.append(_modulatesOpacity);

      // Check if we have a SdrShaderNode with this topology
      auto cachedShader = _cachedShaders.find(hash);
      if (cachedShader == _cachedShaders.end()) {
        std::stringstream shaderFunctions;
        std::stringstream mainFunction;
        std::stringstream jsonHeader;
        std::unordered_map<std::string, fdk::StringList> shaderInputs;
        std::unordered_map<std::string, fdk::StringSet> shaderTextures;

        _emitShaderFunctions(shaderFunctions, shaderInputs, shaderTextures);
        _emitMainFunction(mainFunction, shaderInputs, shaderTextures);
        _emitJsonHeader(jsonHeader);

        std::string glslfxCode = jsonHeader.str() + shaderFunctions.str() + mainFunction.str();

        // Finally register our shader code in the SdrRegistry
        PXR_NS::SdrRegistry& sdrreg = PXR_NS::SdrRegistry::GetInstance();
        PXR_NS::SdrShaderNodeConstPtr sdrNode = sdrreg.GetShaderNodeFromSourceCode(
          glslfxCode, PXR_NS::HioGlslfxTokens->glslfx, PXR_NS::NdrTokenMap());

        if (gDebugGlslfxShaderBuilding) {
          std::cout << "----------------------------------------" << std::endl;
          std::cout << "Glslfx Source" << std::endl;
          std::cout << glslfxCode << std::endl;
          std::cout << "----------------------------------------" << std::endl;
        }

        cachedShader = _cachedShaders.emplace(hash, sdrNode).first;
      }

      // Now we create a material network
      // TODO: This can also be cached based on the parameters hash 
      //       but that hasn't been implemented.
      _materialNetwork = _createHdMaterialNetwork(cachedShader->second, materialId);

      if (gDebugGlslfxShaderBuilding) {
        PXR_NS::HdMaterialNetwork& materialNetwork = _materialNetwork.map[PXR_NS::HdMaterialTerminalTokens->surface];
        std::cout << "----------------------------------------" << std::endl;
        std::cout << "HdMaterialNetwork" << std::endl;
        for (const auto& node : materialNetwork.nodes) {
          std::cout << "Node: " << node.path.GetText() << " id:" << node.identifier.GetString() << std::endl;
          for (const auto& par : node.parameters) {
            std::cout << "  par: " << par.first << "(" << par.second.GetTypeName() << ") = ";
            if (par.second.IsHolding<PXR_NS::SdfAssetPath>()) { std::cout << par.second.UncheckedGet<PXR_NS::SdfAssetPath>().GetAssetPath(); }
            else if (par.second.IsHolding<PXR_NS::TfToken>()) { std::cout << par.second.UncheckedGet<PXR_NS::TfToken>().GetText(); }
            else if (par.second.IsHolding<PXR_NS::GfVec4d>()) { std::cout << par.second.UncheckedGet<PXR_NS::GfVec4d>(); }
            else if (par.second.IsHolding<PXR_NS::GfVec4f>()) { std::cout << par.second.UncheckedGet<PXR_NS::GfVec4f>(); }
            else if (par.second.IsHolding<PXR_NS::GfVec3d>()) { std::cout << par.second.UncheckedGet<PXR_NS::GfVec3d>(); }
            else if (par.second.IsHolding<PXR_NS::GfVec3f>()) { std::cout << par.second.UncheckedGet<PXR_NS::GfVec3f>(); }
            else if (par.second.IsHolding<int>()) { std::cout << par.second.UncheckedGet<int>(); }
            else if (par.second.IsHolding<bool>()) { std::cout << par.second.UncheckedGet<bool>(); }
            std::cout << std::endl;
          }
        }
        for (const auto& conn : materialNetwork.relationships) {
          std::cout << "Rel: " << conn.inputId.GetString()  << " : " << conn.inputName.GetString() << std::endl;
          std::cout << "     " << conn.outputId.GetString() << " : " << conn.outputName.GetString() << std::endl;
        }
        for (const auto& primvar : materialNetwork.primvars) {
          std::cout << "PrimVar: " << primvar.GetString() << std::endl;
        }
        std::cout << "----------------------------------------" << std::endl;
      }
    }

    const PXR_NS::HdMaterialNetworkMap& materialNetwork() const { return _materialNetwork; }

  private:

    class GlslfxShaderSource : public usg::ShaderSource
    {
      public:
        GlslfxShaderSource(const ShaderSchema* schema, const std::string& schemaName, std::stringstream& sourceDest)
          : _shaderSchema(schema), _schemaName{schemaName}, _sourceDest{sourceDest}
        {
          const auto& shaderProperties = _shaderSchema->shaderProperties();
          for (const auto& prop : shaderProperties) {
            if (prop.isOutput()) {
              _outputs.emplace(prop.getName(), prop);
            }
            else {
              _inputs.emplace(prop.getName(), prop);
            }
          }
        }

        const usg::Token& sourceType() const override { return usg::GeomTokens.customGlslfxSourceType; }

        void startFunction() override
        {
        }

        void emitLine(const std::string& line) override
        {
          _functionBody << "  " << line << "\n";
        }

        void endFunction() override
        {
          // Emit the output struct definition
          _sourceDest << "struct " << _schemaName << "_Out {\n";
          for (const auto& output : _outputValues) {
            const std::string& name = output.first;
            if (_outputs.find(name) == _outputs.end()) {
              return;
            }
            const ShaderProperty& prop = _outputs.at(name);
            Value::Type propType = prop.getValueType();
            _sourceDest << "  " << getTypeFor(propType) << " " << name << ";\n";
          }
          emitDefaultOutputStruct("rgba", "vec4");
          emitDefaultOutputStruct("rgb", "vec3");
          emitDefaultOutputStruct("r", "float");
          emitDefaultOutputStruct("g", "float");
          emitDefaultOutputStruct("b", "float");
          emitDefaultOutputStruct("opacity", "float");
          emitDefaultOutputStruct("a", "float");
          _sourceDest << "};\n";

          // Emit the function signature
          _sourceDest << _schemaName << "_Out " << _schemaName << "(";
          std::size_t count = 0;
          // Start with the node inputs
          for (const auto& inputName : _shaderInputs) {
            if (inputName == "Peye") {
              _sourceDest << "vec4 " << inputName;
            }
            else if (inputName == "Neye") {
              _sourceDest << "vec3 " << inputName;
            }
            else {
              assert(_inputs.find(inputName) != _inputs.end());
              const ShaderProperty& prop = _inputs.at(inputName);
              const Value::Type propType = prop.getValueType();

              _sourceDest << getTypeFor(propType) << " " << inputName;
            }

            if (count != _shaderInputs.size() - 1) {
              _sourceDest << ", ";
            }
            ++count;
          }

          // Emit the texture inputs
          for (const auto& texture : _textureInputs) {
            _sourceDest << ", ";
#ifdef __APPLE__
            _sourceDest << "texture2d<float> " << texture << "_textureBind, ";
            _sourceDest << "sampler " << texture << "_samplerBind";
#else
            _sourceDest << "sampler2D " << texture << "_samplerBind";
#endif
          }
          _sourceDest << ") {\n";

          // Emit the output struct variable
          _sourceDest << "  " << _schemaName << "_Out result;\n";

          // Emit the function body
          _sourceDest << _functionBody.str();

          // Finish the function
          for (const auto& output : _outputValues) {
            _sourceDest << "  result." << output.first << " = " << output.second << ";\n";
          }

          emitDefaultOutput("rgba", "vec4(0,0,0,1)");
          emitDefaultOutput("rgb", "result.rgba.xyz");
          emitDefaultOutput("r", "result.rgba.x");
          emitDefaultOutput("g", "result.rgba.y");
          emitDefaultOutput("b", "result.rgba.z");
          emitDefaultOutput("opacity", "result.rgba.w");
          emitDefaultOutput("a", "result.rgba.w");
          
          _sourceDest << "  return result;\n";
          _sourceDest << "}\n";
        }

        std::string input(const std::string& name) override
        {
          const auto iter = _inputs.find(name);
          if (iter == _inputs.end() && name != "Peye" && name != "Neye") {
            std::cerr << "Shader input '" << name << "' requested in source code but not defined by the '" << _schemaName << "' schema." << std::endl;
            return "";
          }
          _shaderInputs.insert(name);
          return name;
        }

        void output(const std::string& name, const std::string& expression) override
        {
          if (_outputs.find(name) == _outputs.end()) {
            std::cerr << "Shader output '" << name << "' requested in source code but not defined by the '" << _schemaName << "' schema." << std::endl;
            return;
          }
          _outputValues.emplace(name, expression);
        }

        fdk::StringList inputs() const
        {
          fdk::StringList inputNames;
          inputNames.reserve(_shaderInputs.size());
          for (const auto& i : _shaderInputs) {
            inputNames.push_back(i);
          }
          return inputNames;
        }

        fdk::StringSet& textures() { return _textureInputs; }

        std::string sampleTexture(const std::string& inputName, const std::string& texCoord) override
        {
          _textureInputs.insert(inputName);
#ifdef __APPLE__
          return inputName + "_textureBind.sample(" + inputName + "_samplerBind, texCoord)";
#else
          return "texture(" + inputName + "_samplerBind, " + texCoord + ")";
#endif
        }

      private:
        std::string getTypeFor(const Value::Type type)
        {
          switch (Value::getBaseType(type)) {
            case Value::Bool:
              return "bool";
            case Value::UInt:
            case Value::UInt64:
              return "uint";
            case Value::Int:
            case Value::Int64:
              return "int";
            case Value::Int2:
              return "ivec2";
            case Value::Int3:
              return "ivec3";
            case Value::Int4:
              return "ivec4";
            case Value::Float:
              return "float";
            case Value::Float2:
              return "vec2";
            case Value::Float3:
              return "vec3";
            case Value::Float4:
              return "vec4";
            case Value::Double:
              return "double";
            case Value::Double2:
              return "dvec2";
            case Value::Double3:
              return "dvec3";
            case Value::Double4:
              return "dvec4";
            case Value::Matrix2d:
              return "mat2";
            case Value::Matrix3d:
              return "mat3";
            case Value::Matrix4d:
              return "mat4";
            case Value::Quatf:
              return "vec4";
            case Value::Quatd:
              return "dvec4";
            default:
              return Value::getTypeName(type);
          }
        }

        inline void emitDefaultOutputStruct(const std::string& outputName, const std::string& type)
        {
          if (shouldEmitDefaultOutput(outputName)) {
            _sourceDest << "  " << type << " " << outputName << ";\n";
          }
        }

        inline void emitDefaultOutput(const std::string& outputName, const std::string& source)
        {
          if (shouldEmitDefaultOutput(outputName)) {
            _sourceDest << "  result." << outputName << " = " << source << ";\n";
          }
        }

        inline bool shouldEmitDefaultOutput(const std::string& outputName)
        {
          // The schema defines the output but the shader code didn't write to it.
          return _outputs.find(outputName) != _outputs.end() && _outputValues.find(outputName) == _outputValues.end();
        }

        const ShaderSchema* _shaderSchema;
        std::string _schemaName;
        std::stringstream& _sourceDest;
        std::stringstream _functionBody;
        std::unordered_map<std::string, ShaderProperty> _outputs;
        std::unordered_map<std::string, ShaderProperty> _inputs;
        fdk::StringSet _textureInputs;
        fdk::StringSet _shaderInputs;
        fdk::KeyValueMap _outputValues;
    };

    static inline bool IsHdStormNode(const usg::ShaderDesc* shaderDesc)
    {
      const std::string& schemaName = shaderDesc->schemaName();
      return schemaName == "UsdUVTexture"
          || schemaName == "UsdPrimvarReader_float2"
          || schemaName == "UsdPreviewSurface"
          || schemaName == "UsdTransform2d";
    }

    inline std::string CompoundInputName(const usg::ShaderDesc* shaderDesc, const std::string& inputName)
    {
      return _renamedShaders.at(shaderDesc) + "_" + inputName;
    }

    static inline void JsonParameter(std::stringstream& out, const std::string& name, const std::string& value)
    {
      out << "    \"" << name << "\": {\n";
      out << "        \"default\": " << value << "\n";
      out << "    }";
    }

    static void JsonValue(std::stringstream& out, const std::string& name, const usg::Value& value)
    {
      const usg::Value::Type propType = usg::Value::getBaseType(value.getType());
      std::stringstream defaultValueStr;
      switch (propType) {
        case Value::Bool:
          JsonParameter(out, name, value.get<bool>() ? "true" : "false");
          break;
        case Value::UInt:
          JsonParameter(out, name, std::to_string(value.get<uint32_t>()));
          break;
        case Value::UInt64:
          JsonParameter(out, name, std::to_string(value.get<uint64_t>()));
          break;
        case Value::Int:
          JsonParameter(out, name, std::to_string(value.get<int32_t>()));
          break;
        case Value::Int64:
          JsonParameter(out, name, std::to_string(value.get<int64_t>()));
          break;
        case Value::Int2:
        {
          fdk::Vec2i v = value.get<fdk::Vec2i>();
          defaultValueStr << std::fixed << "[" << v.x << ", " << v.y << "]";
          JsonParameter(out, name, defaultValueStr.str());
          break;
        }
        case Value::Int3:
        {
          fdk::Vec3i v = value.get<fdk::Vec3i>();
          defaultValueStr << std::fixed << "[" << v.x << ", " << v.y << ", " << v.z << "]";
          JsonParameter(out, name, defaultValueStr.str());
          break;
        }
        case Value::Int4:
        {
          fdk::Vec4i v = value.get<fdk::Vec4i>();
          defaultValueStr << std::fixed << "[" << v.x << ", " << v.y << ", " << v.z << ", " << v.w << "]";
          JsonParameter(out, name, defaultValueStr.str());
          break;
        }
        case Value::Float:
          JsonParameter(out, name, std::to_string(value.get<float>()));
          break;
        case Value::Float2:
        {
          fdk::Vec2f v = value.get<fdk::Vec2f>();
          defaultValueStr << std::fixed << "[" << v.x << ", " << v.y << "]";
          JsonParameter(out, name, defaultValueStr.str());
          break;
        }
        case Value::Float3:
        {
          fdk::Vec3f v = value.get<fdk::Vec3f>();
          defaultValueStr << std::fixed << "[" << v.x << ", " << v.y << ", " << v.z << "]";
          JsonParameter(out, name, defaultValueStr.str());
          break;
        }
        case Value::Float4:
        {
          fdk::Vec4f v = value.get<fdk::Vec4f>();
          defaultValueStr <<  std::fixed << "[" << v.x << ", " << v.y << ", " << v.z << ", " << v.w << "]";
          JsonParameter(out, name, defaultValueStr.str());
          break;
        }
        case Value::Double:
          JsonParameter(out, name, std::to_string(value.get<double>()));
          break;
        case Value::Double2:
        {
          fdk::Vec2d v = value.get<fdk::Vec2d>();
          defaultValueStr <<  std::fixed << "[" << v.x << ", " << v.y << "]";
          JsonParameter(out, name, defaultValueStr.str());
          break;
        }
        case Value::Double3:
        {
          fdk::Vec3d v = value.get<fdk::Vec3d>();
          defaultValueStr <<  std::fixed << "[" << v.x << ", " << v.y << ", "<< v.z << "]";
          JsonParameter(out, name, defaultValueStr.str());
          break;
        }
        case Value::Double4:
        {
          fdk::Vec4d v = value.get<fdk::Vec4d>();
          defaultValueStr <<  std::fixed << "[" << v.x << ", " << v.y << ", "<< v.z << ", " <<v.w << "]";
          JsonParameter(out, name, defaultValueStr.str());
          break;
        }
        case Value::Matrix4d:
        {
          const auto& matrix = value.get<fdk::Mat4d>();
          JsonValue(out, name + "_col1", Value(fdk::Vec4f(matrix.col0())));
          out << ",\n";
          JsonValue(out, name + "_col2", Value(fdk::Vec4f(matrix.col1())));
          out << ",\n";
          JsonValue(out, name + "_col3", Value(fdk::Vec4f(matrix.col2())));
          out << ",\n";
          JsonValue(out, name + "_col4", Value(fdk::Vec4f(matrix.col3())));
          break;
        }
        case Value::Quatf:
        {
          fdk::Quatf q = value.get<fdk::Quatf>();
          defaultValueStr <<  std::fixed << "[" << q.v.x << ", " << q.v.y << ", " << q.v.z << ", " << q.s << "]";
          JsonParameter(out, name, defaultValueStr.str());
          break;
        }
        case Value::Quatd:
        {
          fdk::Quatd q = value.get<fdk::Quatd>();
          defaultValueStr <<  std::fixed << "[" << q.v.x << ", " << q.v.y << ", " << q.v.z << ", " << q.s << "]";
          JsonParameter(out, name, defaultValueStr.str());
          break;
        }

        default:
#ifdef NDEBUG
          const std::string msg = "Unhandled type " + usg::Value::getTypeName(propType);
          assert(false && msg.c_str());
#else
          break;
#endif
      }
    }

    static bool CanHandleValue(const usg::Value& value) {
      switch (usg::Value::getBaseType(value.getType())) {
        case Value::Bool:
        case Value::UInt:
        case Value::UInt64:
        case Value::Int:
        case Value::Int64:
        case Value::Int2:
        case Value::Int3:
        case Value::Int4:
        case Value::Float:
        case Value::Float2:
        case Value::Float3:
        case Value::Float4:
        case Value::Double:
        case Value::Double2:
        case Value::Double3:
        case Value::Double4:
        case Value::Matrix4d:
        case Value::Quatf:
        case Value::Quatd:
          return true;
        default:
          return false;
      }
    }

    static bool ModulatesOpacity(const usg::ShaderDesc* shaderDesc, const usg::ShaderProperty& property)
    {
      const fdk::KeyValueMap& hints = property.getHints();
      const auto opacityIt = hints.find("opacity");
      const bool isTranslucent = opacityIt != hints.end() &&
                                 opacityIt->second == "1";
      if (isTranslucent) {
        const std::string propertyName = property.getName();
        const usg::ShaderConnectionArray* connections = shaderDesc->getInputConnection(propertyName);
        const usg::ShaderProperty* over = shaderDesc->getInputOverride(propertyName);
        // If the input has a connection or is overriden twe need to assume
        // the opacity is modulated.
        return connections || (over && over->isOverriden());
      }
      return false;
    }

    static bool IsMatrix4(const usg::ShaderProperty& property)
    {
      return property.getValue().getType() == usg::Value::Matrix4d;
    }

    void _processInputs(const usg::ShaderDesc* shaderDesc, const usg::ShaderSchema* schema)
    {
      for (const auto& property : schema->shaderProperties()) {
        if (!property.isOutput()) {
          _modulatesOpacity |= ModulatesOpacity(shaderDesc, property);
          if (IsMatrix4(property)) {
            _matrixInputs.emplace(CompoundInputName(shaderDesc, property.getName()), property);
          }
        }
      }
    }

    void _sortTopologically(const usg::ShaderDesc* shaderDesc, std::unordered_set<const usg::ShaderDesc*>& visitedShaders)
    {
      if (visitedShaders.find(shaderDesc) != visitedShaders.end()) {
        return;
      }

      // Traverse the graph upstream
      for (const auto& i : shaderDesc->getInputConnections()) {
        const usg::ShaderConnectionArray* connections = i.second;
        if (!connections) {
          continue;
        }
        for (const usg::ShaderConnection& c : *connections) {
          const usg::ShaderDesc* upstreamNode = _shaderGroup.getShaderNode(c.upstreamNodePath.asString());
          _sortTopologically(upstreamNode, visitedShaders);
        }
      }

      // Don't visit the same shader twice
      visitedShaders.insert(shaderDesc);

      _topologicalOrder.push_back(shaderDesc);
      const auto& schemaName = shaderDesc->schemaName();
      const usg::SchemaDescription* desc = usg::SchemaDescription::find(shaderDesc->schemaName().c_str());
      if (desc) {
        usg::Schema* schema = desc->builder_method();
        if (auto shaderSchema = schema->asShaderSchema()) {
          _shaderSchemas.emplace(schemaName, shaderSchema);
          // Rename the shader nodes so the shader can be reused for networks
          // with the same schema topology but different Nuke nodes.
          _renamedShaders.emplace(shaderDesc, schemaName + "_" + std::to_string(_renamedShaders.size()));

          _shaderBySchemas[schemaName].push_back(shaderDesc);
          _processInputs(shaderDesc, shaderSchema);
        }
      }
    }

    void _emitShaderFunctions(std::stringstream& shaderFunctions, std::unordered_map<std::string, fdk::StringList>& shaderInputs, std::unordered_map<std::string, fdk::StringSet>& shaderTextures)
    {
      for (const auto& schema : _shaderSchemas) {
        const auto& schemaName = schema.first;
        const auto& shaderSchema = schema.second;

        GlslfxShaderSource shaderSource(shaderSchema, schemaName, shaderFunctions);
        shaderSchema->sourceCode(shaderSource);
        shaderInputs[schemaName] = shaderSource.inputs();
        shaderTextures[schemaName] = shaderSource.textures();
      }
    }

    void _emitMainFunction(std::stringstream& mainFunction, std::unordered_map<std::string, fdk::StringList>& shaderInputs, std::unordered_map<std::string, fdk::StringSet>& shaderTextures)
    {
      // Emit matrix build functions
      for (const auto& pair : _matrixInputs) {
        const std::string name = std::get<0>(pair);
        const usg::ShaderProperty& property = std::get<1>(pair);

        mainFunction << "mat4 UsgGet_" << name << "() {\n";
        mainFunction << "  return transpose(mat4(HdGet_" << name << "_col1(),\n";
        mainFunction << "              HdGet_" << name << "_col2(),\n";
        mainFunction << "              HdGet_" << name << "_col3(),\n";
        mainFunction << "              HdGet_" << name << "_col4()));\n";
        mainFunction << "}\n";
      }

      mainFunction << "vec4 surfaceShader(vec4 Peye, vec3 Neye, vec4 color, vec4 patchCoord) {\n";
      for (const auto& shaderDesc : _topologicalOrder) {
        const std::string& schemaName = shaderDesc->schemaName();
        const std::string& nodeName = _renamedShaders[shaderDesc];

        // Skip nodes that are handled by HdStorm and do not contribute to our
        // glslfx code
        if (IsHdStormNode(shaderDesc)) {
          continue;
        }

        const auto& inputs = shaderInputs[schemaName];
        const auto& textures = shaderTextures[schemaName];

        // Emit a line that calls the schema shader function with the right
        // inputs and stores the results in a struct.
        //   <SchemaName>_Out <RenamedShaderName> = <SchemaFunction>(<input1>, ..., <inputN>);
        // the inputs can either be fetched from the results of a previous call
        // to a schema function or to a 'HdGet_*' function that grabs the value
        // of a parameter as defined in the json header.
        mainFunction << "  " << schemaName << "_Out " << nodeName << " = " << schemaName << "(";
        for (std::size_t i = 0; i < inputs.size(); ++i) {
          const std::string& input = inputs[i];

          if (input == "Peye" || input == "Neye") {
            mainFunction << input;
          }
          else {
            const usg::ShaderProperty& property = shaderDesc->getInput(input);
            const usg::ShaderConnectionArray* connections = shaderDesc->getInputConnection(input);
            usg::ShaderConnection connection;
            usg::ShaderDesc* inputNode = nullptr;

            if (connections && !connections->empty()) {
              connection = connections->front();
              inputNode = _shaderGroup.getShaderNode(connection.upstreamNodePath.asString());
            }

            if (inputNode && !IsHdStormNode(inputNode)) {
              const std::string fullOutputName = connection.upstreamOutputName;
              const std::string outputName = fullOutputName.substr(fullOutputName.find(":") + 1);
              mainFunction << _renamedShaders[inputNode] << "." << outputName;
            }
            else if (IsMatrix4(property)) {
              mainFunction << "UsgGet_" << CompoundInputName(shaderDesc, input) << "()";
            }
            else {
              mainFunction << "HdGet_" << CompoundInputName(shaderDesc, input) << "()";
            }
          }

          if (i != inputs.size() - 1) {
            mainFunction << ", ";
          }
        }

        // Texture inputs
        for (const auto& textureInput : textures) {
          const auto inputName = CompoundInputName(shaderDesc, textureInput);
          mainFunction << ", ";
#ifdef __APPLE__
          mainFunction << "textureBind_" << inputName << ", samplerBind_" << inputName;
#else
          mainFunction << "sampler2d_" << inputName;
#endif
        }
        mainFunction << ");\n";
      }

      // Return last shader's surface output.
      const usg::ShaderDesc* terminalShader = _topologicalOrder.back();
      if (terminalShader->hasOutput("surface")) {
        mainFunction << "  vec4 surfaceColor = " << _renamedShaders[terminalShader] << ".rgba;\n";
      }
      else {
        mainFunction << "  vec4 surfaceColor = vec4(1,1,0,1);";
      }

      mainFunction << "  return surfaceColor;\n";

      mainFunction << "}";
    }

    void _emitJsonHeader(std::stringstream& header)
    {
      std::vector<std::pair<std::string, usg::ShaderProperty>> inputs;
      for (const auto& shader : _topologicalOrder) {
        const usg::ShaderSchema* schema = _shaderSchemas[shader->schemaName()];
        for (const auto& property : schema->shaderProperties()) {
          if (!property.isOutput() && CanHandleValue(property.getValue())) {
            const auto& name = CompoundInputName(shader, property.getName());
            inputs.emplace_back(name, property);
          }
        }
      }

      header << "-- glslfx version 0.1\n";
      header << "\n";
      header << "-- configuration\n";
      header << "{\n";

      if (_modulatesOpacity) {
        header << "  \"metadata\": {\n";
        header << "    \"materialTag\": \"translucent\"\n";
        header << "  },\n";
      }

      header << "  \"parameters\": {\n";

      for (std::size_t i = 0; i < inputs.size(); ++i) {
        const auto& pair = inputs[i];
        const auto& input = std::get<1>(pair);
        const std::string& name = std::get<0>(pair);
        const bool isLastInput = (i == inputs.size() - 1);

        JsonValue(header, name, input.getValue());
        header << (isLastInput ? "\n" : ",\n");
      }

      header << "  },\n";
      header << "  \"techniques\": {\n";
      header << "    \"default\": {\n";
      header << "      \"surfaceShader\": {\n";
      header << "        \"source\": [ \"Surface\" ]\n";
      header << "      }\n";
      header << "    }\n";
      header << "  }\n";
      header << "}\n";
      header << "\n";
      header << "-- glsl Surface\n\n";
    }

    PXR_NS::HdMaterialNetworkMap _createHdMaterialNetwork(const PXR_NS::SdrShaderNodeConstPtr& sdrNode, const PXR_NS::SdfPath& materialId)
    {
      PXR_NS::HdMaterialNetworkMap material;
      PXR_NS::HdMaterialNetwork& materialNetwork = material.map[PXR_NS::HdMaterialTerminalTokens->surface];
      PXR_NS::HdMaterialNode terminalNode;
      terminalNode.identifier = sdrNode->GetIdentifier();
      terminalNode.path = materialId.AppendPath(PXR_NS::SdfPath(_shaderGroup.output->getMaterialRelativeName()));

      for (const usg::ShaderDesc* shader : _topologicalOrder) {
        PXR_NS::HdMaterialNode* parameterNode = &terminalNode;

        // We need to create an HdStorm specific node for these cases since
        // they are not included in our glslfx code
        if (IsHdStormNode(shader)) {
          // create an HdMaterialNode
          PXR_NS::HdMaterialNode materialNode;
          materialNode.identifier = PXR_NS::TfToken(shader->schemaName());
          materialNode.path = materialId.AppendPath(PXR_NS::SdfPath(shader->getMaterialRelativeName()));
          materialNetwork.nodes.push_back(materialNode);

          // The parameters are set in this node not in the terminal node
          parameterNode = &materialNetwork.nodes.back();

          const auto& schema = _shaderSchemas[shader->schemaName()];
          fdk::StringPtrList inputNames;
          shader->getInputNames(inputNames);
          for (const auto& inputName : inputNames) {
            const auto& input = shader->getInput(*inputName);
            const PXR_NS::VtValue value = input.getValue().impl()->vtvalue;
            parameterNode->parameters[PXR_NS::TfToken(*inputName)] = value;
          }
        }

        // Set the parameters in the respective node
        for (const auto &input : shader->getInputOverrides()) {
          std::string inputName;
          if (parameterNode == &terminalNode) {
            inputName = CompoundInputName(shader, input.first);
          }
          else {
            inputName = input.first;
          }

          const PXR_NS::VtValue value = input.second->getValue().impl()->vtvalue;

          if (IsMatrix4(*input.second)) {
            PXR_NS::GfMatrix4d matrix(1.0);
            if (value.IsHolding<fdk::Mat4d>()) {
              matrix = reinterpret_cast<const PXR_NS::GfMatrix4d&>(value.UncheckedGet<fdk::Mat4d>());
            }
            else if (value.IsHolding<PXR_NS::GfMatrix4d>()) {
              matrix = value.UncheckedGet<PXR_NS::GfMatrix4d>();
            }
            parameterNode->parameters[PXR_NS::TfToken(inputName + "_col1")] = PXR_NS::VtValue(PXR_NS::GfVec4f(matrix.GetColumn(0)));
            parameterNode->parameters[PXR_NS::TfToken(inputName + "_col2")] = PXR_NS::VtValue(PXR_NS::GfVec4f(matrix.GetColumn(1)));
            parameterNode->parameters[PXR_NS::TfToken(inputName + "_col3")] = PXR_NS::VtValue(PXR_NS::GfVec4f(matrix.GetColumn(2)));
            parameterNode->parameters[PXR_NS::TfToken(inputName + "_col4")] = PXR_NS::VtValue(PXR_NS::GfVec4f(matrix.GetColumn(3)));
          }
          else {
            parameterNode->parameters[PXR_NS::TfToken(inputName)] = value;
          }
        }

        // Deal with creating the relationships (connections) between nodes
        for (const auto& input : shader->getInputConnections()) {
          const usg::ShaderConnectionArray* connections = input.second;
          if (!connections && connections->size() != 1) {
            continue;
          }

          const usg::ShaderConnection connection = connections->front();
          const usg::ShaderDesc* upstreamNode = _shaderGroup.getShaderNode(connection.upstreamNodePath.asString());
          if (upstreamNode && IsHdStormNode(upstreamNode)) {
            // create a relationship
            const std::string fullOutputName = connection.upstreamOutputName;
            const std::string outputName = fullOutputName.substr(fullOutputName.find(":") + 1);
            PXR_NS::HdMaterialRelationship relationship;
            relationship.inputId = materialId.AppendPath(PXR_NS::SdfPath(upstreamNode->getMaterialRelativeName()));
            relationship.inputName = PXR_NS::TfToken(outputName);
            relationship.outputId = parameterNode->path;
            if (parameterNode == &terminalNode) {
              relationship.outputName = PXR_NS::TfToken(CompoundInputName(shader, input.first));
            }
            else {
              relationship.outputName = PXR_NS::TfToken(input.first);
            }
            materialNetwork.relationships.push_back(relationship);
          }
        }
      }

      // The terminal node needs to be added last or HdStorm will get confused.
      materialNetwork.nodes.push_back(terminalNode);
      material.terminals.push_back(terminalNode.path);
      return material;
    }

    static std::unordered_map<fdk::Hash, PXR_NS::SdrShaderNodeConstPtr> _cachedShaders;

    const usg::ShaderDescGroup& _shaderGroup;
    const PXR_NS::SdfPath& _materialId;
    std::vector<const usg::ShaderDesc*> _topologicalOrder;
    std::unordered_map<const usg::ShaderDesc*, std::string> _renamedShaders;

    std::unordered_map<std::string, const usg::ShaderSchema*> _shaderSchemas;
    std::unordered_map<std::string, std::vector<const usg::ShaderDesc*>> _shaderBySchemas;

    PXR_NS::HdMaterialNetworkMap _materialNetwork;

    bool _modulatesOpacity{false};
    std::unordered_map<std::string, const usg::ShaderProperty> _matrixInputs;
};
/*static*/ std::unordered_map<fdk::Hash, PXR_NS::SdrShaderNodeConstPtr> GlslfxGenerator::_cachedShaders;

//--------------------------------------------------------------------------------------------------
class UsgSceneDelegate::Impl
{
public:
  Impl(UsgSceneDelegate* delegate)
    : delegate(delegate)
    , renderIndex(delegate->GetRenderIndex())
    , changeTracker(renderIndex.GetChangeTracker())
  {
    shadowCollection.SetName(PXR_NS::HdLightTokens->shadowCollection);
    shadowCollection.SetRootPath(PXR_NS::SdfPath("/"));
    changeTracker.AddCollection(PXR_NS::HdLightTokens->shadowCollection);
  }

  // Returns a Prim from the delegate UsgStage. It's assumed that sdfPath is not namespaced with the delegate id.
  usg::Prim getStagePrim(const PXR_NS::SdfPath& sdfPath) {
    return delegate->_stage->getPrimAtPath(reinterpret_cast<const Path::Handle&>(sdfPath));
  }

  // Returns a LightPrim from the delegate UsgStage. It's assumed that sdfPath is namespaced with the delegate id.
  usg::LightPrim getShadowPrim(const PXR_NS::SdfPath& sdfPath) {
    if (sdfPath.GetName() != kShadowsPath) {
      return usg::LightPrim();
    }
    // sdfPath is the path inserted in populateShadowWorkaround(): "{DelegateId}/{light.GetPath()}/{kShadowsPath}"
    auto path = sdfPath.GetParentPath() // Strip {kShadowsPath}
        .MakeRelativePath(delegate->GetDelegateID()) // Strip {DelegateId}
        .MakeAbsolutePath(PXR_NS::SdfPath("/")); // -> /{light path}
    return delegate->_stage->getPrimAtPath(reinterpret_cast<const Path::Handle&>(path));
  }


  void populateShadowWorkaround(const PXR_NS::SdfPath lightPath)
  {
    if (!usg::lux::ShadowAPI::isValidPrim(getStagePrim(lightPath))) {
      return;
    }

    // "{DelegateId}/{light path}/{kShadowsPath}"
    auto path = PXR_NS::SdfPath(kShadowsPath)
        .MakeAbsolutePath(lightPath)
        .MakeRelativePath(PXR_NS::SdfPath("/"))
        .MakeAbsolutePath(delegate->GetDelegateID());
    renderIndex.InsertSprim(PXR_NS::HdPrimTypeTokens->simpleLight, delegate, path);
    changeTracker.MarkSprimDirty(path, PXR_NS::HdChangeTracker::AllDirty);
  }

  UsgSceneDelegate* delegate;
  PXR_NS::HdRenderIndex& renderIndex;
  PXR_NS::HdChangeTracker& changeTracker;
  PXR_NS::HdRprimCollection shadowCollection;
};


UsgSceneDelegate::UsgSceneDelegate(PXR_NS::HdRenderIndex* renderIndex, const PXR_NS::SdfPath& delegateId)
  : PXR_NS::UsdImagingDelegate(renderIndex, delegateId)
  , _pImpl(std::make_unique<Impl>(this))
{
}

UsgSceneDelegate::~UsgSceneDelegate() = default;

void UsgSceneDelegate::populate(usg::StageRef stage)
{
  _stage = stage;
  auto wibble = stage->getUsdStageRefPtr(usg::usdAPIVersion());
  PXR_NS::UsdStageRefPtr pxrStage = *reinterpret_cast<PXR_NS::UsdStageRefPtr*>(wibble);
  if (gDebugHydra) {
    std::cout << "    UsgSceneDelegate::populate()..." << std::endl;
  }
  UsdImagingDelegate::Populate(pxrStage->GetPseudoRoot());
  if (gDebugHydra) {
    std::cout << "      ...populate() done" << std::endl;
  }

#ifdef __APPLE__
  const auto defaultMaterialId = GetDelegateID().AppendPath(PXR_NS::SdfPath("materials/default"));
  auto& renderIndex = GetRenderIndex();
  auto& changeTracker = renderIndex.GetChangeTracker();
  renderIndex.InsertSprim(PXR_NS::HdPrimTypeTokens->material, this, defaultMaterialId);
  changeTracker.MarkSprimDirty(defaultMaterialId, PXR_NS::HdChangeTracker::AllDirty);
#endif
}

PXR_NS::VtValue UsgSceneDelegate::GetLightParamValue(PXR_NS::SdfPath const& id, PXR_NS::TfToken const& key)
{
  auto path = id
      .MakeRelativePath(_pImpl->delegate->GetDelegateID()) // Strip {DelegateId}
      .MakeAbsolutePath(PXR_NS::SdfPath("/")); // -> /{light path}
  const usg::LightPrim& light = _pImpl->getStagePrim(path);

  // The shadows workaround adds a SimpleLight, which duplicates the intensity of
  // non shadowed areas and casting lights over the areas where we should have shadows.
  if (light.hasApi(usg::lux::LuxTokens.types.ShadowAPI) && key == PXR_NS::HdLightTokens->intensity) {
    return PXR_NS::VtValue(0.0f);
  }
  return PXR_NS::UsdImagingDelegate::GetLightParamValue(id, key);
}

// We have to use Get for params, shadowParams, shadowCollection
PXR_NS::VtValue UsgSceneDelegate::Get(PXR_NS::SdfPath const& id, PXR_NS::TfToken const& key)
{
  const usg::LightPrim& light = _pImpl->getShadowPrim(id);
  if (!light.isValid()) {
    return PXR_NS::UsdImagingDelegate::Get(id, key);
  }
  if (key == PXR_NS::HdLightTokens->params) {
    PXR_NS::GlfSimpleLight simpleLight;
    auto pixel = light.getColor() * light.getIntensity();
    PXR_NS::GfVec4f diffuse(pixel[0], pixel[1], pixel[2], 1.0f);
    PXR_NS::GfVec4f specular(pixel[0], pixel[1], pixel[2], 1.0f);
    simpleLight.SetDiffuse(light.getDiffuse() * diffuse);
    simpleLight.SetSpecular(light.getSpecular() * specular);
    simpleLight.SetIsCameraSpaceLight(false);

    fdk::Mat4f matrix(light.getConcatenatedXform());
    simpleLight.SetPosition({matrix.a03, matrix.a13, matrix.a23, matrix.a33});

    if (light.hasApi(usg::lux::LuxTokens.types.ShadowAPI)) {
      // We should be using
      //
      //    usg::lux::ShadowAPI shadow(light);
      //    simpleLight.SetHasShadow(shadow.getShadowEnable());
      //
      // But we seem to have a bug there. Getting the attributes directly from
      // Prim works fine.
      int32_t enabled;
      light.getAttr(usg::lux::LuxTokens.inputsShadowEnable).getValue<int32_t>(enabled);
      simpleLight.SetHasShadow(enabled);
    } else {
      simpleLight.SetHasShadow(false);
    }

    const auto lightType = light.getTypeName();
    if (lightType == usg::lux::LuxTokens.types.DistantLight) {
      PXR_NS::GfVec3f direction = ConvertLightDirection(light, {0, 0, 1});

      simpleLight.SetPosition({direction[0], direction[1], direction[2], 0});
    } else if (lightType == usg::lux::LuxTokens.types.DiskLight) {
      usg::lux::ShapingAPI shape(light);
      simpleLight.SetSpotCutoff(shape.getShapingConeAngle());
      simpleLight.SetSpotFalloff(0.0f);
      PXR_NS::GfVec3f direction = ConvertLightDirection(light, {0, 0, -1});
      simpleLight.SetSpotDirection(direction);
    }

    simpleLight.SetAttenuation({1.0f, 0.0f, 0.0f});

    return  PXR_NS::VtValue(simpleLight);
  }
  else if (key == PXR_NS::HdLightTokens->shadowCollection) {
    return PXR_NS::VtValue{_pImpl->shadowCollection};
  } else  if (key == PXR_NS::HdLightTokens->shadowParams) {
    fdk::Mat4f matrix(light.getConcatenatedXform());
    PXR_NS::GfVec3f lightPosition(PXR_NS::GfVec3f(matrix.a03, matrix.a13, matrix.a23));

    PXR_NS::HdxShadowParams shadowParams;
    if (light.hasApi(usg::lux::LuxTokens.types.ShadowAPI)) {
      usg::lux::ShadowAPI shadow(light);
      int32_t enabled;
      light.getAttr(usg::lux::LuxTokens.inputsShadowEnable).getValue<int32_t>(enabled);
      shadowParams.enabled = enabled;
    } else {
      shadowParams.enabled = false;
    }
    auto direction = ConvertLightDirection(light, {0,0,1});
    const bool ortho = light.isA(usg::lux::LuxTokens.types.DistantLight);
    auto shadowMatrix = new ShadowMatrix(ortho, lightPosition, direction);
    shadowParams.shadowMatrix = PXR_NS::HdxShadowMatrixComputationSharedPtr(shadowMatrix);

    shadowParams.resolution = 1024;
    shadowParams.bias = -0.001f;
    shadowParams.blur =  0.001f;
    return PXR_NS::VtValue(shadowParams);
  }

  return PXR_NS::UsdImagingDelegate::Get(id, key);
}

PXR_NS::SdfPath UsgSceneDelegate::GetMaterialId(const PXR_NS::SdfPath &rprim)
{
  auto materialId = PXR_NS::UsdImagingDelegate::GetMaterialId(rprim);

#ifdef __APPLE__
  // Work around HdStorm bug on Metal with AMD cards
  if (materialId.IsEmpty()) {
    materialId = GetDelegateID().AppendPath(PXR_NS::SdfPath("materials/default"));
  }
#endif

  return materialId;
}

PXR_NS::VtValue UsgSceneDelegate::GetMaterialResource(const PXR_NS::SdfPath &materialId)
{
  auto pxrNewPath = materialId.MakeRelativePath(GetDelegateID()).MakeAbsolutePath(PXR_NS::SdfPath("/"));
  usg::Path newPath(reinterpret_cast<const usg::Path::Handle&>(pxrNewPath));
  usg::MaterialPrim material(_stage->getPrimAtPath(newPath));


  // Handle render targets we need to convert to custom glslfx code:
  if (material.isValid() && material.hasCustomData("nuke_material_op")) {
    PXR_NS::VtValue resource;

    // glslfx render target:
    usg::ShaderDescGroup shaderDescGroup;
    material.getShaderDescGroup(usg::GeomTokens.universalRenderContext, "surface", GetTime().GetValue(), shaderDescGroup);
    if (shaderDescGroup.output) {
      if (shaderDescGroup.output->schemaName() == "UsdPreviewSurface") {
        // Special case for UsdPreviewSurface since we don't need to do any
        // conversions to glslfx
        resource = UsdImagingDelegate::GetMaterialResource(materialId);
      }
      else {
        GlslfxGenerator glslfxGen(shaderDescGroup, materialId);
        PXR_NS::HdMaterialNetworkMap hdMaterialNetwork = glslfxGen.materialNetwork();
        shaderDescGroup.deleteShaders();
        resource = PXR_NS::VtValue::Take(hdMaterialNetwork);
      }
    }

    // - Custom handle other targets here -

    if (resource.IsHolding<PXR_NS::HdMaterialNetworkMap>()) {
      auto hdMaterialNetwork = resource.UncheckedGet<PXR_NS::HdMaterialNetworkMap>();
      for (const auto& node : hdMaterialNetwork.map[PXR_NS::HdMaterialTerminalTokens->surface].nodes) {
        if (node.identifier == "UsdUVTexture") {
          static const PXR_NS::TfToken fileToken("file");
          auto file = node.parameters.find(fileToken);
          if (file != node.parameters.end()) {
            const auto assetPath = file->second.Get<PXR_NS::SdfAssetPath>().GetAssetPath();
            _texturesUsed[assetPath].insert(fromSdfPath(materialId));
          }
        }
      }
    }

    return resource;
  }
#ifdef __APPLE__
  // Work around HdStorm bug on Metal with AMD cards
  else if (materialId == GetDelegateID().AppendPath(PXR_NS::SdfPath("materials/default"))) {
    static PXR_NS::SdrShaderNodeConstPtr _sdrSurfaceNode = nullptr;
    static constexpr const char* sShaderSource =
R"(-- glslfx version 0.1

-- configuration
{
  "metadata": {
    "materialTag": "translucent"
  },
  "parameters": {
  },
  "techniques" : {
    "default" : {
      "surfaceShader" : {
        "source": [ "Default.SurfaceShader" ]
      }
    }
  }
}

-- glsl Default.SurfaceShader

vec4 surfaceShader(vec4 Peye, vec3 Neye, vec4 color, vec4 patchCoord)
{
  color.rgb = FallbackLighting(Peye.xyz, Neye, color.rgb);
  return color;
}
)";
    if (!_sdrSurfaceNode) {
      PXR_NS::SdrRegistry &shaderReg = PXR_NS::SdrRegistry::GetInstance();
      _sdrSurfaceNode =
          shaderReg.GetShaderNodeFromSourceCode(sShaderSource, PXR_NS::HioGlslfxTokens->glslfx, PXR_NS::NdrTokenMap());
    }
    PXR_NS::HdMaterialNetworkMap material;
    PXR_NS::HdMaterialNetwork& materialNetwork = material.map[PXR_NS::HdMaterialTerminalTokens->surface];
    PXR_NS::HdMaterialNode node;
    node.identifier = _sdrSurfaceNode->GetIdentifier();
    node.path = materialId;

    materialNetwork.nodes.push_back(node);
    material.terminals.push_back(node.path);

    return PXR_NS::VtValue::Take(material);
  }
#endif

  return UsdImagingDelegate::GetMaterialResource(materialId);
}

PXR_NS::HdReprSelector UsgSceneDelegate::GetReprSelector(PXR_NS::SdfPath const &id)
{
  auto pxrNewPath = id.MakeRelativePath(GetDelegateID()).MakeAbsolutePath(PXR_NS::SdfPath("/"));
  usg::Path newPath(reinterpret_cast<const usg::Path::Handle&>(pxrNewPath));

  if (_selected.find(newPath) != _selected.end()) {
    return PXR_NS::HdReprSelector(PXR_NS::HdReprTokens->refinedWireOnSurf, PXR_NS::HdReprTokens->refinedWire);
  }
  return PXR_NS::UsdImagingDelegate::GetReprSelector(id);
}

void UsgSceneDelegate::updateSelection(const usg::PathSet& selected, const usg::PathSet& deselected)
{
  _selected = selected;
  _deselected = deselected;
  const auto& delegateId = GetDelegateID();
  auto& renderIndex = GetRenderIndex();

  const auto dirtyRprim = [&delegateId, &renderIndex](const usg::Path& p) {
    auto pxrNewPath = toSdfPath(p).MakeRelativePath(PXR_NS::SdfPath("/")).MakeAbsolutePath(delegateId);
    renderIndex.GetChangeTracker().MarkRprimDirty(pxrNewPath, PXR_NS::HdChangeTracker::DirtyRepr);
  };

  std::for_each(_selected.begin(), _selected.end(), dirtyRprim);
  std::for_each(_deselected.begin(), _deselected.end(), dirtyRprim);
}

const usg::PathSet& UsgSceneDelegate::materialsUsingTexture(const std::string& texture)
{
  return _texturesUsed[texture];
}

void UsgSceneDelegate::clearMaterialUsingTexture(const std::string& texture)
{
  _texturesUsed.erase(texture);
}

//--------------------------------------------------------------------------------------------------


/*static*/ void
StormRenderInterface::getGLBufferFormat(const DataFormat& inFormat,
                                        uint32_t*         outGLFormat,
                                        uint32_t*         outGLType,
                                        uint32_t*         outGLInternalFormat)
{
  switch (static_cast<PXR_NS::HdFormat>(inFormat)) {
  case PXR_NS::HdFormatUNorm8: *outGLFormat = GL_RED; *outGLType = GL_UNSIGNED_BYTE; *outGLInternalFormat = GL_R8; break;
  case PXR_NS::HdFormatUNorm8Vec2: *outGLFormat = GL_RG; *outGLType = GL_UNSIGNED_BYTE; *outGLInternalFormat = GL_RG8; break;
  case PXR_NS::HdFormatUNorm8Vec4: *outGLFormat = GL_RGBA; *outGLType = GL_UNSIGNED_BYTE; *outGLInternalFormat = GL_RGBA8; break;
  case PXR_NS::HdFormatSNorm8: *outGLFormat = GL_RED; *outGLType = GL_BYTE; *outGLInternalFormat = GL_R8_SNORM; break;
  case PXR_NS::HdFormatSNorm8Vec2: *outGLFormat = GL_RG; *outGLType = GL_BYTE; *outGLInternalFormat = GL_RG8_SNORM; break;
  case PXR_NS::HdFormatSNorm8Vec4: *outGLFormat = GL_RGBA; *outGLType = GL_BYTE; *outGLInternalFormat = GL_RGBA8_SNORM; break;
  case PXR_NS::HdFormatFloat16: *outGLFormat = GL_RED; *outGLType = GL_HALF_FLOAT; *outGLInternalFormat = GL_R16F; break;
  case PXR_NS::HdFormatFloat16Vec2: *outGLFormat = GL_RG; *outGLType = GL_HALF_FLOAT; *outGLInternalFormat = GL_RG16F; break;
  case PXR_NS::HdFormatFloat16Vec3: *outGLFormat = GL_RGB; *outGLType = GL_HALF_FLOAT; *outGLInternalFormat = GL_RGB16F; break;
  case PXR_NS::HdFormatFloat16Vec4: *outGLFormat = GL_RGBA; *outGLType = GL_HALF_FLOAT; *outGLInternalFormat = GL_RGBA16F; break;
  case PXR_NS::HdFormatFloat32: *outGLFormat = GL_RED; *outGLType = GL_FLOAT; *outGLInternalFormat = GL_R32F; break;
  case PXR_NS::HdFormatFloat32Vec2: *outGLFormat = GL_RG; *outGLType = GL_FLOAT; *outGLInternalFormat = GL_RG32F; break;
  case PXR_NS::HdFormatFloat32Vec3: *outGLFormat = GL_RGB; *outGLType = GL_FLOAT; *outGLInternalFormat = GL_RGB32F; break;
  case PXR_NS::HdFormatFloat32Vec4: *outGLFormat = GL_RGBA; *outGLType = GL_FLOAT; *outGLInternalFormat = GL_RGBA32F; break;
  case PXR_NS::HdFormatInt16: *outGLFormat = GL_RED_INTEGER; *outGLType = GL_SHORT; *outGLInternalFormat = GL_R16I; break;
  case PXR_NS::HdFormatInt16Vec2: *outGLFormat = GL_RG_INTEGER; *outGLType = GL_SHORT; *outGLInternalFormat = GL_RG16I; break;
  case PXR_NS::HdFormatInt16Vec3: *outGLFormat = GL_RGB_INTEGER; *outGLType = GL_SHORT; *outGLInternalFormat = GL_RGB16I; break;
  case PXR_NS::HdFormatInt16Vec4: *outGLFormat = GL_RGBA_INTEGER; *outGLType = GL_SHORT; *outGLInternalFormat = GL_RGBA16I; break;
  case PXR_NS::HdFormatUInt16:*outGLFormat = GL_RED_INTEGER; *outGLType = GL_UNSIGNED_SHORT; *outGLInternalFormat = GL_R16UI; break;
  case PXR_NS::HdFormatUInt16Vec2:*outGLFormat = GL_RG_INTEGER; *outGLType = GL_UNSIGNED_SHORT; *outGLInternalFormat = GL_RG16UI; break;
  case PXR_NS::HdFormatUInt16Vec3:*outGLFormat = GL_RGB_INTEGER; *outGLType = GL_UNSIGNED_SHORT; *outGLInternalFormat = GL_RGB16UI; break;
  case PXR_NS::HdFormatUInt16Vec4:*outGLFormat = GL_RGBA_INTEGER; *outGLType = GL_UNSIGNED_SHORT; *outGLInternalFormat = GL_RGBA16UI; break;
  case PXR_NS::HdFormatInt32: *outGLFormat = GL_RED_INTEGER; *outGLType = GL_INT; *outGLInternalFormat = GL_R32I; break;
  case PXR_NS::HdFormatInt32Vec2: *outGLFormat = GL_RG_INTEGER; *outGLType = GL_INT; *outGLInternalFormat = GL_RG32I; break;
  case PXR_NS::HdFormatInt32Vec3: *outGLFormat = GL_RGB_INTEGER; *outGLType = GL_INT; *outGLInternalFormat = GL_RGB32I; break;
  case PXR_NS::HdFormatInt32Vec4: *outGLFormat = GL_RGBA_INTEGER; *outGLType = GL_INT; *outGLInternalFormat = GL_RGBA32I; break;
  case PXR_NS::HdFormatFloat32UInt8:*outGLFormat = GL_DEPTH_STENCIL; *outGLType = GL_FLOAT; *outGLInternalFormat = GL_DEPTH32F_STENCIL8; break;
  default: *outGLFormat = GL_RGBA; *outGLType = GL_UNSIGNED_BYTE; *outGLInternalFormat = GL_RGBA8; break;
  }
}


/*static*/ StormRenderInterface::GLContextSharedPtr
StormRenderInterface::getCurrentGLContext()
{
  PXR_NS::GlfGLContextSharedPtr ctx = PXR_NS::GlfGLContext::GetCurrentGLContext();
  return reinterpret_cast<GLContextSharedPtr&>(ctx);
}

StormRenderInterface::GLContextSharedPtr
StormRenderInterface::getSharedGLContext()
{
  PXR_NS::GlfGLContextSharedPtr ctx = PXR_NS::GlfGLContext::GetSharedGLContext();
  return reinterpret_cast<GLContextSharedPtr&>(ctx);
}

/*static*/ void
StormRenderInterface::makeCurrentGLContext(const GLContextSharedPtr& glcontext)
{
  PXR_NS::GlfGLContext::MakeCurrent(reinterpret_cast<const PXR_NS::GlfGLContextSharedPtr&>(glcontext));
}


//--------------------------------------------------------------------------------------------------



StormRenderInterface::StormRenderInterface() :
  RenderInterface(),
  //
  _flipFrontFacing(false),
  _applyRenderState(false),
  _enableIdRender(false),
  _enableSampleAlphaToCoverage(false),
  _sampleCount(1),
  _drawMode(1)
{
  // Remove the ifdef to silence the voluminous diagnostic output from Hydra on
  // all OSes. Currently enabled for windows to suppress the OpenGL error
  // messages being output by.
#ifdef _WIN32
  PXR_NS::TfDiagnosticMgr::GetInstance().SetQuiet(true);
#endif

  // Initialize SdrRegistry now as it can deadlock if initialized from the scene delegate
  PXR_NS::SdrRegistry::GetInstance();
}


/*virtual*/ void
StormRenderInterface::_initHydraGL()
{
  static std::once_flag sInitializedGL;
  std::call_once(sInitializedGL, [] {
    PXR_NS::GlfContextCaps::InitInstance();
  });
}


bool
StormRenderInterface::_initRenderer(const Token& rendererId)
{
  if (rendererId.empty() || !_pImpl) {
    return false;
  }

  Token rendererToUse = rendererId;
  // Check for the chosen renderer not being available
  if (!rendererPluginExists(rendererId)) {
    // List of all available Hydra renderer plugins:
    const auto& plugins = rendererPlugins();
    const std::string rendererName = rendererId.asString();

    // Special case for hdStorm which has three different IDs, one for GL, one for Metal and one for old versions.
    // If we can't find the specified hdStorm render, switch silently to any available hdStorm to avoid alarming the user.
    if (rendererName.substr(0, 7) == "HdStorm") {
      auto it = std::find_if(plugins.begin(), plugins.end(), [](const RenderInterface::RenderPluginDesc& plugin) {
        return plugin.id.asString().substr(0, 7) == "HdStorm";
      });
      if (it != plugins.end()) {
        rendererToUse = it->id;
      }
    }

    // Renderer not available: show an error
    if (!plugins.empty()) {
      std::cerr << "StormRenderInterface::initialize: Hydra renderer '" << rendererId << "' is not available, ignoring" << std::endl;
    }
    else {
      std::cerr << "StormRenderInterface::initialize: No Hydra renderers are available!" << std::endl;
    }

    return false;
  }

  if (_pImpl->initHydra(toTfToken(rendererToUse))) {
    _rendererId = rendererToUse;
    _pImpl->taskController->SetEnableSelection(false);

    // To disable viewport rendering
    _pImpl->taskController->SetViewportRenderOutput(PXR_NS::TfToken());

    _pImpl->selTracker.reset(new PXR_NS::HdxSelectionTracker());

    const auto& renderSettings = _pImpl->renderDelegate->GetRenderSettingDescriptors();
    _rendererSettings.reserve(renderSettings.size());
    for (const auto& setting : renderSettings) {
      RenderSettingsDescriptor desc;
      desc.name = setting.name;
      desc.key = fromTfToken(setting.key);
      desc.defaultValue = Value(Value::Impl(setting.defaultValue));
      _rendererSettings.push_back(desc);
    }
  }

  return true;
}

void
StormRenderInterface::_configureScene(const usg::StageRef usgStage, fdk::TimeValue frame)
{
  if (!rendererInitialized() || !usgStage) {
    return;
  }

// TODO: all this is temp - it's dumped in here from HydraViewer so we can
// avoid exposing all the methods for now.

  PXR_NS::HdStRenderDelegate* hdStRenderDelegate = dynamic_cast<PXR_NS::HdStRenderDelegate*>(_pImpl->renderDelegate);

  {
    PXR_NS::HdxRenderTaskParams renderTaskParams;
    renderTaskParams.enableLighting       = _enableLighting;
    renderTaskParams.enableSceneMaterials = _enableSceneMaterials;
    renderTaskParams.cullStyle            = static_cast<PXR_NS::HdCullStyle>(_cullStyle);
    renderTaskParams.wireframeColor       = reinterpret_cast<PXR_NS::GfVec4f&>(_wireframeColor);
    renderTaskParams.pointColor           = reinterpret_cast<PXR_NS::GfVec4f&>(_pointColor);
    renderTaskParams.pointSize            = _pointSize;

    _pImpl->taskController->SetRenderParams(renderTaskParams);
  }

  if (hdStRenderDelegate != nullptr)
  {
    // Set the color correction to sRGB to make the display colors match Nuke's viewer.
    // Nuke's viewer actually does sRGB by converting all material colours to sRGB,
    // which is incorrect as lighting needs to be done in linear space.

    // TODO: We would like to track the viewer's display LUT, but this
    // will require an API to be added to ViewerContext to obtain this.
    // We'll also need to enable OCIO support in USD.
    PXR_NS::HdxColorCorrectionTaskParams colorCorrectParams;
    colorCorrectParams.colorCorrectionMode = PXR_NS::HdxColorCorrectionTokens->sRGB;

    _pImpl->taskController->SetColorCorrectionParams(colorCorrectParams);
  }


  {
    static PXR_NS::TfToken sReprTokens[] =
    {
      PXR_NS::HdReprTokens->hull,
      PXR_NS::HdReprTokens->smoothHull,
      PXR_NS::HdReprTokens->wire,
      PXR_NS::HdReprTokens->wireOnSurf,
      PXR_NS::HdReprTokens->refined,
      PXR_NS::HdReprTokens->refinedWire,
      PXR_NS::HdReprTokens->refinedWireOnSurf,
      PXR_NS::HdReprTokens->points,
    };

    _pImpl->taskController->SetCollection(PXR_NS::HdRprimCollection(PXR_NS::HdTokens->geometry,
                                          PXR_NS::HdReprSelector(sReprTokens[_drawMode])));
  }

  _pImpl->taskController->SetEnableShadows(_enableShadows);

  {
    setOutputLayers(TokenArray{ GeomTokens.colorAov });

    // Which output should be rendered to the viewport.
    _pImpl->taskController->SetViewportRenderOutput(PXR_NS::HdAovTokens->color);

    PXR_NS::HdAovDescriptor desc;
    desc.format       = PXR_NS::HdFormatFloat16Vec4;
    desc.clearValue   = PXR_NS::VtValue(PXR_NS::GfVec4d(0.0, 0.0, 0.0, 0.0));
    desc.multiSampled = true;

    _pImpl->taskController->SetRenderOutputSettings(PXR_NS::HdAovTokens->color, desc);
  }

  // Build the scene delegate from the input stage:
  {
    const usg::Stage::Handle* stageHandle = usgStage->getUsdStageRefPtr(usdAPIVersion());

    const PXR_NS::UsdStageRefPtr stage = *reinterpret_cast<const PXR_NS::UsdStageRefPtr*>(stageHandle);
    if ((stage && !_pImpl->sceneDelegate) || _pImpl->stage != stage || usgStage->sessionState() != _lastStageSessionHash) {
      PXR_NS::SdfPath path("/NukeSceneDelegates");
      if (!_sceneName.empty()) {
        path.AppendChild(PXR_NS::TfToken(_sceneName));
      }
      else {
        path.AppendChild(PXR_NS::TfToken("untitledScene"));
      }
      if (_pImpl->sceneDelegate) {
        _pImpl->renderIndex->RemoveSubtree(path, _pImpl->sceneDelegate.get());
      }

      _pImpl->sceneDelegate = std::make_unique<UsgSceneDelegate>(_pImpl->renderIndex, path);
      _pImpl->sceneDelegate->populate(usgStage);

      _pImpl->stage = stage;
      _lastStageSessionHash = usgStage->sessionState();
    }
  }

  // TODO ndk14: This block should not be executed for each drawn frame.
  // Because the scene graph makes an assumption that it owns the session
  // layer, replacing its contents based a string, we can't rely on it's
  // correctness between frames. Fixing this in a better way probably means
  // rethinking how the session layer is best shared with the scene graph.
  // However, this will add too much risk before the RC.
  {
    _lastStageProcessHash = usgStage->processState();

    auto sessionLayer = usgStage->getSessionLayer();
    auto stage = _pImpl->stage;

    std::set<std::string> layerIds;
    std::set<PXR_NS::SdfPath> blockedPrims;
    for (auto layer : stage->GetLayerStack(false)) {
      auto customLayerData = layer->GetCustomLayerData();
      auto displayMaterialsData = customLayerData.find(toTfToken(GeomTokens.usg.displayMaterials));

      if (displayMaterialsData != customLayerData.end()
          && !displayMaterialsData->second.GetWithDefault<bool>(true)) {

        std::stack<PXR_NS::SdfLayerRefPtr> layers;
        layers.push(layer);

        while (!layers.empty()) {
          auto layer = layers.top();
          layers.pop();
          std::vector<std::string> sublayers = layer->GetSubLayerPaths();
          for (const auto &s : sublayers) {
            if (auto sublayer = PXR_NS::SdfLayer::FindRelativeToLayer(layer, s)) {
              layers.push(sublayer);
            }
          }

          layerIds.insert(layer->GetRealPath());
          for (const auto& p : layer->GetRootPrims()) {
            blockedPrims.insert(p->GetPath());
          }
        }
      }
    }

    const PXR_NS::TfToken materialBinding = toTfToken(GeomTokens.material.binding);
    const PXR_NS::TfToken nukeMaterialOp("nuke_material_op");
    for (const auto& p : stage->Traverse()) {
      _pImpl->sceneDelegate->impl()->populateShadowWorkaround(p.GetPath());
      for (const auto& blockedPrim : blockedPrims) {
        if (p.GetPath().HasPrefix(blockedPrim)) {
          if (p.HasRelationship(materialBinding)) {
            PXR_NS::SdfPathVector targets;
            p.GetRelationship(materialBinding).GetTargets(&targets);
            bool overrideMaterialBinding = true;
            for (const auto& t : targets) {
              auto prim = stage->GetPrimAtPath(t);
              if (prim.IsValid() && prim.HasCustomDataKey(nukeMaterialOp)) {
                overrideMaterialBinding = false;
                break;
              }
            }

            if (overrideMaterialBinding) {
              auto primOverride = sessionLayer->overridePrim(Prim::Impl(p));
              if (primOverride.getRelationship(GeomTokens.material.binding)) {
                continue;
              }
              auto rel = primOverride.createRelationship(GeomTokens.material.binding, true);
              if (rel.impl()->sdfrel) {
                rel.impl()->sdfrel->GetTargetPathList().ClearEditsAndMakeExplicit();
              }
            }
          }
          break;
        }
      }
    }
  }

  // Set the render tags
  PXR_NS::TfTokenVector renderTags;
  renderTags.push_back(PXR_NS::HdRenderTagTokens->geometry);
  if (_displayRender) {
    renderTags.push_back(PXR_NS::HdRenderTagTokens->render);
  }
  if (_displayProxy) {
    renderTags.push_back(PXR_NS::HdRenderTagTokens->proxy);
  }
  if (_displayGuides) {
    renderTags.push_back(PXR_NS::HdRenderTagTokens->guide);
  }
  _pImpl->taskController->SetRenderTags(renderTags);

  _pImpl->sceneDelegate->updateSelection(_selectedPrims, _deselectedPrims);
  _deselectedPrims.clear();

  _pImpl->sceneDelegate->SetTime(PXR_NS::SdfTimeCode(frame));

  _pImpl->sceneDelegate->SetRefineLevelFallback(_refineLevel);
  _pImpl->sceneDelegate->SetReprFallback(
      PXR_NS::HdReprSelector(PXR_NS::HdReprTokens->refined));

  {
    if (!_pImpl->lightingContext) {
      _pImpl->lightingContext = PXR_NS::GlfSimpleLightingContext::New();
    }
    PXR_NS::GlfSimpleMaterial defaultMaterial = _pImpl->lightingContext->GetMaterial();
    if (_flatShading) {
      defaultMaterial.SetAmbient({0.2f, 0.2f, 0.2f, 1.0f});
    }
    else {
      defaultMaterial.SetAmbient({_ambientColor.x, _ambientColor.y, _ambientColor.z, 1.0f});
    }
    _pImpl->lightingContext->SetMaterial(defaultMaterial);

    PXR_NS::GlfSimpleLight cameraLight;
    cameraLight.SetPosition({_cameraLightPosition.x,
                             _cameraLightPosition.y,
                             _cameraLightPosition.z,
                             1.0f});
    if (_cameraLightEnabled) {
      cameraLight.SetAmbient({0, 0, 0, 1});
      cameraLight.SetDiffuse({1, 1, 1, 1});
      cameraLight.SetSpecular({1, 1, 1, 1});
    }
    else {
      cameraLight.SetAmbient({1, 1, 1, 1});
      cameraLight.SetDiffuse({0, 0, 0, 1});
      cameraLight.SetSpecular({0, 0, 0, 1});
    }
    _pImpl->lightingContext->SetLights({cameraLight});

    _pImpl->taskController->SetLightingState(_pImpl->lightingContext);
  }

  PXR_NS::VtValue selectionValue(_pImpl->selTracker);
  _pImpl->engine->SetTaskContextData(PXR_NS::HdxTokens->selectionState, selectionValue);

  if (!hdStRenderDelegate) {
    // Make sure we get a color buffer. Some renderers require this, some don't.
    _pImpl->taskController->SetRenderOutputs(PXR_NS::TfTokenVector{PXR_NS::HdAovTokens->color, PXR_NS::HdAovTokens->depth});
  }
}

void StormRenderInterface::_enableCameraLight(bool enable)
{
  _cameraLightEnabled = enable;
}

void StormRenderInterface::_setCameraLightPosition(const fdk::Vec3f &pos)
{
  _cameraLightPosition = pos;
}

void StormRenderInterface::_reloadTextures(const std::vector<std::string>& textures)
{
  if (!textures.empty()) {
    auto& changeTracker = _pImpl->sceneDelegate->GetRenderIndex().GetChangeTracker();
    auto resourceRegistry = _pImpl->renderIndex->GetResourceRegistry();
    for (const auto& t : textures) {
      resourceRegistry->ReloadResource(PXR_NS::HdResourceTypeTokens->texture, t);
      // We also need to mark the materials as dirty
      for (const auto& m : _pImpl->sceneDelegate->materialsUsingTexture(t)) {
        changeTracker.MarkSprimDirty(toSdfPath(m), PXR_NS::HdChangeTracker::AllDirty);
      }
      _pImpl->sceneDelegate->clearMaterialUsingTexture(t);
    }
  }
}

void
StormRenderInterface::executeRenderer(int  rendererDrawsWithGLVersion,
                                      bool rendererDrawsWithGLCore)
{
  if (!rendererInitialized()) {
    return;
  }

  mFnTimeTracedFunction();
  executeTasks(!_progressiveRendering);
}


void
StormRenderInterface::finalizeRender()
{
  if (!rendererInitialized()) {
    return;
  }
}




} //namespace usg

